# -*- encoding: utf-8 -*-
"""
This module contents initialization of the Celery instance - a distributed task queue.
Depends on `settings`.`USE_CELERY` configuration parameter the system will use one of the task managers:
    * the `Celery` instance when `USE_CELERY` set to `True` or
    * the `TaskManager` instance (`uds.core.managers.TaskManager:TaskManager`)
      when `USE_CELERY` set to `False`

The `Celery` file structure consists of a few files.
termidesk/
        src/
            server/
                celery.py - this module
                celeryconfig.py - configuration file for the `Celery` instance
                celery_utils.py - contents helpers utilities

Configuration parameter reads from `settings.py`, and then from `celery_utils.py`.

"""
from __future__ import absolute_import

import os
from django.conf import settings

from celery import Celery

# set the default Django settings module for the 'celery' program.
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'server.settings')

# TODO: Delete if not needed
# sys.path.append('/'.join((os.path.dirname(settings.BASE_DIR), 'tests')))
# sys.path.append('/'.join((os.path.dirname(settings.BASE_DIR), 'tests/celery')))

if settings.USE_CELERY:
    # initialize the celery
    app = Celery('server', )

    # Using a string here means the worker doesn't have to serialize
    # the configuration object to child processes.
    # - namespace='CELERY' means all celery-related configuration keys
    #   should have a `CELERY_` prefix.
    app.config_from_object('django.conf:settings', namespace='CELERY')
    app.config_from_object('server.celeryconfig')

    # Load task modules from all registered Django app configs.
    app.autodiscover_tasks()

else:
    class FakeCeleryApp(object):
        """
        FakeCeleryApp imitates the `Celery` structure
        and functionality (app.conf, app.task decorator, etc.)
        """
        def __init__(self):
            self.Task = self.FakeStructure
            self.tasks = self.FakeStructure
            self.conf = self.FakeStructure.FakeConf

        @staticmethod
        def task(*d_args, **d_kwargs):
            """
            Fake app.task decorator

            *d_args, **d_kwargs content the decorators arguments
            For example, `name`, `bind` for a record like

            @app.task(name='my.task', bind=True)
            """
            def _task(*f_args, **f_kwargs):
                """
                *f_args, **f_kwargs content the decorated function arguments.
                For example, `object`, `flag` for a record like

                @app.task(name='my.task', bind=True)
                def my_function(object, flag)
                """
                def wrapper(func):
                    # Call decorated function with original arguments '*f_args', '**f_kwargs'

                    func(*f_args, **f_kwargs)

                return wrapper
            return _task

        # Inner class for FakeCeleryApp class
        class FakeStructure(object):
            """
            This class implements a fake structure of original `Celery` object
            with attributes, methods, and substructures
            """

            @staticmethod
            def register(*args, **kwargs):
                pass

            # Inner class for fake_structure class
            class FakeConf(object):
                beat_schedule = {}

        # End of FakeCeleryApp class

    # Make the app instance behavior like the `Celery`
    app = FakeCeleryApp()
