import smtplib

from django.apps import apps
from django.core.mail import get_connection
from django.db.utils import OperationalError, ProgrammingError
from django.utils.log import AdminEmailHandler
from tdsklog.helpers import EmailThread, get_config_value
from uds.core.managers.CryptoManager import CryptoManager


class TermideskEmailHandler(AdminEmailHandler):

    def handle(self, record):
        if apps.ready is False:
            return

        from uds.core.util.Config import NOTIFICATIONS_SECTION, GlobalConfig

        try:
            enable_notification = get_config_value(
                GlobalConfig.NOTIFICATIONS_ENABLE.key(), NOTIFICATIONS_SECTION
            )
        except (OperationalError, ProgrammingError):
            # Обработка ошибки недоготовности таблиц в БД
            return

        if enable_notification != '1':
            return

        try:
            super().handle(record)
        except (ConnectionRefusedError, smtplib.SMTPServerDisconnected):
            # Обработка ошибки неверного конфигурирования параметров уведомления
            return

    def connection(self):
        from uds.core.util.Config import NOTIFICATIONS_SECTION, GlobalConfig

        host = get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_HOST.key(), NOTIFICATIONS_SECTION
        )
        port = int(get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_PORT.key(), NOTIFICATIONS_SECTION
        ))
        username = get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_USERNAME.key(), NOTIFICATIONS_SECTION
        )
        password = get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_PASSWORD.key(), NOTIFICATIONS_SECTION
        )
        password = CryptoManager.manager().decrypt(password)

        timeout = get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_TIMEOUT.key(), NOTIFICATIONS_SECTION
        )
        timeout = int(timeout) if timeout != '0' else None

        use_tls, use_ssl = self.get_tls_ssl()

        return get_connection(backend=self.email_backend,
                              host=host,
                              port=port,
                              username=username,
                              password=password,
                              use_tls=use_tls,
                              use_ssl=use_ssl,
                              timeout=timeout)

    def send_mail(self, subject, message, *args, **kwargs):
        self._mail_admins(subject, message, *args,
                          connection=self.connection(), **kwargs)

    @staticmethod
    def _mail_admins(subject, message, fail_silently=False, connection=None,
                     html_message=None):
        """Send a message to the admins."""
        from uds.core.util.Config import NOTIFICATIONS_SECTION, GlobalConfig

        admins = get_config_value(
            GlobalConfig.NOTIFICATIONS_TO_EMAILS.key(), NOTIFICATIONS_SECTION
        ).split(',')

        subject_prefix = get_config_value(
            GlobalConfig.NOTIFICATIONS_SUBJECT_PREFIX.key(),
            NOTIFICATIONS_SECTION
        )
        server_mail = get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_SERVER_EMAIL.key(), NOTIFICATIONS_SECTION
        )

        if not admins:
            return

        EmailThread(subject='%s%s' % (subject_prefix, subject),
                    message=message, server_mail=server_mail,
                    admins=admins, connection=connection,
                    html_message=html_message, fail_silently=fail_silently).start()

    @staticmethod
    def get_tls_ssl():
        """ Получение статусов активации протоколов.
        Оба протокола работать одновременно не могут.
        Для двух включенных протоколов приоритет отдается TLS. """

        from uds.core.util.Config import NOTIFICATIONS_SECTION, GlobalConfig

        use_tls = get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_TLS.key(), NOTIFICATIONS_SECTION
        )

        use_tls = True if use_tls == '1' else False

        if use_tls:
            return use_tls, False

        use_ssl = get_config_value(
            GlobalConfig.NOTIFICATIONS_SMTP_SSL.key(), NOTIFICATIONS_SECTION
        )
        use_ssl = True if use_ssl == '1' else False

        return False, use_ssl
