# -*- coding: utf-8 -*-

import socket
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING, Handler, getLevelName
from logging.handlers import SYSLOG_UDP_PORT, SysLogHandler

from django.apps import apps
from django.db.utils import OperationalError, ProgrammingError

from termidesk.core.util.cache import timed_lru_cache

UDP = 'UDP'
TCP = 'TCP'
MONITORING_SOCKETS = [UDP, TCP]

level_name = [
    getLevelName(CRITICAL),
    getLevelName(ERROR),
    getLevelName(WARNING),
    getLevelName(INFO),
    getLevelName(DEBUG),
]

facilities = list(SysLogHandler.facility_names.items())
facilities.sort(key=lambda i: i[1])
facilities_name = [facility[0] for facility in facilities]


def get_socktype(socket_val):
    """ Получение кода типа для выбранного протокола."""
    if socket_val == UDP:
        return socket.SOCK_DGRAM
    if socket_val == TCP:
        return socket.SOCK_STREAM
    return None


@timed_lru_cache()
def _get_syslog_param(key):
    """ Получение параметров мониторинга из модели. """
    from uds.core.util.Config import MONITORING_SECTION
    from uds.models.Config import Config

    try:
        return Config.objects.get(
            section=MONITORING_SECTION, key=key
        ).value
    except Config.DoesNotExist:
        return None


@timed_lru_cache()
def _get_syslog_params(keys: tuple):
    """ Получение нескольких параметров мониторинга из модели. """
    from uds.core.util.Config import MONITORING_SECTION
    from uds.models.Config import Config

    try:
        params = Config.objects.filter(section=MONITORING_SECTION, key__in=keys).all()
        return {elem.key: elem.value for elem in params}
    except Config.DoesNotExist:
        return None


def _send_record_to_server(address, socktype, facility, record, formatter):
    """ Отправка уведомления на сервер. """
    handler = SysLogHandler(address, socktype=socktype, facility=facility)
    handler.formatter = formatter
    handler.handle(record)
    handler.close()


class TermideskSysLogDispatcher(Handler):
    """Класс-диспетчер обеспечивает отправку сообщений на несколько syslog серверов"""
    _fix_time = False
    socktype = None

    servers_list = []

    def handle(self, record):
        if apps.ready is False:
            return

        from uds.core.util.Config import GlobalConfig

        if GlobalConfig.initDone is False:
            return

        try:
            enable_syslog = _get_syslog_param(GlobalConfig.ENABLE_SYSLOG.key())
        except (OperationalError, ProgrammingError):
            # Обработка ошибки недоготовности таблиц в БД
            return

        if enable_syslog is not None and int(enable_syslog) == 1:
            self.set_common_params()

            # перекрытие уровня критичности уведомления
            if record.levelno < self.levelno:
                return

            self.send_record(record)

    def set_clean_servers(self, servers_list):
        """
        Валидация и запись полученных адресов.
        Удаление повторяющихся и пустых адресов из списка syslog серверов.
        """
        servers = []
        for server_address in servers_list:
            # Адрес сервера должен отcутствовать в `servers_list`
            # и должен отличаться от пустой строки '' и `None`
            if server_address not in servers and server_address not in ('', None,):
                servers.append(server_address)

        self.servers_list = servers

    def send_record(self, record):
        """Отправка логов на syslog сервер в отдельных потоках"""

        # Список для работы с созданными потоками
        threads = []
        from threading import Thread
        # Для каждого сервера создается отдельный поток
        for syslog_server_address in self.servers_list:
            address = self.get_syslog_server_address(syslog_server_address)

            thread = Thread(target=_send_record_to_server,
                            args=(address, self.socktype, self.facility, record, self.formatter))

            # Запуск потока на выполнение
            thread.start()
            # Добавление текущего потока в общий список
            threads.append(thread)

    def set_common_params(self):
        """ Установка общих параметров для будущей отправки уведомлений. """
        from uds.core.util.Config import GlobalConfig

        main_params = _get_syslog_params(
            (GlobalConfig.SYSLOG_SOCKET.key(),
             GlobalConfig.SYSLOG_LEVEL.key(),
             GlobalConfig.SYSLOG_FACILITY.key(),
             GlobalConfig.SYSLOG_HOST.key(),
             GlobalConfig.SYSLOG_HOST_2.key(),
             GlobalConfig.SYSLOG_HOST_3.key())
        )

        self.socktype = get_socktype(main_params.get(GlobalConfig.SYSLOG_SOCKET.key()))
        self._set_level(main_params.get(GlobalConfig.SYSLOG_LEVEL.key()))
        self._set_facility(main_params.get(GlobalConfig.SYSLOG_FACILITY.key()))

        self.set_clean_servers(
            (
                main_params.get(GlobalConfig.SYSLOG_HOST.key()),
                main_params.get(GlobalConfig.SYSLOG_HOST_2.key()),
                main_params.get(GlobalConfig.SYSLOG_HOST_3.key()),
            )
        )

    @staticmethod
    def get_syslog_server_address(server_address_str):
        """ Получение имени хоста и порта syslog-сервера. """
        # Ожидается, что переданный адрес будет в формате `hostname:port_number`
        _address = server_address_str.split(':')

        try:
            host = _address[0]
            port = int(_address[1])
        except IndexError:
            # Если адрес сервера не содержит номера порта, то используется номер порта по умолчанию
            host = server_address_str
            port = int(SYSLOG_UDP_PORT)

        return (host, port,)

    def _set_level(self, level_name: str):
        """ Установка уровня важности сообщения (Severity). """
        self.levelname = level_name
        self.levelno = getLevelName(level_name)

    def _set_facility(self, facility_name: str):
        """ Установка категории сформировавшего сообщение (Facility). """
        self.facility = int(SysLogHandler.facility_names[facility_name])
