from django.utils.translation import gettext_lazy as _
from rest_framework import serializers
from uds.core.auths.auth import authenticate
from uds.models.Authenticator import Authenticator
from uds.core.util.stats import events


class AuthTokenSerializer(serializers.Serializer):
    username = serializers.CharField(
        label=_("Login"),
        write_only=True
    )
    password = serializers.CharField(
        label=_("Password"),
        style={'input_type': 'password'},
        trim_whitespace=False,
        write_only=True
    )
    auth_id = serializers.IntegerField(
        label=_("Authenticator ID"),
        write_only=True
    )
    token = serializers.CharField(
        label=_("Token"),
        read_only=True
    )

    def validate(self, attrs):

        try:
            authenticator = Authenticator.objects.get(id=attrs.get('auth_id'))
        except Authenticator.DoesNotExist:
            msg = _('Accept invalid authenticator ID')
            raise serializers.ValidationError(msg, code='authorization')

        username = attrs.get('username')
        password = attrs.get('password')

        if not username or not password:
            msg = _('Must include "username" and "password".')
            raise serializers.ValidationError(msg, code='authorization')

        user = authenticate(username, password, authenticator)
        request = self.context["request"]
        # The authenticate call simply returns None for is_active=False
        # users. (Assuming the default ModelBackend authentication
        # backend.)
        if not user:
            msg = _('Unable to log in with provided credentials.')
            raise serializers.ValidationError(msg, code='authorization')
        else:
            # Add an login event
            events.addEvent(authenticator, events.ET_LOGIN, username=username,
                            srcip=request.ip)
            events.addEvent(authenticator, events.ET_PLATFORM,
                            platform=request.os.OS,
                            browser=request.os.Browser,
                            version=request.os.Version)

        attrs['user'] = user
        return attrs
