from functools import wraps
import logging

from rest_framework import exceptions, permissions
from rest_framework.exceptions import PermissionDenied, APIException

from termidesk.api.permissions import is_auth_swagger_path
from termidesk.core.auths.utils import check_exists_permission_for_user
from termidesk.models.User import AnonymousUser
from termidesk.models.UserPermissionApi import PREFIX_PERMISSION_API
from uds.core.util.Config import GlobalConfig

logger = logging.getLogger(__name__)


class PermissionRequired(permissions.BasePermission):
    def __init__(self, permission_name: str):
        self.permission_name = permission_name

    def has_permission(self, request, view):
        if is_auth_swagger_path(request.user, request.path):
            return True

        path_permissions = self.permission_name.split('.')

        if PREFIX_PERMISSION_API != path_permissions[0]:
            raise APIException(
                "The wrong permission was passed. PermissionRequired accept only API permissions."
            )

        if not request.user.has_perm(self.permission_name):
            raise exceptions.PermissionDenied("Don't have permission")
        return True


class ApiPermission(permissions.BasePermission):
    """ Класс-декоратор для установки api-permissions. """

    def __init__(self, name):
        self.name = name

    def __call__(self):
        return PermissionRequired(self.name)


def extra_permission_classes(permission_names: tuple):
    """ Добавление и проверка всех разрешений на определенные методы. """
    function_permissions = [PermissionRequired(name) for name in permission_names]

    def decorator(func):
        def decorated_func(self, *args, **kwargs):
            class_permissions = [permission() for permission in self.permission_classes]
            all_permissions = class_permissions + function_permissions

            for permission in all_permissions:
                if not permission.has_permission(self.request, self):
                    self.permission_denied(
                        self.request,
                        message=getattr(permission, 'message', None),
                        code=getattr(permission, 'code', None)
                    )

            return func(self, *args, **kwargs)

        return decorated_func

    return decorator


def has_role_permission(permission_name: str):
    """ Добавление и проверка всех разрешений на определенные методы. """
    def decorator(func):
        @wraps(func)
        def decorated_func(self, *args, **kwargs):
            if not isinstance(self.request.user, AnonymousUser):
                # У анонимного пользователя нет никаких разрешений
                if not self.request.user.has_perm(permission_name):
                    # У пользователя отсутствует разрешение
                    # В ответ на запрос придет 403 Forbidden
                    logger.debug('Checking permission "{}" failed!'.format(permission_name))
                    raise PermissionDenied

            return func(self, *args, **kwargs)

        return decorated_func

    return decorator


def has_object_role_permission(permission_name: str, model, name_uuid_field):
    """ Проверка наличия разрешения для доступа к операции над объектом конкретного пользователя. """

    def decorator(func):
        def decorated_func(self, *args, **kwargs):
            if not isinstance(self.request.user, AnonymousUser) and not self.request.user.is_superuser:

                uuid = kwargs.get(name_uuid_field)

                if not check_exists_permission_for_user(self.request.user, model, uuid, permission_name):
                    # У пользователя отсутствует разрешение
                    # В ответ на запрос придет 403 Forbidden
                    logger.debug('Checking permission "{}" failed!'.format(permission_name))
                    raise PermissionDenied

            return func(self, *args, **kwargs)

        return decorated_func

    return decorator


def enabled_rbac_required(func):
    """ Требует включенной ролевой модели. """
    def decorator(self, *args, **kwargs):
            rbac_enabled = GlobalConfig.EXPERIMENTAL_RBAC_ENABLED.getBool(True)
            if not rbac_enabled:
                # Ролевая модель не активна
                # В ответ на запрос придет 403 Forbidden
                logger.debug('Checking RBAC enabled: "{}"'.format(rbac_enabled))
                raise PermissionDenied
            else:
                logger.debug('Checking RBAC enabled: "{}"'.format(rbac_enabled))

            return func(self, *args, **kwargs)

    return decorator
