import re

from rest_framework.response import Response
from rest_framework.schemas.openapi import SchemaGenerator
from rest_framework.viewsets import ViewSet
from .serializers.api import (
    ApiSerializer, ModulesSerializer, SchemaSerializer, ServerSerializer, VersionsSerializer)
from version import __version__


class ApiSchema(ViewSet):
    # модули, которые должны быть исключены из выдачи
    IGNORE_MODULES = ['discover', 'test', 'webui']
    MODULE_EXTERNAL = 'external'

    def get(self, request):
        api_serializer = ApiSerializer(data={
            'schema': self.__get_schema(),
            'server': self.__get_server(),
            'modules': self.__get_modules(request),
        })
        api_serializer.is_valid(raise_exception=True)

        return Response(api_serializer.data)

    def __get_schema(self):
        """ Получение версия структуры обзорного ресурса. """
        versions_serializer = VersionsSerializer(data=[
            {'version': '1.0', 'url': '/api/discover'}
        ], many=True)
        versions_serializer.is_valid(raise_exception=True)

        schema = SchemaSerializer(data={'version': '1.0', 'versions': versions_serializer.data})
        schema.is_valid(raise_exception=True)
        return schema.data

    def __get_server(self):
        """ Получение версии сервера для определения доступных api. """
        version_serializer = ServerSerializer(data={'version': __version__})
        version_serializer.is_valid(raise_exception=True)
        return version_serializer.data

    def __get_modules(self, request):
        """ Получение списка модулей REST API. """
        document = SchemaGenerator().get_schema(request=request, public=True)
        modules = []
        mod_cache = dict()
        for name, elements in document['paths'].items():
            _, base, module, path, *rest = name.split('/')
            if module in self.IGNORE_MODULES:
                continue

            if module == self.MODULE_EXTERNAL:
                _, base, external, submodule, path, *rest = name.split('/')
                module = "{}/{}".format(external, submodule)

            if module not in mod_cache:
                mod_cache[module] = {'versions': {}}

            versions = mod_cache[module]['versions']
            version = self.__get_version(path)
            versions[version] = {'version': version, 'url': '/api/{}/{}'.format(module, path)}

        for module in mod_cache:
            versions = [v for k, v in mod_cache[module]['versions'].items()]
            versions = sorted(versions, key=lambda item: item['version'])
            versions_serializer = VersionsSerializer(data=versions, many=True)
            versions_serializer.is_valid(raise_exception=True)
            modules.append({'id': module, 'versions': versions_serializer.data})

        modules = sorted(modules, key=lambda item: item['id'])
        module_serializer = ModulesSerializer(data=modules, many=True)
        module_serializer.is_valid(raise_exception=True)
        return module_serializer.data

    @staticmethod
    def __get_version(path: str):
        if path == 'draft':
            return path

        pattern = r'v(\w+\.\w+|\w+)'
        result = re.findall(pattern, path)
        if len(result) != 0:
            return result[0]

        return ''
