"""Модуль содержит обработчик ошибок HTTP запросов"""
# -*- encoding: utf-8 -*-
import logging

from django.conf import settings

from rest_framework.exceptions import ValidationError
from rest_framework.response import Response
from rest_framework.status import HTTP_500_INTERNAL_SERVER_ERROR
from rest_framework.views import exception_handler


logger = logging.getLogger(__name__)


def api_exception_handler(exc, context):
    """Собственная реализация обработчика исключений для формирования json-ответа"""

    # Обработка штатных исключений, например, если функция `get_object_or_404()` вернет 404 ошибку,
    # то `exception_handler` вернет соответсвующий объект ответа (Response)
    response = exception_handler(exc, context)

    if isinstance(exc, ValidationError):
        # Обрабатываем только случай, когда переданная ошибка содержит наименования полей
        # Эти имена связаны либо с полями, либо с типом ошибки non_field_errors
        # Ошибки, которые падают списком, обрабатываются общим обработчиком
        if isinstance(exc.detail, dict):
            response.data = get_data_for_validation_error(exc)

    if response is None:
        # Если `exception_handler` не смог обработать исключение и вернул None, то считаем,
        # что произошла непредвиденная ошибка и формируем ответ `HTTP_500_INTERNAL_SERVER_ERROR`
        logger.exception(exc)
        response = Response(
            data={'detail': get_exception_detail(exc), },
            status=HTTP_500_INTERNAL_SERVER_ERROR,
        )

    # Тело ответа может отсутствовать
    response_data = getattr(response, "data", "")

    if not isinstance(response_data, dict):
        # Если тело ответа не является словарем, выполняется унификация структуры
        # * тело ответа считается детальным описанием ошибки (`detail`),
        # * код ответа дублируется в тело ответа (`status_code`)
        response.data = {
            'detail': response_data,
        }

    # Добавление кода ошибки в тело ответа
    response.data['status_code'] = getattr(response, "status_code", HTTP_500_INTERNAL_SERVER_ERROR)

    return response


def get_exception_detail(_exc):
    """Функция формирует строковое описание для переданного объекта исключения"""
    exception_detail = "{exc_type}: {exc_desc}".format(exc_type=type(_exc).__name__, exc_desc=str(_exc))
    return exception_detail


def get_data_for_validation_error(exc: ValidationError):
    """
    Приведение ошибки валидации к специальному виду,
    для правильного отображения на лицевой части.
    """
    first_mistake = None

    non_field_key = settings.REST_FRAMEWORK.get('NON_FIELD_ERRORS_KEY', 'non_field_errors')

    non_field_values = exc.detail.pop(non_field_key, [])
    if isinstance(non_field_values, list) and len(non_field_values) > 0:
        first_mistake = non_field_values[0]

    field_errors = {}
    for key, value in exc.detail.items():
        if first_mistake is None:
            first_mistake = value[0] if isinstance(value, list) and len(value) >= 1 else value

        field_errors[key] = value

    return {
        "status_code": exc.status_code,
        "detail": first_mistake,
        "validation": {
            non_field_key: non_field_values,
            "field_errors": field_errors
        }
    }
