"""Модуль содержит логику работы с разрешениями для REST API на базе DRF"""
# -*- encoding: utf-8 -*-
from django.utils.translation import gettext_lazy as _
from rest_framework.permissions import BasePermission

from termidesk.models.User import AnonymousUser
from uds.REST import Handler
from uds.REST.handlers import AUTH_TOKEN_HEADER
from uds.core.util.model import processUuid
from uds.models import UserService

NAME_DOCUMENTATION = 'docs'

NAME_PATH_DOCUMENTATION = '/api/{}/'.format(NAME_DOCUMENTATION)


def is_auth_swagger_path(user, path: str):
    """ Проверка, что текущий пользователь авторизован и запращивает swagger. """
    return not isinstance(user, AnonymousUser) and user.is_admin and path == NAME_PATH_DOCUMENTATION


class AccessByServicePoolPermission(BasePermission):
    message = 'Only useable with the correct uuid of agent is available'

    def has_permission(self, request, view):
        if is_auth_swagger_path(request.user, request.path):
            return True
        service_uuid = view.kwargs.get('uuid_actor', None)
        return UserService.objects.filter(uuid=processUuid(service_uuid)).exists()


class IsAuthAdmin(BasePermission):
    message = 'Access only for authorized staff members'

    def has_permission(self, request, view):
        user = request.user

        if isinstance(user, AnonymousUser):
            return False

        if any([user.is_staff, user.is_admin]) is False:
            # Для доступа пользователь должен быть:
            # * либо Персоналом (is_staff/staff_member)
            # * либо Суперадминистратором (is_admin/is_superuser)
            # * либо всё вместе.
            return False

        return True


class AccessPermission(BasePermission):
    message = 'Available only for staff members'

    def has_permission(self, request, view):
        user = request.user

        if isinstance(user, AnonymousUser):
            auth_token = request.META.get(AUTH_TOKEN_HEADER)
            if auth_token is not None:
                handler = Handler(request=request,
                                  path=request.get_full_path(),
                                  operation=request.method,
                                  params={})
                user = handler.getUser()

        return not isinstance(user, AnonymousUser) and any([user.is_staff, user.is_admin])


class EnabledPoliciesRequiredPermission(BasePermission):
    """ Проверка доступа к политикам."""
    message = _("You do not have permission to perform this action.")

    def has_permission(self, request, view):
        from uds.core.util.Config import GlobalConfig
        user = request.user

        if isinstance(user, AnonymousUser):
            return False

        return GlobalConfig.EXPERIMENTAL_POLICIES_ENABLED.getBool(True) and user.is_admin
