# -*- encoding: utf-8 -*-

import logging
from abc import ABCMeta, abstractmethod

from django.contrib.auth.models import Group, Permission
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from guardian.models.models import GroupObjectPermission
from guardian.shortcuts import assign_perm
from rest_framework import serializers

logger = logging.getLogger(__name__)


class EntitySerializer(serializers.ModelSerializer):
    id = serializers.IntegerField(label='ID', read_only=True)
    group_id = serializers.CharField(required=True, label='Роль')
    object_uuid = serializers.CharField(source='content_object.uuid', required=True, label='Уникальный идентификатор объекта')
    data_type = serializers.CharField(required=True, label='Тип объекта', write_only=True)

    group_name = serializers.CharField(source='group.name', required=False, label='Роль', read_only=True)
    name_object = serializers.CharField(source='content_object.name',
                                        required=False,
                                        label='Имя объекта',
                                        read_only=True)
    type_object = serializers.SerializerMethodField()
    permission_name = serializers.CharField(source='permission.name', required=False, label='Имя разрешения',
                                            read_only=True)

    object_pk = serializers.IntegerField(required=False, read_only=True)
    permission_id = serializers.CharField(required=True)
    content_type_id = serializers.IntegerField(required=False, read_only=True)

    # наименование группы объектов, которое будет отображаться в списке
    title = None

    # описание группы объектов, для заполнения информации из списка объектов
    description = ''

    # модель для работы с данными
    model = None

    __metaclass__ = ABCMeta

    class Meta:
        fields = ['id', 'data_type', 'group_id', 'group_name', 'object_uuid',
                  'name_object', 'type_object', 'permission_name', 'permission_id',
                  'object_pk', 'content_type_id']
        model = GroupObjectPermission

    @staticmethod
    @abstractmethod
    def get_gui_objects():
        return ()

    @staticmethod
    @abstractmethod
    def get_gui_allow_permissions():
        return ()

    @staticmethod
    @abstractmethod
    def get_allow_permissions():
        return []

    def get_type_object(self, instance):
        if not isinstance(instance, GroupObjectPermission):
            return ''

        return instance.content_object.__class__.__name__

    def validate(self, attrs):
        group_id = attrs.get('group_id')
        if group_id is None or Group.objects.filter(id=group_id).exists() is False:
            raise serializers.ValidationError({
                'group': 'Роли с идентификатором "{}" не существует'.format(group_id)
            })

        permission_id = attrs.get('permission_id')
        allow_permission_ids = [perm.id for perm in self.get_allow_permissions()]
        if permission_id is None or int(permission_id) not in allow_permission_ids:
            raise serializers.ValidationError({
                'permissions': ('Не все указанные разрешения "{}" допустимы. '
                                'Разрешены разрешения с идентификаторами: {}'.format(permission_id,
                                                                                     allow_permission_ids))
            })

        object_uuid = attrs.get('content_object').get('uuid')
        if self.model.objects.filter(uuid=object_uuid).exists() is False:
            raise serializers.ValidationError({
                'object_uuid': 'Такого объекта типа "{}" с идентификатором "{}" не существует'.format(
                    attrs.get('data_type'), object_uuid
                )
            })
        del attrs['content_object']

        object_pk = self.get_object_by_uuid(object_uuid).id
        content_type_id = self.get_content_type().id
        if self.__exists_object_permission(object_pk, content_type_id, group_id, permission_id) is True:
            raise serializers.ValidationError({
                'permission_id': 'Выбранная роль уже имеет такое разрешение для указанного объекта'
            })

        attrs['object_pk'] = object_pk

        return super().validate(attrs)

    def __exists_object_permission(self, object_pk, content_type_id, group_id, permission_id):
        if self.instance is not None:
            return GroupObjectPermission.objects.filter(
                object_pk=object_pk, content_type_id=content_type_id,
                group_id=group_id, permission_id=permission_id).exclude(id=self.instance.id).exists()

        return GroupObjectPermission.objects.filter(
            object_pk=object_pk, content_type_id=content_type_id,
            group_id=group_id, permission_id=permission_id).exists()

    def create(self, validated_data):
        elem = self.model.objects.get(id=validated_data.get('object_pk'))
        group = get_object_or_404(Group, id=validated_data.get('group_id'))
        permission = get_object_or_404(Permission, id=validated_data.get('permission_id'))

        return assign_perm(permission, group, elem)

    def get_content_type(self):
        return ContentType.objects.get_for_model(self.model)

    def get_object_by_uuid(self, uuid: str):
        """ Получить объект по uuid. """
        return get_object_or_404(self.model, uuid=uuid)
