import logging

from django.db import IntegrityError
from django.utils.translation import ugettext as _
from rest_framework import serializers
from rest_framework.exceptions import APIException
from rest_framework.fields import empty
from termidesk.rest.helpers import GroupUI
from uds.core.auths.Exceptions import AuthenticatorException
from uds.models.Group import Group

logger = logging.getLogger(__name__)


class AuthGroupSerializer(serializers.ModelSerializer):
    """ Сериализатор для работы с пользователями домена аутентификации. """
    id = serializers.CharField(source='uuid', required=False, read_only=True)
    type = serializers.CharField(source='data_type')
    comments = serializers.CharField(max_length=255, allow_blank=True)
    state = serializers.ChoiceField(choices=[value['id'] for value in GroupUI.state._data['values']])
    groups = serializers.SlugRelatedField(queryset=Group.objects.filter(is_meta=False).all(),
                                          slug_field='uuid', many=True, required=False)

    class Meta:
        model = Group
        fields = ['id', 'name', 'comments', 'state', 'type', 'meta_if_any', 'groups']

    def __init__(self, instance=None, data=empty, **kwargs):
        self.authenticator = kwargs.pop('authenticator')
        super().__init__(instance=instance, data=data, **kwargs)

    def validate(self, attrs):
        logger.debug('Before validate {0}'.format(attrs))
        data_type = attrs.get('data_type', self.instance.get_data_type() if self.instance else '')
        if data_type not in Group.DATA_TYPES:
            raise serializers.ValidationError({
                'data_type': 'This is not exists «{0}» type. Choose from available: {1}'.format(
                    data_type, Group.DATA_TYPES
                )})

        if self.instance is not None:
            # имя группы нельзя изменять
            del attrs['name']

        return attrs

    def save(self, **kwargs):
        meta_if_any = self.validated_data.get('meta_if_any', False)

        logger.debug('Saving group {0} / {1}'.format(self.authenticator, self.validated_data.get('name')))
        logger.debug('Meta any {}'.format(meta_if_any))

        return super().save(**kwargs)

    def create(self, validated_data: dict):
        return self.__validate_and_save(validated_data)

    def update(self, instance: Group, validated_data: dict):
        return self.__validate_and_save(validated_data, instance)

    def __has_no_meta_and_no_pattern(self, validated_data: dict):
        """ Группа не является метограппой и не имеет указания на патерн в имени. """
        is_meta = validated_data.get('data_type') == Group.DATA_TYPE_META
        is_pattern = validated_data.get('name', '').find('pat:') == 0
        return not is_meta and not is_pattern

    def __validate_and_save(self, validated_data, group=None):
        try:
            if self.__has_no_meta_and_no_pattern(validated_data):
                try:
                    if group is None:
                        self.authenticator.getInstance().createGroup(validated_data)
                    else:
                        self.authenticator.getInstance().modifyGroup(validated_data)
                except AuthenticatorException:
                    raise APIException(
                        'Этот тип домена аутентификации не поддерживает создание групп.')

            groups = validated_data.pop('groups') if validated_data.get('groups') is not None else []

            if group is None:
                group = self.authenticator.groups.create(**validated_data)
            else:
                group.__dict__.update(validated_data)

            if group.is_meta:
                group.groups.set(groups)

            group.save()
        except IntegrityError:  # Duplicate key probably
            raise APIException(
                _('User already exists (duplicate key error)')
            )
        except Exception as err:
            logger.exception('Saving group')
            raise APIException(str(err))

        return group

    def to_representation(self, instance: Group):
        """
        Object instance -> Dict of primitive datatypes.
        """
        ret = super().to_representation(instance)

        # только мета-группы имеют связей с другими группами
        if not instance.is_meta and ret.get('pop') is not None:
            ret.pop('groups')
        return ret
