import logging

from django.contrib.auth.models import Permission
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.utils.translation import ugettext as _
from rest_framework import serializers
from rest_framework.exceptions import APIException
from rest_framework.fields import empty
from termidesk.core.auths.login_manager import LoginAttemptsManager
from termidesk.license.base import checkRegistrationPossible
from termidesk.license.exceptions import LicenseMaxRegisteredUsersReachedError
from termidesk.notifications.entities.LicenceEventsHelper import LicenseEventsHelper
from termidesk.rest.helpers import UsersUI
from uds.core.auths.Exceptions import AuthenticatorException
from uds.core.util.State import State
from uds.core.managers import cryptoManager
from uds.models.User import User


logger = logging.getLogger(__name__)


class AuthUserSerializer(serializers.ModelSerializer):
    """ Сериализатор для работы с пользователями домена аутентификации. """
    id = serializers.CharField(source='uuid', required=False, read_only=True)
    last_access = serializers.SerializerMethodField()
    parent = serializers.CharField(read_only='True', required=False)
    state = serializers.ChoiceField(choices=[value['id'] for value in UsersUI.state._data['values']])
    comments = serializers.CharField(allow_blank=True)
    api_permissions = serializers.PrimaryKeyRelatedField(
        queryset=Permission.objects.all(), many=True, required=False
    )
    auth_groups = serializers.ListField(source='group_ids', required=False)

    class Meta:
        model = User
        fields = ['id', 'name', 'real_name', 'comments', 'state', 'staff_member', 'is_admin', 'password',
                  'last_access', 'parent', 'auth_groups', 'api_permissions', 'roles']
        extra_kwargs = {
            'password': {'write_only': True, 'required': False, "allow_blank": True},
        }

    def get_last_access(self, instance):
        return int(instance.last_access.timestamp())

    def __init__(self, instance=None, data=empty, **kwargs):
        self.authenticator = kwargs.pop('authenticator')
        super().__init__(instance=instance, data=data, **kwargs)

    def validate(self, attrs):
        groups = attrs.get('group_ids', [])

        if groups == []:
            raise serializers.ValidationError(
                {'auth_groups': 'Field "auth_groups" is required'}
            )

        if not self.authenticator.groups.filter(uuid__in=groups).exists():
            raise serializers.ValidationError(
                {'auth_groups': 'Not all groups from your list "{}" exists'.format(groups)}
            )
        return super().validate(attrs)

    def save(self, **kwargs):
        meta_if_any = self.validated_data.get('meta_if_any', False)
        logger.debug('Saving user {0} / {1}'.format(self.authenticator, self.validated_data.get('name')))
        logger.debug('Meta any {}'.format(meta_if_any))

        if self.validated_data.get('password') is not None:
            self.validated_data['password'] = cryptoManager().hash(self.validated_data['password'])

        if self.instance is not None and self.instance.state != self.validated_data['state']:
            login_attempts = LoginAttemptsManager(self.authenticator, self.instance.name)
            if self.validated_data['state'] == State.USER_TEMPORARY_BLOCKED:
                login_attempts.block_user()
            else:
                login_attempts.clear()

        return super().save(**kwargs)

    def create(self, validated_data: dict):
        return self.__validate_and_save(validated_data)

    def update(self, instance: User, validated_data: dict):
        return self.__validate_and_save(validated_data, instance)

    def __validate_and_save(self, validated_data, user=None):
        try:
            permissions = (validated_data.pop('api_permissions')
                           if validated_data.get('api_permissions') is not None
                           else [])
            roles = (validated_data.pop('roles')
                     if validated_data.get('roles') is not None
                     else None)
            groups = validated_data.pop('group_ids')

            try:
                auth = self.authenticator.getInstance()
                if user is None:
                    auth.createUser(validated_data)
                    checkRegistrationPossible()
                else:
                    auth.modifyUser(validated_data)
            except AuthenticatorException:
                raise APIException(
                    'Этот тип домена аутентификации не поддерживает создание пользователей.')
            except LicenseMaxRegisteredUsersReachedError:
                LicenseEventsHelper().users_limit_reached()
                raise APIException(
                    _('Достигнуто максимальное количество зарегистрированных пользователей'))

            if user is None:
                user = self.authenticator.users.create(**validated_data)
                if user.state == State.USER_TEMPORARY_BLOCKED:
                    login_attempts = LoginAttemptsManager(self.authenticator, self.instance.name)
                    login_attempts.block_user()
            else:
                user.__dict__.update(validated_data)

            logger.debug('User parent: {}'.format(user.parent))

            if auth.isExternalSource is False and (user.parent is None or user.parent == ''):
                logger.debug('Groups: {}'.format(groups))
                logger.debug('Got Groups {}'.format(self.authenticator.groups.filter(uuid__in=groups)))
                user.get_auth_groups().set(self.authenticator.groups.filter(uuid__in=groups))

            user.save()
            user.api_permissions.set(permissions)
            if roles is None:
                # Если список ролей не передан в запросе, то попытаемся восстановить роли из БД
                roles = [role.id for role in user.roles.all()]

            user.roles.set(roles)

        except IntegrityError:  # Duplicate key probably
            raise APIException(
                _('User already exists (duplicate key error)')
            )
        except ValidationError as e:
            raise APIException(e.message)
        except Exception as err:
            logger.exception('Saving group')
            raise APIException(str(err))

        return user

    def to_representation(self, instance):
        ret = super(AuthUserSerializer, self).to_representation(instance)
        ret['api_permissions'] = [str(permission) for permission in ret['api_permissions']]
        ret['roles'] = [str(role) for role in ret['roles']]

        return ret
