# -*- encoding: utf-8 -*-

import logging

from rest_framework import serializers
from rest_framework.exceptions import APIException
from rest_framework.fields import empty
from uds.core import auths
from uds.core.util import permissions
from uds.models import Tag
from uds.models.Authenticator import Authenticator

logger = logging.getLogger(__name__)


class AuthenticatorSerializer(serializers.ModelSerializer):
    id = serializers.CharField(source='uuid', required=False, read_only=True)
    name = serializers.CharField(
        required=True,
        label='Название',
        help_text='Название домена',
    )
    comments = serializers.CharField(required=False, allow_blank=True)
    type = serializers.CharField(source="data_type", required=False)
    data_type = serializers.CharField(write_only=True, required=False)
    type_name = serializers.CharField(source="getInstance.typeName", read_only=True)

    permission = serializers.SerializerMethodField()
    tags = serializers.SerializerMethodField()
    users_count = serializers.IntegerField(source="users.count", read_only=True)

    # Обязательные поля для сохранения в отдельные аттрибуты
    COMMON_FIELDS = ['name', 'comments', 'priority', 'small_name', 'data_type']

    class Meta:
        model = Authenticator
        fields = ['id', 'name', 'comments', 'priority', 'small_name', 'type', 'data_type', 'type_name',
                  'permission', 'tags', 'users_count']

    def __init__(self, instance=None, data=empty, **kwargs):
        super().__init__(instance=instance, data=data, **kwargs)

        # проверка на попытку обновления
        # если обновление - то нужно заполнить данными из instance
        if instance is not None and data is not empty:
            self.initial_data = self.get_full_data(self.initial_data)

    def get_permission(self, instance):
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)

    def get_full_data(self, data: dict):
        """ Получить список всех значений полей. """
        for name, value in self.instance.getInstance().valuesDict().items():
            if data.get(name) is None:
                data[name] = value
        return data

    def get_tags(self, instance):
        if not isinstance(instance, Authenticator):
            return []
        return [tag.tag for tag in instance.tags.all()]

    def validate(self, attrs):
        logger.debug('Before {0}'.format(attrs))

        data_type = attrs.get('data_type', self.instance.data_type if self.instance else '')
        type = auths.factory().lookup(data_type)
        if type is None:
            raise serializers.ValidationError(
                {'data_type': 'This is not exists «{}» type'.format(data_type)})

        return super().validate(attrs)

    def create(self, validated_data):
        osmanager = Authenticator.objects.create(
            **{name: validated_data.get(name) for name in self.COMMON_FIELDS}
        )
        return self.validate_object_and_save(osmanager)

    def update(self, instance: Authenticator, validated_data: dict):
        for name in self.COMMON_FIELDS:
            setattr(instance, name, validated_data.get(name, getattr(instance, name)))

        return self.validate_object_and_save(instance)

    def validate_object_and_save(self, auth: Authenticator):
        try:
            self.check_set_required_fields(auth)

            auth.data = auth.getInstance(self.initial_data).serialize()
            auth.save()
        except KeyError as key_error:
            raise APIException(
                'The field {} required for «{}» type'.format(key_error, auth.data_type))
        except auths.Authenticator.ValidationException as validate_error:
            raise APIException(validate_error)

        # Now if tags, update them
        tags = self.initial_data.get('tags')
        if tags is not None:
            logger.debug('Updating tags: {}'.format(tags))
            auth.tags.set([Tag.objects.get_or_create(tag=val)[0] for val in tags])

        return auth

    def check_set_required_fields(self, auth: Authenticator):
        """ Проверка наличия всех обязательных полей в данных, которые были переданы. """
        try:
            auth_type = auths.factory().lookup(auth.data_type)
            # инициализации сущности, при недостатке данных выпадет ошибка
            auth_obj = auth_type(None, None, self.initial_data)
        except KeyError as key_error:
            self.__show_error_required_field(key_error, auth.data_type)

        # Дополнительная проверка обязательных полей нужна для случая перехода
        # с обращения по ключу на обращение через метод get().
        # При установке свойств объекта заданного типа, мы все-равно должны быть увенрены,
        # что все обязательные поля заданы.
        for key in auth_obj.get_required_fields():
            if key not in self.initial_data:
                self.__show_error_required_field(key, auth.data_type)

    def __show_error_required_field(self, key, data_type):
        raise serializers.ValidationError(
            {key: 'The field {} required for «{}» type'.format(key, data_type)}
        )
