# -*- encoding: utf-8 -*-

import logging

from django.utils.translation import ugettext as _
from rest_framework import serializers
from rest_framework.validators import UniqueValidator
from uds.core.util import net, permissions
from uds.models.Network import Network

logger = logging.getLogger(__name__)

NET_STRING_TOOLTIP = "Диапазон адресов. Допустимы различные форматы, например: A.B.C.*, A.B.C.D/N, A.B.C.D - E.F.G.D"

class NetworkSerializer(serializers.ModelSerializer):
    id = serializers.CharField(source='uuid', required=False)
    name = serializers.CharField(
        required=True,
        validators=[
            UniqueValidator(queryset=Network.objects.all(), message=_('Element already exists (duplicate key error)'))],
        label='Название',
        help_text='Название сети',
    )
    permission = serializers.SerializerMethodField()
    tags = serializers.SerializerMethodField()
    networks_count = serializers.SerializerMethodField()

    class Meta:
        model = Network
        fields = ['id', 'name', 'net_string', 'permission', 'tags', 'networks_count']
        extra_kwargs = {
            'net_string': {
                'label': 'Диапазон',
                'help_text': NET_STRING_TOOLTIP,
            },
        }

    def get_permission(self, instance):
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)

    def get_tags(self, instance):
        if not isinstance(instance, Network):
            return []
        return [tag.tag for tag in instance.tags.all()]

    def get_networks_count(self, instance):
        if not isinstance(instance, Network):
            return 0
        return instance.transports.count()

    def validate(self, attrs):
        logger.debug('Before {0}'.format(attrs))
        try:
            nr = net.networksFromString(attrs['net_string'], False)
            attrs['net_start'] = nr[0]
            attrs['net_end'] = nr[1]
        except Exception as e:
            raise serializers.ValidationError({'net_string': ['Invalid network: ' + str(e)]})

        return super().validate(attrs)

    def create(self, validated_data):
        network_name = validated_data.get('name')
        network_net_string = validated_data.get('net_string')
        return Network.create(network_name, network_net_string)

    def update(self, instance, validated_data):
        network_name = validated_data.get('name')
        network_net_string = validated_data.get('net_string')
        instance.update(network_name, network_net_string)
        return instance
