import logging
import re

from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _
from termidesk.models.UserPermission import DEPLOYED_SERVICE_CHANGE, DEPLOYED_SERVICE_DELETE
from termidesk.core.auths.utils import check_exists_permission_for_user
from rest_framework import serializers
from uds.core.managers import userServiceManager
from uds.core.ui.images import DEFAULT_THUMB_BASE64
from uds.core.util import permissions
from uds.core.util.model import processUuid
from uds.core.util.State import State
from uds.models import Image, OSManager, Service, ServicesPoolGroup
from uds.models.ServicesPool import ServicePool
from uds.REST.methods.services import Services

logger = logging.getLogger(__name__)

DEFAULT_VALUE_FOR_OPTIONAL_FIELDS = "-1"


class ServicePoolSerializer(serializers.ModelSerializer):
    """ Сериализатор фонда рабочих мест. """
    id = serializers.CharField(source='uuid', required=False, read_only=True)
    parent = serializers.SerializerMethodField()
    parent_type = serializers.SerializerMethodField()
    restrained = serializers.BooleanField(source='isRestrained', read_only=True)
    service_id = serializers.CharField(source='service_uuid')
    provider_id = serializers.SerializerMethodField()
    osmanager_id = serializers.CharField(source='osmanager_uuid')
    image_id = serializers.CharField(source='image_uuid', allow_null=True)
    servicesPoolGroup_id = serializers.CharField(source='servicesPoolGroup_uuid', allow_null=True)
    pool_group_name = serializers.SerializerMethodField()
    pool_group_thumb = serializers.SerializerMethodField()
    comments = serializers.CharField(allow_blank=True)
    help_url = serializers.CharField(allow_blank=True)
    thumb = serializers.SerializerMethodField()
    state = serializers.SerializerMethodField()
    info = serializers.SerializerMethodField()
    user_services_count = serializers.SerializerMethodField()
    user_services_in_preparation = serializers.SerializerMethodField()
    permission = serializers.SerializerMethodField()
    tags = serializers.SerializerMethodField()

    editable = serializers.SerializerMethodField()
    removable = serializers.SerializerMethodField()

    # Хранение состояния необходимости публикации после сохранения
    publish_on_save = False

    class Meta:
        model = ServicePool
        fields = ['id', 'name', 'tags', 'parent', 'parent_type', 'comments', 'state', 'thumb',
                  'service_id', 'provider_id', 'osmanager_id', 'image_id', 'help_url',
                  'servicesPoolGroup_id', 'pool_group_name', 'pool_group_thumb',
                  'cache_l1_srvs', 'cache_l2_srvs', 'max_srvs', 'user_services_count',
                  'user_services_in_preparation', 'restrained', 'show_transports',
                  'debug_mode', 'backup', 'fallbackAccess', 'permission', 'info',
                  'editable', 'removable']

    def get_thumb(self, instance):
        if not isinstance(instance, ServicePool):
            return ''
        return instance.image.thumb64 if instance.image is not None else DEFAULT_THUMB_BASE64

    def get_user_services_count(self, instance):
        return (instance.userServices.filter(state=State.USABLE).count()
                if isinstance(instance, ServicePool) else 0)

    def get_user_services_in_preparation(self, instance):
        return (instance.userServices.filter(state=State.PREPARING).count()
                if isinstance(instance, ServicePool) else 0)

    def get_info(self, instance):
        if not isinstance(instance, ServicePool) or instance.service is None:
            return ''
        return Services.serviceInfo(instance.service)

    def get_tags(self, instance):
        if not isinstance(instance, ServicePool):
            return []
        return [tag.tag for tag in instance.tags.all()]

    def get_permission(self, instance):
        if not isinstance(instance, ServicePool):
            return ''
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)

    def get_pool_group_name(self, instance):
        if not isinstance(instance, ServicePool) or instance.servicesPoolGroup is None:
            return _('Основная')
        return instance.servicesPoolGroup.name

    def get_pool_group_thumb(self, instance):
        if (not isinstance(instance, ServicePool)
                or instance.servicesPoolGroup is None
                or instance.servicesPoolGroup.image is None):
            return None

        return instance.servicesPoolGroup.image.thumb64

    def get_parent(self, instance):
        if not isinstance(instance, ServicePool) or instance.service is None:
            return None
        return instance.service.name

    def get_parent_type(self, instance):
        if not isinstance(instance, ServicePool) or instance.service is None:
            return None
        return instance.service.data_type

    def get_provider_id(self, instance):
        if (not isinstance(instance, ServicePool)
                or instance.service is None
                or instance.service.provider is None):
            return None
        return instance.service.provider.uuid

    def get_state(self, instance):
        if not isinstance(instance, ServicePool) or instance.service is None:
            return ''

        state = instance.state
        if instance.isInMaintenance():
            state = State.MAINTENANCE
        elif userServiceManager().canInitiateServiceFromDeployedService(instance) is False:
            state = State.SLOWED_DOWN
        return state

    def get_editable(self, instance):
        if not isinstance(instance, ServicePool):
            return ''

        user = self.context.get('request').user
        return check_exists_permission_for_user(user, ServicePool, instance.uuid, DEPLOYED_SERVICE_CHANGE)

    def get_removable(self, instance):
        if not isinstance(instance, ServicePool):
            return ''

        user = self.context.get('request').user
        return check_exists_permission_for_user(user, ServicePool, instance.uuid, DEPLOYED_SERVICE_DELETE)

    def save(self, **kwargs):
        self.validate_before_save(self.validated_data)

        super().save(**kwargs)

        if self.publish_on_save is True:
            self.instance.publish()

        return self.instance

    def validate_before_save(self, attrs):
        logger.debug('Before {0}'.format(attrs))

        if attrs['service_uuid'] == '-1':
            raise serializers.ValidationError(
                {'service_id': ugettext('Необходимо выбрать шаблон рабочего места')}
            )

        # Название фонда обычно используется в формировании имени шаблона в поставщиках ресурсов,
        # поэтому на него накладываются ограничения.
        if re.search("(?u)[^ a-zA-Z0-9_-]", attrs['name']):
            raise serializers.ValidationError(
                {'name': ('Название фонда может содержать только латинские буквы, '
                          'цифры, пробел, тире ( - ) и подчеркивание ( _ )')}
            )

        try:
            service = Service.objects.get(uuid=processUuid(attrs['service_uuid']))
            attrs['service_uuid'] = service.id
        except:
            raise serializers.ValidationError({'service_id': ugettext('Указанный шаблон не найден')})

        try:
            serviceType = service.getType()

            if serviceType.publicationType is None:
                self.publish_on_save = False

            if serviceType.needsManager is True:
                osmanager = OSManager.objects.get(uuid=processUuid(attrs['osmanager_uuid']))
                attrs['osmanager_uuid'] = osmanager.id
            else:
                del attrs['osmanager_uuid']

            if serviceType.usesCache is False:
                for key in ('cache_l1_srvs', 'cache_l2_srvs', 'max_srvs'):
                    attrs[key] = 0

        except Exception:
            raise serializers.ValidationError(
                {'osmanager_id': ugettext('Необходимо выбрать параметры гостевой ОС')}
            )

        # If max < cache_1 or cache_l2
        attrs['max_srvs'] = max((int(attrs['cache_l1_srvs']), int(attrs['max_srvs'])))

        # TSRV-418: текущая версия Диспетчера подключений
        # не позволяет назначать ВРМ пользователю напрямую из кеша L2.
        # Чтобы использовать кеш L2, необходимо иметь хотя бы 1 ВРМ в кеше L1
        if int(attrs['cache_l1_srvs']) == 0 and int(attrs['cache_l2_srvs']) > 0:
            raise serializers.ValidationError({
                'cache_l1_srvs': ugettext('Для использования кеша L2, необходимо хотя бы 1 ВРМ в кеше L1')
            })

        imgId = attrs['image_uuid']
        attrs['image_uuid'] = None
        logger.debug('Image id: {}'.format(imgId))
        try:
            if imgId != DEFAULT_VALUE_FOR_OPTIONAL_FIELDS:
                image = Image.objects.get(uuid=processUuid(imgId))
                attrs['image_uuid'] = image.id
        except Exception:
            logger.exception('At image recovering')
            raise serializers.ValidationError(
                {'image_id': ("Необходимо указать id уществующего изображения или указать '{}', если нужно "
                              "создать без изображения").format(DEFAULT_VALUE_FOR_OPTIONAL_FIELDS)}
            )

        # Servicepool Group
        spgrpId = attrs['servicesPoolGroup_uuid']
        attrs['servicesPoolGroup_uuid'] = None
        logger.debug('servicesPoolGroup_uuid: {}'.format(spgrpId))
        try:
            if spgrpId != DEFAULT_VALUE_FOR_OPTIONAL_FIELDS:
                service_pool_group = ServicesPoolGroup.objects.get(uuid=processUuid(spgrpId))
                attrs['servicesPoolGroup_uuid'] = service_pool_group.id
        except Exception:
            logger.exception('At service pool group recovering')
            raise serializers.ValidationError(
                {'servicesPoolGroup_id': ("Необходимо указать id уществующей группы или указать '{}', "
                                          "если нужно создать без группы".format(
                    DEFAULT_VALUE_FOR_OPTIONAL_FIELDS))
                }
            )

        return attrs
