# -*- encoding: utf-8 -*-

import logging

from django.utils.translation import ugettext as _
from rest_framework import serializers
from rest_framework.validators import UniqueValidator, ValidationError
from uds.core.util import permissions
from uds.models.Transport import Transport

logger = logging.getLogger(__name__)


class TransportSerializer(serializers.ModelSerializer):
    """Адаптер для работы с таблицей протоколов доставки"""

    id = serializers.CharField(source='uuid', required=False)
    name = serializers.CharField(
        required=True,
        validators=[
            UniqueValidator(queryset=Transport.objects.all(),
                            message=_('Element already exists (duplicate key error)'))],
        label='Название',
        help_text='Название протокола доставки',
    )
    permission = serializers.SerializerMethodField()
    tags = serializers.SerializerMethodField()
    deployed_count = serializers.SerializerMethodField()

    priority = serializers.CharField()
    comments = serializers.CharField(allow_blank=True)

    allowed_oss = serializers.SerializerMethodField()
    protocol = serializers.SerializerMethodField()
    networks = serializers.SerializerMethodField()
    type = serializers.SerializerMethodField()
    type_name = serializers.SerializerMethodField()

    class Meta:
        model = Transport
        fields = ['id', 'deployed_count', 'name', 'allowed_oss', 'protocol', 'nets_positive',
                  'type_name', 'networks', 'priority', 'tags', 'type', 'permission', 'comments']

    def validate(self, attrs):
        """
        Предварительная проверка переданных значений

        В метод передается словарь attr с простыми полями класса.
        self.initial_data содержит данные, переданные при инициализации класса через аттрибут `data`.

        :param attrs: - словарь, представляющий простые поля класса (CharField, IntegerField, etc.)
        """
        logger.debug('Before {0}'.format(attrs))

        attrs['data_type'] = self.initial_data['data_type']

        # При обновлении (update) протокола доставки приходит список, но ожидается строка
        if isinstance(self.initial_data['allowed_oss'], list):
            attrs['allowed_oss'] = ','.join(self.initial_data['allowed_oss'])
        else:
            attrs['allowed_oss'] = self.initial_data['allowed_oss']

        attrs['networks'] = self.initial_data['networks']
        attrs['tags'] = self.initial_data['tags']

        custom_fields_list = self.get_custom_fields(attrs['data_type'])
        for custom_field_name in custom_fields_list:
            try:
                attrs[custom_field_name] = self.validate_custom_field(custom_field_name)
            except ValidationError as validation_exception:
                raise validation_exception
            except KeyError:
                continue

        return super().validate(attrs)

    def get_permission(self, instance):
        """Получение разрешений для пользователя из запроса"""
        request = self.context.get('request')
        return permissions.getEffectivePermission(request.user, instance)

    def get_tags(self, instance):
        """Получение меток (тэгов) для протокола доставки"""
        if not isinstance(instance, Transport):
            return []
        return [tag.tag for tag in instance.tags.all()]

    def get_deployed_count(self, instance):
        """Получение количества сервисов, использующих данный протокол доставки"""
        if not isinstance(instance, Transport):
            return 0
        return instance.deployedServices.count()

    def get_allowed_oss(self, instance):
        """Получение списка разрешенных операционных систем"""
        if not isinstance(instance, Transport):
            return []
        result = [{'id': x} for x in instance.allowed_oss.split(',')] if instance.allowed_oss != '' else []
        return result

    def get_protocol(self, instance):
        """Получение названия протокола"""
        if not isinstance(instance, Transport):
            return ''
        return instance.getType().protocol

    def get_networks(self, instance):
        """Получение перечня сетей, в которых может работать данный протокол"""
        if not isinstance(instance, Transport):
            return []
        return [{'id': n.id} for n in instance.networks.all()]

    def get_type(self, instance):
        """Получение типа протокола доставки"""
        if not isinstance(instance, Transport):
            return ''
        return instance.getType().type()

    def get_type_name(self, instance):
        """Получение имени типа протокола доставки"""
        if not isinstance(instance, Transport):
            return ''
        return instance.getType().name()

    def create(self, validated_data):
        """
        Создание в БД записи о новом протоколе доставки

        :param validated_data: словарь с проверенными (validated) данными
        :return: объект-модель протокола доставки
        """
        params = {
            'name': validated_data['name'],
            'data_type': validated_data['data_type'],
            'comments': validated_data['comments'],
            'nets_positive': validated_data['nets_positive'],
            'priority': validated_data['priority'],
            'allowed_oss': validated_data['allowed_oss'],
        }
        transport = self.Meta.model.objects.create(**params)

        # Сети и метки имеют связь многие-ко-многим, поэтому добавляются иначе
        networks = validated_data['networks']
        tags = validated_data['tags']

        from uds.models import Network, Tag
        transport.networks.set(Network.objects.filter(id__in=networks))
        transport.tags.set([Tag.objects.get_or_create(tag=val)[0] for val in tags])

        return transport

    def update(self, instance, validated_data):
        """
        Обновление сведений в БД о протоколе доставке

        :param instance: объект-запись из БД о протоколе доставки
        :param validated_data: словарь с проверенными (validated) данными о протоколе
        :return: объект-модель протокола доставки
        """
        instance.name = validated_data['name']
        instance.data_type = validated_data['data_type']
        instance.comments = validated_data['comments']
        instance.nets_positive = validated_data['nets_positive']
        instance.priority = validated_data['priority']
        instance.allowed_oss = validated_data['allowed_oss']

        # Сети и метки имеют связь многие-ко-многим, поэтому добавляются иначе
        networks = validated_data['networks']
        tags = validated_data['tags']

        from uds.models import Network, Tag
        instance.networks.set(Network.objects.filter(id__in=networks))
        instance.tags.set([Tag.objects.get_or_create(tag=val)[0] for val in tags])

        instance.save()

        # В БД хранятся дополнительные данные для инициализации объекта протокола доставки
        instance.data = self.get_serialized_data()

        instance.save()
        return instance

    def get_additional_data(self, uuid):
        """
        Добавление дополнительных данных для отправки в качестве ответа

        :param uuid: uuid протокола доставки
        :return: словарь, дополненный данными для передачи в качестве ответа на запрос
        """
        # Некоторые запросы возвращают дополнительные параметры, например,
        # "Время ожидания соединения", которые можно получить из объекта протокола доставки.
        additional_data = {}
        transport_instance = self.get_transport_instance(uuid)
        custom_fields_list = self.get_custom_fields(transport_instance.type())

        transport_instance_fields = transport_instance.valuesDict()

        for custom_field_name in custom_fields_list:
            custom_field_value = transport_instance_fields[custom_field_name]
            # Translate "true" & "false" to True & False (booleans)
            validated_value = {"true": True, "false": False}.get(custom_field_value, custom_field_value)
            additional_data[custom_field_name] = validated_value

        return additional_data

    def get_transport_instance(self, uuid, params=None):
        """
        Получение объекта протокола доставки

        :param uuid: - строка с uuid протокола доставки
        :param params: словарь с параметрами для инициализации объекта
        :return: объект протокола доставки
        """
        transport = self.Meta.model.objects.get(uuid=uuid)
        transport_instance = transport.getInstance(params)

        return transport_instance

    def get_custom_fields(self, transport_type=None):
        """
        Получение списка полей, уникальных для данного типа протокола доставки
        :param transport_type: строка с типом протокола доставки
        :return: список имен уникальных полей для данного типа протокола доставки.
        """
        custom_fields_list = []
        if transport_type:
            from uds.core.transports import factory
            transport_instance = factory().providers().get(transport_type.lower())
            custom_fields_list = [x['name'] for x in transport_instance.guiDescription()]

        return custom_fields_list

    def get_serialized_data(self):
        """
        Вспомогательный метод для получения значений полей, уникальных для данного протокола доставки,
        в формате пригодном к записи в БД.

        :return: строка с данными, подготовленными к записи в БД
        """

        # У протоколов доставки могут быть определены индивидуальные поля.
        # Их необходимо добавить.

        custom_fields_list = self.get_custom_fields(self.validated_data['data_type'])

        _params = {}
        for custom_field_name in custom_fields_list:
            # В TSRV-346 обнаружилась особенность - если не задать пароль (оставить поле пустым),
            # то данное поле не попадет в запрос и будет отсутствовать в self.validated_data.
            # Возможно, такое поведение зависит от браузера.
            # Здесь есть косвенное подтверждение https://www.w3.org/TR/html401/interact/forms.html#h-17.13.2
            #   `If a control doesn't have a current value when the form is submitted,`
            #   `user agents are not required to treat it as a successful control.`
            # Таким образом, все отсутствующие поля считаются пустой строкой
            _params[custom_field_name] = self.validated_data.get(custom_field_name, "")

        transport_instance = self.get_transport_instance(self.data.get('id'), _params)

        serialized_data = transport_instance.serialize()

        return serialized_data

    def validate_custom_field(self, field_name):
        """Проверка полей, специфичных для протокола"""

        from uds.core.transports import factory
        transport_gui_description = factory().lookup(self.initial_data['data_type']).guiDescription()
        target_field = next((x for x in transport_gui_description if x["name"] == field_name), {})
        is_required_field = target_field.get('gui', {}).get('required', False)
        target_field_value = self.initial_data.get(field_name)

        if is_required_field and target_field_value == '':
            raise ValidationError(
                "The field {required_field} is empty.".format(required_field=field_name))

        elif is_required_field and target_field_value is None:
            raise ValidationError("The field {required_field} not found."
                                  .format(required_field=field_name))

        else:
            return self.initial_data[field_name]
