# -*- encoding: utf-8 -*-
import logging

from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from uds.models import UserService
from uds.models.ServicesPool import ServicePool

logger = logging.getLogger(__name__)


class UserServiceSerializer(serializers.ModelSerializer):
    """ Сериализатор рабочего места. """
    id = serializers.CharField(source='uuid', required=False, read_only=True)
    state_date = serializers.SerializerMethodField()
    creation_date = serializers.SerializerMethodField()
    id_deployed_service = serializers.SerializerMethodField()
    revision = serializers.SerializerMethodField()
    ip = serializers.SerializerMethodField()
    actor_version = serializers.SerializerMethodField()

    class Meta:
        model = UserService
        fields = [
            'id', 'friendly_name', 'unique_id', 'state', 'os_state', 'state_date',
            'revision', 'ip', 'creation_date', 'id_deployed_service', 'actor_version'
        ]

    def get_state_date(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return int(instance.state_date.timestamp())

    def get_creation_date(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return int(instance.creation_date.timestamp())

    def get_id_deployed_service(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return instance.deployed_service.uuid

    def get_revision(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return instance.publication and instance.publication.revision or ''

    def get_ip(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        props = instance.getProperties()
        return props.get('ip', _('unknown'))

    def get_actor_version(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        props = instance.getProperties()
        return props.get('actor_version', _('unknown'))


class CachedUserServiceSerializer(UserServiceSerializer):
    """ Сериализатор рабочего места. """

    class Meta:
        model = UserService
        fields = [
            'id', 'friendly_name', 'unique_id', 'state', 'os_state', 'state_date',
            'revision', 'ip', 'creation_date', 'id_deployed_service', 'actor_version',
            'cache_level'
        ]


class AssignedUserServiceSerializer(UserServiceSerializer):
    """ Сериализатор рабочего места. """
    source_host = serializers.CharField(source='src_hostname', required=False, read_only=True)
    source_ip = serializers.CharField(source='src_ip', required=False, read_only=True)
    owner = serializers.SerializerMethodField()
    owner_info = serializers.SerializerMethodField()

    class Meta:
        model = UserService
        fields = [
            'id', 'friendly_name', 'unique_id', 'state', 'os_state', 'state_date',
            'revision', 'ip', 'creation_date', 'id_deployed_service', 'actor_version',
            'owner', 'owner_info', 'in_use', 'in_use_date', 'source_host', 'source_ip',
        ]

    def get_owner(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return instance.user.manager.name + "-" + instance.user.name

    def get_owner_info(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        _owner_info = {
           'auth_id': instance.user.manager.uuid,
           'user_id': instance.user.uuid
        }
        return _owner_info


class SessionUserServiceSerializer(AssignedUserServiceSerializer):
    """ Сериализатор активных сессий пользователей. """
    workplace_name = serializers.SerializerMethodField()
    workplace_uuid = serializers.SerializerMethodField()
    in_use_date = serializers.SerializerMethodField()

    class Meta:
        model = UserService
        fields = [
            'id', 'friendly_name', 'unique_id', 'ip', 'in_use_date',
            'owner', 'source_host', 'source_ip', 'workplace_name', 'workplace_uuid'
        ]

    def get_workplace_name(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return instance.deployed_service.name

    def get_workplace_uuid(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return instance.deployed_service.uuid

    def get_in_use_date(self, instance):
        if not isinstance(instance, self.Meta.model):
            return None
        return instance.in_use_date.isoformat()
