# -*- encoding: utf-8 -*-

import logging

from django.contrib.auth.models import Group
from django.shortcuts import get_object_or_404
from guardian.models.models import GroupObjectPermission
from rest_framework.response import Response
from rest_framework.status import HTTP_201_CREATED, HTTP_400_BAD_REQUEST
from termidesk.api.decorators import has_role_permission, enabled_rbac_required
from termidesk.api.webui.serializers.acl.entity import EntitySerializer
from termidesk.api.webui.serializers.acl.workplace import WorkplaceEntitySerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.api.webui.serializers.type import TypeSerializer
from termidesk.models.UserPermission import USER_ROLES_MANAGE
from uds.core.ui.UserInterface import gui
from uds.REST.model import OK

logger = logging.getLogger(__name__)


class AccessControlList(CommonViewSet):
    """
    gui:
        Спецификация полей GUI для пользовательских разрешений к конкретным объектам.
    """
    model = GroupObjectPermission
    serializer_class = EntitySerializer

    ENTITIES = {
        'workplace': WorkplaceEntitySerializer,
    }

    def get_queryset(self):
        return GroupObjectPermission.objects.select_related('group', 'permission', 'content_type').all()

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def retrieve(self, request, *args, **kwargs):
        """ Получить информацию об объектном разрешении. """
        permission = get_object_or_404(GroupObjectPermission, id=kwargs.get('id'))
        serializer = self.get_serializer(permission, context={'request': request})

        return Response(serializer.data)

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def get_types(self, request, *args, **kwargs):
        """ Получить типы объектов, доступных для ассоциации. """
        data = []
        for entity, obj in self.ENTITIES.items():
            data.append({'name': obj.title,
                         'type': entity,
                         'description': obj.description,
                         'icon': ''})

        serializer = TypeSerializer(data=data, many=True)

        return Response(sorted(serializer.initial_data, key=lambda f: f['name']))

    def get_gui_params(self, **kwargs):
        type_ = kwargs.get('data_type')

        class_serializer = self.get_instance_by_type(type_)
        objects = class_serializer.get_gui_objects()

        groups = (gui.choiceItem(group.id, group.name) for group in Group.objects.order_by('name').all())
        gui_represent = self.add_field(self.add_default_fields([], []),
                                       {'name': 'group_id',
                                        'label': 'Роль',
                                        'values': groups,
                                        'tooltip': '',
                                        'type': gui.ChoiceField.CHOICE_TYPE,
                                        'order': 101,
                                        'required': True, })

        permissions = class_serializer.get_gui_allow_permissions()
        self.add_field(gui_represent, {'name': 'permission_id',
                                       'label': 'Пользовательское разрешение',
                                       'values': permissions,
                                       'tooltip': '',
                                       'type': gui.ChoiceField.CHOICE_TYPE,
                                       'order': 102,
                                       'required': True, })

        self.add_field(gui_represent,
                       {'name': 'object_uuid',
                        'label': 'Объект',
                        'values': objects,
                        'tooltip': '',
                        'type': gui.ChoiceField.CHOICE_TYPE,
                        'order': 103,
                        'required': True, },
                       )

        return gui_represent

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def create(self, request, *args, **kwargs):
        """ Создание разрешения к конкретному объекту в БД. """
        try:
            serializer = self.__get_serializer(request)

            if not serializer.is_valid():
                for error in serializer.errors.values():
                    return Response(error[0], status=HTTP_400_BAD_REQUEST)

            if serializer.save() is None:
                return Response('Не удалось создать разрешение.', status=HTTP_400_BAD_REQUEST)

        except Exception as err:
            return Response(str(err), status=HTTP_400_BAD_REQUEST)

        headers = self.get_success_headers(serializer.data)
        return Response(OK, status=HTTP_201_CREATED, headers=headers)

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def update(self, request, *args, **kwargs):
        """ Обновить параметры объектного разрешения. """
        obj_permission = get_object_or_404(self.model, id=kwargs.get('id', ''))

        serializer = self.__get_serializer(request, obj_permission)

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        return Response(serializer.data)

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def overview(self, request, *args, **kwargs):
        """ Список всех доступных объектных разрешений. """
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)

        # добавление удобочитаемого типа объекта
        permissions = serializer.data

        types = self.__get_list_types_object_with_title()

        for permission in permissions:
            permission['type_title'] = types[permission['type_object']]['title']
            permission['type_object'] = types[permission['type_object']]['type']

        return Response(permissions)

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def delete(self, request, id: int):
        """ Удалить объектное разрешение. """
        group = get_object_or_404(self.model, id=id)

        try:
            group.delete()
            res = OK
        except Exception as error:
            return Response(error, status=HTTP_400_BAD_REQUEST)

        return Response(res)

    def get_instance_by_type(self, entity: str):
        class_serialise = self.ENTITIES.get(entity)

        if class_serialise is None:
            raise ValueError(
                'Переданный тип объекта "{}" не доступен для создания разрешения'.format(entity)
            )

        return class_serialise

    def __get_list_types_object_with_title(self):
        """
        Формирование списка объектов следующего вида:
        {'название класса' : {'title': 'Подпись для отображения в списке', 'type': 'Тип для получения gui'}, }
        """
        entities = {}
        for type_, serialize_class in self.ENTITIES.items():
            entities[serialize_class.model.__name__] = {
                'title': serialize_class.title, 'type': type_
            }

        return entities

    def __get_serializer(self, request, instance=None):
        class_serializer = self.get_instance_by_type(request.data.get('data_type'))
        if instance is not None:
            return class_serializer(data=request.data, instance=instance, context={'request': request})

        return class_serializer(data=request.data)
