"""Модуль с реализацией Представления (View) событий аудита из локальной БД"""
# -*- encoding: utf-8 -*-

import logging

from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.viewsets import GenericViewSet

from termidesk.api.permissions import AccessPermission
from termidesk.audit.events import EVENT_PREFIX
from termidesk.models import AuditEvent
from termidesk.models.UserPermission import LOG_VIEW
from ..pagination import ListResultsPagination
from ..serializers.audit_event import AuditEventsSerializer
from ...decorators import has_role_permission

logger = logging.getLogger(__name__)

START_DATE_FILTER = 'from_date'
END_DATE_FILTER = 'to_date'
EVENT_NAME_FILTER = 'event_name'
FILTER = 'filter'


class AuditEventsView(GenericViewSet):
    """Представление для получения списка событий аудита из локальной БД"""
    pagination_class = ListResultsPagination
    permission_classes = [AccessPermission]
    model = AuditEvent
    serializer_class = AuditEventsSerializer
    schema = AutoSchema()

    # Метод позволяет частично сгенерировать документацию на базе сериализатора
    def get_serializer_context(self):
        """Получение контекста сериализатора"""
        return {
            'request': self.request,
            'format': self.format_kwarg,
            'view': self
        }

    # Метод позволяет частично сгенерировать документацию на базе сериализатора
    def get_serializer(self, *args, **kwargs):
        """Получение сериализатора"""
        kwargs['context'] = self.get_serializer_context()
        return self.serializer_class(*args, **kwargs)

    def get_serializer_data(self, _queryset):
        """Получение данных, подготовленных для передачи в качестве ответа на запрос"""
        serializer = self.get_serializer(_queryset, many=True)
        _serializer_data = [
            {
                'timestamp': x.get("timestamp"),
                'event': x.get("event"),
                'description': x.get("description"),
            } for x in serializer.data]

        return _serializer_data

    @has_role_permission(LOG_VIEW)
    def get_audit_events(self, request):
        """
        Получить сведения о событиях аудита из локальной БД
        """
        queryset = self.get_queryset()

        page = self.paginate_queryset(queryset)

        if page is not None:
            serializer_data = self.get_serializer_data(page)
            response_handler = self.get_paginated_response
        else:
            serializer_data = self.get_serializer_data(queryset)
            response_handler = Response

        return response_handler(serializer_data)

    def get_queryset(self):
        request = self.request
        # Выбор всех записей с сортировкой записей от новых к старым.
        queryset = self.model.objects.all().order_by('-timestamp')

        start_date = request.query_params.get(START_DATE_FILTER)
        if start_date not in ['', None]:
            queryset = queryset.filter(timestamp__gte=start_date)

        end_date = request.query_params.get(END_DATE_FILTER)
        if end_date not in ['', None]:
            queryset = queryset.filter(timestamp__lte=end_date)

        event_name = request.query_params.get(EVENT_NAME_FILTER)
        if event_name not in ['', None]:
            queryset = queryset.filter(event=event_name)

        filter = request.query_params.get(FILTER)
        if filter not in ['', None]:
            queryset = queryset.filter(data__icontains=filter)

        return queryset

    def get_audit_events_types(self, request):
        """Получение типов событий аудита"""
        event_field_name = 'event'
        event_types = self.model.objects.values(event_field_name).distinct()
        types_dict = {}

        event_prefix = ''.join((EVENT_PREFIX, '.'))
        for event_name in event_types:
            full_name = event_name[event_field_name]
            short_name = full_name.replace(event_prefix, '')
            types_dict.update({full_name: short_name})
        return Response(types_dict)
