# -*- encoding: utf-8 -*-

import logging

from coreapi import Field
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext_lazy as _
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.status import HTTP_201_CREATED, HTTP_400_BAD_REQUEST, HTTP_403_FORBIDDEN

from termidesk.api.decorators import has_role_permission
from termidesk.api.webui.serializers.authenticator import AuthenticatorSerializer
from termidesk.api.webui.serializers.log import LogSerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.core.auths.Authenticator import BaseAuthenticator2Fa
from termidesk.core.util.utils import is_2fa_feature_enabled
from termidesk.models.UserPermission import AUTHENTICATOR_VIEW, AUTHENTICATOR_CHANGE, AUTHENTICATOR_CREATE, \
    AUTHENTICATOR_DELETE
from termidesk.rest.handlers import CREATE_ACTION, DELETE_ACTION, READ_ACTION, UPDATE_ACTION, set_crud_event
from uds.core import auths
from uds.core.auths import factory
from uds.core.Environment import Environment
from uds.core.util import log, permissions
from uds.models import Authenticator
from uds.REST.model import OK

from rest_framework import serializers

logger = logging.getLogger(__name__)


class Authenticators(CommonViewSet):
    """
    list:
        Список всех доступных доменов аутентификации.
    create:
        Создание домена аутентификации.
        Для просмотра доступных доменов аутентификации обратитесь по следующему пути: /api/webui/authenticators/types
    """
    model = Authenticator
    serializer_class = AuthenticatorSerializer

    def get_queryset(self):
        available_types = [auth.type() for auth in factory().providers().values()]
        queryset = Authenticator.objects.filter(data_type__in=available_types).all()
        
        filter = self.request.query_params.get('name')
        if filter not in ['', None]:
            query = Q(name__icontains=filter)
            query.add(Q(comments__icontains=filter), Q.OR)
            queryset = Authenticator.objects.filter(query).all()
        return queryset 

    @has_role_permission(AUTHENTICATOR_VIEW)
    def retrieve(self, request, *args, **kwargs):
        """
        Получить информацию о домене аутентификации.
        """
        uuid = kwargs.get('uuid', '').lower()
        domain = get_object_or_404(self.model, uuid=uuid)
        serializer = AuthenticatorSerializer(domain, context={'request': request})

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            READ_ACTION
        )
        response = self.fill_intance_fields(serializer.instance, serializer.get_full_data(serializer.data))
        return Response(response)

    def gui(self, request, *args, **kwargs):
        """
        Спецификация полей GUI определенного типа для домена аутентификации.
        """
        type_auth = kwargs.get('type')
        try:
            # Получение полей конкретного аутентификатора
            gui_fields = factory().lookup(type_auth).guiDescription()
        except AttributeError:
            raise APIException('The type "{}" does not exist'.format(type_auth))

        # Дополнение списка полей, полями по-умолчанию.
        gui_fields = self.add_default_fields(
            gui_fields,
            ['name', 'comments', 'tags', 'priority', 'small_name']
        )

        # Сортировка полей по атрибуту `order`
        return Response(sorted(gui_fields, key=lambda f: f['gui']['order']))

    @has_role_permission(AUTHENTICATOR_CHANGE)
    def update(self, request, *args, **kwargs):
        """
        Обновить параметры домена аутентификации.
        """
        uuid = kwargs.get('uuid', '').lower()
        auth = get_object_or_404(self.model, uuid=uuid)

        serializer = AuthenticatorSerializer(instance=auth, data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            UPDATE_ACTION
        )

        return Response(serializer.get_full_data(serializer.data))

    @has_role_permission(AUTHENTICATOR_CREATE)
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            CREATE_ACTION
        )
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.get_full_data(serializer.data), status=HTTP_201_CREATED, headers=headers)

    @has_role_permission(AUTHENTICATOR_VIEW)
    def overview(self, request):
        """
        Список доменов аутентификации.
        """
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = AuthenticatorSerializer(page, many=True, context={'request': request})
            return self.get_paginated_response(serializer.data)

        serializers = AuthenticatorSerializer(queryset, many=True, context={'request': request})
        return Response(serializers.data)

    @has_role_permission(AUTHENTICATOR_DELETE)
    def delete(self, request, uuid):
        """
        Удалить домен аутентификации.
        """
        uuid = uuid.lower()
        if request.user.manager.uuid == uuid:
            return Response('Удаление домена внутренним пользователем запрещено', status=HTTP_403_FORBIDDEN)

        domain = get_object_or_404(self.model, uuid=uuid)

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            domain.uuid,
            domain.getType().typeType,
            domain.name,
            DELETE_ACTION
        )

        try:
            domain.delete()
            res = OK
        except Exception as error:
            return Response(error, status=HTTP_400_BAD_REQUEST)

        return Response(res)

    @has_role_permission(AUTHENTICATOR_VIEW)
    def get_logs(self, request, *args, **kwargs):
        """ Получение журнала. """
        try:
            uuid = kwargs.get('uuid', '').lower()
            authenticator = get_object_or_404(Authenticator, uuid=uuid)
        except Exception as error:
            return Response(str(error), status=HTTP_400_BAD_REQUEST)

        filter = self.request.query_params.get('search')

        logs = log.getLogs(authenticator)

        if filter not in ['', None]:
            filtered = []
            for i in logs:
                if filter.lower() in i["message"].lower():
                    filtered.append(i)
            logs = filtered

        serializer = LogSerializer(logs, many=True)
        return Response(serializer.data)

    def test(self, request, *args, **kwargs):
        """ Тестирование подключения по переданным данным. """
        type_auth = kwargs.get('type')
        authType = factory().lookup(type_auth)
        params = request.data.copy()
        params['_request'] = request
        res = authType.test(Environment.getTempEnv(), params)

        return Response(self.success() if res[0] is True else res[1])

    def type_info(self, type_):
        return {
            'canSearchUsers': type_.searchUsers != auths.Authenticator.searchUsers,
            'canSearchGroups': type_.searchGroups != auths.Authenticator.searchGroups,
            'needsPassword': type_.needsPassword,
            'userNameLabel': _(type_.userNameLabel),
            'groupNameLabel': _(type_.groupNameLabel),
            'passwordLabel': _(type_.passwordLabel),
            'canCreateUsers': type_.createUser != auths.Authenticator.createUser,
            'isExternal': type_.isExternalSource,
        }

    @has_role_permission(AUTHENTICATOR_VIEW)
    def get_types(self, request):
        """ Получение списка доступных типов доменов аутентификации. """
        types = []
        for type_ in auths.factory().providers().values():
            # TSRV-362: Если системная настройка `EXPERIMENTAL_2FA_ENABLED` выключена,
            # то необходимо исключить из выдачи домены аутентификации 2FA
            if issubclass(type_, BaseAuthenticator2Fa) and not is_2fa_feature_enabled():
                continue
            types.append(self.type_as_dict(type_))
        return Response(sorted(types, key=lambda f: f['name']))


class SearchResultSerializer(serializers.Serializer):
    id = serializers.CharField(max_length=200)
    name = serializers.CharField(max_length=200)


class AuthenticatorSearch(CommonViewSet):
    schema = AutoSchema()
    serializer_class = SearchResultSerializer

    @has_role_permission(AUTHENTICATOR_VIEW)
    def list(self, request, *args, **kwargs):
        """Получение списка доступных групп или пользователей для домена аутентификации."""
        try:
            uuid = kwargs.get('uuid', '').lower()
            authenticator = get_object_or_404(Authenticator, uuid=uuid)
        except Exception as error:
            return Response(str(error), status=HTTP_400_BAD_REQUEST)

        try:
            type_ = request.GET.get('type')

            if type_ not in ('user', 'group'):
                return Response('Not supported type {}'.format(type_), status=HTTP_400_BAD_REQUEST)

            term = request.GET.get('term')

            auth = authenticator.getInstance()

            canDoSearch = type_ == 'user' and (auth.searchUsers != auths.Authenticator.searchUsers) or (
                    auth.searchGroups != auths.Authenticator.searchGroups)

            if canDoSearch is False:
                return Response('Operation not supported', status=HTTP_400_BAD_REQUEST)

            if type_ == 'user':
                return Response(auth.searchUsers(term))
            else:
                return Response(auth.searchGroups(term))

        except Exception as error:
            return Response(str(error), status=HTTP_400_BAD_REQUEST)
