import logging

from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST

from termidesk.api.decorators import has_role_permission
from termidesk.api.webui.serializers.image import ImageOverviewSerializer, ImageSerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.models.UserPermission import IMAGE_VIEW, IMAGE_CHANGE, IMAGE_DELETE, IMAGE_CREATE
from termidesk.rest.handlers import CREATE_ACTION, DELETE_ACTION, READ_ACTION, UPDATE_ACTION, set_crud_event
from uds.core.ui.UserInterface import gui
from uds.models.Image import Image
from uds.REST.model import OK

logger = logging.getLogger(__name__)


class Images(CommonViewSet):
    """
    list:
        Список всех доступных изображений.
    create:
        Создание изображения.
    gui:
        Спецификация полей GUI для изображения.
    """
    model = Image
    queryset = Image.objects.all()
    serializer_class = ImageSerializer

    @has_role_permission(IMAGE_VIEW)
    def retrieve(self, request, *args, **kwargs):
        """
        Получить информацию об изображении.
        """
        uuid = kwargs.get('uuid', '').lower()
        image = get_object_or_404(self.model, uuid=uuid)
        serializer = ImageSerializer(image, context={'request': request})

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            READ_ACTION
        )

        return Response(serializer.data)

    def get_gui_params(self, **kwargs):
        return self.add_field(self.add_default_fields([], ['name']), {
            'name': 'data',
            'value': '',
            'label': ugettext('Image'),
            'tooltip': ugettext('Image object'),
            'type': gui.InputField.IMAGECHOICE_TYPE,
            'order': 100,  # At end
        })

    def perform_create(self, serializer):
        serializer.save()

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            CREATE_ACTION
        )

    @has_role_permission(IMAGE_CHANGE)
    def update(self, request, *args, **kwargs):
        """
        Обновить параметры изображения.
        """
        uuid = kwargs.get('uuid', '').lower()
        image = get_object_or_404(self.model, uuid=uuid)

        serializer = ImageSerializer(instance=image, data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            UPDATE_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(IMAGE_DELETE)
    def delete(self, request, uuid):
        """
        Удалить изображение.
        """
        uuid = uuid.lower()
        image = get_object_or_404(self.model, uuid=uuid)

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            image.uuid,
            '',
            image.name,
            DELETE_ACTION
        )

        try:
            image.delete()
            res = OK
        except Exception as error:
            return Response(error, status=HTTP_400_BAD_REQUEST)

        return Response(res)

    @has_role_permission(IMAGE_VIEW)
    def overview(self, request, *args, **kwargs):
        """ Получение списка изображений с расширенным набором полей. """
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = ImageOverviewSerializer(page, many=True, context={'request': request})
            return self.get_paginated_response(serializer.data)

        serializer = ImageOverviewSerializer(queryset, many=True, context={'request': request})
        return Response(serializer.data)

    @has_role_permission(IMAGE_CREATE)
    def create(self, request, *args, **kwargs):
        """ Добавление изображения в БД. """
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().create(request, *args, **kwargs)

    @has_role_permission(IMAGE_VIEW)
    def list(self, request, *args, **kwargs):
        """Получение списка изображений из галереи"""
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)
