# -*- encoding: utf-8 -*-

from rest_framework.status import HTTP_400_BAD_REQUEST
from rest_framework.response import Response
from django.shortcuts import get_object_or_404

from termidesk.models.UserPermission import NETWORK_VIEW, NETWORK_CREATE, NETWORK_CHANGE, NETWORK_DELETE
from ..serializers.network import NetworkSerializer
from termidesk.rest.handlers import set_crud_event, CREATE_ACTION, READ_ACTION, UPDATE_ACTION, DELETE_ACTION
from uds.REST.model import OK
from uds.models import Network
from termidesk.api.webui.views.common_viewset import CommonViewSet

import logging

from ...decorators import has_role_permission

logger = logging.getLogger(__name__)


class Networks(CommonViewSet):
    model = Network
    serializer_class = NetworkSerializer

    @has_role_permission(NETWORK_VIEW)
    def get_network(self, request, uuid):
        """
        Получить информацию о сети
        """
        uuid = uuid.lower()
        network = get_object_or_404(self.model, uuid=uuid)
        serializer = NetworkSerializer(network, context={'request': request})

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            '',
            serializer.data.get('name', ''),
            READ_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(NETWORK_VIEW)
    def networks_overview(self, request):
        """
        Список сетей
        """
        serializers = NetworkSerializer(self.model.objects.all(), many=True, context={'request': request})

        return Response(serializers.data)

    def networks_gui(self, request):
        """
        Спецификация полей GUI для сетей
        """
        res = self.network_gui()

        return Response(res)

    @has_role_permission(NETWORK_CREATE)
    def create_network(self, request):
        """
        Создать сеть
        """
        serializer = NetworkSerializer(data=request.data, context={'request': request})
        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            '',
            serializer.data.get('name', ''),
            CREATE_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(NETWORK_CHANGE)
    def update_network(self, request, uuid):
        """
        Обновить сеть
        """
        uuid = uuid.lower()
        network = get_object_or_404(self.model, uuid=uuid)
        serializer = NetworkSerializer(network, data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            '',
            serializer.data.get('name', ''),
            UPDATE_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(NETWORK_DELETE)
    def delete_network(self, request, uuid):
        """
        Удалить сеть
        """
        uuid = uuid.lower()
        network = get_object_or_404(self.model, uuid=uuid)

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            network.uuid,
            '',
            network.name,
            DELETE_ACTION
        )

        try:
            network.delete()
            res = OK
        except:
            res = None

        return Response(res)
