import logging

from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST

from termidesk.api.decorators import has_role_permission
from termidesk.api.webui.serializers.osmanager import OSManagerOverviewSerializer, OSManagerSerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.core.util.utils import max_session_time_enabled
from termidesk.models.UserPermission import OSMANAGER_VIEW, OSMANAGER_CHANGE, OSMANAGER_DELETE, \
    OSMANAGER_CREATE
from termidesk.rest.handlers import CREATE_ACTION, DELETE_ACTION, READ_ACTION, UPDATE_ACTION, set_crud_event
from termidesk.rest.helpers import remove_field
from uds.core.osmanagers import factory
from uds.models.OSManager import OSManager
from uds.REST.model import OK

logger = logging.getLogger(__name__)


class OSManagers(CommonViewSet):
    """
    list:
        Список всех доступных параметров гостевых ОС.
    create:
        Создание параметров гостевой ОС.
        Для просмотра доступных типов гостевой ОС обратитесь по следующему пути: /api/webui/osmanagers/types
    gui:
        Спецификация полей GUI определенного типа для параметров гостевой ОС.
    """
    model = OSManager
    serializer_class = OSManagerSerializer

    def get_queryset(self):
        available_types = [osmanager.type() for osmanager in factory().providers().values()]
        return OSManager.objects.filter(data_type__in=available_types).all()

    @has_role_permission(OSMANAGER_VIEW)
    def retrieve(self, request, *args, **kwargs):
        """
        Получить информацию о параметрах гостевой ОС.
        """
        uuid = kwargs.get('uuid', '').lower()
        osmanager = get_object_or_404(self.model, uuid=uuid)
        serializer = OSManagerSerializer(osmanager, context={'request': request})

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            READ_ACTION
        )

        return Response(serializer.get_full_data(serializer.data))

    def get_gui_params(self, **kwargs):
        type_osmanager = kwargs.get('type')
        try:
            fields = factory().lookup(type_osmanager).guiDescription()
        except AttributeError:
            raise APIException('The type "{}" does not exist'.format(type_osmanager))

        if not max_session_time_enabled():
            # TSRV-447: Если экспериментальный параметр не включен,
            # то необходимо убрать его из интерфейса
            remove_field(fields, 'maxSessionTime')

        fields = self.add_default_fields(fields, ['name', 'comments', 'tags'])
        return fields

    def perform_create(self, serializer):
        serializer.save()

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            CREATE_ACTION
        )

    @has_role_permission(OSMANAGER_CHANGE)
    def update(self, request, *args, **kwargs):
        """
        Обновить параметры гостевых ОС.
        """
        uuid = kwargs.get('uuid', '').lower()
        osmanager = get_object_or_404(self.model, uuid=uuid)

        serializer = OSManagerSerializer(instance=osmanager, data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            UPDATE_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(OSMANAGER_DELETE)
    def delete(self, request, uuid):
        """
        Удалить параметры гостевых ОС.
        """
        uuid = uuid.lower()
        osmanager = get_object_or_404(self.model, uuid=uuid)

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            osmanager.uuid,
            osmanager.getType().typeType,
            osmanager.name,
            DELETE_ACTION
        )

        if osmanager.deployedServices.count() > 0:
            return Response(
                ugettext('Can\'t delete an OS Manager with services pools associated'),
                status=HTTP_400_BAD_REQUEST
            )

        try:
            osmanager.delete()
            res = OK
        except Exception as error:
            return Response(error, status=HTTP_400_BAD_REQUEST)

        return Response(res)

    @has_role_permission(OSMANAGER_VIEW)
    def get_types(self, request):
        """ Получение списка доступных типов для параметров гостевой ОС. """
        types = [self.type_as_dict(type_) for type_ in factory().providers().values()]
        return Response(sorted(types, key=lambda f: f['name']))

    @has_role_permission(OSMANAGER_VIEW)
    def overview(self, request, *args, **kwargs):
        """ Получение списка доступных параметров гостевых ОС (без дополнительных полей под каждый тип). """
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = OSManagerOverviewSerializer(page, many=True, context={'request': request})
            return self.get_paginated_response(serializer.data)

        serializer = OSManagerOverviewSerializer(queryset, many=True, context={'request': request})
        return Response(serializer.data)

    @has_role_permission(OSMANAGER_CREATE)
    def create(self, request, *args, **kwargs):
        """ Добавление параметра гостевых ОС в БД. """
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().create(request, *args, **kwargs)

    @has_role_permission(OSMANAGER_VIEW)
    def list(self, request, *args, **kwargs):
        """ Получение параметров гостевых ОС из БД. """
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)
