import logging

from rest_framework.response import Response
from termidesk.api.permissions import AccessPermission, EnabledPoliciesRequiredPermission
from termidesk.api.webui.serializers.policy import PolicySerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.models.PolicyGlobal import PolicyGlobal
from termidesk.core.policies.utils import (PolicyDTO, registry, get_policy_class_by_type,
                                           get_policy_values_by_type)

logger = logging.getLogger(__name__)


class Policies(CommonViewSet):
    """
    gui:
        Спецификация полей GUI для политики конкретного типа.
    list:
        Список всех доступных политик.
    """

    serializer_class = PolicySerializer
    permission_classes = [AccessPermission, EnabledPoliciesRequiredPermission]

    lookup_field = "data_type"

    def get_queryset(self):
        db_policies = {policy.data_type: policy for policy in PolicyGlobal.objects.all()}
        code_policies = registry().policies()
        elems = []
        for code_policy in code_policies.values():
            db_policy = db_policies.get(code_policy.typeType)
            if db_policy is None:
                value = code_policy.value.defValue
                values = PolicyDTO(
                    code_policy.typeName, code_policy.typeDescription, code_policy.typeType, value,
                    code_policy.get_view_value(value)
                ).get_values()
            else:
                values = db_policy.getInstance().valuesDict()

            elems.append(values)

        return elems

    def get_by_type(self, request, data_type: str):
        """Получить информацию о политике."""

        policy_values = get_policy_values_by_type(data_type)
        serializer = self.get_serializer(policy_values, context={"request": request})

        return Response(serializer.data)

    def update_by_type(self, request, data_type: str):
        """Обновить параметры политики."""
        data = request.data

        # для отработки процессов в сериализаторе необходимо передать data_type
        data["data_type"] = data_type

        if PolicyGlobal.objects.filter(data_type=data_type).exists():
            instance = PolicyGlobal.objects.get(data_type=data_type)
            serializer = self.get_serializer(instance=instance, data=data)
        else:
            serializer = self.get_serializer(data=data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return Response(serializer.instance.getInstance().valuesDict())

    def get_gui_params(self, **kwargs):
        type_policy = kwargs.get("data_type", "")

        policy = get_policy_class_by_type(type_policy)
        return policy.guiDescription()
