# -*- encoding: utf-8 -*-

import logging

from django.contrib.auth.models import Group
from django.shortcuts import get_object_or_404
from rest_framework.response import Response
from rest_framework.status import HTTP_201_CREATED, HTTP_400_BAD_REQUEST

from termidesk.api.decorators import has_role_permission, enabled_rbac_required
from termidesk.api.webui.serializers.group import GroupSerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.models.UserPermission import get_user_permissions, USER_ROLES_MANAGE
from uds.core.ui.UserInterface import gui
from uds.REST.model import OK

logger = logging.getLogger(__name__)


class Rbac(CommonViewSet):
    """
    list:
        Список всех доступных ролей.
    gui:
        Спецификация полей GUI для роли.
    """
    model = Group
    serializer_class = GroupSerializer

    def get_queryset(self):
        return self.model.objects.all()

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def retrieve(self, request, *args, **kwargs):
        """
        Получить информацию о роли.
        """
        group = get_object_or_404(self.model, id=kwargs.get('id', ''))
        serializer = self.serializer_class(group, context={'request': request})

        return Response(serializer.data)

    def get_gui_params(self, **kwargs):
        permission_values = (gui.choiceItem(permission.id, permission.name) for permission in get_user_permissions())

        return self.add_field(self.add_default_fields([], ['name']), {
            'name': 'permissions',
            'label': 'Пользовательские разрешения',
            'values': permission_values,
            'tooltip': '',
            'type': gui.MultiChoiceField.MULTI_CHOICE_TYPE,
            'order': 102,
        })

    @has_role_permission(USER_ROLES_MANAGE)
    @enabled_rbac_required
    def create(self, request, *args, **kwargs):
        """ Создание роли. """
        serializer = self.get_serializer(data=request.data)

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        self.perform_create(serializer)

        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=HTTP_201_CREATED, headers=headers)

    @enabled_rbac_required
    @has_role_permission(USER_ROLES_MANAGE)
    def update(self, request, *args, **kwargs):
        """
        Обновить параметры роли.
        """
        group = get_object_or_404(self.model, id=kwargs.get('id', ''))

        serializer = self.serializer_class(instance=group, data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        return Response(serializer.data)

    @has_role_permission(USER_ROLES_MANAGE)
    @enabled_rbac_required
    def overview(self, request):
        """
        Список всех доступных ролей.
        """
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.serializer_class(page, many=True, context={'request': request})
            return self.get_paginated_response(serializer.data)

        serializers = self.serializer_class(queryset, many=True, context={'request': request})
        return Response(serializers.data)

    @has_role_permission(USER_ROLES_MANAGE)
    @enabled_rbac_required
    def delete(self, request, id: int):
        """
        Удалить роль.
        """
        group = get_object_or_404(self.model, id=id)

        try:
            group.delete()
            res = OK
        except Exception as error:
            return Response(error, status=HTTP_400_BAD_REQUEST)

        return Response(res)

    @has_role_permission(USER_ROLES_MANAGE)
    @enabled_rbac_required
    def list(self, request, *args, **kwargs):
        """Получение списка ролей"""
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)

