import logging

import config
from django.db.models import Q
from django.contrib.contenttypes.models import ContentType
from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext
from django.utils.translation import ugettext_lazy as _
from guardian.models.models import GroupObjectPermission
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST

from termidesk.api.decorators import has_role_permission, has_object_role_permission
from termidesk.api.webui.serializers.log import LogSerializer
from termidesk.api.webui.serializers.service_pool import ServicePoolSerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.core.util.acl import get_objects_for_groups
from termidesk.models.UserPermission import DEPLOYED_SERVICE_VIEW, DEPLOYED_SERVICE_CHANGE, \
    DEPLOYED_SERVICE_DELETE, DEPLOYED_SERVICE_CREATE
from termidesk.rest.handlers import CREATE_ACTION, DELETE_ACTION, READ_ACTION, UPDATE_ACTION, set_crud_event
from uds.core.ui.images import DEFAULT_THUMB_BASE64
from uds.core.ui.UserInterface import gui
from uds.core.util import log
from uds.core.util.Config import GlobalConfig
from uds.core.util.model import processUuid
from uds.models import Image, OSManager, Service, ServicesPoolGroup
from uds.models.ServicesPool import ServicePool
from uds.REST import ResponseError
from uds.REST.model import OK

logger = logging.getLogger(__name__)

SEARCH = 'search'


class ServicesPools(CommonViewSet):
    """
    list:
        Список всех доступных фондов рабочих мест.
    create:
        Создание фонда рабочих мест.
        Для просмотра доступных типов фондов рабочих мест обратитесь по следующему пути: /api/webui/servicespools/types
    gui:
        Спецификация полей GUI определенного типа для фонда рабочих мест.
    get_logs:
        Получение журнала фонда рабочих мест.
    """
    model = ServicePool
    serializer_class = ServicePoolSerializer

    def get_queryset(self):
        rbac_enabled = GlobalConfig.EXPERIMENTAL_RBAC_ENABLED.getBool(True)
        if self.request.user.is_superuser or not rbac_enabled:
            queryset = self.model.objects.all()
        else:
            groups = self.request.user.roles.all()
            if len(groups) == 0:
                return None

            queryset = get_objects_for_groups(groups, 'uds.{}'.format(DEPLOYED_SERVICE_VIEW))

        search = self.request.query_params.get(SEARCH)
        if search not in ['', None]:
            queryset = queryset.filter(Q(name__icontains=search) | Q(service__name__icontains=search))

        return queryset

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    @has_object_role_permission(DEPLOYED_SERVICE_VIEW, ServicePool, 'uuid')
    def retrieve(self, request, *args, **kwargs):
        """
        Получить информацию о фонде рабочих мест.
        """
        uuid = kwargs.get('uuid', '').lower()
        service_pool = get_object_or_404(self.model, uuid=uuid)
        serializer = ServicePoolSerializer(service_pool, context={'request': request})

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            READ_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def get_gui_params(self, **kwargs):
        gui_default_fields = self.add_default_fields([], ['name', 'comments', 'tags'])

        for field in [{
            'name': 'service_id',
            'values': [gui.choiceItem(-1, '')] + gui.sortedChoices(
                [gui.choiceItem(v.uuid, v.provider.name + '\\' + v.name) for v in
                 Service.objects.all()]),
            'label': ugettext('Шаблон'),
            'tooltip': ugettext('Шаблон рабочего места для формирования фонда'),
            'type': gui.InputField.CHOICE_TYPE,
            'rdonly': True,
            'required': True,
            'order': 100,  # Ensueres is At end
        }, {
            'name': 'osmanager_id',
            'values': [gui.choiceItem(-1, '')] + gui.sortedChoices(
                [gui.choiceItem(v.uuid, v.name) for v in OSManager.objects.all()]),
            'label': ugettext('Параметры гостевой ОС'),
            'tooltip': ugettext('Параметры гостевой ОС для рабочих мест фонда'),
            'type': gui.InputField.CHOICE_TYPE,
            'rdonly': True,
            'required': True,
            'order': 101,
        }, {
            'name': 'image_id',
            'values': [gui.choiceImage(-1, '--------', DEFAULT_THUMB_BASE64)] + gui.sortedChoices(
                [gui.choiceImage(v.uuid, v.name, v.thumb64) for v in Image.objects.all()]),
            'label': ugettext('Изображение'),
            'tooltip': ugettext('Изображение, ассоциированное с фондом'),
            'type': gui.InputField.IMAGECHOICE_TYPE,
            'order': 102,
            'tab': ugettext('Display'),
        }, {
            'name': 'servicesPoolGroup_id',
            'values': [gui.choiceImage(-1, _('Основная'), DEFAULT_THUMB_BASE64)] + gui.sortedChoices(
                [gui.choiceImage(v.uuid, v.name, v.thumb64) for v in ServicesPoolGroup.objects.all()]),
            'label': ugettext('Группа'),
            'tooltip': ugettext('Группа рабочих мест'),
            'type': gui.InputField.IMAGECHOICE_TYPE,
            'order': 103,
            'tab': ugettext('Display'),
        }, {
            'name': 'help_url',
            'label': _(config.APP_CUSTOMIZE_WORKPLACE_HELP_URL_LABEL),
            'tooltip': _(config.APP_CUSTOMIZE_WORKPLACE_HELP_URL_TOOLTIP),
            'length': 256,
            'order': 104,
        }, {
            'name': 'cache_l1_srvs',
            'value': '0',
            'minValue': '0',
            'label': ugettext('Кеш рабочих мест 1-го уровня'),
            # 'tooltip': ugettext('Services kept in cache for improved user service assignation'),
            'tooltip': ugettext(
                'Сколько рабочих мест поддерживать наготове (запущенными) для быстрого подключения пользователей'),
            'type': gui.InputField.NUMERIC_TYPE,
            'order': 111,
            'tab': ugettext('Availability'),
        }, {
            'name': 'cache_l2_srvs',
            'value': '0',
            'minValue': '0',
            'label': ugettext('Кеш рабочих мест 2-го уровня'),
            # 'tooltip': ugettext('Services kept in cache of level2 for improved service generation'),
            'tooltip': ugettext(
                'Сколько рабочих мест поддерживать в спящем режиме для экономии ресурсов'),
            'type': gui.InputField.NUMERIC_TYPE,
            'order': 112,
            'tab': ugettext('Availability'),
        }, {
            'name': 'max_srvs',
            'value': '0',
            'minValue': '1',
            'label': ugettext('Максимальное количество рабочих мест'),
            # 'tooltip': ugettext('Maximum number of service (assigned and L1 cache) that can be created for this service'),
            'tooltip': ugettext(
                'Максимум рабочих мест (назначенных пользователям и в кеше 1-го уровня) которые могут быть развернуты в фонде'),
            'type': gui.InputField.NUMERIC_TYPE,
            'order': 113,
            'tab': ugettext('Availability'),
        }, {
            'name': 'show_transports',
            'value': True,
            'label': ugettext('Выбор протоколов доставки'),
            'tooltip': ugettext('Показывать пользователю альтернативные протоколы для подключения'),
            'type': gui.InputField.CHECKBOX_TYPE,
            'order': 120,
        }, {
            'name': 'debug_mode',
            'value': False,
            'label': ugettext('Режим отладки'),
            'tooltip': ugettext('Включение режима отладки'),
            'type': gui.InputField.CHECKBOX_TYPE,
            'order': 121,
        }, {
            'name': 'backup',
            'value': False,
            'label': ugettext('Разрешить резервные копии'),
            'tooltip': ugettext('Включить создание резервных копий'),
            'type': gui.InputField.CHECKBOX_TYPE,
            'order': 121,
        }]:
            self.add_field(gui_default_fields, field)

        return gui_default_fields

    def perform_create(self, serializer):
        serializer.save()

        set_crud_event(
            self.request.user.manager.name,
            self.request.user.name,
            self.request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            CREATE_ACTION
        )

    @has_role_permission(DEPLOYED_SERVICE_CHANGE)
    @has_object_role_permission(DEPLOYED_SERVICE_CHANGE, ServicePool, 'uuid')
    def update(self, request, *args, **kwargs):
        """
        Обновить фонд рабочих мест.
        """
        uuid = kwargs.get('uuid', '').lower()
        service_pool = get_object_or_404(self.model, uuid=uuid)

        serializer = ServicePoolSerializer(instance=service_pool, data=request.data,
                                           context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            UPDATE_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(DEPLOYED_SERVICE_DELETE)
    @has_object_role_permission(DEPLOYED_SERVICE_DELETE, ServicePool, 'uuid')
    def delete(self, request, uuid):
        """
        Удалить фонд рабочих мест.
        """
        uuid = uuid.lower()
        service_pool = get_object_or_404(self.model, uuid=uuid)

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            service_pool.uuid,
            '',
            service_pool.name,
            DELETE_ACTION
        )

        try:
            content_type = ContentType.objects.get_for_model(self.model)
            deleted_permissions_count, _ = GroupObjectPermission.objects.filter(
                object_pk=service_pool.id, content_type_id=content_type.id).delete()

            if deleted_permissions_count > 0:
                logger.debug('Number of permissions removed: {}'.format(deleted_permissions_count))

            service_pool.remove()
            res = OK
        except Exception as error:
            return Response(error, status=HTTP_400_BAD_REQUEST)

        return Response(res)

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def get_types(self, request):
        """ Получение списка доступных типов для фондов рабочих мест. """
        return Response([])

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def overview(self, request, *args, **kwargs):
        """ Получение списка доступных фондов рабочих мест (без дополнительных полей под каждый тип). """
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = ServicePoolSerializer(page, many=True, context={'request': request})
            return self.get_paginated_response(serializer.data)

        serializer = ServicePoolSerializer(queryset, many=True, context={'request': request})
        return Response(serializer.data)

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def get_logs(self, request, *args, **kwargs):
        """ Получение журнала. """
        try:
            uuid = kwargs.get('uuid', '').lower()
            service_pool = get_object_or_404(ServicePool, uuid=processUuid(uuid))
        except Exception as error:
            raise ResponseError(str(error))

        serializer = LogSerializer(log.getLogs(service_pool), many=True)
        return Response(serializer.data)

    @has_role_permission(DEPLOYED_SERVICE_CREATE)
    def create(self, request, *args, **kwargs):
        """Создание фонда рабочих мест"""
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().create(request, *args, **kwargs)

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def list(self, request, *args, **kwargs):
        """Получение списка фондов рабочих мест"""
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)
