import logging

from django.shortcuts import get_object_or_404
from django.utils.translation import ugettext_lazy as _
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST

from termidesk.api.decorators import has_object_role_permission, has_role_permission
from termidesk.api.webui.serializers.log import LogSerializer
from termidesk.api.webui.serializers.user_service import CachedUserServiceSerializer
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet
from termidesk.models.UserPermission import DEPLOYED_SERVICE_VIEW, DS_USER_SERVICE_MANAGE
from termidesk.rest.handlers import DELETE_ACTION, set_crud_event
from uds.core.util import log
from uds.core.util.model import processUuid
from uds.core.util.State import State
from uds.models.ServicesPool import ServicePool
from uds.REST import ResponseError
from uds.REST.model import OK

logger = logging.getLogger(__name__)


class CachedUserService(ChildCommonViewSet):
    """Класс обработки запросов для кэшированных рабочих мест."""

    model = ServicePool
    serializer_class = CachedUserServiceSerializer

    @has_object_role_permission(DS_USER_SERVICE_MANAGE, ServicePool, "uuid")
    def delete_cached(self, request, uuid, service_uuid):
        """Удаление кэшированного рабочего места"""
        workplace_uuid = uuid.lower()
        workplace = get_object_or_404(self.model, uuid=workplace_uuid)

        user_service = get_object_or_404(workplace.cachedUserServices(), uuid=service_uuid.lower())

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            user_service.uuid,
            "",
            user_service.name,
            DELETE_ACTION,
        )

        logger.debug("Deleting cached service")
        if user_service.state in (State.USABLE, State.REMOVING):
            user_service.remove()
        elif user_service.state == State.PREPARING:
            user_service.cancel()
        elif user_service.state == State.REMOVABLE:
            return Response(_("Item already being removed"), status=HTTP_400_BAD_REQUEST)
        else:
            return Response(_("Item is not removable"), status=HTTP_400_BAD_REQUEST)

        return Response(OK)

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def get_cached(self, request, uuid):
        """Получение списка рабочих мест в кэше"""
        uuid = uuid.lower()
        workplace = get_object_or_404(self.model, uuid=processUuid(uuid))
        queryset = workplace.cachedUserServices()
        serializer = CachedUserServiceSerializer(queryset, many=True, context={"request": request})
        return Response(serializer.data)

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def get_cached_log(self, request, uuid, service_uuid):
        """Получение логов для рабочего места из кэша"""
        try:
            uuid = uuid.lower()
            workplace = get_object_or_404(self.model, uuid=processUuid(uuid))

            user_service = get_object_or_404(workplace.cachedUserServices(), uuid=service_uuid.lower())
        except Exception as error:
            raise ResponseError(str(error))

        serializer = LogSerializer(log.getLogs(user_service), many=True)
        return Response(serializer.data)
