import logging
import uuid

from django.shortcuts import get_object_or_404
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK
from rest_framework.viewsets import ModelViewSet
from termidesk.api.decorators import has_object_role_permission, has_role_permission
from termidesk.api.webui.serializers.servicespool.group import (GroupServicePoolSerializer,
                                                                GroupListSerializer)
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet
from termidesk.api.webui.views.servicespool import SERVICE_POOL_UUID
from termidesk.models.UserPermission import DEPLOYED_SERVICE_VIEW, DS_GROUP_MANAGE
from termidesk.rest.helpers import ServicePoolsGroupUI
from uds.core.Environment import Environment
from uds.core.util.model import processUuid
from uds.models.ServicesPool import ServicePool
from uds.models.Group import Group

logger = logging.getLogger(__name__)


class GroupsOverviewServicePool(ModelViewSet):
    """
    list:
        Список всех доступных групп фонда рабочих мест.
    """
    model = Group
    serializer_class = GroupListSerializer

    def get_queryset(self):
        service_pool_uuid = str(self.kwargs.get(SERVICE_POOL_UUID, '')).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        return service_pool.assignedGroups.all()

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def list(self, request, *args, **kwargs):
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)


class GroupsServicePool(ChildCommonViewSet):
    """
    add:
        Добавление группы в фонд рабочих мест.
    gui:
        Спецификация полей интерфейса для групп фонда рабочих мест.
    delete:
        Удалить группу из фонда рабочих мест.
    """
    serializer_class = GroupServicePoolSerializer

    @has_object_role_permission(DS_GROUP_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def add(self, request, servicespool_uuid: uuid):
        service_pool = get_object_or_404(ServicePool, uuid=str(servicespool_uuid).lower())

        group = get_object_or_404(Group,
                                  uuid=processUuid(request.data.get('id')),
                                  manager__uuid=processUuid(request.data.get('domain')))

        if service_pool.assignedGroups.filter(uuid=group.uuid).exists():
            raise APIException('Group "{}" has already attached by this workplace'.format(group))

        service_pool.assignedGroups.add(group)

        self._set_update_parents_audit(service_pool)

        return Response(self.success(), status=HTTP_200_OK)

    def get_gui_params(self, **kwargs):
        service_pool_uuid = str(kwargs.get(SERVICE_POOL_UUID, '')).lower()
        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        env = Environment.getTempEnv()
        return ServicePoolsGroupUI.guiDescription(ServicePoolsGroupUI(env, service_pool))

    @has_object_role_permission(DS_GROUP_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def delete(self, request, servicespool_uuid: uuid, uuid: uuid):
        service_pool_uuid = str(servicespool_uuid).lower()
        uuid = str(uuid).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)
        group = get_object_or_404(Group, uuid=uuid)
        if not service_pool.assignedGroups.filter(uuid=uuid).exists():
            raise APIException('Group "{}" is not attached by this workplace'.format(group))

        service_pool.assignedGroups.remove(group)

        self._set_update_parents_audit(service_pool)

        return Response(self.success())

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def gui(self, request, *args, **kwargs):
        return super().gui(request, *args, **kwargs)
