import logging
import uuid

from django.shortcuts import get_object_or_404
from rest_framework.response import Response
from termidesk.api.permissions import AccessPermission, EnabledPoliciesRequiredPermission
from termidesk.api.webui.serializers.servicespool.policy import (DEPLOYED_SERVICE_ID,
                                                                 PolicyDeployedServiceSerializer)
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.api.webui.views.servicespool import SERVICE_POOL_UUID
from termidesk.core.policies.utils import (PolicyDTO, get_policy_class_by_type,
                                           get_policy_values_by_type, registry)
from termidesk.models.PolicyDeployedService import PolicyDeployedService
from termidesk.models.PolicyGlobal import PolicyGlobal
from uds.models.ServicesPool import DeployedService

logger = logging.getLogger(__name__)


class PoliciesDeployedServices(CommonViewSet):
    """
    list:
        Список всех политик фонда рабочих мест.
    """

    serializer_class = PolicyDeployedServiceSerializer
    permission_classes = [AccessPermission, EnabledPoliciesRequiredPermission]

    lookup_field = "data_type"

    deployed_service = None

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)

        servicespool_uuid = kwargs.get(SERVICE_POOL_UUID)
        self.deployed_service = get_object_or_404(DeployedService, uuid=servicespool_uuid)

    def get_queryset(self):
        deployed_service_policies = PolicyDeployedService.get_policies_by_types(self.deployed_service.id)
        global_policies = PolicyGlobal.get_policies_by_types()
        code_policies = registry().policies()
        elems = []
        for code_policy in code_policies.values():
            db_policy = deployed_service_policies.get(
                code_policy.typeType, global_policies.get(code_policy.typeType)
            )
            if db_policy is None:
                value = code_policy.value.defValue
                values = PolicyDTO(
                    code_policy.typeName, code_policy.typeDescription, code_policy.typeType, value,
                    code_policy.get_view_value(value)
                ).get_values()
            else:
                values = db_policy.getInstance().valuesDict()
            elems.append(values)

        return elems

    def get_serializer(self, *args, **kwargs):
        kwargs['context'] = self.get_serializer_context()
        kwargs[DEPLOYED_SERVICE_ID] = self.deployed_service.id if self.deployed_service is not None else None

        return self.serializer_class(*args, **kwargs)

    def get_by_type(self, request, servicespool_uuid: uuid, data_type: str):
        """Получить информацию о политиках фонда рабочих мест."""
        policy_values = get_policy_values_by_type(data_type, self.deployed_service.id)
        serializer = self.get_serializer(policy_values, context={"request": request})

        return Response(serializer.data)

    def update_by_type(self, request, servicespool_uuid: uuid, data_type: str):
        """ Обновить параметры политики фонда рабочих мест. """
        data = request.data
        # для отработки процессов в сериализаторе необходимо передать data_type
        data["data_type"] = data_type

        instance = None
        if PolicyDeployedService.is_policy_exists(data_type, self.deployed_service.id):
            instance = PolicyDeployedService.get_policy(data_type, self.deployed_service.id)

        serializer = self.get_serializer(instance=instance, data=data)
        serializer.is_valid(raise_exception=True)
        serializer.save()

        return Response(serializer.instance.getInstance().valuesDict())

    def get_gui_params(self, **kwargs):
        type_policy = kwargs.get("data_type", "")

        policy = get_policy_class_by_type(type_policy)
        return policy.guiDescription()
