import logging
import uuid

from django.shortcuts import get_object_or_404
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema
from rest_framework.status import HTTP_200_OK
from rest_framework.viewsets import ModelViewSet
from termidesk.api.decorators import has_object_role_permission, has_role_permission
from termidesk.api.webui.serializers.servicespool.publication import (PublicationListSerializer,
                                                                      PublicationSerializer)
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet
from termidesk.api.webui.views.servicespool import SERVICE_POOL_UUID
from termidesk.models.UserPermission import DEPLOYED_SERVICE_VIEW, DS_PUBLICATION_MANAGE
from termidesk.rest.helpers import ServicePoolsPublicationUI
from uds.core.Environment import Environment
from uds.models import DeployedServicePublication
from uds.models.ServicesPool import ServicePool
from uds.models.ServicesPoolPublication import ServicePoolPublication

logger = logging.getLogger(__name__)


class PublicationsOverviewServicePool(ModelViewSet):
    """
    list:
        Список всех доступных публикаций фонда рабочих мест.
    """
    model = ServicePoolPublication
    serializer_class = PublicationListSerializer

    def get_queryset(self):
        service_pool_uuid = str(self.kwargs.get(SERVICE_POOL_UUID, '')).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        return service_pool.publications.all()

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def list(self, request, *args, **kwargs):
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)


class PublicationsServicePool(ChildCommonViewSet):
    """
    gui:
        Спецификация полей интерфейса для публикации фонда рабочих мест.
    cancel:
        Отмена публикации фонда рабочих мест.
    """
    serializer_class = PublicationSerializer

    def get_gui_params(self, **kwargs):
        service_pool_uuid = str(kwargs.get(SERVICE_POOL_UUID, '')).lower()
        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        env = Environment.getTempEnv()
        return ServicePoolsPublicationUI.guiDescription(ServicePoolsPublicationUI(env, service_pool))

    @has_object_role_permission(DS_PUBLICATION_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def cancel(self, request, servicespool_uuid: uuid, uuid: uuid):
        """
        Invoked to cancel a running publication
        Double invocation (this means, invoking cancel twice) will mean that is a "forced cancelation"
        """
        service_pool_uuid = str(servicespool_uuid).lower()
        uuid = str(uuid).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        if not any((request.user.is_superuser, request.user.is_staff,)):
            logger.debug('Management Permission failed. User {} is not admin or staff member.'
                         .format(request.user))
            self.access_denied()

        try:
            deployed_service = DeployedServicePublication.objects.get(uuid=uuid,
                                                                      deployed_service=service_pool)
            deployed_service.cancel()
        except Exception as e:
            raise APIException(e)

        return Response(self.success())

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def gui(self, request, *args, **kwargs):
        return super().gui(request, *args, **kwargs)


class PublishSchema(AutoSchema):
    def get_operation(self, path, method):
        operation = super().get_operation(path, method)
        operation['parameters'].append({
            "name": "changelog",
            "in": "query",
            "required": False,
            "description": "Причины публикации",
            'schema': {'type': 'string'}
        })
        operation['parameters'].append({
            "name": "tags",
            "in": "query",
            "required": False,
            "description": "Теги",
            'schema': {'type': 'list'}
        })
        return operation


class PublishServicePool(ChildCommonViewSet):
    """
    publish:
        Публикация фонда рабочих мест.
    """
    serializer_class = PublicationSerializer
    schema = PublishSchema()

    @has_object_role_permission(DS_PUBLICATION_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def publish(self, request, servicespool_uuid: uuid):
        """
        Custom method "publish", provided to initiate a publication of a deployed service
        """
        service_pool = get_object_or_404(ServicePool, uuid=str(servicespool_uuid).lower())

        changelog = self.request.query_params.get('changelog', '')

        if not any((request.user.is_superuser, request.user.is_staff,)):
            logger.debug('Management Permission failed. User {} is not admin or staff member.'
                         .format(request.user))
            self.access_denied()

        logger.debug('Custom "publish" invoked for {}'.format(service_pool))
        service_pool.publish(changelog)  # Can raise exceptions that will be processed on response

        self._set_update_parents_audit(service_pool)

        return Response(self.success(), status=HTTP_200_OK)
