import logging
import uuid

from django.shortcuts import get_object_or_404
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK
from termidesk.api.decorators import has_object_role_permission, has_role_permission
from termidesk.api.webui.serializers.servicespool.transport import (
    TransportListSerializer, TransportServicePoolSerializer)
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet, CommonViewSet
from termidesk.api.webui.views.servicespool import SERVICE_POOL_UUID
from termidesk.models.UserPermission import DEPLOYED_SERVICE_VIEW, DS_TRANSPORT_MANAGE
from termidesk.rest.helpers import ServicePoolsTransportUI
from uds.core.Environment import Environment
from uds.core.util.model import processUuid
from uds.models.ServicesPool import ServicePool
from uds.models.Transport import Transport

logger = logging.getLogger(__name__)


class TransportsOverviewServicePool(CommonViewSet):
    """
    list:
        Список всех доступных протоколов доставки фонда рабочих мест.
    """
    model = Transport
    serializer_class = TransportListSerializer

    def get_queryset(self):
        service_pool_uuid = str(self.kwargs.get(SERVICE_POOL_UUID, '')).lower()
        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        return service_pool.transports.all()

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)

        transports = []
        for transport in serializer.data:
            if transport['type'] is None:
                continue
            transport['type'] = self.type_as_dict(transport['type'])
            transports.append(transport)

        return Response(transports)


class TransportsServicePool(ChildCommonViewSet):
    """
    add:
        Добавление протокола доставки в фонд рабочих мест.
    gui:
        Спецификация полей интерфейса для протокола доставки фонда рабочих мест.
    delete:
        Удалить протокол доставки из фонда рабочих мест.
    """
    serializer_class = TransportServicePoolSerializer

    @has_object_role_permission(DS_TRANSPORT_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def add(self, request, servicespool_uuid: uuid):
        service_pool = get_object_or_404(ServicePool, uuid=str(servicespool_uuid).lower())

        transport = get_object_or_404(Transport, uuid=processUuid(request.data.get('id')))
        if service_pool.transports.filter(uuid=transport.uuid).exists():
            raise APIException('Transport "{}" has already attached by this workplace'.format(transport))

        service_pool.transports.add(transport)

        self._set_update_parents_audit(service_pool)

        return Response(self.success(), status=HTTP_200_OK)

    def get_gui_params(self, **kwargs):
        service_pool_uuid = str(kwargs.get(SERVICE_POOL_UUID, '')).lower()
        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        env = Environment.getTempEnv()
        return ServicePoolsTransportUI.guiDescription(ServicePoolsTransportUI(env, service_pool))

    @has_object_role_permission(DS_TRANSPORT_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def delete(self, request, servicespool_uuid: uuid, uuid: uuid):
        service_pool_uuid = str(servicespool_uuid).lower()
        uuid = str(uuid).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)
        transport = get_object_or_404(Transport, uuid=uuid)
        if not service_pool.transports.filter(uuid=uuid).exists():
            raise APIException('Transport "{}" is not attached by this workplace'.format(transport))

        service_pool.transports.remove(transport)

        self._set_update_parents_audit(service_pool)

        return Response(self.success())

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def gui(self, request, *args, **kwargs):
        return super().gui(request, *args, **kwargs)
