import logging
import uuid

from django.shortcuts import get_object_or_404
from rest_framework.exceptions import APIException
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK
from rest_framework.viewsets import ModelViewSet
from termidesk.api.decorators import has_object_role_permission, has_role_permission
from termidesk.api.webui.serializers.servicespool.user import UserListSerializer, UserSerializer
from termidesk.api.webui.views.common_viewset import ChildCommonViewSet
from termidesk.api.webui.views.servicespool import SERVICE_POOL_UUID
from termidesk.models.UserPermission import DEPLOYED_SERVICE_VIEW, DS_USER_MANAGE
from termidesk.rest.helpers import ServicePoolsUserUI
from uds.core.Environment import Environment
from uds.core.util.model import processUuid
from uds.models.ServicesPool import ServicePool
from uds.models.User import User

logger = logging.getLogger(__name__)


class UsersOverviewServicePool(ModelViewSet):
    """
    list:
        Список всех доступных пользователей фонда рабочих мест.
    """
    model = User
    serializer_class = UserListSerializer

    def get_queryset(self):
        service_pool_uuid = str(self.kwargs.get(SERVICE_POOL_UUID, '')).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        return service_pool.assigned_users.all()

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def list(self, request, *args, **kwargs):
        # Метод переопределяется только для реализации проверок согласно ролевой модели
        # (@has_role_permission())
        return super().list(request, *args, **kwargs)


class UsersServicePool(ChildCommonViewSet):
    """
    add:
        Добавление пользователя в фонд рабочих мест.
    gui:
        Спецификация полей интерфейса для пользователей фонда рабочих мест.
    delete:
        Удалить пользователя из фонда рабочих мест.
    """
    serializer_class = UserSerializer

    @has_object_role_permission(DS_USER_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def add(self, request, servicespool_uuid: uuid):
        service_pool = get_object_or_404(ServicePool, uuid=str(servicespool_uuid).lower())

        user = get_object_or_404(User,
                                 uuid=processUuid(request.data.get('id')),
                                 manager__uuid=processUuid(request.data.get('domain')))

        if service_pool.assigned_users.filter(uuid=user.uuid).exists():
            raise APIException('User "{}" has already attached by this workplace'.format(user))

        service_pool.assigned_users.add(user)

        self._set_update_parents_audit(service_pool)

        return Response(self.success(), status=HTTP_200_OK)

    def get_gui_params(self, **kwargs):
        service_pool_uuid = str(kwargs.get(SERVICE_POOL_UUID, '')).lower()
        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)

        env = Environment.getTempEnv()
        return ServicePoolsUserUI.guiDescription(ServicePoolsUserUI(env, service_pool))

    @has_object_role_permission(DS_USER_MANAGE, ServicePool, SERVICE_POOL_UUID)
    def delete(self, request, servicespool_uuid: uuid, uuid: uuid):
        service_pool_uuid = str(servicespool_uuid).lower()
        uuid = str(uuid).lower()

        service_pool = get_object_or_404(ServicePool, uuid=service_pool_uuid)
        user = get_object_or_404(User, uuid=uuid)
        if not service_pool.assigned_users.filter(uuid=uuid).exists():
            raise APIException('User "{}" is not attached by this workplace'.format(user))

        service_pool.assigned_users.remove(user)

        self._set_update_parents_audit(service_pool)

        return Response(self.success())

    @has_role_permission(DEPLOYED_SERVICE_VIEW)
    def gui(self, request, *args, **kwargs):
        return super().gui(request, *args, **kwargs)
