# -*- encoding: utf-8 -*-

from rest_framework.status import HTTP_400_BAD_REQUEST
from rest_framework.response import Response
from django.shortcuts import get_object_or_404

from termidesk.models.UserPermission import SERVICE_POOL_GROUP_VIEW, SERVICE_POOL_GROUP_CHANGE, \
    SERVICE_POOL_GROUP_DELETE, SERVICE_POOL_GROUP_CREATE
from ..serializers.spgroups import ServicePoolGroupSerializer
from termidesk.rest.handlers import (
    set_crud_event, CREATE_ACTION, READ_ACTION, UPDATE_ACTION, DELETE_ACTION
)
from uds.REST.model import OK
from uds.core.ui import gui as GUI
from uds.models import ServicesPoolGroup
from termidesk.api.webui.views.common_viewset import CommonViewSet
from django.utils.translation import ugettext

import logging

from ...decorators import has_role_permission

logger = logging.getLogger(__name__)


class ServicesPoolGroup(CommonViewSet):
    model = ServicesPoolGroup
    serializer_class = ServicePoolGroupSerializer

    @has_role_permission(SERVICE_POOL_GROUP_VIEW)
    def retrieve(self, request, uuid):
        """
        Получить информацию о группе
        """
        uuid = uuid.lower()
        group = get_object_or_404(self.model, uuid=uuid)
        serializer = ServicePoolGroupSerializer(
            group, context={'request': request}
        )

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            '',
            serializer.data.get('name', ''),
            READ_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(SERVICE_POOL_GROUP_CHANGE)
    def update(self, request, uuid):
        """
        Обновить группу
        """
        uuid = uuid.lower()
        group = get_object_or_404(self.model, uuid=uuid)
        serializer = ServicePoolGroupSerializer(
            instance=group, data=request.data, context={'request': request}
        )
        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)
        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            '',
            serializer.data.get('name', ''),
            UPDATE_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(SERVICE_POOL_GROUP_DELETE)
    def delete(self, request, uuid):
        """
        Удалить группу
        """
        uuid = uuid.lower()
        group = get_object_or_404(self.model, uuid=uuid)

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            group.uuid,
            '',
            group.name,
            DELETE_ACTION
        )

        try:
            group.delete()
            res = OK
        except Exception:
            res = None

        return Response(res)

    @has_role_permission(SERVICE_POOL_GROUP_CREATE)
    def create(self, request):
        """
        Создать группу
        """
        serializer = ServicePoolGroupSerializer(
            data=request.data, context={'request': request}
        )
        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            '',
            serializer.data.get('name', ''),
            CREATE_ACTION
        )

        return Response(serializer.data)

    @has_role_permission(SERVICE_POOL_GROUP_VIEW)
    def overview(self, request):
        """
        Список групп
        """
        serializers = ServicePoolGroupSerializer(
            self.model.objects.all(), many=True, context={'request': request}
        )

        return Response(serializers.data)

    def gui(self, request):

        image_id_custom_fields = {
                'name': 'image_id',
                'label': 'Associated Image',
                'value': '',
                'tooltip': 'Image assocciated with this service',
                'type': GUI.InputField.HIDDEN_TYPE,
                'order': 102,
                }
        gui_default_fields = self.add_default_fields(
            [], ['name', 'comments', 'priority']
        )
        res = self.add_field(gui_default_fields, image_id_custom_fields)
        return Response(res)
