# -*- encoding: utf-8 -*-

import logging
import sys

from rest_framework.generics import get_object_or_404
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST
from termidesk.api.decorators import has_role_permission
from termidesk.api.webui.exceptions import NotImplemented
from termidesk.api.webui.serializers.user_service import SessionUserServiceSerializer
from termidesk.api.webui.views.common_viewset import CommonViewSet
from termidesk.audit.events import WEB_LOGOFF_SESSION_USERSERVICE
from termidesk.models.UserPermission import USER_WORKPLACE_SESSION_MANAGE, USER_WORKPLACE_SESSION_VIEW
from termidesk.notifications.delayed_task.logoff_userservice import SUGGESTED_TIME, TAG, LogoffUserService
from uds.core.jobs.DelayedTaskRunner import DelayedTaskRunner
from uds.core.util import log
from uds.core.util.State import State
from uds.models.UserService import UserService

logger = logging.getLogger(__name__)


class SpSessions(CommonViewSet):
    model = UserService
    serializer_class = SessionUserServiceSerializer

    # минимальная версия агента для прерывания активной сессии
    MIN_VERSION_AGENT = (2, 3, 0)

    def get_queryset(self):
        return self.model.objects.filter(cache_level=0, in_use=True, state=State.USABLE).order_by('-in_use_date').all()

    @has_role_permission(USER_WORKPLACE_SESSION_VIEW)
    def overview(self, request):
        """
        Список активных сессий пользователей.
        """
        queryset = self.filter_queryset(self.get_queryset())

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.serializer_class(page, many=True, context={'request': request})
            return self.get_paginated_response(serializer.data)

        serializers = self.serializer_class(queryset, many=True, context={'request': request})
        return Response(serializers.data)

    @has_role_permission(USER_WORKPLACE_SESSION_MANAGE)
    def logoff(self, request, uuid):
        user_service = get_object_or_404(self.model, uuid=uuid)

        actor_version = tuple([int(elem) for elem in user_service.getProperty('actor_version', '0').split('.')])

        if actor_version < self.MIN_VERSION_AGENT:
            raise NotImplemented(
                'Невозможно остановить выбранную активную сессию пользователя из-за устаревшей версии агента',
            )

        tag = TAG + uuid
        if DelayedTaskRunner.runner().checkExists(tag):
            return Response('Выбранная вами сессия уже находится в статусе отключения', HTTP_400_BAD_REQUEST)

        message = 'Session logoff requested by {}/{}'.format(request.user.name, request.user.manager.name)
        log.doLog(user_service, log.INFO, message, log.SERVICE)
        logger_message = message + 'for user service {} ({})'.format(user_service.friendly_name, user_service.uuid)
        logger.info(logger_message)

        userservice_view = '{} ({})'.format(user_service.friendly_name, user_service.uuid)
        user_view = '{}/{}'.format(request.user.name, request.user.manager.name)
        event = WEB_LOGOFF_SESSION_USERSERVICE(user=user_view, userservice=userservice_view).generate_event()
        sys.audit(*event)

        delayed_task = LogoffUserService(uuid)
        delayed_task.register(SUGGESTED_TIME, tag)

        return Response('ok')
