import json
import logging
import sys
from collections import OrderedDict

from rest_framework.response import Response
from rest_framework import viewsets

from termidesk.api.decorators import has_role_permission
from termidesk.api.permissions import AccessPermission
from termidesk.audit.events import WEB_SYSTEM_CONFIG_CHANGED_V1
from termidesk.models.UserPermission import CONFIG_VIEW, CONFIG_CHANGE
from termidesk.sys_audit_patch import add_audit_to_sys
from uds.REST.methods.actor import actorKey
from uds.core import auths, transports
from uds.core.util.Config import (AUDIT_SECTION, GLOBAL_SECTION, IPAUTH, MONITORING_SECTION, NX_PLUGIN,
                                  SECURITY_SECTION, NOTIFICATIONS_SECTION)
from uds.core.util.Config import Config as ConfigParams
from uds.core.util.Config import GlobalConfig

add_audit_to_sys()

logger = logging.getLogger(__name__)

TRANSLATED_SECTIONS = {
    GLOBAL_SECTION: 'Общие',
    SECURITY_SECTION: 'Безопасность',
    MONITORING_SECTION: 'Мониторинг',
    AUDIT_SECTION: 'Аудит',
    NOTIFICATIONS_SECTION: 'Уведомления',
    IPAUTH: 'Аутентификация',
}
# 2to3: TRANSLATED_SECTIONS.items() passed to list()
REVERSED_SECTIONS = {v: k for k, v in list(TRANSLATED_SECTIONS.items())}

FIELD_INFO = [
    {
        'tab_name': GLOBAL_SECTION,
        'tab_items': [
            {
                'key': GlobalConfig.THEME.key(),
                'label': 'Тема оформления',
                'tooltip': 'Тема оформления',
            },
            {
                'key': GlobalConfig.AUTORUN_SERVICE.key(),
                'label': 'Автозапуск рабочего места',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.CACHE_CHECK_DELAY.key(),
                'label': 'Интервал проверок кэша рабочих мест',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.CHECK_UNUSED_TIME.key(),
                'label': 'Интервал проверок неиспользуемых рабочих мест',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.CLEANUP_CHECK.key(),
                'label': 'Интервал очистки информационных объектов',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.DELAYED_TASKS_THREADS.key(),
                'label': 'Количество потоков фоновых задач',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.DISALLOW_GLOBAL_LOGIN.key(),
                'label': 'Запретить глобальный вход',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.IGNORE_LIMITS.key(),
                'label': 'Не учитывать максимальные ограничения',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.KEEP_INFO_TIME.key(),
                'label': 'Время хранения информационных объектов',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.LOGIN_BLOCK.key(),
                'label': 'Время блокировки входа',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.LOGIN_URL.key(),
                'label': 'URL входа',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.MAX_INITIALIZING_TIME.key(),
                'label': 'Максимальное время инициализации рабочего места',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.MAX_LOGS_PER_ELEMENT.key(),
                'label': 'Максимум записей в журнале для объектов',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.MAX_LOGIN_TRIES.key(),
                'label': 'Максимум попыток входа',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.REDIRECT_TO_HTTPS.key(),
                'label': 'Перенаправлять на HTTPS',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.REMOVAL_CHECK.key(),
                'label': 'Интервал проверки для удаления объектов',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.RESTRAINT_COUNT.key(),
                'label': 'Количество ошибок для ограничения фонда',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.RESTRAINT_TIME.key(),
                'label': 'Интервал отслеживания ошибок в фонде',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SCHEDULER_THREADS.key(),
                'label': 'Количество потоков планировщика задач',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SESSION_EXPIRE_TIME.key(),
                'label': 'Срок действия устаревшей публикации',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.STATS_DURATION.key(),
                'label': 'Срок хранения статистики',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.USER_SERVICE_CLEAN_NUMBER.key(),
                'label': 'Количество удаляемых рабочих мест за один проход',
                'tooltip': '',
            },
        ]
    },
    {
        'tab_name': SECURITY_SECTION,
        'tab_items': [
            {
                'key': GlobalConfig.EXCLUSIVE_LOGOUT.key(),
                'label': 'Эксклюзивный выход',
                'tooltip': '',
            },
            {
                'key': actorKey.key(),
                'label': 'Мастер ключ',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.TRUSTED_SOURCES.key(),
                'label': 'Доверенные хосты',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.ADMIN_IDLE_TIME.key(),
                'label': 'Длительность сессии администратора',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SUPER_USER_ALLOW_WEBACCESS.key(),
                'label': 'Доступ администратора к веб-части',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.HONOR_CLIENT_IP_NOTIFY.key(),
                'label': 'Использовать анонсируемый IP клиента',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SUPER_USER_PAM.key(),
                'label': 'Системная аутентификация администратора',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SUPER_USER_PASS.key(),
                'label': 'Пароль администратора',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SUPER_USER_LOGIN.key(),
                'label': 'Логин администратора',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.USER_SESSION_LENGTH.key(),
                'label': 'Длительность сессии пользователя',
                'tooltip': '',
            },
        ],
    },
    {
        'tab_name': MONITORING_SECTION,
        'tab_items': [
            {
                'key': GlobalConfig.ENABLE_SYSLOG.key(),
                'label': 'Логирование Syslog',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SYSLOG_HOST.key(),
                'label': 'Хост 1',
                'tooltip': 'Адрес syslog-сервера в формате host:port',
            },
            {
                'key': GlobalConfig.SYSLOG_HOST_2.key(),
                'label': 'Хост 2',
                'tooltip': 'Адрес syslog-сервера в формате host:port',
            },
            {
                'key': GlobalConfig.SYSLOG_HOST_3.key(),
                'label': 'Хост 3',
                'tooltip': 'Адрес syslog-сервера в формате host:port',
            },
            {
                'key': GlobalConfig.SYSLOG_SOCKET.key(),
                'label': 'Протокол',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SYSLOG_FACILITY.key(),
                'label': 'Категория сообщения',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.SYSLOG_LEVEL.key(),
                'label': 'Уровень логирования',
                'tooltip': '',
            },
        ]
    },
    {
        'tab_name': AUDIT_SECTION,
        'tab_items': [
            {
                'key': GlobalConfig.AUDIT_STRICT_MODE_ENABLE.key(),
                'label': 'Использовать "строгий" режим аудита',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_DB_LOG_ENABLE.key(),
                'label': 'Сохранение в БД',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_DB_LOG_DELETE_TIMEOUT.key(),
                'label': 'Время хранения записи в БД (дней)',
                'tooltip': 'От 0 до 356. 0 - не удалять записи',
            },
            {
                'key': GlobalConfig.AUDIT_DB_DELETE_RECORDS_AT_ONCE.key(),
                'label': 'Максимум удаляемых событий',
                'tooltip': 'От 100 до 100 000. По умолчанию, 1 000',
            },
            {
                'key': GlobalConfig.AUDIT_FILEHANDLER_ENABLE.key(),
                'label': 'Сохранение в файл',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_FILEHANDLER_PATH.key(),
                'label': 'Файл хранения событий',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_FILEHANDLER_BACKUPS.key(),
                'label': 'Количество архивных файлов (7-30)',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_SYSLOG_ENABLE.key(),
                'label': 'Отправка в Syslog',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_SYSLOG_HOST.key(),
                'label': 'Хост',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_SYSLOG_SOCKET.key(),
                'label': 'Протокол',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_SYSLOG_PORT.key(),
                'label': 'Порт',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.AUDIT_SYSLOG_FACILITY.key(),
                'label': 'Категория сообщения',
                'tooltip': '',
            },
        ],
    },
    {
        'tab_name': NOTIFICATIONS_SECTION,
        'tab_items': [
            {
                'key': GlobalConfig.NOTIFICATIONS_ENABLE.key(),
                'label': 'Вкл/выкл почтовых уведомлений',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_HOST.key(),
                'label': 'Хост',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_PORT.key(),
                'label': 'Порт',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_SERVER_EMAIL.key(),
                'label': 'Email отправителя',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_USERNAME.key(),
                'label': 'Пользователь',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_PASSWORD.key(),
                'label': 'Пароль',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_TLS.key(),
                'label': 'Поддержка TLS',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_SSL.key(),
                'label': 'Поддержка SSL',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SMTP_TIMEOUT.key(),
                'label': 'Таймаут',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_TO_EMAILS.key(),
                'label': 'Email получателей (через запятую)',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_SUBJECT_PREFIX.key(),
                'label': 'Префикс для темы письма',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_EVENTS_PROVIDER_MAINTENANCE_CHANGED.key(),
                'label': 'Уведомление о смене режима техобслуживания в поставщике ресурсов',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_EVENTS_USER_SERVICE_STATE_CHANGED_TO_ERROR.key(),
                'label': 'Уведомление о возникновении ошибок с рабочими местами',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_EVENTS_LICENSE_CONNECTION_LIMIT_REACHED.key(),
                'label': 'Уведомление о превышении лицензированного количества подключений',
                'tooltip': '',
            },
            {
                'key': GlobalConfig.NOTIFICATIONS_EVENTS_LICENSE_USER_LIMIT_REACHED.key(),
                'label': 'Уведомление о превышении лицензированного количества пользователей',
                'tooltip': '',
            },
        ]
    },
    {
        'tab_name': IPAUTH,
        'tab_items': [
            {
                'key': GlobalConfig.AUTHENTICATION_ERROR_CUSTOM_MESSAGE.key(),
                'label': 'Доп. информация при ошибке входа',
                'tooltip': '',
            },
        ],
    },
]

SYSLOG_HOSTS_KEYS = [
    GlobalConfig.SYSLOG_HOST.key(),
    GlobalConfig.SYSLOG_HOST_2.key(),
    GlobalConfig.SYSLOG_HOST_3.key(),
]

PORT_CONSIST_NOT_ONLY_DIGITS_ERROR_MESSAGE = 'Порт "%s" должен содержать только цифры'
PORT_NUMBER_NOT_IN_RANGE_ERROR_MESSAGE = 'Номер порта "%s" должен быть в диапазоне 1-65535'

# Маркеры для скриптов подготовки редакции Termidesk For Astra.
# Не удалять!
# ROOT PARAMS SUPPRESSION BEGIN
# ROOT PARAMS SUPPRESSION END


class TConfig(viewsets.ViewSet):
    """
    View to list all params for showing.
    """
    permission_classes = [AccessPermission]

    @has_role_permission(CONFIG_VIEW)
    def get(self, request, format=None):
        self.add_fields_all_plugins()

        res = []
        for config in FIELD_INFO:
            config_tab = {
                'tab_name': TRANSLATED_SECTIONS.get(config['tab_name'], config['tab_name']),
                'tab_items': config['tab_items'],
            }
            res.append(config_tab)

        for param in ConfigParams.enumerate():

            section = param.section()
            key = param.key()

            # 2to3: has_key() changed to 'in'
            if section in TRANSLATED_SECTIONS:
                section = TRANSLATED_SECTIONS[section]

            try:
                section_index = TConfig.get_index_of_dict_in_list(res, 'tab_name', section)
                _config_section = res[section_index]
                key_index = TConfig.get_index_of_dict_in_list(_config_section['tab_items'], 'key', key)
                _config_item = _config_section['tab_items'][key_index]
                _config_item.update({
                    'value': param.get(),
                    'crypt': param.isCrypted(),
                    'longText': param.isLongText(),
                    'type': param.getType(),
                    'params': param.getParams(),
                })

            except IndexError:
                pass
            except TypeError:
                pass

        logger.debug('Configuration: {0}'.format(res))
        return Response(res)

    @has_role_permission(CONFIG_CHANGE)
    def put(self, request, *args, **kwargs):
        """ Сохранение параметров конфигурации. """

        params = request.data.copy()
        for section, secDict in params.items():
            # 2to3: has_key() changed to 'in'
            if section in REVERSED_SECTIONS:
                section = REVERSED_SECTIONS[section]

            # 2to3: iteritems() changed to items()
            for key, vals in secDict.items():
                param_value = vals['value']
                if section == MONITORING_SECTION and key in SYSLOG_HOSTS_KEYS:
                    try:
                        self.validate_port_from_host_port_address(param_value)
                    except ValueError as e:
                        return Response(str(e))

                ConfigParams.update(section, key, param_value)

        # Регистрация события изменения системных параметров Диспетчера подключений
        event_params = dict(
            authenticator=request.user.manager.name,
            username=request.user.name,
            ip=self.request.ip,
        )
        evt = WEB_SYSTEM_CONFIG_CHANGED_V1(**event_params).generate_event()
        sys.audit(*evt)

        return Response('ok')

    @staticmethod
    def add_fields_all_plugins():
        """ Добавляем поля установленных плагинов. """
        nx_trans_plugin = transports.factory().lookup('NXTransport')
        if nx_trans_plugin is not None:
            FIELD_INFO.append({
                'tab_name': NX_PLUGIN,
                'tab_items': [
                    {
                        'key': '.'.join((NX_PLUGIN.lower(), 'downloadUrl')),
                        'label': 'Download url',
                        'tooltip': '',
                    },
                    {
                        'key': '.'.join((NX_PLUGIN.lower(), 'downloadUrlMACOS')),
                        'label': 'Download url for MAC OS',
                        'tooltip': '',
                    }
                ],
            })

        ip_auth_plugin = auths.factory().lookup('IPAuth')
        if ip_auth_plugin is not None:
            ip_auth_index = TConfig.get_index_of_dict_in_list(FIELD_INFO, 'tab_name', IPAUTH)
            auth_items = FIELD_INFO[ip_auth_index]['tab_items']
            from uds.auths.IP import autologin
            auto_login_key = autologin.key()
            auto_login_item_index = TConfig.get_index_of_dict_in_list(auth_items, 'key', auto_login_key)
            try:
                auth_items[auto_login_item_index].update({
                    'key': auto_login_key,
                    'label': 'IP аутентификация',
                    'tooltip': '',
                })

            except TypeError:
                auth_items.append({
                    'key': auto_login_key,
                    'label': 'IP аутентификация',
                    'tooltip': '',
                })

    @staticmethod
    def get_index_of_dict_in_list(_list, _tab_key, _tab_name):
        """ Получение индекса словаря в списке по определенному ключу """
        logger.debug("[TSRV-511] get_index %s, %s, %s" % (str(_tab_key), str(_tab_name), str(_list), ))
        try:
            tab_indexes = [i for i, x in enumerate(_list) if x[_tab_key] == _tab_name]
        except KeyError:
            logger.exception("[TSRV-511] KeyError for key = '%s'" % str(_tab_name))
            tab_indexes = []

        # FIXME: Маловероятно, но теоретически возможна ситуация,
        #  когда в списке будет два словаря с `'tab': _tab_name`

        # Ожидаемое количество вхождений - 0 (указанный таб отсутствует в списке)
        # или 1 - в списке присутствует единственный словарь с `'tab': _tab_name`
        assert len(tab_indexes) in (0, 1)
        try:
            tab_index = tab_indexes[0]
        except IndexError:
            logger.exception("[TSRV-511] IndexError for key = '%s'" % str(_tab_name))
            tab_index = None

        return tab_index

    @staticmethod
    def validate_port_from_host_port_address(value):
        """ Валидация номера порта из строки вида `hostname:port_number` """
        try:
            port = value.split(':', maxsplit=1)[1]
        except IndexError:
            return

        try:
            port = int(port)
        except ValueError:
            raise ValueError(PORT_CONSIST_NOT_ONLY_DIGITS_ERROR_MESSAGE % str(port))

        if not (0 < port < 65536):
            raise ValueError(PORT_NUMBER_NOT_IN_RANGE_ERROR_MESSAGE % str(port))
