# -*- encoding: utf-8 -*-

from rest_framework import viewsets
from rest_framework.status import HTTP_400_BAD_REQUEST
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema
from django.shortcuts import get_object_or_404

from termidesk.api.permissions import AccessPermission
from termidesk.models.UserPermission import TRANSPORT_CREATE, TRANSPORT_VIEW, TRANSPORT_CHANGE, \
    TRANSPORT_DELETE
from uds.core.transports import factory
from uds.core.util import OsDetector
from ..serializers.transport import TransportSerializer
from termidesk.rest.handlers import set_crud_event, CREATE_ACTION, READ_ACTION, UPDATE_ACTION, DELETE_ACTION
from uds.REST.model import OK
from uds.core.ui import gui as core_gui
from uds.models import Transport, Network
from django.utils.translation import ugettext_lazy as _, ugettext

import logging

from ...decorators import has_role_permission

logger = logging.getLogger(__name__)


class BaseTransports(viewsets.ViewSet):
    """
    Класс содержит общую логику для отображений протоколов доставки

    Общая логика была выделена в отдельный класс, так как API принимает три вида запросов:
    1. без параметров
    2. с uuid протокола доставки
    3. с типом протокола доставки
    Для отображения в интерфейсе `Swagger UI` описания параметров запросов 2 и 3
    используются разные `schema`.
    """
    permission_classes = [AccessPermission]
    model = Transport
    serializer_class = TransportSerializer
    schema = AutoSchema()

    # Метод позволяет частично сгенерировать документацию на базе сериализатора
    def get_serializer_context(self):
        return {
            'request': self.request,
            'format': self.format_kwarg,
            'view': self
        }

    # Метод позволяет частично сгенерировать документацию на базе сериализатора
    def get_serializer(self, *args, **kwargs):
        kwargs['context'] = self.get_serializer_context()
        return self.serializer_class(*args, **kwargs)


class Transports(BaseTransports):
    """Работа с запросами без параметров"""

    @has_role_permission(TRANSPORT_CREATE)
    def create_transport(self, request):
        """
        Создать протокол доставки

        :param request: объект-запрос
        """
        serializer = TransportSerializer(data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        # В БД хранятся дополнительные данные для инициализации объекта протокола доставки
        transport = self.model.objects.get(uuid=serializer.data.get('id', ''))
        transport.data = serializer.get_serialized_data()
        transport.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            CREATE_ACTION
        )

        # Добавление в ответ дополнительных данных
        additional_data = serializer.get_additional_data(serializer.data.get('id', ''))
        additional_data.update(serializer.data)
        return Response(additional_data)

    @has_role_permission(TRANSPORT_VIEW)
    def transports_overview(self, request):
        """
        Список протоколов доставки

        :param request: объект-запрос
        """
        # `TransportSerializer` должен проводить выборку
        # только для установленных в системе протоколов доставки.
        installed_transports = [transport.type() for transport in factory().providers().values()]
        serializer = TransportSerializer(self.model.objects.filter(data_type__in=installed_transports),
                                         many=True, context={'request': request})

        return Response(serializer.data)

    @has_role_permission(TRANSPORT_VIEW)
    def transports_types(self, request):
        """
        Спецификация доступных протоколов доставки

        :param request: объект-запрос
        """
        transports_list = []
        transports = list(factory().providers().values())

        # TSRV-319: Если системная настройка `EXPERIMENTAL_LOUDPLAY_TRANSPORTS_ENABLED` выключена,
        # то необходимо исключить из выдачи транспорты Loudplay
        from termidesk.transports.loudplay import EXPERIMENTAL_LOUDPLAY_TRANSPORTS_ENABLED
        loudplay_transports_enabled = EXPERIMENTAL_LOUDPLAY_TRANSPORTS_ENABLED.getBool(force=True)

        if not loudplay_transports_enabled:
            from termidesk.transports.loudplay import loudplay_transports_list
            transports = \
                [_transport for _transport in transports if _transport not in loudplay_transports_list]

        for transport in transports:
            item = {
                'type': transport.type(),
                'icon': transport.icon(),
                'name': transport.name(),
                'group': transport.group,
                'description': transport.description()
            }
            transports_list.append(item)
        # Для удобства отображения в интерфейсе список отсортирован по названию протокола
        sorted_transports_list = sorted(transports_list, key=lambda _transport: _transport['name'])
        return Response(sorted_transports_list)


class TransportsByUUID(BaseTransports):
    """Работа с запросами, которым передается uuid протокола доставки"""

    schema = AutoSchema()

    @has_role_permission(TRANSPORT_VIEW)
    def get_transport(self, request, uuid):
        """
        Получить информацию о протоколе доставки по указанному `uuid`

        :param request: объект-запрос
        :param uuid: строка, `uuid` протокола доставки
        """
        uuid = uuid.lower()
        transport = get_object_or_404(self.model, uuid=uuid)
        serializer = TransportSerializer(transport, context={'request': request})

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            READ_ACTION
        )

        # Добавление в ответ дополнительных данных
        additional_data = serializer.get_additional_data(serializer.data.get('id', ''))
        additional_data.update(serializer.data)
        return Response(additional_data)

    @has_role_permission(TRANSPORT_CHANGE)
    def update_transport(self, request, uuid):
        """
        Обновить сведения о протоколе доставки

        :param request: объект-запрос
        :param uuid: строка, `uuid` протокола доставки
        """
        uuid = uuid.lower()
        transport = get_object_or_404(self.model, uuid=uuid)
        serializer = TransportSerializer(transport, data=request.data, context={'request': request})

        if not serializer.is_valid():
            for error in serializer.errors.values():
                return Response(error[0], status=HTTP_400_BAD_REQUEST)

        serializer.save()

        # В БД хранятся дополнительные данные для инициализации объекта протокола доставки
        transport = self.model.objects.get(uuid=serializer.data.get('id', ''))
        transport.data = serializer.get_serialized_data()
        transport.save()

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            serializer.data.get('id', ''),
            serializer.data.get('type', ''),
            serializer.data.get('name', ''),
            UPDATE_ACTION
        )

        # Добавление в ответ дополнительных данных
        additional_data = serializer.get_additional_data(serializer.data.get('id', ''))
        additional_data.update(serializer.data)
        return Response(additional_data)

    @has_role_permission(TRANSPORT_DELETE)
    def delete_transport(self, request, uuid):
        """
        Удалить протокол доставки

        :param request: объект-запрос
        :param uuid: строка, `uuid` протокола доставки
        """
        uuid = uuid.lower()
        network = get_object_or_404(self.model, uuid=uuid)

        set_crud_event(
            request.user.manager.name,
            request.user.name,
            request.ip,
            self.model.__name__,
            network.uuid,
            network.data_type,
            network.name,
            DELETE_ACTION
        )

        try:
            network.delete()
            res = OK
        except Exception:  # TODO: Заменить `Exception` на перечисление конкретных исключений
            res = None

        return Response(res)


class TransportsByType(BaseTransports):
    """Работа с запросами, которым передается наименование типа протокола доставки"""

    schema = AutoSchema()

    def transports_gui(self, request, transport_type):
        """
        Спецификация полей GUI для протоколов доставки

        :param request: объект-запрос
        :param transport_type: строка, тип протокола доставки
        """
        res = self.generate_transport_gui(transport_type)

        return Response(res)

    def transports_test(self, request, transport_type):
        """
        Тестирование работы выбранного протокола доставки

        :param request: объект-запрос
        :param transport_type: строка, тип протокола доставки
        """
        from uds.core.Environment import Environment

        _transport_type = factory().lookup(transport_type)

        dct = self.request.data.copy()
        dct['_request'] = request
        try:
            test_result = _transport_type.test(Environment.getTempEnv(), dct)
        except AttributeError:
            # Если протокол доставки был установлен из плагина, но впоследствии плагин был удален,
            # а запись в БД осталась, то в таком случае `_transport_type` будет иметь значение `None`.
            # Примечание. Такая ситуация маловероятна, но лучше ее предусмотреть.
            test_result = [
                False,
                'The transport "{transport_type}" is not installed'.format(transport_type=transport_type)
            ]

        if test_result[0]:
            result = self.success()
        else:
            result = test_result[1]

        return Response(result)

    def generate_transport_gui(self, transport_type):
        """
        Генерация спецификации интерфейса для указанного протокола доставки

        :param transport_type: строка, тип протокола доставки
        :return: словарь с описанием элементов интерфейса
        """
        # Описание полей для переключателя "Доступ из сетей"
        nets_positive_custom_fields = {
            'name': 'nets_positive',
            'value': True,
            'label': ugettext('Доступ из сетей'),
            'tooltip': ugettext('При выборе, протокол будет активен только для перечисленных сетей. '
                                'Иначе, протокол будет заблокирован для выбранных сетей'),
            'type': 'checkbox',
            'order': 100,  # At end
        }

        # Описание полей для выпадающего списка доступных сетей
        networks_custom_fields = {
            'name': 'networks',
            'value': [],
            # TODO: We will fix this behavior after fully removed current admin client
            'values': sorted(
                [{'id': x.id, 'text': x.name} for x in Network.objects.all()],
                key=lambda x: x['text'].lower()),

            'label': ugettext('Сети'),
            'tooltip': ugettext('Сети, ассоциированные с данным протоколом. '
                                'Пустой список означает "все сети"'),
            'type': 'multichoice',
            'order': 101
        }

        # Описание полей для выпадающего списка разрешенных операционных систем
        allowed_oss_custom_fields = {
            'name': 'allowed_oss',
            'value': [],
            # TODO: We will fix this behavior after fully removed current admin client
            'values': sorted(
                [{'id': x, 'text': x} for x in OsDetector.knownOss], key=lambda x: x['text'].lower()),
            'label': ugettext('Разрешенные устройства'),
            'tooltip': ugettext('Пустой список разрешает любое совместимое устройство. '
                                'Иначе, только перечисленные устройства будут разрешены'),
            'type': 'multichoice',
            'order': 102
        }
        # Описание базовых элементов
        transports_gui = factory().lookup(transport_type).guiDescription()
        # Генерация списка полей на основе базовых элементов протокола доставки
        # и элементов по умолчанию ('name', 'comments', 'tags', 'priority')
        fields = self.add_default_fields(transports_gui, ['name', 'comments', 'tags', 'priority'])
        # Добавление необходимых элементов интерфейса
        fields = self.add_field(fields, nets_positive_custom_fields)
        fields = self.add_field(fields, networks_custom_fields)
        fields = self.add_field(fields, allowed_oss_custom_fields)

        # Элементы интерфейса сортируются согласно указанному в них порядку расположения
        return sorted(fields, key=lambda field: field['gui']['order'])

    def add_default_fields(self, gui, fields):
        """
        Добавляет поля по умолчанию (на основании списка) в описание интерфейса
        :param gui: Список со спецификацией интерфейса, в который будет проводиться добавление.
        :param fields: Список с именами полей, для добавления в интерфейс
                       Корректными значениями считаются 'name', 'comments', 'tags', 'priority'.
        """
        if 'tags' in fields:
            self.add_field(gui, {
                'name': 'tags',
                'label': _('Теги'),
                'type': 'taglist',
                'tooltip': _('Теги для данного объекта'),
                'order': 0 - 101,
            })

        if 'name' in fields:
            self.add_field(gui, {
                'name': 'name',
                'required': True,
                'label': _('Название'),
                'length': 128,
                'tooltip': _('Имя объекта'),
                'order': 0 - 100,
            })
        if 'comments' in fields:
            self.add_field(gui, {
                'name': 'comments',
                'label': _('Комментарий'),
                'tooltip': _('Комментарий для объекта'),
                'length': 256,
                'order': 0 - 99,
            })
        if 'priority' in fields:
            self.add_field(gui, {
                'name': 'priority',
                'type': 'numeric',
                'label': _('Приоритет'),
                'tooltip': _('Приоритет для данного объекта (чем меньше число, тем выше приоритет)'),
                'required': True,
                'value': 1,
                'length': 4,
                'order': 0 - 97,
            })

        return gui

    @staticmethod
    def add_field(gui, field):
        """
        Добавление поля в описание интерфейса
        Этот метод гарантирует наличие всех необходимых элементов
        для описания добавляемого поля. В случае, если значение не задано,
        оно заменяется значением по умолчанию.

        :param gui: Список со спецификацией интерфейса, в который будет проводиться добавление.
        :param field: Словарь, с данными о добавляемом поле.
        """
        v = {
            'name': field.get('name', ''),
            'value': '',
            'gui': {
                'required': field.get('required', False),
                'defvalue': field.get('value', ''),
                'value': field.get('value', ''),
                'minValue': field.get('minValue', '987654321'),
                'label': field.get('label', ''),
                'length': field.get('length', 128),
                'multiline': field.get('multiline', 0),
                'tooltip': field.get('tooltip', ''),
                'rdonly': field.get('rdonly', False),
                'type': field.get('type', core_gui.InputField.TEXT_TYPE),
                'order': field.get('order', 0),
                'values': field.get('values', [])
            }
        }
        if 'tab' in field:
            v['gui']['tab'] = field['tab']
        gui.append(v)
        return gui

    @staticmethod
    def success():
        return "ok"
