# -*- coding: utf-8 -*-
"""
Данный модуль содержит определения событий аудита и вспомогательные классы.

Классы событий фактически являются контейнерами аттрибутов на базе именованых 
кортежей.

Значения аттрибутов должны быть простыми типами (строка, целое и т.п.) 
для упрощения поддержки сериализации.

Каждое событие содержит два обязательных аттрибута:
  * уникальное название
  * версия схемы события

В схему события входят:
  * названия аттрибутов
  * типы аттрибутов
  * количество аттрибутов
  * порядок аттрибутов
  
При изменении вышеперечисленного необходимо увеличивать версию схемы.
Для этого создается новое событие с тем же именем, но новым набором аттрибутов 
и увеличенной версией схемы.

Например:
class WEB_USER_LOGIN_V1(
        BaseEventMixin,
        namedtuple(
            'WEB_USER_LOGIN_V1',
            ['authenticator', 'username', 'ip'])):
    __slots__ = ()
    _NAME = build_name('web.UserLogin')
    _VERSION = 1

class WEB_USER_LOGIN_V2(
        BaseEventMixin,
        namedtuple(
            'WEB_USER_LOGIN_V2',
            ['authenticator', 'username', 'srcip', 'user_agent'])):
    __slots__ = ()
    _NAME = build_name('web.UserLogin')
    _VERSION = 2

Классы событий предназначены для генерации событий аудита с помощью 
механизма sys.audit python. Например:

import sys

evt = WEB_USER_LOGIN_V1(
        authenticator=auth_name, 
        username=user.name, 
        ip=request.ip).generate_event()
# или
evt = WEB_USER_LOGIN_V1(auth_name, request.user.name, request.ip).generate_event()  
 
sys.audit(*evt)

Вызов generate_event() требуется для обхода ошибки в python 3.5, когда оператор 
распаковки не вызывает переопределенный метод __iter__.

"""
from __future__ import unicode_literals

__author__ = 'Dmitry Dvoeglazov'

import sys
import warnings
from collections import namedtuple
from itertools import chain

# Общий префикс для всех событий Termidesk
EVENT_PREFIX = 'termidesk.audit.events'


def build_name(name):
    """
    Вспомогательная функция для создания уникального имени события
    """
    return '.'.join([EVENT_PREFIX, name])

class BaseEventMixin:
    """
    Вспомогательный класс для добавления общей для всех событий функциональности.
    """

    __slots__ = ()
    def __iter__(self):
        """
        Название события и версия схемы должны идти первыми при перечислении 
        элементов кортежа.
        """
        return chain(
            [self.event, self.version],
            super().__iter__()
        )
    
    @property
    def event(self):
        return self._NAME
    
    @property
    def version(self):
        return self._VERSION
    
    def generate_event(self):
        if sys.version_info > (3, 6):
            warnings.warn("Method generate_event deprecated for python > 3.6 and can be removed.", stacklevel=2)
        return iter(self)
    
class WEB_USER_LOGIN_V1(
        # Наследуем общую для всех классов событий функциональность
        BaseEventMixin,
        # Создаем родительский класс на базе кортежа с нужными аттрибутами
        namedtuple(
            # Имя родительского типа именованого кортежа
            'WEB_USER_LOGIN_V1',
            # Определение аттрибутов события 
            ['authenticator', 'username', 'ip'])):
    '''
       Событие: Вход пользователя в систему через веб-интерфейс
       Данные:
         :param authenticator: Название домена аутентифкации пользователя - args[0]
         :param username: Логин пользователя - args[1]
         :param ip: IP адрес, с которого был сделан запрос - args[2]
    '''
    __slots__ = ()
    _NAME = build_name('web.UserLogin') # Задаем имя события
    _VERSION = 1 # Версия схемы события
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" вошел в систему с ip-адреса {args[2]}'

class WEB_USER_LOGOUT_V1(
        BaseEventMixin, 
        namedtuple(
            'WEB_USER_LOGOUT_V1', 
            ['authenticator', 'username', 'ip'])):
    '''
       Событие: Выход пользователя из веб-интерфейса (logout)
       Данные:
         :param authenticator: Название домена аутентифкации пользователя - args[0]
         :param username: Логин пользователя - args[1]
         :param ip: IP адрес, с которого был сделан запрос - args[2]
    '''
    __slots__ = ()
    _NAME = build_name('web.UserLogout')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" вышел из системы (ip-адрес {args[2]})'


class USER_LOGGED_IN_V1(
    BaseEventMixin,
    namedtuple(
        'USER_LOGGED_IN_V1',
        ['authenticator', 'username', 'guest_os_username', 'ip', 'workplace', 'vm_name', ])):
    """
        Событие: Вход пользователя в ОС рабочего места
        Данные:
            :param authenticator: Название домена аутентификации пользователя - args[0]
            :param username: Логин пользователя - args[1]
            :param guest_os_username: Имя пользователя совершающего вход в гостевую ОС - args[2]
            :param ip: IP адрес, с которого был сделан запрос - args[3]
            :param workplace": имя фонда рабочих мест - args[4]
            :param vm_name": имя выданной машины - args[5]
    """
    __slots__ = ()
    _NAME = build_name('workplace.UserLogin')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" вошел в гостевую ОС ВМ {args[5]} ' \
                           'фонда {args[4]} как пользователь {args[2]} с ip-адреса {args[3]}'


class USER_LOGGED_OUT_V1(
    BaseEventMixin,
    namedtuple(
        'USER_LOGGED_OUT_V1',
        ['authenticator', 'username', 'guest_os_username', 'ip', 'workplace', 'vm_name', ])):
    """
        Событие: Выход пользователя из ОС рабочего места
        Данные:
            :param authenticator: Название домена аутентификации пользователя - args[0]
            :param username: Логин пользователя - args[1]
            :param guest_os_username: Имя пользователя совершающего выход из гостевой ОС - args[2]
            :param ip: IP адрес, с которого был сделан запрос - args[3]
            :param workplace": имя фонда рабочих мест - args[4]
            :param vm_name": имя выданной машины - args[5]
    """
    __slots__ = ()
    _NAME = build_name('workplace.UserLogout')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" вышел из гостевой ОС ВМ {args[5]} ' \
                           'фонда {args[4]} как пользователь {args[2]} с ip-адреса {args[3]}'


class USER_WORKPLACE_CONNECTION_REQUEST_V1(
    BaseEventMixin,
    namedtuple(
        'USER_WORKPLACE_CONNECTION_REQUEST_V1',
        ['authenticator', 'username', 'ip', 'workplace', 'vm_name', 'transport', ])):
    """
        Событие: Подключение пользователя к рабочему месту и начало работы
        Данные:
            :param authenticator: Название домена аутентификации пользователя - args[0]
            :param username: Логин пользователя - args[1]
            :param ip: IP адрес, с которого был сделан запрос - args[2]
            :param workplace: название фонда рабочих мест - args[3]
            :param vm_name": имя выданной машины - args[4]
            :param transport: название протокола доставки - args[5]
    """
    __slots__ = ()
    _NAME = build_name('user.WorkplaceConnectionRequest')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" подключился к ВМ {args[4]} ' \
                           'фонда {args[3]} по протоколу {args[5]} с ip-адреса {args[2]}'


class WEB_SYSTEM_CONFIG_CHANGED_V1(
    BaseEventMixin,
    namedtuple(
        'WEB_SYSTEM_CONFIG_CHANGED_V1',
        ['authenticator', 'username', 'ip', ])):
    """
        Событие: Изменение системных параметров Диспетчера подключений
        Данные:
            :param authenticator: Название домена аутентификации пользователя - args[0]
            :param username: Логин пользователя - args[1]
            :param ip: IP адрес, с которого был сделан запрос - args[2]
    """
    __slots__ = ()
    _NAME = build_name('web.SystemConfigChanged')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" изменил системные параметры ' \
                           '(ip-адрес {args[2]})'


class WEB_ENTITY_ACTION(
    BaseEventMixin,
    namedtuple(
        'WEB_CRUD_ENTITY_V1',
        ['authenticator', 'username', 'ip', 'entity', 'uuid', 'subtype', 'name', 'action'])):
    """Событие: CRUD операции с объектами через REST API.

       Данные:
       :param authenticator: название домена аутентифкации пользователя - args[0]
       :param username: логин пользователя - args[1]
       :param ip: IP адрес, с которого был сделан запрос - args[2]
       :param entity: тип сущности - args[3]
       :param uuid: уникальный идентификатор - args[4]
       :param subtype: тип объекта - args[5]
       :param name: название объекта - args[6]
       :param action: действие над объектом (CRUD) - args[7]
    """
    _NAME = build_name('web.EntityAction')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" выполнил операцию {args[7]} для объекта ' \
                           '{args[3]} ({args[4]}) {args[5]} "{args[6]}" ' \
                           '(ip-адрес {args[2]})'


class WEB_LICENSE_UPDATED_V1(
    BaseEventMixin,
    namedtuple(
        'WEB_LICENSE_UPDATED_V1',
        ['authenticator', 'username', 'ip', 'license_file_name'])):
    """Событие: Загрузка файла лицензии через REST API.

       Данные:
       :param authenticator: название домена аутентификации пользователя - args[0]
       :param username: логин пользователя - args[1]
       :param ip: IP адрес, с которого был сделан запрос - args[2]
       :param license_file_name: имя файла лицензии - args[3]
    """
    _NAME = build_name('web.LicenseUpdated')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[1]}({args[0]})" загрузил новый файл лицензии {args[3]} ' \
                           'с ip-адреса {args[2]}'


class CLI_SYSTEM_CONFIG_CHANGED_V1(
    BaseEventMixin,
    namedtuple(
        'CLI_SYSTEM_CONFIG_CHANGED_V1',
        ['username', 'section_name', 'parameter_key', 'parameter_value'])):
    """
        Событие: Изменение системных параметров Диспетчера подключений через командную строку
        Данные:
            :param username: Логин пользователя - args[0]
            :param section_name: Название секции - args[1]
            :param parameter_key: Название изменяемого параметра - args[2]
            :param parameter_value: Новое значение параметра - args[3]
    """
    __slots__ = ()
    _NAME = build_name('cli.SystemConfigChanged')
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[0]}" изменил системный параметр ' \
                           '{args[1]}.{args[2]}={args[3]}'


class CLI_ENTITY_ACTION(
    BaseEventMixin,
    namedtuple(
        'CLI_ENTITY_ACTION_V1',
        ['username', 'entity', 'uuid', 'subtype', 'name', 'action'])):
    """
    Событие: CRUD операции с объектами через CLI.

       Данные:
       :param username: имя системного пользователя (который запустил команду) - args[0]
       :param entity: тип сущности - args[1]
       :param uuid: уникальный идентификатор - args[2]
       :param subtype: тип объекта - args[3]
       :param name: название объекта - args[4]
       :param action: действие над объектом (CRUD) - args[5]
    """
    _NAME = build_name('cli.EntityAction')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[0]}" выполнил операцию {args[5]} для объекта ' \
                           '{args[1]} ({args[2]}) {args[3]} "{args[4]}" '


class WEB_LOGOFF_SESSION_USERSERVICE(
    BaseEventMixin,
    namedtuple(
        'WEB_LOGOFF_SESSION_USERSERVICE',
        ['user', 'userservice'])):
    """Событие: Прекращение сессии пользователя по команде с сервера.

       Данные:
       :param user: логин пользователя - args[0]
       :param userservice: данные гостевой ВМ, сессию которой прекратили - args[1]
    """
    _NAME = build_name('web.LogoffUserservice')
    __slots__ = ()
    _VERSION = 1
    DESCRIPTION_TEMPLATE = 'Пользователь "{args[0]}" отправил запрос на прекращение сессии {args[1]}'
