import datetime as dt
import json
import logging
from datetime import datetime
from datetime import timezone
from logging import LogRecord
from logging.handlers import TimedRotatingFileHandler

from termidesk.audit.dataclasses import RecordDTO, RecordSyslog
from termidesk.audit.events import EVENT_PREFIX
from termidesk.audit.exceptions import AuditEventRecordException
from termidesk.audit.rotators import TarRotator
from logging.handlers import SysLogHandler
from tdsklog.handlers import get_socktype

logger = logging.getLogger(__name__)

MIN_NUMBER_BACKUP_FILE = 7
MAX_NUMBER_BACKUP_FILE = 30


class BaseHandler(object):
    """
    Базовый класс для обработчиков событий аудита

    Содержит шаблонный метод __record_event(), который не должен переопределяться в наследниках.
    Каждый наследник должен описывать свою логику в методе __process_record()

    При реализации данного класса использован шаблон *Template method*.
    Возможно, что с ростом количества обработчиков, удобнее будет использовать шаблон *Strategy*.
    """
    def __call__(self, event, *args):
        """
        Шаблонный метод. Содержит общую для всех логику обработки вызова.

        @param event:
        @param args:
        @return:
        """
        # Общая для всех наследников логика
        if EVENT_PREFIX not in event:
            return

        # Если в настройках системы обработчик "включен", то выполняется запись события
        if self._is_enabled():
            self._record_event(event, args)

    def _is_enabled(self) -> bool:
        """
        Проверяет включен ли в настройках текущий обработчик событий аудита

        @return: True - обработчик включен, иначе False
        """
        return False

    def _process_record(self, event: str, args: tuple):
        """
        Обработка события аудита

        В этом методе описываются действия по обработке события аудита,
        индивидуальные для каждого обработчика.
        @param event str: Наименование события аудита
        @param args tuple: аргументы события аудита
        """
        raise NotImplementedError("Must provide an '__process_record' method for {} class".format(self.__class__))

    def _record_event(self, event: str, args: tuple):
        """
        Шаблонный метод, реализующий базовый алгоритм выполнения записи о событии аудита.
        @param event str: Наименование события аудита
        @param args tuple: аргументы события аудита
        """
        try:
            self._process_record(event, args)
        except Exception as e:
            logger.exception(e)

            from uds.core.util.Config import GlobalConfig
            if GlobalConfig.AUDIT_STRICT_MODE_ENABLE.getBool():
                # Если установлен строгий режим обработки событий аудита,
                # то возбуждается исключение AuditEventRecordException
                raise AuditEventRecordException(e)


class FileHandler(BaseHandler):
    """ Основной обработчик сообщений аудита. """

    # Проверка выполнения отложенного конфигурирования параметров.
    _fix_time = False

    rotating_file_handler = False
    path_to_file = False

    def _is_enabled(self):
        """
        Проверяет включена ли в настройках запись в файл событий аудита

        @return: True - в настройках включена запись в файл, иначе False
        """
        from uds.core.util.Config import GlobalConfig
        return GlobalConfig.AUDIT_FILEHANDLER_ENABLE.getBool(force=True)

    def _process_record(self, event, args):
        """ Обработка записи события. """

        if self._need_updating_time():
            if not self.set_params():
                return False

        version, *params = args[0]
        record = RecordDTO(
            timestamp=datetime.utcnow().isoformat(),
            event=event,
            version=version,
            data=params
        )

        log_record = LogRecord('', 0, '', 0, record.get_json_params(), (), None)

        assert isinstance(self.rotating_file_handler, TimedRotatingFileHandler)
        self.rotating_file_handler.handle(log_record)

    def set_params(self):
        from uds.core.util.Config import GlobalConfig

        path_to_file = GlobalConfig.AUDIT_FILEHANDLER_PATH.get(force=True)

        if not path_to_file:
            logger.error('The file path did not set in parameters')
            return False

        backup_count = self.__get_backup_count()

        if self.rotating_file_handler is False or self.path_to_file != path_to_file:
            self.path_to_file = path_to_file

            log_formatter = logging.Formatter('%(message)s')
            self.rotating_file_handler = TimedRotatingFileHandler(
                filename=path_to_file,
                when="midnight",
                backupCount=backup_count,
                encoding='utf8',
            )
            self.rotating_file_handler.setFormatter(log_formatter)
            self.rotating_file_handler.rotator = TarRotator()
        elif backup_count != self.rotating_file_handler.backupCount:
            self.rotating_file_handler.backupCount = backup_count

        self._fix_time = dt.datetime.now()
        return True

    @staticmethod
    def __get_backup_count():
        from uds.core.util.Config import GlobalConfig

        backup_count = GlobalConfig.AUDIT_FILEHANDLER_BACKUPS.getInt(force=True)
        if backup_count < MIN_NUMBER_BACKUP_FILE:
            return MIN_NUMBER_BACKUP_FILE

        if backup_count > MAX_NUMBER_BACKUP_FILE:
            return MAX_NUMBER_BACKUP_FILE

        return backup_count

    def _need_updating_time(self):
        if self._fix_time is False:
            return True

        time_for_removing = self._fix_time + dt.timedelta(seconds=30)

        return dt.datetime.today() > time_for_removing


class SyslogAuditHandler(BaseHandler):
    """ Основной обработчик отправки событий аудита на syslog сервер. """

    # Проверка выполнения отложенного конфигурирования параметров.
    _fix_time = False

    def _is_enabled(self):
        """
        Проверяет включена ли в настройках отправки событий аудита на syslog сервер

        @return: True - в настройках включена отправка событий на syslog сервер, иначе False.
        """
        from uds.core.util.Config import GlobalConfig
        return GlobalConfig.AUDIT_SYSLOG_ENABLE.getBool(force=True)

    def _process_record(self, event: str, args: tuple):
        """ Обработка записи события. """
        if self._need_updating_time():
            if not self.set_params():
                return False

        version, *params = args[0]
        record = RecordSyslog(
            event=event,
            version=version,
            data=params
        )

        log_record = LogRecord('', logging.INFO, '', 0, record.get_message(), (), None)

        assert isinstance(self.syslog_handler, SysLogHandler)
        self.syslog_handler.handle(log_record)

    def set_params(self):
        from uds.core.util.Config import GlobalConfig

        address = (GlobalConfig.AUDIT_SYSLOG_HOST.get(force=True), GlobalConfig.AUDIT_SYSLOG_PORT.getInt(force=True))

        socktype = get_socktype(GlobalConfig.AUDIT_SYSLOG_SOCKET.get(force=True))

        facility_name = GlobalConfig.AUDIT_SYSLOG_FACILITY.get(force=True)
        facility = int(SysLogHandler.facility_names[facility_name])

        self.syslog_handler = SysLogHandler(address, socktype=socktype, facility=facility)

        log_formatter = logging.Formatter('%(message)s\n')
        self.syslog_handler.setFormatter(log_formatter)

        # Включение свойства append_nul приводит к добавлению символа пустого
        # байта в конце сообщения, из-за которого graylog не распознает
        # необходимые свойства
        self.syslog_handler.append_nul = False

        self._fix_time = dt.datetime.now()

        return True

    def _need_updating_time(self):
        if self._fix_time is False:
            return True

        time_for_removing = self._fix_time + dt.timedelta(seconds=30)

        return dt.datetime.today() > time_for_removing


class DatabaseHandler(BaseHandler):
    """Класс обработчика записи в БД события системы аудита"""

    def _is_enabled(self):
        """
        Проверяет включена ли в настройках запись событий аудита в БД

        @return: True - в настройках включена запись в БД, иначе False
        """
        from uds.core.util.Config import GlobalConfig
        return GlobalConfig.AUDIT_DB_LOG_ENABLE.getBool(force=True)

    def _process_record(self, event, args):
        """Обработка записи события в БД"""

        version, *params = args[0]

        from termidesk.models.AuditEvent import AuditEvent

        db_record = AuditEvent(
            timestamp=datetime.now(timezone.utc),
            event=event,
            version=version,
            data=json.dumps(params, ensure_ascii=False)
        )
        db_record.save()
