"""
Модуль содержит логику генерации описаний событий аудита в различных форматах
"""
# -*- encoding: utf-8 -*-


class AuditEventTemplates:
    """
    Вспомогательный класс, который генерирует словарь-справочник с шаблонами описаний
    для каждой версии события аудита.
    """
    def __init__(self):
        """Инициализация объекта класса"""
        self.templates_map = {}

    def _add_description_template(self, event_class):
        """Добавление шаблона описания в общий словарь-справочник"""
        # Для каждой версии события аудита необходимо хранить свой шаблон описания события
        # Доступ к версии возможен только через защищенный атрибут, так как `event_class`
        # является классом, а не экземпляром класса (объектом)
        versioned_template = {str(event_class._VERSION): event_class.DESCRIPTION_TEMPLATE}

        # У каждого события может быть несколько версий, поэтому каждой версии каждого события
        # соответствует единственная строка с шаблоном описания события
        event_templates = self.templates_map.get(event_class._NAME, {})
        event_templates.update(versioned_template)
        self.templates_map.update({event_class._NAME: event_templates})

    def get_templates_map(self):
        """Генерация словаря-справочника для всех установленных событий аудита"""
        from termidesk.audit.events import BaseEventMixin
        for subclass in BaseEventMixin.__subclasses__():
            self._add_description_template(subclass)

        return self.templates_map


class AuditEventTextRenderer:
    """Класс для генерации описания события аудита в простом текстовом формате"""
    template_not_found = 'There is no template for event "{name}" in templates map'

    def __init__(self, templates_map=None):
        """Инициализация объекта"""

        self._event_templates_map = AuditEventTemplates().get_templates_map()
        # Дополнение/обновление словаря шаблонов описаний
        if isinstance(templates_map, dict):
            self._event_templates_map.update(templates_map)

    def render(self, event_obj):
        """Получение строки с описанием события аудита"""
        import json
        event_params = {
            "name": event_obj.event,
            "timestamp": event_obj.timestamp,
            "version": event_obj.version,
            "args": json.loads(event_obj.data),
        }
        event_templates = self._event_templates_map.get(event_obj.event, {})
        event_template = event_templates.get(event_obj.version, self.template_not_found)

        event_description = event_template.format(**event_params)

        return event_description
