# -*- coding: utf-8 -*-

from __future__ import unicode_literals

import logging

import dns
from django.utils.translation import ugettext_noop as _
from termidesk.auths.Authenticator2FA.views import Views
from termidesk.core.auths.Authenticator import AUTH_METHODS, BaseAuthenticator2Fa
from uds.core.managers import cryptoManager
from uds.core.ui import gui
from uds.core.util.request import getRequest
from uds.core.util.State import State
from uds.models import Authenticator as dbAuthenticator

__updated__ = '2021-09-09'

logger = logging.getLogger(__name__)


class Authenticator2FA(BaseAuthenticator2Fa):
    typeName = _('Внутренняя БД (2FA, эксперим.)')
    typeType = 'InternalDBAuth2Fa'
    typeDescription = _('Внутренняя БД (2FA, эксперим.)')
    iconFile = 'auth.png'

    # If we need to enter the password for this user
    needsPassword = True

    # This is the only internal source
    isExternalSource = False

    differentForEachHost = gui.CheckBoxField(label=_('Разные пользователи для хостов'), order=1,
                                             tooltip=_('Для пользователя, входящего с разных хостов '
                                                       'будут создаваться разные учетные записи'),
                                             defvalue="false", rdonly=True, tab=gui.ADVANCED_TAB)
    reverseDns = gui.CheckBoxField(label=_('Обратный просмотр DNS'), order=2,
                                   tooltip=_('Для IP подключающихся хостов будет производится обратный '
                                             'просмотр DNS'),
                                   defvalue="false", rdonly=True, tab=gui.ADVANCED_TAB)
    acceptProxy = gui.CheckBoxField(label=_('Разрешить проксирование'), order=3,
                                    tooltip=_('Запросы через прокси сервер будут от имени пересылаемого '
                                              '(FORWARDED) IP '
                                              '(Осторожно! Возможен прием запросов с внутренними IP '
                                              'из Интернет)'),
                                    tab=gui.ADVANCED_TAB)

    def initialize(self, values):
        if values is None:
            return

    def getIp(self):
        # pylint: disable=maybe-no-member
        ip = getRequest().ip_proxy if self.acceptProxy.isTrue() else getRequest().ip
        if self.reverseDns.isTrue():
            try:
                return str(dns.resolver.query(dns.reversename.from_address(ip), 'PTR')[0])
            except Exception:
                pass
        return ip

    def transformUsername(self, username):
        if self.differentForEachHost.isTrue():
            newUsername = self.getIp() + '-' + username  # pylint: disable=maybe-no-member
            # Duplicate basic user into username.
            auth = self.dbAuthenticator()
            # "Derived" users will belong to no group at all, because we will extract groups from "base" user
            # This way also, we protect from using forged "ip" + "username",
            # because those will belong in fact to no group
            # and access will be denied
            try:
                usr = auth.users.get(name=username, state=State.ACTIVE)
                parent = usr.uuid
                usr.id = usr.uuid = None  # Empty "key" fields for replication
                if usr.real_name.strip() == '':
                    usr.real_name = usr.name
                usr.name = newUsername
                usr.parent = parent
                usr.save()
            except Exception:
                pass  # User already exists
            username = newUsername

        return username

    def authenticate(self, username, credentials, groupsManager):
        logger.debug('Username: {0}, Password: {1}'.format(username, credentials))
        auth = self.dbAuthenticator()
        try:
            try:
                usr = auth.users.get(name=username, state=State.ACTIVE)
            except Exception:
                return False

            if usr.parent is not None and usr.parent != '':  # Direct auth not allowed for "derived" users
                return False

            # Internal Db Auth has its own groups, and if it active it is valid
            if usr.password == cryptoManager().hash(credentials):
                #  hashlib.sha1(credentials.encode('utf-8')).hexdigest():
                groupsManager.validate([g.name for g in usr.get_auth_groups().all()])
                return True
            return False
        except dbAuthenticator.DoesNotExist:  # @UndefinedVariable
            return False

    def createUser(self, usrData):
        pass

    @staticmethod
    def test(env, data):
        return [True, _("Internal structures seems ok")]

    def check(self):
        return _("All seems fine in the authenticator.")

    def __str__(self):
        return "Internal DB Authenticator 2FA"

    def get_custom_view(self):
        return Views(self.dbAuthenticator())

    def get_auth_methods(self):
        return AUTH_METHODS
