from functools import wraps

from django.http.response import HttpResponseForbidden
from django.utils.translation import ugettext as _
from uds import REST
from uds.core.auths.auth import ROOT_ID, USER_KEY, get_language, getRootUser
from uds.core.util.Config import GlobalConfig
from uds.models.User import User

USER_KEY_2FA = 'uk_2fa'


def login_2fa_first_step(request, user):
    """
    Сохранение промежуточных сессионных параметров для двухфакторной аутентификации,
    которые позволяют перейти ко второму этапу аутентификации.
    """

    user.updateLastAccess()
    request.session.clear()
    request.session[USER_KEY_2FA] = user.id

    request.session.set_expiry(GlobalConfig.USER_SESSION_LENGTH.getInt(force=True))
    return True


def get_user_from_session(request):
    """
    Получение пользователя из сохраненных данных
    с первого шага двухфакторной аутентификации.
    """
    user = request.session.get(USER_KEY_2FA)

    if user is not None:
        try:
            if user == ROOT_ID:
                user = getRootUser()
            else:
                user = User.objects.get(pk=user)
        except User.DoesNotExist:
            user = None

    return user


def set_login_session_data(request, user):
    """
    Установка пользовательских данных в основные сессионные параметры.
    """

    if user.id != ROOT_ID:  # If not ROOT user (this user is not inside any authenticator)
        manager_id = user.manager.id
    else:
        manager_id = -1

    user.updateLastAccess()
    request.session.clear()
    request.session[USER_KEY] = user.id
    # Ensures that this user will have access through REST api if logged in through web interface
    REST.Handler.storeSessionAuthdata(request.session, manager_id, user.name, get_language(), user.is_admin,
                                      user.staff_member)
    request.session.set_expiry(GlobalConfig.USER_SESSION_LENGTH.getInt(force=True))
    return True


def check_access_on_second_step(method):
    """ Проверка доступа на второй шаг двух-факторной аутентификации. """

    @wraps(method)
    def _wrapped_view(self, request, *args, **kwargs):
        if request.session.get(USER_KEY_2FA) is None:
            return HttpResponseForbidden(_('Forbidden'))
        return method(self, request, *args, **kwargs)

    return _wrapped_view
