from termidesk.core.auths.exceptions import UserLoginAttemptsExceeded
from termidesk.core.util.cache import timed_lru_cache
from uds.core.auths.auth import authLogLogin
from uds.core.util.Cache import Cache
from uds.core.util.Config import GlobalConfig
from uds.core.util.State import State
from uds.models.Authenticator import Authenticator


@timed_lru_cache()
def get_default_login_block():
    return GlobalConfig.LOGIN_BLOCK.getInt(force=True)


@timed_lru_cache()
def get_default_max_attempts():
    return GlobalConfig.MAX_LOGIN_TRIES.getInt(force=True)


class LoginAttemptsManager:
    """
    Сервис для работы с кешем,
    в котором хранятся попытки входа пользователя в систему.
    """

    def __init__(self, authenticator: Authenticator, username: str, login_block_time=None,
                 max_login_attempts=None):
        self.cache = Cache('auth')
        self.authenticator = authenticator
        self.username = username

        self.cache_key = str(authenticator.id) + username
        self.tries = self.cache.get(self.cache_key)
        if self.tries is None:
            # Нормируем значение неудачных попыток входа
            self.tries = 0

        self.__login_block = login_block_time if login_block_time is not None else get_default_login_block()
        self.__max_attempts = \
            max_login_attempts if max_login_attempts is not None else get_default_max_attempts()

    def _clear_expired_block(self, user):
        if user is None:
            # Пользователь не задан, разблокировать некого
            # Скорее всего, попытка входа была неуспешная
            return

        checks = [
            # В кэше не было записей о неудачных попытках или они истекли
            self.tries == 0,

            user.state == State.USER_TEMPORARY_BLOCKED,
        ]

        if all(checks):
            user.setState(State.ACTIVE)

    def validate(self, request, user=None):
        """ Проверка допустимого состояния попыток входа. """

        self._clear_expired_block(user)

        if self.authenticator.getInstance().blockUserOnLoginFailures is False:
            # Аутентификатор не предполагает блокировку пользователя
            # при неудачных попытках входа
            return

        if self.is_exceeded():
            # Too many authentication errors. User temporarily  blocked.
            if user is not None:
                user.setState(State.USER_TEMPORARY_BLOCKED)

            authLogLogin(request, self.authenticator, self.username, 'Temporarily blocked')
            raise UserLoginAttemptsExceeded(
                'Превышено количество неудачных попыток входа. Пользователь временно заблокирован.'
            )

    def save(self):
        self.cache.put(self.cache_key, self.tries, self.__login_block)

    def clear(self):
        self.cache.remove(self.cache_key)

    def increase(self):
        self.tries += 1
        self.save()

    def is_exceeded(self):
        return self.tries >= self.__max_attempts

    def block_user(self):
        """
        Перевод пользователя во временно заблокированное состояние
        за счет установки максимального количества неверных попыток входа.
        """
        self.tries = self.__max_attempts
        self.save()
