import logging
import sys

from django.http import HttpResponseBadRequest
from django.shortcuts import get_object_or_404, render
from termidesk.audit.events import WEB_USER_LOGIN_V1
from termidesk.core.auths.helpers import get_redirect_for_2fa_method
from termidesk.core.auths.auth_2fa import login_2fa_first_step
from termidesk.core.auths.login_manager import LoginAttemptsManager
from termidesk.web.mods import chooseSuccessRedirectForUser
from uds.core.auths.auth import ROOT_ID, authenticate, authLogLogin, webLogin
from uds.core.util.stats import events
from uds.core.ui import theme
from uds.core.util.Config import GlobalConfig
from uds.models import Authenticator, User

logger = logging.getLogger(__name__)

TEMPLATE_FOLDER = 'termidesk/auths/2fa/'


class CommonViews:
    def __init__(self, auth):
        self._auth = auth

    def login(self, request):
        """ Аутентификация пользователя с полной валидацией полей. """
        if request.method != 'POST':
            return HttpResponseBadRequest()

        authenticator = get_object_or_404(Authenticator, uuid=self._auth.uuid)
        auth_instance = authenticator.getInstance()
        error = ''
        try:
            username = request.POST.get('user')
            password = request.POST.get('password')

            # предварительный блок получения пользователя
            try:
                user = authenticator.users.get(name=username)
            except User.DoesNotExist:
                user = None

            # если аккаунт пользователя не найден,
            # то и разблокировать его из админ.панели
            # раньше срока истечения блокировки не получится
            login_attempts = LoginAttemptsManager(authenticator, username)
            login_attempts.validate(request, user)

            user = authenticate(username, password, authenticator)
            logger.debug('User: {}'.format(user))

            if user is None:
                logger.debug("Invalid credentials for user {0}".format(username))
                login_attempts.increase()
                authLogLogin(request, authenticator, username, 'Invalid credentials')
                raise ValueError('Неверно введены учетные данные')
            else:
                # Add an login event
                events.addEvent(authenticator, events.ET_LOGIN,
                                username=username,
                                srcip=request.ip)
                events.addEvent(authenticator, events.ET_PLATFORM,
                                platform=request.os.OS,
                                browser=request.os.Browser,
                                version=request.os.Version)
                login_attempts.clear()

                if (GlobalConfig.SUPER_USER_ALLOW_WEBACCESS.getBool(True) and
                        username == GlobalConfig.SUPER_USER_LOGIN.get(True) and
                        password == GlobalConfig.SUPER_USER_PASS.get(True)):
                    response = chooseSuccessRedirectForUser(user)

                    webLogin(request, response, user)
                    authLogLogin(request, authenticator, user.name)
                    return response

                logger.debug('User {} has logged in'.format(user))

                type_2fa = request.POST.get('type_2fa')
                response = get_redirect_for_2fa_method(authenticator.uuid, type_2fa)

                auth_name = authenticator.name
                if user.id == ROOT_ID:
                    auth_name = ''
                event_params = dict(authenticator=auth_name, username=user.name, ip=request.ip)
                evt = WEB_USER_LOGIN_V1(**event_params).generate_event()
                sys.audit(*evt)

                login_2fa_first_step(request, user)
                return response

        except Exception as err:
            logger.exception(error)
            error = err

        form = auth_instance.getHtml(request)
        return render(request,
                      theme.template('custom_login.html'),
                      {'form': form, 'error': error, 'id_auth': self._auth.uuid})
