# -*- encoding: utf-8 -*-
import six
import time
from functools import wraps
import logging

from django.db import connection, reset_queries

from termidesk.management import SENSITIVE_DATA_MASK

logger = logging.getLogger(__name__)


def to_str(str_):
    # type: (any) -> str
    """
    Convert unicode, bytes, str and integer to str
    """
    allowed_types = (six.text_type, six.binary_type, int)

    if not isinstance(str_, allowed_types):
        raise TypeError(
            "The argument must be one of {al_t}, not {t}".format(al_t=allowed_types, t=type(str_)))

    if isinstance(str_, int):
        str_ = str(str_)

    elif not isinstance(str_, str):
        if six.PY3:
            str_ = str_.decode('utf-8')
        else:
            str_ = str_.encode('utf-8')

    return str_


def query_profiler(logger=logger):
    """
    Вспомогательный декоратор для оценки количества и продолжительности
    запросов к БД в методе/функции. Для более точных результатов не должен
    использоваться в вызовах вложенных методов/функций.

    :param logger: Объект с помощью которого выводить статистику.
        Профилирование будет выполняться при уровне `logging.DEBUG`.
    :type logger: class: `logging.Logger`, optional
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if logger.isEnabledFor(logging.DEBUG):
                # Сбрасываем историю подключений к БД
                reset_queries()

                time_start = time.monotonic()
                result = func(*args, **kwargs)
                time_end = time.monotonic()
                time_delta = int((time_end-time_start)*1000)

                name = func.__name__
                queries = connection.queries
                qtime = int(sum([float(q['time']) for q in queries]) * 1000)
                qpercent = (qtime/time_delta)*100

                log_tpl = "[{}] took {}ms DB queries: {} Query time: {}ms ({:.1f}%)"
                logger.debug(log_tpl.format(name, time_delta, len(queries), qtime, qpercent))
                return result

            return func(*args, **kwargs)
        return wrapper
    return decorator


# Список имен ключей, которые содержат чувствительные данные
SENSITIVE_DATA_KEYS_LIST = ['password', ]


def hide_sensitive_data(data_object):
    """Скрывает (заменяет на '*') чувствительные данные в полученном объекте"""

    # Если обрабатываемый объект это словарь, то пробуем найти ключи,
    # которые могут указывать на чувствительные данные
    if isinstance(data_object, dict):
        _data_object = {}

        # Список регистронезависимых ключей
        _sensitive_data_keys_list = [x.lower() for x in SENSITIVE_DATA_KEYS_LIST]

        for key, value in data_object.items():
            if isinstance(value, dict):
                # Если значение данного ключа это словарь, то запускаем рекурсию
                _data_object[key] = hide_sensitive_data(value)

            elif key.lower() in _sensitive_data_keys_list:
                # Проверяем ключ в регистронезависимом режиме
                _data_object[key] = SENSITIVE_DATA_MASK

            else:
                # Переносим как есть
                _data_object[key] = value
    else:
        # Так как неизвестно, какие объекты находятся внутри,
        # то для подстраховки выполняется глубокое копирование
        import copy
        _data_object = copy.deepcopy(data_object)

    return _data_object


def is_2fa_feature_enabled():
    from uds.core.util.Config import GlobalConfig
    return GlobalConfig.EXPERIMENTAL_2FA_ENABLED.getInt(force=True) != 0


def max_session_time_enabled():
    # TSRV-447 Функция возвращает True если включен параметр `Ограничение времени сессии пользователя`
    from uds.core.util.Config import GlobalConfig
    return GlobalConfig.EXPERIMENTAL_MAX_SESSION_TIME_ENABLED.getInt(force=True) != 0
