# -*- coding: utf-8 -*-
"""
Created on 16 янв. 2019 г.

@author: dvd
"""
from __future__ import unicode_literals

import os
import base64
import datetime
import json

import six
from Crypto.Signature import PKCS1_PSS
from Crypto.Hash import SHA256
from Crypto.PublicKey import RSA
import binascii

import logging

from termidesk.core.util.utils import to_str

logger = logging.getLogger(__name__)


def generate_salt():
    # type: () -> str
    """
    Generates salt value - 16 random bytes
    """
    salt = os.urandom(16)
    return str(base64.b64encode(salt))


class TokenEncoder(object):
    """
    Uses to create rsa signature and sha256 digest for data
    """
    @staticmethod
    def _formatDate(d):
        # type: (datetime.datetime) -> str
        """
        Formats datetime.datetime object to str-format like "20200721131401" for 'July 21, 2020 13:14.01'
        """
        return d.strftime("%Y%m%d%H%M%S")

    def __init__(self, sign_key, lifetime=20):
        # type: (object, int) -> ()
        """
        Inits object with private key and timeout value

        @param sign_key: private rsa key for signing
        @param lifetime: timeout for sign validity
        """
        self._sign_key = sign_key
        self._lifetime = lifetime

    def encode(self, data):
        # type: (str) -> str
        """
        Calculate the sign for 'data' on 'self._sign_key' rsa private key

        @param data: message to signing
        @return: base64 encoded json string of dictionary with keys:
            'salt': 16 random bytes
            'digest': digests algorithm (default, sha256)
            'validFrom': utc date sign validity began
            'validTo': utc date sign validity ended
            'data': clear data require to sign
            'signature': sign for 'data'
        """

        d = {
            'salt': generate_salt(),
            'digest': 'sha256',
            'validFrom': self._formatDate(datetime.datetime.utcnow()),
            'validTo': self._formatDate(
                datetime.datetime.utcnow() + datetime.timedelta(
                    seconds=self._lifetime
                )
            ),
            'data': data
        }
        digest_str = json.dumps(d, sort_keys=True)

        if six.PY3 and isinstance(digest_str, str):
            digest_str = digest_str.encode('utf-8')

        # logger.info("digest_str: {}".format(digest_str))
        key = RSA.importKey(self._sign_key)
        h = SHA256.new(digest_str)

        signer = PKCS1_PSS.new(key)
        signature = signer.sign(h)

        d.update({'signature': to_str(binascii.hexlify(signature))})

        # logger.info("signature: {}".format(d['signature']))
        jsoned_d = json.dumps(d)

        if six.PY3 and isinstance(jsoned_d, str):
            jsoned_d = jsoned_d.encode('utf-8')

        return to_str(base64.b64encode(jsoned_d))


class TokenDecoder(object):
    """
    Uses to verify rsa signature for data
    """

    def __init__(self, verification_key, lifetime=20):
        # type: (object, int) -> ()
        """
        Inits object with private key and timeout value

        @param verification_key: public rsa key corresponding private key used to create signature.
                                 Used to verify the signature.
        @param lifetime: timeout for sign validity
        """
        self._verification_key = verification_key
        self._lifetime = lifetime

    @staticmethod
    def _parseDate(d):
        # type: (str) -> datetime.datetime
        """
        Parses datetime string to 'datetime.datetime' object
        @param d: string like "20200721131401" for 'July 21, 2020 13:14.01'
        @return: 'datetime.datetime' object represented to datetime string
        """
        return datetime.datetime.strptime(d, '%Y%m%d%H%M%S')

    def decode(self, ticket):
        # type: (str) -> str
        """
        Verify the sign for 'data' with 'self._sign_key' rsa public key

        @param ticket: base64 encoded json string of dictionary with keys:
            'salt': 16 random bytes
            'digest': digests algorithm (default, sha256)
            'validFrom': utc date sign validity began
            'validTo': utc date sign validity ended
            'data': clear data require to sign
            'signature': sign for 'data'
        @return: signing message if signature verified successful,
                 ValueError if a signature is invalid or ticket lifetime is expired.
        """

        try:
            debase64 = base64.b64decode(ticket)
            decoded = json.loads(to_str(debase64))

            # logger.info("signature: {}".format(decoded['signature']))

            signature = binascii.unhexlify(decoded['signature'])
            del(decoded['signature'])

            key = RSA.importKey(self._verification_key)
            digest_str = json.dumps(decoded, sort_keys=True)
            # logger.info("digest_str: {}".format(digest_str))
            if six.PY3 and isinstance(digest_str, str):
                digest_str = digest_str.encode('utf-8')

            h = SHA256.new(digest_str)

            verifier = PKCS1_PSS.new(key)
            if not verifier.verify(h, signature):
                raise ValueError('Invalid ticket signature')

            if not (
                self._parseDate(decoded['validFrom']) <=
                datetime.datetime.utcnow() <=
                self._parseDate(decoded['validTo'])
            ):
                raise ValueError('Ticket lifetime expired')

            return decoded['data']

        except Exception as e:
            import traceback
            print(traceback.format_exc())
            logger.exception(str(e))
