# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from termidesk.models import AuditEvent
from uds.core.jobs.Job import Job
from server.celery import app as celery
from uds.core.util.Config import GlobalConfig

import logging
import datetime as dt

logger = logging.getLogger(__name__)


class DBAuditLogCleaner(celery.Task, Job):
    """
    Удаление из БД устаревших записей о событиях аудита
    """
    TIMEDELTA_HOURS = 2
    frecuency = 24*60*60  # каждые 24 часа
    friendly_name = 'DB audit logs cleaner'

    celery_name = 'J.termidesk.core.workers.DBAuditLogCleaner.DBAuditLogCleaner'

    name = celery_name

    # Возможность выполнения в celery
    celery_compatible = True

    def __init__(self, environment=None):
        """
        Инициализация окружения для задачи.
        :param environment:
        Для TaskManager, environment инициализируется из модели.
        В celery модель не привязана к задаче, поэтому для обеспечения
        уникальности инициализируется environment
        c celery_name в качестве unique_key
        """
        if environment is None:
            from uds.core.Environment import Environment
            environment = Environment(self.celery_name)

        super(DBAuditLogCleaner, self).__init__(environment)

    def __call__(self, *args, **kwargs):
        """
        Необходимо переопределить __call__(), если метод self.run()
        не принимает аргументы.
        В противном случае выполнение задачи вызывает исключение.
        :param args:
        :param kwargs:
        :return:
        """
        return self.run()

    def run(self):
        """
        Выполнение удаления из БД устаревших записей о событиях аудита

        Из таблицы UserService выбираются записи старше `AUDIT_DB_LOG_DELETE_TIMEOUT` дней и удаляются.
        Если `AUDIT_DB_LOG_DELETE_TIMEOUT` равно 0, то записи не удаляются.
        За один раз могут быть удалены не более `AUDIT_DB_DELETE_RECORDS_AT_ONCE` записей.
        Если количество записей превышает ограничение, то процедура удаления повторяется,
        пока все записи не будут удалены.
        """
        delete_timeout = GlobalConfig.AUDIT_DB_LOG_DELETE_TIMEOUT.getInt(force=True)

        # Если `AUDIT_DB_LOG_DELETE_TIMEOUT` равно 0, то записи не удаляются.
        if delete_timeout == 0:
            logger.debug('Records will not delete, because `AUDIT_DB_LOG_DELETE_TIMEOUT` = {delete_timeout}'
                         .format(delete_timeout=delete_timeout))
            return

        time_for_removing = dt.datetime.utcnow() - dt.timedelta(days=delete_timeout)

        outdated_db_logs = AuditEvent.objects.get_outdated_records(time_for_removing.isoformat())

        logger.info('{count} outdated (before {time}) audit records will be delete from DB.'
                    .format(count=outdated_db_logs.count(), time=time_for_removing.isoformat()))

        while outdated_db_logs.count():
            # Для регулирования нагрузки на БД за один раз удаляется ограниченное количество записей.
            # Максимальное количество записей для удаления указано в `AUDIT_DB_DELETE_RECORDS_AT_ONCE`

            delete_limit = GlobalConfig.AUDIT_DB_DELETE_RECORDS_AT_ONCE.getInt(force=True)

            deleted = AuditEvent.objects.delete_with_limit(outdated_db_logs, delete_limit)[0]

            logger.info('Deleted {count} record(s).'.format(count=deleted))

            # Обновление перечня устаревших записей
            outdated_db_logs = AuditEvent.objects.get_outdated_records(time_for_removing.isoformat())


# Регистрация класса в качестве задачи в реестре celery.
celery.tasks.register(DBAuditLogCleaner)
