# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from io import StringIO

from celery.utils.log import get_task_logger
from django.contrib.sessions.models import Session
from django.core.management import call_command
from django.utils import timezone
from server.celery import app as celery
from uds.core.jobs.DelayedTask import DelayedTask
from uds.core.jobs.Job import Job

CHUNK = 1000
COMMAND_NAME = 'tdsk_clearsessions'

TAG = 'clear-session'

logger = get_task_logger(__name__)


def _run_command(chunk):
    sessions_count = Session.objects.filter(
        expire_date__lt=timezone.now()
    ).count()
    logger.debug("Found {} expired HTTP sessions.".format(sessions_count))

    if sessions_count > chunk:
        suggested_time = 300

        delayed_task = DelayedExpiredSessionsCleaner()
        delayed_task.register(suggested_time, TAG)
        logger.debug("Registered delayed task for additional expired HTTP sessions cleanup in {}s."
                     .format(suggested_time))

    out = StringIO()
    call_command(
        COMMAND_NAME,
        '--chunk={}'.format(chunk),
        stdout=out
    )
    logger.info(out.getvalue())


class ExpiredHttpSessionsCleaner(celery.Task, Job):
    """
    Class for cleaning up stale sessions.
    """

    frecuency = 86400
    friendly_name = 'Expired http sessions cleaner'

    celery_name = 'J.termidesk.core.workers.ExpiredHttpSessionsCleaner.ExpiredHttpSessionsCleaner'

    name = celery_name

    # Возможность выполнения в celery
    celery_compatible = True

    def __init__(self, environment=None, chunk=CHUNK):
        """
        Инициализация окружения для задачи.
        :param environment:
        Для TaskManager, environment инициализируется из модели.
        В celery модель не привязана к задаче, поэтому для обеспечения
        уникальности инициализируется environment
        c celery_name в качестве unique_key
        """
        self.chunk = chunk

        if environment is None:
            from uds.core.Environment import Environment
            environment = Environment(self.celery_name)

        super(ExpiredHttpSessionsCleaner, self).__init__(environment)

    def __call__(self, *args, **kwargs):
        """
        Необходимо переопределить __call__(), если метод self.run()
        не принимает аргументы.
        В противном случае, выполнение задачи вызывает исключение.
        :param args:
        :param kwargs:
        :return:
        """
        return self.run()

    def run(self):
        _run_command(self.chunk)


class DelayedExpiredSessionsCleaner(DelayedTask):
    '''
    This delayed task is responsible of cleaning expired session.
    '''

    celery_task_name = 'T.termidesk.core.workers.ExpiredHttpSessionsCleaner.DelayedExpiredSessionsCleaner'

    def __init__(self, chunk=CHUNK):
        self.chunk = chunk

        super(DelayedExpiredSessionsCleaner, self).__init__()

    @celery.task(name=celery_task_name)
    def run_in_celery(self):
        self.execute()

    def run(self):
        _run_command(self.chunk)
