# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from uds.core.jobs.Job import Job
from server.celery import app as celery
from uds.models.UserService import UserService
from uds.core.jobs.DelayedTaskRunner import DelayedTaskRunner
from uds.core.util.State import State
from uds.core.managers import taskManager
from uds.core.managers.userservice.opchecker import (USERSERVICE_TAG,
                                                     UserServiceOpChecker)
import logging
import datetime as dt

logger = logging.getLogger(__name__)


class FrozenJobs(celery.Task, Job):
    """
    Class for finding frozen workplaces in "REMOVING" or "CANCELING" states.
    """
    TIMEDELTA_HOURS = 2
    frecuency = 3600
    friendly_name = 'Frozen jobs cleaner'

    celery_name = 'J.termidesk.core.workers.FrozenJobs.FrozenJobs'

    name = celery_name

    # Возможность выполнения в celery
    celery_compatible = True

    def __init__(self, environment=None):
        """
        Инициализация окружения для задачи.
        :param environment:
        Для TaskManager, environment инициализируется из модели.
        В celery модель не привязана к задаче, поэтому для обеспечения
        уникальности инициализируется environment
        c celery_name в качестве unique_key
        """
        if environment is None:
            from uds.core.Environment import Environment
            environment = Environment(self.celery_name)

        super(FrozenJobs, self).__init__(environment)

    def __call__(self, *args, **kwargs):
        """
        Необходимо переопределить __call__(), если метод self.run()
        не принимает аргументы.
        В противном случае, выполнение задачи вызывает исключение.
        :param args:
        :param kwargs:
        :return:
        """
        return self.run()

    def run(self):
        time_for_removing = dt.datetime.today() - dt.timedelta(
            hours=self.TIMEDELTA_HOURS)
        user_services = UserService.objects.filter(
            state__in=[State.REMOVING, State.CANCELING],
            state_date__lt=time_for_removing
        )

        for user_service in user_services:
            tag = USERSERVICE_TAG + user_service.uuid
            if DelayedTaskRunner.runner().checkExists(tag):
                continue
            ci = user_service.getInstance()

            taskManager().register_task(UserServiceOpChecker(user_service),
                                        ci.suggestedTime, tag)

            state_map = {
                State.REMOVING: "REMOVING",
                State.CANCELING: "CANCELING"
            }
            msg = 'Restarted OpChecker for user service {0} stuck in state {1}'
            logger.warning(msg.format(
                user_service.friendly_name,
                state_map.get(user_service.state, user_service.state))
            )


# Регистрация класса в качестве задачи в реестре celery.
celery.tasks.register(FrozenJobs)
