# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from termidesk.core.util.utils import to_str
from termidesk.management.commands.common.subcommands.subcommand import Subcommand
from termidesk.management.commands.common.subcommands.mixins.create import CreateMixin
from abc import ABCMeta, abstractmethod
import six


class CreateSubcommand(Subcommand, CreateMixin):
    __metaclass__ = ABCMeta
    # Имя сервиса в родительном падеже, для вывода сообщений
    name_of_service = ''

    # Типы сервиса
    types = []

    # Базовые аргументы для модели
    BASE_ARGS = []

    SUBCOMMAND = 'create'

    def handle_subcommand(self, options, stdout, stderr):
        service_type = options['type']

        if options['list_params']:
            self._show_params_for_current_type(service_type, stdout)
            return

        self.validate_base_data(options)

        params = self._get_params_with_default_value(
            service_type,
            self._get_parse_params(options)
        )

        test_success, test_msg = self.testing_service(service_type, params)
        loud_work = not options['quiet']
        if test_success:
            if options['test']:
                if loud_work:
                    stdout.write(
                        self.style.SUCCESS('Проверка прошла успешно')
                    )
                return
        else:
            # Приведение сообщения к строковому виду необходимо для обработки случая,
            # когда ошибка приходит в виде байт строки (задача TSRV-322)
            stderr.write(self.style.ERROR(to_str(test_msg)))
            return

        try:
            main_args = {arg: options[arg] for arg in self.BASE_ARGS
                         if arg in options}
            self.create_service(options['type'], main_args, params)
            if loud_work:
                self.show_success(stdout, self.name_of_service)
        except Exception as error:
            self.show_error(stdout, error, self.name_of_service)

    def add_arguments_subcommand(self, parser):
        """ Установка параметров для команды создания сервиса. """
        parser.add_argument(
            '--type', choices=self.types, required=True,
            help='Тип {}'.format(self.name_of_service)
        )

        parser.add_argument(
            '--params', nargs='*', type=str,
            help=('Параметры {}. '.format(self.name_of_service)
                  + 'Формат: имя1=значeние1 имя2=значение2 ...'),
            default=[]
        )

        parser.add_argument(
            '--list', action='store_true', dest='list_params', default=False,
            help='Показать параметры {}'.format(self.name_of_service),
        )

        parser.add_argument(
            '--test', action='store_true', dest='test', default=False,
            help='Проверка параметров без создания {}'.format(
                self.name_of_service
            ),
        )

        parser.add_argument(
            '--quiet', action='store_true', dest='quiet', default=False,
            help='Выводить только сообщения об ошибках',
        )

    @abstractmethod
    def get_service_factory(self):
        """
        Получение фабрики для работы с текущим сервисом.
        Подклассы должны обязательно содержать этот метод.
        """

    @abstractmethod
    def create_service(self, service_type, main_args, params):
        """
        Базовая логика по созданию сервиса. Подклассы должны обязательно
        содержать этот метод.
        :param service_type: Выбранный тип сервиса
        :param main_args: Основные аргументы для модели сервиса
        :param params: Список переданных дополнительных параметров
        :return:
        """

    @abstractmethod
    def validate_base_data(self, data):
        """
        Валидация базовых данных. Подклассы должны обязательно
        содержать этот метод.
        :param data: Базовые данные для проверки
        :return:
        """

    def testing_service(self, service_type, data):
        """
        Проверка корректности данных для заданного типа протокола.
        :param str service_type:
        :param dict data:
        :return: list
        """
        from uds.core.Environment import Environment

        service = self.get_lookup_service_by_type(service_type)
        return service.test(Environment.getTempEnv(), data)

    @staticmethod
    def _get_parse_params(options):
        """
        Разбор параметров, которые переданы через --params.
        :param dict options:
        :return:
        """
        data = {}
        for field in options['params']:
            field = field.decode('utf-8') if six.PY2 else field
            pair = field.split('=')
            if len(pair) == 2:
                name, value = pair
            else:
                name, value = pair[0], '='.join(pair[1:])

            data[name] = value

        return data

    def _get_params_with_default_value(self, prov_type, passed_params):
        """
        Установка дефолтных значений в те параметры,
        которые не были переданы.
        :param str prov_type:
        :param dict passed_params:
        :return:
        """
        def_values = self.get_lookup_service_by_type(
            prov_type).guiDescription()

        full_data = {}
        for param in def_values:
            full_data[param['name']] = (passed_params[param['name']]
                                        if param['name'] in passed_params
                                        else param['gui']['defvalue'])
        return full_data

    def _show_params_for_current_type(self, service_type, stdout):
        gui = self.get_lookup_service_by_type(service_type).guiDescription()
        header = "Тип {} поддерживает следующие параметры:".format(
            service_type)
        stdout.write(
            self.style.SUCCESS(header)
        )
        for param in gui:
            stdout.write(
                self.style.WARNING("    {}".format(param['name']))
            )

            tooltip = param['gui']['tooltip']
            stdout.write("        {}".format(tooltip))

            def_val = param['gui']['defvalue']
            if def_val:
                def_val_str = "        По-умолчанию: {}".format(def_val)
                stdout.write(
                    self.style.SUCCESS(def_val_str)
                )

            if param['gui']['required']:
                stdout.write(self.style.ERROR("        Обязательный"))

            stdout.write(self.style.ERROR(""))

    def get_lookup_service_by_type(self, service_type):
        """ Получение объекта сервиса по его типу. """
        if self.get_service_factory():
            return self.get_service_factory().lookup(service_type)

        return None
