# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import sys
import json
from django.contrib.sessions.backends.db import SessionStore
from django.utils.translation import get_language
from termidesk.management.commands.common.subcommands.subcommand import Subcommand

from uds.core.auths.auth import authenticate
from uds.models import Authenticator
from uds.core.util.Config import GlobalConfig
from uds.REST.handlers import Handler


class LoginCommand(Subcommand):
    ''' Консольная команда аутентификации '''

    SUBCOMMAND = "login"
    JSON_FORMATTING = 'json'
    TYPES_FORMATTING = (JSON_FORMATTING,)

    def add_subparser(self, subparsers):
        login_parser = subparsers.add_parser(
            self.SUBCOMMAND,
            help="Аутентификация"
        )
        self.add_arguments_subcommand(login_parser)

        login_parser.add_argument(
            '--uuid', type=str, required=True,
            help="Id аутентификатора для входа"
        )

        login_parser.add_argument(
            '--output', choices=self.TYPES_FORMATTING, required=False,
            help='Вывод данных в формате для машинной обработки',
            default=False
        )

    @staticmethod
    def validate_base_data(data):
        """ Проверка существования домена """
        try:
            authenticator = Authenticator.objects.get(uuid=data)
            return authenticator
        except Authenticator.DoesNotExist:
            return False

    def auth_login(self, auth_uuid):
        from getpass import getpass

        login = input("Login: ")
        password = getpass()
        authenticator = self.validate_base_data(auth_uuid)

        usr = authenticate(login, password, authenticator)
        if not usr:
            return False

        token = Handler.genAuthToken(self, auth_uuid, login, get_language(),
                                     usr.is_admin, usr.staff_member)

        return token

    def show_formatting_items(self, options,  stdout, stderr):
        """ Форматированный вывод. """
        if options['output'] != self.JSON_FORMATTING:
            stderr.write(self.style.ERROR(
                'Переданный тип форматировани {} не поддерживается'.format(
                    options['output']
                )
            ))

        if self.validate_base_data(options['uuid']):
            token = self.auth_login(options['uuid'])
            if token:
                stdout.write(
                    json.dumps({'success': 'true', 'token': token})
                )
            else:
                stdout.write(
                    json.dumps({'success': 'false',
                                'detail': 'Неверный логин или пароль'},
                               ensure_ascii=False)
                )
                sys.exit(1)
        else:
            stdout.write(
                json.dumps({'success': 'false',
                            'detail': 'Домен не найден'},
                           ensure_ascii=False)
            )
            sys.exit(1)

    def handle_subcommand(self, options, stdout, stderr):
        if options['output']:
            return self.show_formatting_items(
                options,
                stdout,
                stderr
            )

        if self.validate_base_data(options['uuid']):
            token = self.auth_login(options['uuid'])
            if token:
                stdout.write(self.style.SUCCESS(
                    'Аутентификация прошла успешно'
                ))
                stdout.write('Token: '+token)
            else:
                stdout.write(
                    self.style.WARNING('Неверный логин или пароль')
                )
                sys.exit(1)
        else:
            stdout.write(
                self.style.WARNING('Домен не найден')
            )
            sys.exit(1)
