"""Модуль содержит реализацию субкоманды установки лицензии"""
# -*- coding: utf-8 -*-
import os
import sys
from argparse import OPTIONAL

from django.contrib.sessions.models import Session
from django.core.management import CommandError
from django.db import transaction

from termidesk.audit.events import WEB_LICENSE_UPDATED_V1
from termidesk.license.storage import LicenseDBStorage
from termidesk.management.commands.common.subcommands.install import InstallSubcommand
from termidesk.models.UserPermission import LICENSE_UPLOAD
from uds.models import Authenticator, User

MAX_FILE_SIZE = 1024 ** 2  # Максимальный размер файла лицензии 1 Мб
TOKEN_ENV_VAR = "TDSK_AUTH_TOKEN"


def get_ip_address():
    """Получение ip-адреса системы"""
    import socket  # pylint: disable=import-outside-toplevel
    hostname = socket.gethostname()
    return socket.gethostbyname(hostname)


class InstallCommand(InstallSubcommand):
    """Установка лицензии"""
    name_of_service = 'лицензии'
    BASE_ARGS = ("file_name", "license_content", "user",)

    def add_subparser(self, subparsers):
        install_parser = subparsers.add_parser(
            self.subcommand,
            help="Install a license from a file"
        )

        install_parser.add_argument(
            'file_name', type=str, nargs=OPTIONAL,
            help="Файл лицензии (относительный или абсолютный путь)"
        )

    def handle_subcommand(self, options, stdout, stderr):
        file_name = options['file_name']

        if not file_name:
            self.show_error(stdout, "Не указан файл лицензии", self.name_of_service)

        # Проверка токена (прав пользователя)
        valid_user = self.check_access(stdout)

        try:
            with open(file_name, "r") as lic:
                license_content = lic.read()

        except FileNotFoundError:
            self.show_error(stdout, "Указанный файл лицензии отсутствует в ФС", self.name_of_service)

        # Проверка размера
        if len(license_content) > MAX_FILE_SIZE:
            self.show_error(
                stdout, "Размер файла лицензии превышает %d байт" % MAX_FILE_SIZE, self.name_of_service
            )

        extended_options = {
            "license_content": license_content,
            "user": valid_user,
            "quiet": False,
        }
        extended_options.update(options)
        super().handle_subcommand(extended_options, stdout, stderr)

    @transaction.atomic()
    def create_service(self, main_args):
        """Запись лицензии в БД"""
        file_name = os.path.basename(main_args['file_name'])
        license_content = main_args['license_content']
        LicenseDBStorage().save(file_name, license_content)

        # Регистрация события загрузки лицензии в системе аудита
        user = main_args['user']
        event_params = dict(
            authenticator=user.manager.name,
            username=user.name,
            ip=get_ip_address(),
            license_file_name=file_name,
        )

        # Отсутствует событие аудита для загрузки лицензии с помощью cli-команды,
        # поэтому применяется его web аналог
        event = WEB_LICENSE_UPDATED_V1(**event_params).generate_event()
        sys.audit(*event)  # pylint: disable=no-member

    def validate_base_data(self, data):
        """ Проверка существования лицензии по имени. """
        license_file = os.path.basename(data['file_name'])
        if LicenseDBStorage().exists(license_file):
            error = ("Указанный файл лицензии '%s' уже присутствует в БД" % license_file)
            raise CommandError(error)

    def check_access(self, stdout):
        """Проверка прав пользователя для установки лицензии"""
        token = os.environ.get(TOKEN_ENV_VAR, None)

        if token in (None, ''):
            self.show_error(
                stdout, "В переменной окружения '%s' отсутствует токен" % TOKEN_ENV_VAR, self.name_of_service
            )
            return False

        try:
            session = Session.objects.get(session_key=token)
        except Session.DoesNotExist:  # pylint: disable=no-member
            self.show_error(
                stdout, "Неверный токен '%s'" % str(token), self.name_of_service
            )
            return False

        session_data = session.get_decoded()['REST']

        username = session_data['username']
        auth_uuid = session_data['auth']

        auth = Authenticator.objects.get(uuid=auth_uuid)  # pylint: disable=no-member
        user = User.objects.get(name=username, manager_id=auth.pk)

        if user.has_perm(LICENSE_UPLOAD):
            # Возвращаем True только если пользователь имеет необходимые разрешения
            return user

        self.show_error(
            stdout, "У пользователя отсутствует разрешение на установку лицензии", self.name_of_service
        )
        return False

    def show_error(self, stdout, error_exception, name_of_service):
        """Функция помощник для отображения ошибки"""
        super().show_error(stdout, error_exception, name_of_service)
        # Для того чтобы при ошибке код возврата был '1' (sys.exit(1))
        # возбуждается исключение CommandError
        raise CommandError(error_exception)
