"""Модуль содержит реализацию субкоманды отображения установленной лицензии"""
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import collections

from termidesk.license import base as t_license
from termidesk.license.base import OS_LABELS, get_license_type, LICENSE_TYPE_1, LICENSE_TYPE_2

from termidesk.license.base import get_active_license_info
from termidesk.license.exceptions import LicenseTypeNotSupportedError
from termidesk.management.commands.common.subcommands.list import ListSubcommand
from termidesk.models import License


class ShowCommand(ListSubcommand):
    """Отображение установленной лицензии"""
    subcommand = "show"

    def add_subparser(self, subparsers):
        list_parser = subparsers.add_parser(
            self.subcommand,
            help="Show information about the installed license"
        )
        self.add_arguments_subcommand(list_parser)

    @staticmethod
    def has_elements():
        try:
            get_active_license_info()
            return True

        except Exception:  # pylint: disable=broad-except
            return False

    @staticmethod
    def get_table_records():
        """
        Получение сведений об установленной лицензии.
        """
        active_license = get_active_license_info()

        os_types = ', '.join([OS_LABELS.get(v, v) for v in t_license.get_os_types()])

        license_type = get_license_type()

        if license_type == LICENSE_TYPE_1:
            license_info = collections.OrderedDict([
                ('cn', active_license.get('cn', '')),  # Имя
                ('org', active_license.get('org', '')),  # Организация
                ('email', active_license.get('email', '')),  # Email
                # Конкурентные соединения
                ('active_workplaces', active_license.get('active_workplaces', None)),
                ('os_types', os_types),  # Доступные гостевые ОС
            ])
            return license_info

        if license_type == LICENSE_TYPE_2:
            license_info = collections.OrderedDict([
                ('cn', active_license.get('cn', '')),  # Имя
                ('org', active_license.get('org', '')),  # Организация
                ('email', active_license.get('email', '')),  # Email
                ('max_users', active_license.get('max_users', None)),  # Количество пользователей
                ('os_types', os_types),  # Доступные гостевые ОС
            ])
            return license_info

        raise LicenseTypeNotSupportedError("Данный тип лицензии не поддерживается '%s'" % str(license_type))

    def show_elements(self, stdout, stderr):
        pattern = "{0:20}   {1:20}   {2:20}   {3:24}   {4:21}"

        stdout.write(
            self.style.SUCCESS("Установлена лицензия:")
        )

        license_info = self.get_table_records()
        license_type = get_license_type()

        if license_type == LICENSE_TYPE_1:
            stdout.write(
                self.style.SUCCESS(
                    pattern.format(
                        'Имя',
                        'Организация',
                        'Email',
                        # Пробел добавлен для "выравнивания" отображения
                        ' '+'Конкурентные соединения',
                        'Доступные гостевые ОС',
                    )
                )
            )

        if license_type == LICENSE_TYPE_2:
            stdout.write(
                self.style.SUCCESS(
                    pattern.format(
                        'Имя',
                        'Организация',
                        'Email',
                        'Количество пользователей',
                        'Доступные гостевые ОС',
                    )
                )
            )

        stdout.write(
            self.style.WARNING(
                pattern.format(*(license_info[key] for key in license_info))
            )
        )

        from termidesk.rest.handlers import raise_cli_crud_event  # pylint: disable=import-outside-toplevel
        from termidesk.rest.handlers import READ_ACTION  # pylint: disable=import-outside-toplevel
        raise_cli_crud_event(
            entity=License.__name__,
            uuid='',
            subtype='',
            name='',
            action=READ_ACTION,
            username=None  # В качестве имени пользователя будет использоваться
                           # имя системного пользователя, запустившего cli-команду
        )
