# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.core.management.base import CommandError
from django.db import transaction
from uds.models import (
    DeployedService, Service, OSManager, ServicesPoolGroup, Image
)
from termidesk.management.commands.common.subcommands.create_without_type import (
    CreateWithoutType
)
from termidesk.management.commands.common.argtypes import positive_int_or_zero_type


class CreateCommand(CreateWithoutType):
    name_of_service = 'фонда рабочих мест'

    SERVICE_RELATED_FIELD = 'service_id'
    OSMANAGER_RELATED_FIELD = 'osmanager_id'
    IMAGE_RELATED_FIELD = 'image_id'
    SERVICESPOOLGROUP_RELATED_FIELD = 'servicesPoolGroup_id'
    SHOW_TRANSPORTS_FIELD = 'show_transports'

    BASE_ARGS = (
        'name', 'comments',
        SERVICE_RELATED_FIELD, OSMANAGER_RELATED_FIELD,
        IMAGE_RELATED_FIELD, SERVICESPOOLGROUP_RELATED_FIELD,
        'cache_l1_srvs', 'cache_l2_srvs',
        'max_srvs', SHOW_TRANSPORTS_FIELD
    )

    SHOW_TRANSPORTS_TRUE = 'true'
    SHOW_TRANSPORTS_FALSE = 'false'

    def add_subparser(self, subparsers):
        create_parser = subparsers.add_parser(
            self.SUBCOMMAND,
            help="Create service"
        )

        self.add_arguments_subcommand(create_parser)

        create_parser.add_argument(
            '--name', type=str, required=True,
            help="Имя объекта"
        )

        create_parser.add_argument(
            '--comments', type=str,
            help="Комментарий", default=''
        )

        create_parser.add_argument(
            '--service_id', type=self._service_type, required=True,
            help='ID шаблона фонда рабочих мест'
        )

        create_parser.add_argument(
            '--osmanager_id', type=self._osmanager_type, required=True,
            help="ID параметров гостевой ОС"
        )

        create_parser.add_argument(
            '--image_id', type=self._image_type,
            help="ID изображения", default=False
        )

        create_parser.add_argument(
            '--servicesPoolGroup_id', type=self._services_poll_type,
            help="ID группы", default=False
        )

        create_parser.add_argument(
            '--cache_l1_srvs', type=positive_int_or_zero_type,
            help="Кеш рабочих мест 1-го уровня", default=0
        )

        create_parser.add_argument(
            '--cache_l2_srvs', type=positive_int_or_zero_type,
            help="Кеш рабочих мест 2-го уровня", default=0
        )

        create_parser.add_argument(
            '--max_srvs', type=positive_int_or_zero_type,
            help="Максимальное количество рабочих мест", default=0
        )

        create_parser.add_argument(
            '--show_transports', choices=(
                self.SHOW_TRANSPORTS_TRUE, self.SHOW_TRANSPORTS_FALSE),
            help="Выбор протоколов доставки", default=self.SHOW_TRANSPORTS_TRUE
        )

    @transaction.atomic()
    def create_service(self, main_args):
        main_args[self.SHOW_TRANSPORTS_FIELD] = (
            main_args[self.SHOW_TRANSPORTS_FIELD] == self.SHOW_TRANSPORTS_TRUE)

        if main_args[self.SERVICESPOOLGROUP_RELATED_FIELD] is False:
            del main_args[self.SERVICESPOOLGROUP_RELATED_FIELD]

        if main_args[self.IMAGE_RELATED_FIELD] is False:
            del main_args[self.IMAGE_RELATED_FIELD]

        created_deployed_service = DeployedService.objects.create(**main_args)

        from termidesk.rest.handlers import raise_cli_crud_event
        from termidesk.rest.handlers import CREATE_ACTION

        raise_cli_crud_event(
            entity=created_deployed_service.__class__.__name__,
            uuid=created_deployed_service.uuid,
            subtype='',
            name=created_deployed_service.name,
            action=CREATE_ACTION,
            username=None  # В качестве имени пользователя будет использоваться
                           # имя системного пользователя, запустившего cli-команду
        )

    def _service_type(self, uuid):
        return self._get_service_id_by_uuid(Service.__name__, uuid)

    def _osmanager_type(self, uuid):
        return self._get_service_id_by_uuid(OSManager.__name__, uuid)

    def _image_type(self, uuid):
        return self._get_service_id_by_uuid(Image.__name__, uuid)

    def _services_poll_type(self, uuid):
        return self._get_service_id_by_uuid(ServicesPoolGroup.__name__, uuid)

    @staticmethod
    def _get_service_id_by_uuid(model_name, uuid):
        """
        Проверка существования подсервиса по переданному uuid
        и отдача идентификатора в случае, если подсервис найден.
        :param model_name:
        :param uuid:
        :return:
        """
        try:
            command = '{}.objects.get(uuid="{}")'.format(model_name, uuid)
            service = eval(command)
            return service.pk
        except Exception:
            raise CommandError(
                'Not found {} by uuid: {}'.format(model_name, uuid)
            )

    def validate_base_data(self, data):
        """ Проверка существования фонда по имени. """
        try:
            DeployedService.objects.get(name=data['name'])
            error = ('Фонд рабочих мест с таким именем {} '
                     'уже существует').format(data['name'])
            raise CommandError(error)
        except DeployedService.DoesNotExist:
            pass
