# -*- coding: utf-8 -*-
'''
Created on 6 мая 2020 г.

@author: dvd
'''
from __future__ import unicode_literals

from django.core.management.base import CommandError
from .common.common_command import CommonCommand
from termidesk.management.commands.common.subcommands.create import CreateSubcommand
from termidesk.management.commands.subcommands.auth.list import ListCommand
from termidesk.management.commands.subcommands.auth.login import LoginCommand
from termidesk.management.commands.subcommands.auth.remove import RemoveCommand
from django.db import transaction
from uds.models import Authenticator
from uds.core import auths
from termidesk.core.util.utils import to_str
from termidesk.management.commands.common.argtypes import positive_int_type

import urllib3
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

import argparse
import re


def small_name_type(val):
    max_length = Authenticator._meta.get_field('small_name').max_length
    val = to_str(val)

    if len(val) > max_length:
        error = "Length exceeds {} chars".format(max_length)
        raise argparse.ArgumentTypeError(error)

    if re.search("(?u)[^ a-zA-Z0-9_]", val):
        error = "Allowed chars: a-z A-Z 0-9 _"
        raise argparse.ArgumentTypeError(error)

    return val


class Command(CommonCommand):
    help = '''
           Создает домен аутентификации.
           Для просмотра параметров домена аутентификации 
           не указывайте аргумент --params.
           '''

    create_command = None
    list_command = None
    remove_command = None
    login_command = None

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        self.create_command = CreateCommand(self.style)
        self.create_command.add_subparser(self.subparsers)

        self.list_command = ListCommand(self.style)
        self.list_command.add_subparser(self.subparsers)

        self.remove_command = RemoveCommand(self.style)
        self.remove_command.add_subparser(self.subparsers)

        self.login_command = LoginCommand(self.style)
        self.login_command.add_subparser(self.subparsers)

    def handle(self, *args, **options):
        subcommand = options.get('subcommand')

        if subcommand == self.create_command.SUBCOMMAND:
            return self.create_command.handle_subcommand(
                options,
                self.stdout,
                self.stderr
            )

        if subcommand == self.list_command.SUBCOMMAND:
            return self.list_command.handle_subcommand(
                options,
                self.stdout,
                self.stderr
            )

        if subcommand == self.remove_command.SUBCOMMAND:
            return self.remove_command.handle_subcommand(
                options,
                self.stdout,
                self.stderr
            )

        if subcommand == self.login_command.SUBCOMMAND:
            return self.login_command.handle_subcommand(
                options,
                self.stdout,
                self.stderr
            )

        self.stdout.write(
            self.style.ERROR(
                'Не удалось обработать команду '.format(subcommand)
            )
        )


class CreateCommand(CreateSubcommand):
    name_of_service = 'домена аутентификации'

    types = [auth.typeType
             for auth in list(auths.factory().providers().values())]

    BASE_ARGS = ('name', 'small_name', 'priority', 'comments')

    _factory = None

    def add_subparser(self, subparsers):
        create_parser = subparsers.add_parser(
            CreateSubcommand.SUBCOMMAND,
            help="Create service"
        )

        self.add_arguments_subcommand(create_parser)

        create_parser.add_argument(
            '--name', type=str, required=True,
            help="Имя домена аутентификации"
        )

        create_parser.add_argument(
            '--small_name', type=small_name_type, required=True,
            help=("Короткое имя домена аутентификации. "
                  "Допустимые символы: a-z A-Z 0-9 _. "
                  "Максимум 32 символа.")
        )

        create_parser.add_argument(
            '--priority', type=positive_int_type,
            help="Приоритет домена аутентификации. Значение по-умолчанию: 1.",
            default=1
        )

        create_parser.add_argument(
            '--comments', type=str, help="Комментарий", default=''
        )

    @transaction.atomic()
    def create_service(self, auth_type, main_args, params):
        item = Authenticator.objects.create(**main_args)
        item.data_type = auth_type
        item.data = item.getInstance(params).serialize()
        item.save()

        from termidesk.rest.handlers import raise_cli_crud_event
        from termidesk.rest.handlers import CREATE_ACTION

        raise_cli_crud_event(
            entity=item.__class__.__name__,
            uuid=item.uuid,
            subtype=auth_type,
            name=item.name,
            action=CREATE_ACTION,
            username=None  # В качестве имени пользователя будет использоваться
                           # имя системного пользователя, запустившего cli-команду
        )

    def get_service_factory(self):
        if self._factory is None:
            self._factory = auths.factory()
        return self._factory

    def validate_base_data(self, data):
        """ Проверка наличия в БД. """
        try:
            Authenticator.objects.get(name=data['name'])
            error = ('Домен аутентификации {} '
                     'уже существует').format(data['name'])
            raise CommandError(error)
        except Authenticator.DoesNotExist:
            pass

        try:
            Authenticator.objects.get(small_name=data['small_name'])
            error = ('Домен аутентификации c коротким именем {} '
                     'уже существует').format(data['small_name'])
            raise CommandError(error)
        except Authenticator.DoesNotExist:
            pass
