# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from .common.common_command import CommonCommand
from termidesk.management.commands.common.subcommands.create import CreateSubcommand
from django.core.management.base import CommandError
from django.db import transaction
from uds.models import Transport, Network
from uds.core.util import OsDetector
from uds.core import transports
import argparse
from termidesk.core.util.utils import to_str
from .subcommands.trans.list import ListCommand
from .subcommands.trans.remove import RemoveCommand

OSS = sorted([os for os in OsDetector.knownOss], key=lambda x: x.lower())


def _get_networks():
    return {x.name: x.id for x in Network.objects.all()}


def os_str_type(val):
    oss = val.split(',')
    for os_ in oss:
        if os_ not in OSS:
            error = "This OS type is not available: {}".format(os_)
            raise argparse.ArgumentTypeError(error)

    return val


def network_str_type(network):
    networks = _get_networks()
    network_ids = []
    names = to_str(network).split(',')
    for name in names:
        network_id = networks.get(name)
        if network_id is None:
            error = "This network is not available: {}".format(name)
            raise argparse.ArgumentTypeError(error.encode('ascii', 'replace'))

        network_ids.append(network_id)
    return network_ids


class Command(CommonCommand):
    help = '''Управление протоколами доставки.'''

    create_command = None

    def add_arguments(self, parser):
        super(Command, self).add_arguments(parser)
        self.create_command = CreateCommand(self.style)
        self.create_command.add_subparser(self.subparsers)

        self.list_command = ListCommand(self.style)
        self.list_command.add_subparser(self.subparsers)

        self.remove_command = RemoveCommand(self.style)
        self.remove_command.add_subparser(self.subparsers)

    def handle(self, *args, **options):
        subcommand = options.get('subcommand')

        if subcommand == self.create_command.SUBCOMMAND:
            return self.create_command.handle_subcommand(
                options,
                self.stdout,
                self.stderr
            )

        if subcommand == self.list_command.SUBCOMMAND:
            return self.list_command.handle_subcommand(
                options,
                self.stdout,
                self.stderr
            )

        if subcommand == self.remove_command.SUBCOMMAND:
            return self.remove_command.handle_subcommand(
                options,
                self.stdout,
                self.stderr
            )

        self.stdout.write(
            self.style.ERROR(
                'Не удалось обработать команду '.format(subcommand)
            )
        )


class CreateCommand(CreateSubcommand):
    name_of_service = 'протокола доставки'

    types = [transport.typeType
             for transport in list(transports.factory().providers().values())]

    BASE_ARGS = (
        'name', 'data_type', 'comments', 'priority', 'allowed_oss',
        'nets_positive', 'networks'
    )

    _factory = None

    def add_subparser(self, subparsers):
        create_parser = subparsers.add_parser(
            CreateSubcommand.SUBCOMMAND,
            help="Create service"
        )

        self.add_arguments_subcommand(create_parser)

        create_parser.add_argument(
            '--name', type=str, required=True,
            help="Имя объекта"
        )

        create_parser.add_argument(
            '--comments', type=str,
            help="Комментарий для объекта",
            default=''
        )

        create_parser.add_argument(
            '--priority', type=str,
            help="Приоритет",
            default=1
        )

        allowed_oss = ', '.join(OSS)
        create_parser.add_argument(
            '--allowed_oss', type=os_str_type,
            help=("Разрешенные OC из списка: {}. ".format(allowed_oss)
                  + "Формат: имя1,имя2,имя3"),
            default=False
        )

        create_parser.add_argument(
            '--nets_positive', type=bool,
            help="Доступ из сетей",
            default=False
        )

        networks = _get_networks()
        allowed_nets = ', '.join(networks.keys())
        create_parser.add_argument(
            '--networks', type=network_str_type,
            help=("Разрешенны сети из списка: {}. ".format(allowed_nets)
                  + "Формат: имя1,имя2,имя3"),
            default=False
        )

    @transaction.atomic()
    def create_service(self, type_service, main_args, params):
        networks = main_args['networks']
        del main_args['networks']

        transport = Transport(**main_args)
        transport.data_type = type_service
        transport.data = transport.getInstance(params).serialize()
        transport.save()
        if networks:
            transport.networks.set(networks)

        from termidesk.rest.handlers import raise_cli_crud_event
        from termidesk.rest.handlers import CREATE_ACTION

        raise_cli_crud_event(
            entity=transport.__class__.__name__,
            uuid=transport.uuid,
            subtype=transport.data_type,
            name=transport.name,
            action=CREATE_ACTION,
            username=None  # В качестве имени пользователя будет использоваться
                           # имя системного пользователя, запустившего cli-команду
        )

    def get_service_factory(self):
        if self._factory is None:
            self._factory = transports.factory()
        return self._factory

    def validate_base_data(self, data):
        """ Проверка существования протокола по имени. """
        try:
            Transport.objects.get(name=data['name'])
            error = ('Протокол с таким именем {} '
                     'уже существует').format(data['name'])
            raise CommandError(error)
        except Transport.DoesNotExist:
            pass
