# -*- encoding: utf-8 -*-
from django.contrib.postgres.fields import JSONField
from django.db import models

import logging

logger = logging.getLogger(__name__)


class AuditEventManager(models.Manager):
    """Менеджер содержит дополнительные методы для выполнения на уровне таблицы (выборки)"""

    def get_outdated_records(self, before_time=None):
        """
        Возвращает queryset с устаревшими записями

        :param before_time: utc дата и время в формате iso до которого записи считаются устаревшими
        """
        # Если параметр `before_time` не задан, то вернется пустой queryset
        outdated_records = self.none()

        if before_time:
            outdated_records = self.filter(timestamp__lt=before_time)

        return outdated_records

    def delete_with_limit(self, outdated_records=None, delete_limit=None):
        """
        Удаляет ограниченное количество записей из переданного queryset

        :param outdated_records: queryset удаляемых записей
        :param delete_limit: целое число, количество удаляемых записей
        """
        # Если параметр `delete_limit` не задан, то удалять ничего не будем
        if outdated_records is None or delete_limit is None:
            outdated_records = self.none()

        if delete_limit:
            # В Django удаление вместе с limit вызывает ошибку
            # `AssertionError: Cannot use 'limit' or 'offset' with delete.`
            pk_list = outdated_records.values_list('pk')[0:delete_limit]

            outdated_records = self.filter(pk__in=pk_list)

        return outdated_records.delete()


class AuditEvent(models.Model):
    """
    Модель хранения информации о событиях, произошедших в системе аудита
    """

    # Дата с указанием часового пояса UTC
    timestamp = models.DateTimeField(db_index=True)
    # Название события
    event = models.CharField(max_length=256, db_index=True)
    # Версия события
    version = models.CharField(max_length=16, db_index=True)
    # Данные события (json-строка)
    data = JSONField()
    # Менеджер содержит необходимые дополнительные методы для выполнения на уровне таблицы (выборки)
    objects = AuditEventManager()

    class Meta:
        """
        Мета класс для объявления таблицы в БД
        """
        db_table = 'termidesk_audit_event_log'
        app_label = 'termidesk'

    def __unicode__(self):
        # Сообщение будет выводиться в следующем формате:
        #   "Audit event: termidesk.audit.events.web.UserLogin(v1) at 2021-05-06T15:43:12.467113+00:00
        #    with '["root", "10.0.1.13"]'"
        _template = "Audit event: {event}(v{version}) at {timestamp} with '{data}'"
        _template_fields = {
            'event': self.event,
            'version': self.version,
            'timestamp': self.timestamp,
            'data': self.data
        }
        return _template.format(**_template_fields)

    def __str__(self):
        from termidesk.core.util.utils import to_str
        return to_str(self.__unicode__())
