import hashlib
import logging
from time import time

from tdsklog.email_handler import TermideskEmailHandler
from termidesk.notifications.delayed_task.sender import EmailNotification
from uds.core.managers.CryptoManager import CryptoManager
from uds.core.util.Config import GlobalConfig

logger = logging.getLogger(__name__)

DELAYED_TASK_TAG = 'send-notification-event'
DELAYED_TASK_SUGGESTED_TIME = 1


class NotificationsSender:
    """ Класс обработчик рассылки уведомлений. """

    def provider_maintenance_mode_changed(self, event_data: dict):
        """ Обработка события смены режима техобслуживания провайдера. """
        if not GlobalConfig.NOTIFICATIONS_EVENTS_PROVIDER_MAINTENANCE_CHANGED.getBool(True):
            return

        subject = " Изменение режима техобслуживания поставщика ресурсов."
        state_maintenance = "включил" if event_data.get('maintenance_mode') is True else "выключил"
        message = "[{0}] Пользователь '{1}(id:{2})' {3} режим техобслуживания поставщика ресурсов '{4}(id:{5})'.".format(
            event_data.get('time'),
            event_data.get('user_name'),
            event_data.get('user_id'),
            state_maintenance,
            event_data.get('provider_name'),
            event_data.get('provider_id')
        )
        tag = '{}-{}-{}-{}-{}'.format(
            DELAYED_TASK_TAG,
            event_data.get('user_id'),
            event_data.get('provider_id'),
            int(event_data.get('maintenance_mode')),
            time()
        )
        self.send_notification(subject, message, tag)

    def set_user_service_state_to_error(self, event_data: dict):
        """Виртуальное рабочее место установлено в состояние ОШИБКА"""

        event_time = event_data.get('time', '')
        service_pool_uuid = event_data.get('service_pool_uuid', '')
        service_pool_name = event_data.get('service_pool_name', '')
        user_service_uuid = event_data.get('user_service_uuid', '')
        user_service_name = event_data.get('user_service_name', '')
        reason_of_error = event_data.get('reason_of_error', '')

        if not GlobalConfig.NOTIFICATIONS_EVENTS_USER_SERVICE_STATE_CHANGED_TO_ERROR.getBool(True):
            logger.debug(
                'User service {us_name}(uuid: {us_uuid}) '
                'from workplace {sp_name}(uuid: {sp_uuid}) '
                'state is ERROR.\nBut notifications disabled.'.format(
                    us_name=user_service_name, us_uuid=user_service_uuid,
                    sp_name=service_pool_name, sp_uuid=service_pool_uuid,
                )
            )
            return

        subject = "ВРМ %s(%s): в состоянии ОШИБКА." % (user_service_name, user_service_uuid, )

        template = "[{event_time}] ВРМ '{us_name}(id:{us_id})' из фонда '{ds_name}(id:{ds_id})' " \
                   "в состоянии ОШИБКА. \nПричина: {reason}"

        message = template.format(
            event_time=event_time,
            us_name=user_service_name,
            us_id=user_service_uuid,
            ds_name=service_pool_name,
            ds_id=service_pool_uuid,
            reason=reason_of_error,
        )

        # В БД для тега задачи стоит ограничение в 64 байта.
        # Чтобы не превышать данное ограничение составляем тег из префикса и хеша.
        # Для хеша выбран алгоритм md5, так как длина его дайджеста 32 байта.
        # Таким образом, префикс не должен превышать 32 байта.
        task_hash = self.get_md5_hash_digest(
            '{us_name}-{us_id}-{ds_name}-{ds_id}-{cur_time}'.format(
                task_tag=DELAYED_TASK_TAG,
                us_name=user_service_name,
                us_id=user_service_uuid,
                ds_name=service_pool_name,
                ds_id=service_pool_uuid,
                cur_time=time()
            )
        )

        tag = '{task_tag}-{task_hash}'.format(
            task_tag=DELAYED_TASK_TAG,
            task_hash=task_hash
        )

        logger.debug(
            'Send notification to administrator: User service {us_name}(uuid: {us_uuid}) '
            'from workplace {sp_name}(uuid: {sp_uuid}) state is ERROR'.format(
                us_name=user_service_name, us_uuid=user_service_uuid,
                sp_name=service_pool_name, sp_uuid=service_pool_uuid
            )
        )
        self.send_notification(subject, message, tag)

    def license_connection_limit_reached(self, event_data: dict):
        """Обработка события превышения допустимого лицензией количества подключений"""
        if not GlobalConfig.NOTIFICATIONS_EVENTS_LICENSE_CONNECTION_LIMIT_REACHED.getBool(True):
            logger.debug("License connections limit reached - {}.\n"
                         "But notifications disabled."
                         .format(event_data['connections_limit']))
            return

        subject = "Превышение установленного лицензией количества подключений"

        message = \
            "[{0}] Превышено ограничение лицензии на количество подключений. " \
            "Текущее ограничение '{1}' подключений.".format(
                event_data.get('time'),
                event_data.get('connections_limit'),
            )
        tag = '{}-{}-{}'.format(
            DELAYED_TASK_TAG,
            event_data.get('connections_limit'),
            time()
        )
        logger.debug("Sending notification: License connections limit reached - {}."
                     .format(event_data['connections_limit']))
        self.send_notification(subject, message, tag)

    def license_user_limit_reached(self, event_data: dict):
        """Обработка события превышения допустимого лицензией количества пользователей"""
        if not GlobalConfig.NOTIFICATIONS_EVENTS_LICENSE_USER_LIMIT_REACHED.getBool(True):
            logger.debug("License users limit reached - {}.\n"
                         "But notifications disabled."
                         .format(event_data['users_limit']))
            return

        subject = "Превышение установленного лицензией количества пользователей."

        message = \
            "[{0}] Превышено ограничение лицензии на количество пользователей. " \
            "Текущее ограничение '{1}' пользователей.".format(
                event_data.get('time'),
                event_data.get('users_limit'),
            )
        tag = '{}-{}-{}'.format(
            DELAYED_TASK_TAG,
            event_data.get('users_limit'),
            time()
        )
        logger.debug("Sending notification: License users limit reached - {}."
                     .format(event_data['users_limit']))
        self.send_notification(subject, message, tag)

    def send_notification(self, subject: str, message: str, task_tag: str):
        """ Установка отложенной задачи. """
        if self.__is_enable_general_mailing() is False:
            return

        admins = GlobalConfig.NOTIFICATIONS_TO_EMAILS.get(True).split(',')
        if not admins:
            logger.error(
                'Не указаны администраторы для рассылки'
            )
            return

        delayed_task = EmailNotification(**self.get_main_params(subject, message, admins))
        delayed_task.register(DELAYED_TASK_SUGGESTED_TIME, task_tag)

    @staticmethod
    def get_main_params(subject, message, admins):
        """ Получение основных параметров, необходимых для рассылки. """

        server_mail = GlobalConfig.NOTIFICATIONS_SMTP_SERVER_EMAIL.get(True)

        subject_prefix = GlobalConfig.NOTIFICATIONS_SUBJECT_PREFIX.get(True)
        subject = '%s%s' % (subject_prefix, subject)

        # SMTP params
        host = GlobalConfig.NOTIFICATIONS_SMTP_HOST.get(True)
        port = GlobalConfig.NOTIFICATIONS_SMTP_PORT.getInt(True)
        username = GlobalConfig.NOTIFICATIONS_SMTP_USERNAME.get(True)
        password = CryptoManager.manager().encrypt(GlobalConfig.NOTIFICATIONS_SMTP_PASSWORD.get(True))
        timeout = GlobalConfig.NOTIFICATIONS_SMTP_TIMEOUT.getInt(True)

        use_tls, use_ssl = TermideskEmailHandler.get_tls_ssl()

        return {'subject': subject, 'message': message, 'admins': admins, 'server_mail': server_mail,
                'host': host, 'port': port, 'username': username, 'password': password, 'timeout': timeout,
                'use_tls': use_tls, 'use_ssl': use_ssl}

    @staticmethod
    def __is_enable_general_mailing():
        """ Проверка общего включения почтовых уведомлений. """
        return GlobalConfig.NOTIFICATIONS_ENABLE.getBool(True)

    @staticmethod
    def get_md5_hash_digest(_value: str) -> str:
        """Возвращает md5-дайджест переданной строки"""
        try:
            hash_digest = hashlib.md5(_value.encode('utf-8')).hexdigest()
        except AttributeError:
            hash_digest = None

        return hash_digest
