import logging
from abc import ABC, abstractmethod

from termidesk.notifications.entities.EntityEventsHelperInterface import EntityEventsHelperInterface
from termidesk.notifications.entities.LicenceEventsHelper import LicenseEventsHelper, \
    NOTIFICATIONS_EVENTS_LICENSE_CONNECTION_LIMIT_REACHED, NOTIFICATIONS_EVENTS_LICENSE_USER_LIMIT_REACHED
from termidesk.notifications.entities.ProviderEventsHelper import (
    EVENT_PROVIDER_MAINTENANCE_MODE_CHANGED, ProviderEventsHelper)
from termidesk.notifications.entities.UserServiceEventsHelper import EVENT_USER_SERVICE_STATE_SET_TO_ERROR, \
    UserServiceEventsHelper

logger = logging.getLogger(__name__)


class MediatorABC(ABC):

    @abstractmethod
    def notify(self, sender: EntityEventsHelperInterface, event: str, params: dict):
        """ Метод хранит бизнес-логику обработки уведомлений. """
        pass


class Mediator(MediatorABC):
    """
    Класс отвечает передачу информации о событиях заинтерерсованным модулям
    для обработки
    """

    def __init__(self, notifications_sender, *args, **kwargs):
        self._notifications_sender = notifications_sender
        super(Mediator, self).__init__(*args, **kwargs)

    @property
    def observers(self):
        # Карта обработчиков реализована в виде свойства, чтобы методы
        # обработчиков определялись в момент обращения к нему, а не в момент
        # описания класса.
        # В противном случае проблематично заменять их во время тестирования (mock).
        notifications_sender = self._notifications_sender
        observers = {
            ProviderEventsHelper: {
                EVENT_PROVIDER_MAINTENANCE_MODE_CHANGED: [
                    notifications_sender.provider_maintenance_mode_changed
                ]
            },
            UserServiceEventsHelper: {
                EVENT_USER_SERVICE_STATE_SET_TO_ERROR: [
                    notifications_sender.set_user_service_state_to_error
                ]
            },
            LicenseEventsHelper: {
                NOTIFICATIONS_EVENTS_LICENSE_CONNECTION_LIMIT_REACHED: [
                    notifications_sender.license_connection_limit_reached,
                ],
                NOTIFICATIONS_EVENTS_LICENSE_USER_LIMIT_REACHED: [
                    notifications_sender.license_user_limit_reached
                ]
            }
        }
        return observers

    def notify(self, sender: EntityEventsHelperInterface, event: str, params: dict):
        """
        Метод вызывает зарегистрированные обработчики для событий
        """
        try:
            events = self.observers[sender.__class__]
            try:
                handlers = events[event]
                for handler in handlers:
                    handler(params)
            except KeyError:
                msg = "Unsupported event {evt} for sender {sender}" \
                      .format(evt=event, sender=sender)
                logger.error(msg)
        except KeyError:
            logger.error("Unsupported sender: {sender}".format(sender=sender))
