import sys

from termidesk.audit.events import WEB_ENTITY_ACTION, CLI_ENTITY_ACTION
from uds.REST.handlers import NotFound
from uds.REST.model import OK, ModelHandler, OVERVIEW, TYPES, TABLEINFO, GUI

CREATE_ACTION = 'create'
READ_ACTION = 'read'
UPDATE_ACTION = 'update'
DELETE_ACTION = 'delete'


def set_crud_event(authenticator: str, username: str, ip: str,
                   entity: str, uuid: str, subtype: str,
                   name: str, action: str):
    event = WEB_ENTITY_ACTION(
        authenticator=authenticator,
        username=username,
        ip=ip,
        entity=entity,
        uuid=uuid,
        subtype=subtype,
        name=name,
        action=action
    ).generate_event()

    sys.audit(*event)


def raise_cli_crud_event(entity: str, uuid: str,
                         subtype: str, name: str, action: str, username: str = None) -> None:
    """Вызов события аудита при изменении объекта из командной строки"""
    if username is None:
        username = get_system_username()

    event = CLI_ENTITY_ACTION(
        username=username,
        entity=entity,
        uuid=uuid,
        subtype=subtype,
        name=name,
        action=action
    ).generate_event()

    sys.audit(*event)


def get_system_username() -> str:
    """Получение имени системного пользователя"""
    import getpass
    system_user_name = getpass.getuser()
    return system_user_name


class AuditHandler(ModelHandler):
    """ Класс обработчик основных сущностей. """

    def create(self):
        res = super().put()

        if self.__is_detail():
            self.__set_update_parent()
            return res

        set_crud_event(
            self._user.manager.name,
            self._user.name,
            self._request.ip,
            self.model.__name__,
            res.get('id', ''),
            res.get('type', ''),
            res.get('name', ''),
            CREATE_ACTION
        )

        return res

    def put(self):

        if len(self._args) == 0:
            return self.create()
        else:
            return self.update()

    def update(self):
        res = super().put()

        if self.__is_detail():
            self.__set_update_parent()
            return res

        set_crud_event(
            self._user.manager.name,
            self._user.name,
            self._request.ip,
            self.model.__name__,
            res.get('id', ''),
            res.get('type', ''),
            res.get('name', ''),
            UPDATE_ACTION
        )

        return res

    def get(self):
        res = super().get()

        if len(self._args) == 0:
            return res

        value, *_ = self._args
        if len(self._args) == 1 and value not in [OVERVIEW, TYPES, TABLEINFO, GUI]:
            # Фиксация открытия интерфейса редактирования сущности
            set_crud_event(
                self._user.manager.name,
                self._user.name,
                self._request.ip,
                self.model.__name__,
                res.get('id', ''),
                res.get('type', ''),
                res.get('name', ''),
                READ_ACTION
            )

        return res

    def delete(self):
        try:
            item = self.model.objects.get(uuid=self._args[0].lower())
            subtype = item.data_type if hasattr(item, 'data_type') else ''
            name_object = item.name
        except self.model.DoesNotExist:
            raise NotFound('Element do not exists')

        res = super().delete()

        if self.__is_detail():
            self.__set_update_parent()
            return res

        if res == OK:
            set_crud_event(
                self._user.manager.name,
                self._user.name,
                self._request.ip,
                self.model.__name__,
                item.uuid,
                subtype,
                name_object,
                DELETE_ACTION
            )

        return res

    def __is_detail(self):
        """ Проверка является ли запрос действием над детальной. """
        return len(self._args) > 1

    def __set_update_parent(self):
        item = self.model.objects.filter(uuid=self._args[0])[0]
        subtype = item.data_type if hasattr(item, 'data_type') else ''

        set_crud_event(
            self._user.manager.name,
            self._user.name,
            self._request.ip,
            self.model.__name__,
            item.uuid,
            subtype,
            item.name,
            UPDATE_ACTION
        )
