# -*- coding: utf-8 -*-
'''
Created on 21 апр. 2020 г.

@author: dvd
'''

from __future__ import unicode_literals

from termidesk.core.managers.PolicyManager import PolicyManager
from termidesk.policies.ProtocolSelectionPolicy import ProtocolSelectionPolicy
from uds.REST.methods.connection import Connection as UDSConnection
from uds.models import UserService, DeployedService, ServicesPoolGroup
from uds.core.managers.UserServiceManager import UserServiceManager
from uds.core.util.Config import GlobalConfig
from django.urls import reverse
from django.utils.translation import ugettext
from django.utils import formats
import logging

logger = logging.getLogger(__name__)

class Connection(UDSConnection):
    '''
       Наследует и переопределяет обработчик /rest/connection.
       Выдает дополнительную информацию о фондах рабочих мест (сервисах).
       
       Родительский класс не будет зарегистрирован обработчиком, 
       т.к. регистрируются только классы без наследников. 
    '''
    authenticated = True
    needs_admin = False
    needs_staff = False
    
    def serviceList(self):
        # Extract required data to show to user
        services = []
        services.extend(self.__prepare_user_assigned_services(self._user))
        services.extend(self.__prepare_available_services(self._user))

        services = sorted(services, key=lambda s: s['name'].upper())
        services = self._mark_autorun_service(services)

        return Connection.result(result=services)
    
    def _mark_autorun_service(self, services):
        '''
        Помечает рабочее место для автозапуска клиентом при соблюдении определенных условий.
        
        :param services: Список с описанием доступных рабочих мест
        :type services: list
        :returns: Изначальный список с отмеченным для автозапуска рабочим местом.
        '''
        # Условия для отметки рабочего места для автозапуска
        is_single_service_available = len(services) == 1
        is_autorun_enabled = GlobalConfig.AUTORUN_SERVICE.getBool(True)
        has_transports = is_single_service_available and len(services[0]['transports']) > 0

        autorun_conditions = (
            is_single_service_available,
            is_autorun_enabled,
            has_transports
        )

        if all(autorun_conditions):
            services[0]['autorun'] = True

        return services
    
    def __prepare_user_assigned_services(self, user):
        availUserServices = UserService.getUserAssignedServices(user)

        policies_enabled = PolicyManager.is_policies_enabled()
        manager = PolicyManager()

        services = []
        # Select assigned user services
        for svr in availUserServices:
            # Skip maintenance services...
            trans = []
            for t in svr.transports.all().order_by('priority'):
                if t.validForIp(self._request.ip) and t.getType().providesConnetionInfo():
                    trans.append({'id': t.uuid, 'name': t.name})
            
            if svr.deployed_service.image is not None:
                imageId = svr.deployed_service.image.uuid
            else:
                imageId = 'x'  # Invalid
            
            # Extract group
            group = ServicesPoolGroup.default()
            if svr.deployed_service.servicesPoolGroup is not None:
                group = svr.deployed_service.servicesPoolGroup

            if policies_enabled:
                show_transports = manager.get_deployed_service_policy(
                    ProtocolSelectionPolicy, svr.deployed_service
                ).is_allowed()
            else:
                show_transports = svr.deployed_service.show_transports

            services.append({'id': 'A' + svr.uuid,
                             'name': svr['name'],
                             'group': self.__prepare_group_data(group),
                             'transports': trans,
                             'img_id': imageId,
                             'img_url': reverse('serviceImage', kwargs={'idImage':imageId}),
                             'help_url': svr.deployed_service.help_url,
                             'maintenance': svr.isInMaintenance(),
                             'not_accesible': not svr.deployed_service.isAccessAllowed(),
                             'in_use': svr.in_use,
                             'to_be_replaced': False, # Manually assigned will not be autoremoved never
                             'to_be_replaced_text': '',
                             'comments': svr.comments,
                             'show_transports': show_transports,
                             # Флаг клиенту автоматически запустить рабочее место для пользователя
                             # Действителен только в особом случае.
                             'autorun': False,}) 

        logger.debug('User %s assigned services: %s',user, services)
        return services
    
    def __prepare_available_services(self, user):
        # We look for services for this authenticator groups. User is logged in in just 1 authenticator, so his groups must coincide with those assigned to ds
        groups = list(user.getGroups())
        avail_services_for_groups = DeployedService.getDeployedServicesForGroups(groups)
        avail_services_for_user = DeployedService.get_deployed_services_for_user(user.id)
        # join two lists into one list without duplicates
        availServices = list(
            set(
                [ags for ags in avail_services_for_groups] + [aus for aus in avail_services_for_user]
            )
        )

        policies_enabled = PolicyManager.is_policies_enabled()
        manager = PolicyManager()

        services = []
        
        # Now generic user service
        for svr in availServices:
            trans = []
            for t in svr.transports.all().order_by('priority'):
                if t.validForIp(self._request.ip) and t.getType().providesConnetionInfo():
                    trans.append({'id': t.uuid, 'name': t.name})
            
            if svr.image is not None:
                imageId = svr.image.uuid
            else:
                imageId = 'x'
            
            # Locate if user service has any already assigned user service for this
            ads = UserServiceManager.manager().getExistingAssignationForUser(svr, self._user)
            if ads is None:
                in_use = False
            else:
                in_use = ads.in_use
            
            # Extract group
            group = ServicesPoolGroup.default()
            if svr.servicesPoolGroup is not None:
                group = svr.servicesPoolGroup
            
            tbr = svr.toBeReplaced()
            if tbr is not None:
                tbr = formats.date_format(tbr, "SHORT_DATETIME_FORMAT")
                tbrt = ugettext(
                    '''This service is about to be replaced by a new version. 
                       Please, close the session before {} and save all your 
                       work to avoid loosing it.'''
                       ).format(tbr)
            else:
                tbrt = ''

            if policies_enabled:
                show_transports = manager.get_deployed_service_policy(
                    ProtocolSelectionPolicy, svr
                ).is_allowed()
            else:
                show_transports = svr.show_transports

            services.append({'id': 'F' + svr.uuid,
                             'name': svr.name,
                             'group': self.__prepare_group_data(group),
                             'transports': trans,
                             'img_id': imageId,
                             'img_url': reverse('serviceImage', kwargs={'idImage':imageId}),
                             'help_url': svr.help_url,
                             'maintenance': svr.isInMaintenance(),
                             'not_accesible': not svr.isAccessAllowed(),
                             'in_use': in_use,
                             'to_be_replaced': tbr is not None,
                             'to_be_replaced_text': tbrt,
                             'comments': svr.comments,
                             'show_transports': show_transports,
                             # Флаг клиенту автоматически запустить рабочее место для пользователя
                             # Действителен только в особом случае.
                             'autorun': False,}) 

        logger.debug('User %s available services: %s', user, services)
        return services
    
    def __prepare_group_data(self, group):
        img_id = group.image.uuid if group.image is not None else 'x'
        data = {
            'id': group.uuid,
            'name':group.name,
            'priority': group.priority,
            'comments': group.comments,
            'img_id': img_id,
            'img_url': reverse('serviceImage', kwargs={'idImage':img_id}),
        }
        
        return data