# -*- coding: utf-8 -*-
'''
Created on Nov 8, 2019

@author: dvd
'''
from __future__ import unicode_literals

import six
from django.utils.translation import ugettext_lazy as _

from termidesk.services.OVirt.const import (
    VM_STATE_DOWN,
    VM_STATE_IMAGE_ILLEGAL,
    VM_STATE_IMAGE_LOCKED,
    VM_STATE_MIGRATING_FROM,
    VM_STATE_MIGRATING_TO,
    VM_STATE_NOT_RESPONDING,
    VM_STATE_PAUSED,
    VM_STATE_POWERED_DOWN,
    VM_STATE_POWERING_DOWN,
    VM_STATE_POWERING_UP,
    VM_STATE_REBOOT_IN_PROGRESS,
    VM_STATE_RESTORING_STATE,
    VM_STATE_SAVING_STATE,
    VM_STATE_SUSPENDED,
    VM_STATE_UNASSIGNED,
    VM_STATE_UNKNOWN,
    VM_STATE_UP,
    VM_STATE_WAIT_FOR_LAUNCH,
)
from uds.core import services
from uds.core.util.State import State
import logging
import json

logger = logging.getLogger(__name__)


class OVirtStaticDeployment(services.UserDeployment):
    suggestedTime = 5

    def initialize(self):
        self._ip = self.service().getHostname()
        self._vmid = self.service().vmId.value
        self._state = State.FINISHED
        self._reason = ''
    
    # Serializable needed methods
    def marshal(self):
        data = {
            'version': 1,
            'ip': self._ip,
            'vm_id': self._vmid,
            'state': self._state,
            'reason': self._reason,
        }
        
        return json.dumps(data)

    def unmarshal(self, str_):
        if isinstance(str_, six.binary_type):
            str_ = str_.decode('utf-8')
        data = json.loads(str_)
        if data['version']==1:
            self._ip = data.get('ip','')
            self._vmid = data['vm_id']
            self._reason = data['reason']
            self._state = data['state']
        else:
            raise Exception("Unsupported serialized data version {}".format(data['version']))
    
    def setIp(self, ip):
        logger.debug('Setting IP to %s (ignored)' % ip)

    def getIp(self):
        return self._ip

    def getName(self):
        return self._vmid

    def getUniqueId(self):
        return self._vmid

    def setReady(self):
        """
        Mark the machine as ready to use
        """

        # Starts the machine if it down or suspended
        task_state = self.__startMachine()

        # Ensure setReady method returns expected state
        if task_state not in (State.FINISHED, State.RUNNING, State.ERROR):
            logger.error("State of task for VM {vm} must be one of states "
                         "{state_finish}, {state_running} or {state_error}, not '{state}'"
                         .format(vm=self._vmid, 
                                 state_finish=State.string.get(State.FINISHED, None),
                                 state_running=State.string.get(State.RUNNING, None),
                                 state_error=State.string.get(State.ERROR, None),
                                 state=State.string.get(task_state, None)))
            task_state = State.ERROR

        self._state = task_state

        return self._state

    def deployForUser(self, user):
        logger.debug("Starting deploy of {0} for user {1}".format(self._ip, user))
        self.dbservice().setInUse(True)
        self.dbservice().save()
        self._state = State.FINISHED
        return self._state

    def checkState(self):
        # Get the machine state from provider and convert it to expected state
        self._state = self.__checkMachineState()
        return self._state

    def finish(self):
        pass

    def reasonOfError(self):
        '''
        If a publication produces an error, here we must notify the reason why it happened. This will be called just after
        publish or checkPublishingState if they return State.ERROR
        '''
        return self._reason

    def cancel(self):
        return self.destroy()

    def destroy(self):
        self._state = State.FINISHED
        return self._state
    
    def getConsoleConnection(self):
        return self.service().getConsoleConnection(self._vmid)

    def replace(self, data):
        self._ip = ''
        self._vmid = data.get('vm_id')

    def setupQueueForReplacement(self):
        pass

    def __startMachine(self):
        """
        Powers on the machine, when it is in 'down' or 'suspended' state
        """
        new_state = State.FINISHED
        state = self.service().getMachineState(self._vmid)

        # VM_STATE_UNKNOWN means that virtual machine with self._vmid does not exist
        if state == VM_STATE_UNKNOWN:
            raise Exception('Machine not found')

        # User expects to get running machine
        if state in (VM_STATE_DOWN, VM_STATE_SUSPENDED):
            self.service().startMachine(self._vmid)
            new_state = State.RUNNING

        return new_state

    def __checkMachineState(self):
        """
        Get the machine's state from provider and convert it to State format

        Provider returns states:
             unassigned, down, up, powering_up, powered_down,
             paused, migrating_from, migrating_to, unknown, not_responding,
             wait_for_launch, reboot_in_progress, saving_state, restoring_state,
             suspended, image_illegal, image_locked or powering_down

        But the class expects one of State.FINISHED, State.RUNNING, State.ERROR
        """

        vmstate = self.service().getMachineState(self._vmid)

        # state 'unknown' means that virtual machine with self._vmid does not exist
        if vmstate == VM_STATE_UNKNOWN:
            raise Exception('Machine not found')

        if vmstate in (VM_STATE_DOWN, VM_STATE_UP, VM_STATE_PAUSED, VM_STATE_SUSPENDED, VM_STATE_POWERED_DOWN, ):
            # All of ended states are correct to State.FINISHED
            state = State.FINISHED

        elif vmstate in (VM_STATE_POWERING_UP, VM_STATE_MIGRATING_FROM, VM_STATE_MIGRATING_TO, VM_STATE_WAIT_FOR_LAUNCH,
                         VM_STATE_REBOOT_IN_PROGRESS, VM_STATE_SAVING_STATE, VM_STATE_RESTORING_STATE,
                         VM_STATE_IMAGE_LOCKED, VM_STATE_POWERING_DOWN, ):
            # All of continuous states are correct to State.RUNNING
            state = State.RUNNING

        else:
            # All of other states means an error occurs
            state = State.ERROR

        return state
