"""
Created on Feb 20, 2020

@author: dvd
"""
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import json

from uds.core.services import UserDeployment
from uds.core.util import log
from uds.core.util.State import State

from termidesk.services.vair.VAirJobs import VAirDeferredRemoval
from termidesk.services.vair.client.v2 import VM_NOT_FOUND, VM_DOWN

logger = logging.getLogger(__name__)

opError, opFinish, opRequestDeployVm, opWaitDeployVm, opRequestPowerOnVm, opWaitPowerOnVm, opGetVmMac, \
opWaitReadyFromOsManager, opRequestDestroyVm, opWaitDestroyVm, opRequestSuspendVm, opWaitSuspendVm, \
opRequestResumeVm, opWaitResumeVm, opRequestDestroyDisk = range(15)

NO_MORE_NAMES = 'NO-NAME-ERROR'

OP_LABELS = {
    opError: 'error',
    opFinish: 'finish',
    opRequestDeployVm: 'requestDeployVm',
    opWaitDeployVm: 'waitDeployVm',
    opRequestPowerOnVm: 'requestPowerOnVm',
    opWaitPowerOnVm: 'waitPowerOnVm',
    opGetVmMac: 'getVmMac',
    opWaitReadyFromOsManager: 'waitReadyFromOsManager',
    opRequestDestroyVm: 'requestDestroyVm',
    opWaitDestroyVm: 'waitDestroyVm',
    opRequestDestroyDisk: 'requestDestroyDisk',
    opRequestSuspendVm: 'requestSuspendVm',
    opWaitSuspendVm: 'waitSuspendVm',
    opRequestResumeVm: 'requestResumeVm',
    opWaitResumeVm: 'waitResumeVm',
}


class VAirLinkedDeployment(UserDeployment):
    """
    This class generates the user consumable elements of the service tree.

    After creating at administration interface an Deployed Service, UDS will
    create consumable services for users using UserDeployment class as
    provider of this elements.


    At class instantiation, this will receive an environment with"generator",
    that are classes that provides a way to generate unique items.

    The generators provided right now are 'mac' and 'name'. To get more info
    about this, look at py:class:`uds.core.util.UniqueMacGenerator.UniqueNameGenerator`
    and py:class:`uds.core.util.UniqueNameGenerator.UniqueNameGenerator`

    This first sample do not uses cache. To see one with cache, see
    SampleUserDeploymentTwo. The main difference are the "...Cache".." methods,
    that here are not needed.

    As sample also of environment storage usage, wi will use here the provider
    storage to keep all our needed info, leaving marshal and unmarshal (needed
    by Serializble classes, like this) empty (that is, returns '' first and does
    nothing the second one)

    Also Remember, if you don't include this class as the deployedType of the
    SampleServiceOne, or whenever you trie to access a service of SampleServiceOne,
    you will get an excetion that says that you havent included the deployedType.
    """

    suggestedTime = 5  #: Suggested recheck time if publication is unfinished in seconds

    def initialize(self):
        """
        This method will be invoked by default __init__ of base class, so it gives
        us the oportunity to initialize whataver we need here.

        In our case, we setup a few attributes..
        """

        # We do not check anything at marshal method, so we ensure that
        # default values are correctly handled by marshal.
        self._name = ''
        self._ip = ''
        self._mac = ''
        self._vm_id = ''
        self._reason = ''
        self._queue = []
        self._task_id = ''
        self._cancelled = False

        self._initial_name = ""
        self._disk_ids = []

    # Serializable needed methods
    def marshal(self):
        data = {
            'version': 2,
            'name': self._name,
            'ip': self._ip,
            'mac': self._mac,
            'vm_id': self._vm_id,
            'reason': self._reason,
            'queue': self._queue,
            'task_id': self._task_id,
            'cancelled': self._cancelled,
            'initial_name': self._initial_name,
            'disks': self._disk_ids,
        }

        return json.dumps(data)

    def unmarshal(self, str_):
        if isinstance(str_, bytes):
            str_ = str_.decode('utf-8')

        data = json.loads(str_)
        if data['version'] == 2:
            self._name = data['name']
            self._ip = data['ip']
            self._mac = data['mac']
            self._vm_id = data['vm_id']
            self._reason = data['reason']
            self._queue = data['queue']
            self._task_id = data.get('task_id', '')
            self._cancelled = data.get('cancelled', False)
            self._initial_name = data.get('initial_name', '')
            self._disk_ids = data.get('disks', [])
        else:
            raise Exception("Unsupported serialized data version {}".format(data['version']))

    def getName(self):
        """
        We override this to return a name to display. Default inplementation
        (in base class), returns getUniqueIde() value
        This name will help user to identify elements, and is only used
        at administration interface.
        """
        if self._name == '':
            try:
                # gen = self.gidGenerator()
                # gen.setBaseName('')
                # number_suffix = str(int(gen.get()))
                # self._name = self.service().getBaseName() + '-' + number_suffix
                base_name = self.service().getBaseName()
                length = self.service().getLenName()
                _name = self.nameGenerator().get(
                    base_name,
                    length
                )
                self.doLog(log.INFO, 'Set name to "{}"'.format(_name))

                self._initial_name = _name
                self._name = _name

            except KeyError:
                self.doLog(log.ERROR, 'No more names available for machine "{}"'.format(self._name))
                return NO_MORE_NAMES

        return self._name

    def setIp(self, ip):
        """
        In our case, there is no OS manager associated with this, so this method
        will never get called, but we put here as sample.

        Whenever an os manager actor notifies the broker the state of the service
        (mainly machines), the implementation of that os manager can (an probably will)
        need to notify the IP of the deployed service. Remember that UDS treats with
        IP services, so will probable needed in every service that you will create.
        :note: This IP is the IP of the "consumed service", so the transport can
               access it.
        """
        logger.debug('Setting IP to %s', ip)
        self._ip = ip
        self.doLog(log.INFO, 'Set IP to %s' % ip)

    def setReady(self):
        """
        The method is invoked whenever a machine is provided to an user, right
        before presenting it (via transport rendering) to the user.
        """
        if self.cache.get('ready') == '1':
            return State.FINISHED

        logger.debug('setReady Checking vm %s state', self._name)
        try:
            state = self.service().get_vm_state(self._vm_id)
            logger.debug('setReady Vm %s state %s', self._name, state)
            code = state['code']
            if code == VM_NOT_FOUND:
                return self.__error('Machine is not found.')

            if code == VM_DOWN:
                logger.debug('setReady Vm %s needs power on', self._name)
                self._queue = [opRequestPowerOnVm, opWaitPowerOnVm, opFinish]
                # Для ускорения выполняем первый шаг из очереди сейчас
                return self.checkState()

        except Exception as e:
            return self.__error('Machine is not available anymore: {}'.format(e))

        self.cache.put('ready', '1')
        self.doLog(log.INFO, "Machine is ready")
        return State.FINISHED

    def getUniqueId(self):
        """
        Return and unique identifier for this service.
        In our case, we will generate a mac name, that can be also as sample
        of 'mac' generator use, and probably will get used something like this
        at some services.

        The get method of a mac generator takes one param, that is the mac range
        to use to get an unused mac.
        """
        return self._mac.upper()

    def getIp(self):
        """
        We need to implement this method, so we can return the IP for transports
        use. If no IP is known for this service, this must return None

        If our sample do not returns an IP, IP transport will never work with
        this service. Remember in real cases to return a valid IP address if
        the service is accesible and you alredy know that (for example, because
        the IP has been assigend via setIp by an os manager) or because
        you get it for some other method.

        Storage returns None if key is not stored.

        :note: Keeping the IP address is responsibility of the User Deployment.
               Every time the core needs to provide the service to the user, or
               show the IP to the administrator, this method will get called

        """
        return self._ip

    def deployForCache(self, cacheLevel):
        """
        Deploys an service instance for cache
        """
        self.__initQueueForDeploy(cacheLevel == self.L2_CACHE)
        return State.RUNNING

    def deployForUser(self, user):
        """
        Deploys an service instance for an user.
        """
        logger.debug('Deploying for user')
        self.__initQueueForDeploy(False)

        # Execute first step in queue now
        return self.checkState()

    def __initQueueForDeploy(self, forLevel2=False):

        if forLevel2 is False:
            self._queue = [
                opRequestDeployVm, opWaitDeployVm, opRequestPowerOnVm, opWaitPowerOnVm, opGetVmMac,
                opWaitReadyFromOsManager, opFinish
            ]
        else:
            self._queue = [
                opRequestDeployVm, opWaitDeployVm, opRequestPowerOnVm, opWaitPowerOnVm, opGetVmMac,
                opWaitReadyFromOsManager, opRequestSuspendVm, opWaitSuspendVm, opFinish
            ]

    def checkState(self):
        """
        Check what operation is going on, and acts accordingly to it
        """
        self.__debug('checkState')
        operation = self.__getCurrentOp()

        try:
            if operation == opError:
                return State.ERROR

            if operation == opFinish:
                return State.FINISHED

            functions = {
                opRequestDeployVm: self.__request_deploy_vm,
                opWaitDeployVm: self.__wait_deploy_vm,
                opRequestPowerOnVm: self.__request_poweron_vm,
                opWaitPowerOnVm: self.__wait_poweron_vm,
                opGetVmMac: self.__get_vm_mac,
                opWaitReadyFromOsManager: self.__wait_agent_registration,
                opRequestDestroyVm: self.__request_destroy_vm,
                opWaitDestroyVm: self.__wait_destroy_vm,
                opRequestSuspendVm: self.__request_suspend_vm,
                opWaitSuspendVm: self.__wait_suspend_vm,
                opRequestResumeVm: self.__request_resume_vm,
                opWaitResumeVm: self.__wait_resume_vm,
                opRequestDestroyDisk: self.__request_destroy_disk,
            }

            chk_fnc = functions.get(operation, None)

            if chk_fnc is None:
                return self.__error('Unknown operation found at check queue ({0})'.format(operation))

            state = chk_fnc()
            if state == State.FINISHED:
                self.__popCurrentOp()  # Remove runing op
                # Для ускорения выполняем следующий шаг из очереди сейчас
                return self.checkState()

            return state
        except Exception as e:
            logger.exception('checkState Operation %s error', VAirLinkedDeployment.__op2str(operation))
            return self.__error(str(e))

    def finish(self):
        """
        In our case, finish does nothing
        """
        pass

    def reasonOfError(self):
        """
        If a publication produces an error, here we must notify the reason why
        it happened. This will be called just after publish or checkState
        if they return State.ERROR

        Returns an string, in our case, set at checkState
        """
        return self._reason

    def destroy(self):
        '''
        This is called once a VM is no more needed.

        This method do whatever needed to clean up things, such as
        removing created "external" data (environment gets cleaned by core),
        etc..

        The retunred value is the same as when publishing, State.RUNNING,
        State.FINISHED or State.ERROR.
        '''

        try:
            if not self._vm_id:
                self._queue.pop()  # Remove opFinish
                self._queue.extend([opRequestDestroyVm, opWaitDestroyVm, opRequestDestroyDisk, opFinish])
                return State.RUNNING

            if self._vm_id:
                self._queue = [opRequestDestroyVm, opWaitDestroyVm, opRequestDestroyDisk, opFinish]
                return State.RUNNING

        except Exception as e:
            self._state = 'error'
            self._reason = str(e)
            return State.ERROR

        return State.FINISHED

    def cancel(self):
        """
        Do same thing as destroy
        """
        self._cancelled = True

        self.doLog(log.INFO, 'Cancel machine')
        return self.destroy()

    def notifyReadyFromOsManager(self, data):
        # os manager reports vm is ready
        if self.__getCurrentOp() == opWaitReadyFromOsManager:
            logger.debug('Machine %s is ready.', self._name)
            self.doLog(log.INFO, 'OS Manager reports ready')
            self.__popCurrentOp()  # Remove current state

        return State.FINISHED

    def moveToCache(self, newLevel):
        """
        Перемещает ВМ в кеш первого уровня
        """

        if newLevel == self.L1_CACHE:
            logger.debug('Moving VM %s to L1 cache.', self._name)
            self.doLog(log.INFO, "Moving machine to L1 cache")
            self._queue = [opRequestResumeVm, opWaitResumeVm, opFinish]

        return State.RUNNING

    def getConsoleConnection(self):
        return self.service().get_console_connection(self._vm_id)

    def __get_task_state(self):
        task_info = self.service().get_task_info(self._task_id)

        if task_info['data']['progress'] is None or task_info['data']['progress'] < 100.0:
            if task_info['data']['error_code'] == 0:
                return State.RUNNING
            else:
                return self.__error(task_info['data']['message'])

        return State.FINISHED

    def __request_deploy_vm(self):
        vm_name = self.getName()
        if vm_name == NO_MORE_NAMES:
            raise Exception('No more names available for this service.')

        self._task_id = self.service().request_deploy_clone(vm_name)
        return State.FINISHED

    def __wait_deploy_vm(self):
        if self._cancelled:
            return State.FINISHED

        checking_state = self.__get_task_state()
        if checking_state == State.FINISHED:
            task_info = self.service().get_task_info(self._task_id)
            self._vm_id = task_info['data']['parameters']['vms_id'][0]

            # Итоговое имя формируется на стороне vAir - к базовому имени
            # автоматически добавляется счетчик, поэтому необходимо
            # скорректировать имя ВМ на стороне Термидеска
            vm_info = self.service().get_vm_info(self._vm_id)
            vm_name = vm_info.get('vm_name', '')

            self._name = self.service().sanitize_vair_vm_name(vm_name)
            self._disk_ids = vm_info['vm_disks']

            self.doLog(
                log.INFO,
                "Create machine with name {}, vmid {}".format(self._name, self._vm_id)
            )

        return checking_state

    def __request_poweron_vm(self):
        self._task_id = self.service().request_poweron_vm(self._vm_id)
        return State.FINISHED

    def __wait_poweron_vm(self):
        state = self.__get_task_state()
        if state == State.FINISHED:
            self.doLog(log.INFO, "Start machine")

        return state

    def __request_destroy_vm(self):
        self._task_id = self.service().request_delete_vm(self._vm_id)
        return State.FINISHED

    def __wait_destroy_vm(self):
        state = self.__get_task_state()
        if state == State.FINISHED:
            self.doLog(log.INFO, "Remove machine")

            # После удаления ВМ необходимо "освободить" соответствующий id в генераторе.
            self.doLog(log.INFO, 'Free machine name in generator: {} {}'.format(
                self.service().getBaseName(), self._initial_name
            ))
            self.nameGenerator().free(self.service().getBaseName(), self._initial_name)

        return state

    def __request_destroy_disk(self):
        for disk_id in self._disk_ids:
            self.service().request_delete_disk(disk_id)
            self.doLog(log.INFO, 'Remove disk "{}" requested.'.format(disk_id))

        return State.FINISHED

    def __get_vm_mac(self):
        vm_info = self.service().get_vm_info(self._vm_id)
        self._mac = vm_info['mac_address']
        return State.FINISHED

    def __wait_agent_registration(self):
        '''
        Executes opWait, it simply waits something "external" to end
        '''
        if self._cancelled:
            # Если ВМ отменена, то уведомление о регистрации не приходит
            return State.FINISHED
        else:
            # Ожидание регистрации агента ВМ на брокере
            return State.RUNNING

    def __request_suspend_vm(self):
        self._task_id = self.service().request_suspend_vm(self._vm_id)
        return State.FINISHED

    def __wait_suspend_vm(self):
        state = self.__get_task_state()
        if state == State.FINISHED:
            self.doLog(log.INFO, "Suspend machine")

        return state

    def __request_resume_vm(self):
        self._task_id = self.service().request_resume_vm(self._vm_id)
        return State.FINISHED

    def __wait_resume_vm(self):
        return self.__get_task_state()

    def __getCurrentOp(self):
        if len(self._queue) == 0:
            return opFinish

        return self._queue[0]

    def __popCurrentOp(self):
        if len(self._queue) == 0:
            return opFinish

        res = self._queue.pop(0)
        return res

    def __pushFrontOp(self, op):
        self._queue.insert(0, op)

    def __pushBackOp(self, op):
        self._queue.append(op)

    def __error(self, reason: str):
        """
        Internal method to set object as error state

        Returns:
            State.ERROR, so we can do "return self.__error(reason)"
        """
        logger.debug('Setting error state, reason: %s', reason)
        self.doLog(log.ERROR, reason)

        if self._name != '':
            if self._is_dev_mode():
                self.add_error_log_about_trying_delete_in_debug_mode(logger, reason)
            else:
                VAirDeferredRemoval.remove(self.service().parent(), self._vm_id)

        self._queue = [opError]
        self._reason = reason

        return State.ERROR

    @staticmethod
    def __op2str(op):
        return OP_LABELS.get(op, '????')

    def __debug(self, txt):
        try:
            tpl = "[vAir debug] {state} VM id: {id}, name: {name}, ip: {ip}, mac: {mac}, queue: {queue}"
            queue = "[{}]".format(", ".join([VAirLinkedDeployment.__op2str(op) for op in self._queue]))
            msg = tpl.format(state=txt, id=self._vm_id, name=self._name, ip=self._ip, mac=self._mac,
                             queue=str(queue))
            logger.debug(msg)
        except Exception as e:
            logger.exception(e)

    def assignToUser(self, user):
        """
        This method is invoked whenever a cache item gets assigned to a user.
        This gives the User Deployment an opportunity to do whatever actions
        are required so the service puts at a correct state for using by a service.
        """
        self.doLog(log.INFO, "Cache item gets assigned to user: {}".format(user))
        super().assignToUser(user)

    def userLoggedIn(self, user):
        """
        This method must be available so os managers can invoke it whenever
        a user get logged into a service.

        The user provided is just a string, that is provided by actor.
        """
        # We store the value at storage, but never get used, just an example
        self.doLog(log.INFO, 'Login user: {}'.format(user))
        super().userLoggedIn(user)

    def userLoggedOut(self, user):
        """
        This method must be available so os managers can invoke it whenever
        a user get logged out if a service.

        The user provided is just a string, that is provided by actor.
        """
        self.doLog(log.INFO, 'Logout user: {}'.format(user))
        super().userLoggedOut(user)
