"""
Created on Feb 20, 2020

@author: dvd
"""
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging

from django.utils.translation import ugettext_noop as _
from uds.core.services import Service, types as serviceTypes
from uds.core.transports import protocols
from uds.core.ui import gui

from termidesk.services.vair.VAirPublication import VAirPublication
from termidesk.services.vair.VAirLinkedDeployment import VAirLinkedDeployment
from termidesk.services.vair.helpers import VAirHelpers

logger = logging.getLogger(__name__)


class VAirLinkedService(Service):
    """
    Basic service, the first part (variables) include the description of the service.

    Remember to fill all variables needed, but at least you must define:
        * typeName
        * typeType
        * typeDescription
        * iconFile (defaults to service.png)
        * publicationType, type of publication in case it needs publication.
          If this is not provided, core will assume that the service do not
          needs publishing.
        * deployedType, type of deployed user service. Do not forget this!!!

    The rest of them can be ommited, but its recommended that you fill all
    declarations shown in this sample (that in fact, are all)

    This description informs the core what this service really provides,
    and how this is done. Look at description of class variables for more
    information.

    """
    # : Name to show the administrator. This string will be translated BEFORE
    # : sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    typeName = _('Полный клон')
    # : Type used internally to identify this provider
    typeType = 'vairLinkedService'
    # : Description shown at administration interface for this provider
    typeDescription = _('Рабочее место на основе полного клона ВМ')
    # : Icon file used as icon for this provider. This string will be translated
    # : BEFORE sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    iconFile = 'service.png'

    # Functional related data

    # : If the service provides more than 1 "deployed user" (-1 = no limit,
    # : 0 = ???? (do not use it!!!), N = max number to deploy
    maxDeployed = -1
    # : If we need to generate "cache" for this service, so users can access the
    # : provided services faster. Is usesCache is True, you will need also
    # : set publicationType, do take care about that!
    usesCache = True
    # : Tooltip shown to user when this item is pointed at admin interface, none
    # : because we don't use it
    cacheTooltip = _('Желаемое количество активных ВМ в ожидании подключения')
    # : If we need to generate a "Level 2" cache for this service (i.e., L1
    # : could be running machines and L2 suspended machines)
    usesCache_L2 = True
    # : Tooltip shown to user when this item is pointed at admin interface, None
    # : also because we don't use it
    cacheTooltip_L2 = _('Желаемое количество спящих ВМ в ожидании подключения')

    # : If the service needs a s.o. manager (managers are related to agents
    # : provided by services itselfs, i.e. virtual machines with actors)
    needsManager = True
    # : If true, the system can't do an automatic assignation of a deployed user
    # : service from this service
    mustAssignManually = False

    # : Types of publications (preparated data for deploys)
    # : In our case, we do no need a publication, so this is None
    publicationType = VAirPublication
    # : Types of deploys (services in cache and/or assigned to users)
    deployedType = VAirLinkedDeployment

    allowedProtocols = protocols.GENERIC
    servicesTypeProvided = (serviceTypes.VDI,)

    # Параметры формы

    # TSRV-506: параметр убран с формы, до реализации кастомизации создаваемых образов.
    # minSpaceGB = gui.NumericField(
    #     length=3,
    #     label=_('Место (Гб)'),
    #     defvalue='32',
    #     order=105,
    #     tooltip=_('Минимальный размер свободного места в Гб'),
    #     required=True
    # )

    machine = gui.ChoiceField(
        label=_("Базовая ВМ"),
        order=110,
        tooltip=_('Базовый шаблон ВМ для рабочего места'),
        tab=_('Machine'),
        required=True
    )

    base_name = gui.TextField(
        length=20,
        label=_('Базовое имя'),
        rdonly=False,
        order=115,
        tooltip=_('Базовое имя для клонов данной машины. Заглавные латинские, подчеркивание'),
        tab=_('Machine'),
        required=True
    )

    name_suffix_len = gui.TextField(
        length=1,
        label=_('Длина суффикса'),
        rdonly=False,
        defvalue=3,
        order=116,
        tooltip=_('Длина суффикса для формирования имени ВМ, например, 2 - "00", 3 - "000"'),
        tab=_('Machine'),
        required=True
    )

    ov = gui.HiddenField(value=None)
    ev = gui.HiddenField(value=None)  # We need to keep the env so we can instantiate the Provider

    def initGui(self):
        """
        Loads required values inside
        """

        # Here we have to use "default values", cause values aren't used at form initialization
        # This is that value is always '', so if we want to change something, we have to do it
        # at defValue
        self.ov.defValue = self.parent().serialize()
        self.ev.defValue = self.parent().env.key

        machines = self.parent().get_vm_template_list()
        logger.info("Templates: %s", machines)
        vals = []
        for machine in machines:
            vals.append(gui.choiceItem(machine['id'], machine['name']))
        self.machine.setValues(vals)

    def sanitizeVmName(self, name):
        import re
        return re.sub("[^A-Z0-9_-]", "_", name)

    def sanitize_vair_vm_name(self, name):
        """
        Корректирует имя ВМ, полученное со стенда vAir
        """

        # Чтобы имя можно было назначить в качестве hostname
        # оно должно состоять из
        # * латинских букв от "a" до "z"
        # * цифр от "0" до "9"
        # * знака "дефис" - "-"
        import re
        return re.sub("[^A-Z0-9-]", "-", name.upper())

    def getBaseName(self):
        """
        Returns the base name
        """
        return self.base_name.value

    def getLenName(self):
        """
        Returns the length of numbers part
        """
        return int(self.name_suffix_len.value)

    def request_deploy_clone(self, vm_name):
        return self.parent().request_deploy_clone(self.machine.value, vm_name, self.name_suffix_len.value)

    def get_task_info(self, task_id):
        return self.parent().get_task_info(task_id)

    def get_vm_state(self, vm_id):
        return self.parent().get_vm_state(vm_id)

    def get_vm_info(self, vm_id):
        return self.parent().get_vm_info(vm_id)

    def request_poweron_vm(self, vm_id):
        '''
        Invokes request_poweron_vm from parent provider
        '''
        return self.parent().request_poweron_vm(vm_id)

    def request_suspend_vm(self, vm_id):
        return self.parent().request_suspend_vm(vm_id)

    def request_resume_vm(self, vm_id):
        return self.parent().request_resume_vm(vm_id)

    def request_delete_vm(self, vm_id):
        return self.parent().request_delete_vm(vm_id)

    def request_delete_disk(self, disk_id):
        return self.parent().request_delete_disk(disk_id)

    def get_console_connection(self, vm_id):
        return self.parent().get_console_connection(vm_id)
