# -*- coding: utf-8 -*-
'''
Created on Feb 20, 2020

@author: dvd
'''
from __future__ import unicode_literals

from django.utils.translation import ugettext_noop as _
from uds.core.services import ServiceProvider
from uds.core.ui import gui
from uds.core.util import validators

from . import client
from .VAirLinkedService import VAirLinkedService
from .VAirStaticService import VAirStaticService
import logging

logger = logging.getLogger(__name__)

class Provider(ServiceProvider):
    
    # : What kind of services we offer, this are classes inherited from Service
    offers = [VAirLinkedService, VAirStaticService]
    # : Name to show the administrator. This string will be translated BEFORE
    # : sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    typeName = _('Платформа vAir')
    # : Type used internally to identify this provider
    typeType = 'vairPlatform'
    # : Description shown at administration interface for this provider
    typeDescription = _('Поставщик ресурсов на базе vAir')
    # : Icon file used as icon for this provider. This string will be translated
    # : BEFORE sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    iconFile = 'provider.png'

    # now comes the form fields
    # There is always two fields that are requested to the admin, that are:
    # Service Name, that is a name that the admin uses to name this provider
    # Description, that is a short description that the admin gives to this provider
    # Now we are going to add a few fields that we need to use this provider
    # Remember that these are "dummy" fields, that in fact are not required
    # but used for sample purposes
    # If we don't indicate an order, the output order of fields will be
    # "random"
    platformVersion = gui.ChoiceField(order=1,
        label='Версия API платформы',
        tooltip='Версия api vAir',
        # In this case, the choice can have none value selected by default
        required=True,
        rdonly=False,
        values=[
            gui.choiceItem('2', '2'),
        ],
        defvalue='2'  # Default value is the ID of the choicefield
    )

    host = gui.TextField(length=64, label='Адрес сервера', order=2, tooltip=_('IP-адрес или доменное имя сервера vAir'), required=True)
    port = gui.NumericField(length=5, label='Порт', defvalue='443', order=3, tooltip=_('Порт для подключения к api'), required=True)
    ssl = gui.CheckBoxField(label=_('Использовать SSL'), defvalue='true', order=4, tooltip=_('Форсировать использование SSL (не будет работать, если сервер не поддерживает SSL)'))
    
    username = gui.TextField(length=32, label=_('Логин'), order=5, tooltip=_('Пользователь с необходимыми правами в vAir'), required=True, defvalue='')
    password = gui.PasswordField(lenth=32, label=_('Пароль'), order=6, tooltip=_('Пароль пользователя'), required=True)
    
    maxPreparingServices = gui.NumericField(length=7, label=_('Подготавливать ВМ одновременно'), defvalue='10', minValue=1, maxValue=65536, order=50, tooltip=_('Максимальное число одновременно создаваемых виртуальных машин'), required=True, tab=gui.ADVANCED_TAB)
    maxRemovingServices = gui.NumericField(length=8, label=_('Удалять ВМ одновременно'), defvalue='5', minValue=1, maxValue=65536, order=51, tooltip=_('Макимальное число одновременно удаляемых виртуальных машин'), required=True, tab=gui.ADVANCED_TAB)

    # Own variables
    _api = None

    def __getApi(self):
        '''
        Returns the connection API object for VMware
        '''

        if self._api is None:
            client_class = client.get_client(self.platformVersion.value)
            self._api = client_class(
                self.host.value, self.port.value, self.ssl.isTrue(), self.username.value, self.password.value
            )
        return self._api

    # There is more fields type, but not here the best place to cover it
    def initialize(self, values=None):
        '''
        We will use the "autosave" feature for form fields
        '''

        # Just reset _api connection variable
        self._api = None

    def testConnection(self):
        '''
        Test that conection to server is fine

        Returns

            True if all went fine, false if id didn't
        '''

        return self.__getApi().test()
    
    @staticmethod
    def test(env, data):
        '''
        Test VMware Connectivity

        Args:
            env: environment passed for testing (temporal environment passed)

            data: data passed for testing (data obtained from the form
            definition)

        Returns:
            Array of two elements, first is True of False, depending on test
            (True is all right, false is error),
            second is an String with error, preferably internacionalizated..

        '''
        prov = Provider(env, data)
        if prov.testConnection() is True:
            return [True, _("Подключение прошло успешно.")]

        return [False, _("Не удалось подключиться. Проверьте параметры подключения.")]
    
    def get_task_info(self, task_id):
        '''
        Получает информацию о зaдаче в VAir.

        Args:

        Returns
            Dictionary, containing:
        '''
        return self.__getApi().get_task_info(task_id)
    
    def get_vm_list(self):
        '''
        Получает список машин в VAir.

        Args:

        Returns
            An array of dictionaries, containing:
                'name'
                'state'
        '''

        return self.__getApi().get_vm_list()
    
    def get_vm_state(self, vm_id):
        '''
        Получает статус машины в VAir.

        Args:

        Returns
            Dictionary, containing:
                'code'
                'desc'
        '''
        return self.__getApi().get_vm_state(vm_id)
    
    def get_vm_info(self, vm_id):
        '''
        Получает информацию о ВМ в VAir.

        Args:

        Returns
            Dictionary, containing:
                'mac_address'
                'vm_state'
                'vnc_port'
                'node'
        '''
        return self.__getApi().get_vm_info(vm_id)
    
    def get_vm_template_list(self):
        '''
        Получает список шаблонов в VAir.

        Args:

        Returns
            An array of dictionaries, containing:
                'name'
                'state'
        '''

        return self.__getApi().get_vm_template_list()
    
    def request_deploy_clone(self, template_id, vm_name, suffix_len):
        return self.__getApi().request_deploy_clone(template_id, vm_name, suffix_len)
    
    def request_poweron_vm(self, vm_id):
        return self.__getApi().request_poweron_vm(vm_id)
    
    def request_suspend_vm(self, vm_id):
        return self.__getApi().request_suspend_vm(vm_id)
    
    def request_resume_vm(self, vm_id):
        return self.__getApi().request_resume_vm(vm_id)
    
    def request_delete_vm(self, vm_id):
        return self.__getApi().request_delete_vm(vm_id)

    def request_delete_disk(self, disk_id):
        return self.__getApi().request_delete_disk(disk_id)
    
    def get_console_connection(self, vm_id):
        info = self.__getApi().get_display_connection(vm_id)
        
        return {
            'type': info['type'],
            'address': info['host'],
            'port': info['port'],
            'secure_port':-1,
            'monitors': 1,
            'cert_subject': '',
            'ticket': {
                'value': info['password'],
                'expiry': ''
            }
        }
