# -*- coding: utf-8 -*-
'''
Created on Feb 20, 2020

@author: dvd
'''
from __future__ import unicode_literals

from django.utils.translation import ugettext as _
from uds.core.services import Publication
from uds.core.util.State import State
from datetime import datetime
import logging
import json

logger = logging.getLogger(__name__)

opError, opFinish = range(2) 

OP_LABELS = {
    opError: 'error',
    opFinish: 'finish',
}

class VAirPublication(Publication):
    '''
    This class shows how a publication is developed.
    
    In order to a publication to work correctly, we must provide at least the
    following methods:
        * Of course, the __init__
        * :py:meth:`.publish` 
        * :py:meth:`.checkState`
        * :py:meth:`.finish`
        
    Also, of course, methods from :py:class:`uds.core.Serializable.Serializable`
    
    
    Publication do not have an configuration interface, all data contained
    inside an instance of a Publication must be serialized if you want them between
    method calls.
    
    It's not waranteed that the class will not be serialized/deserialized
    between methods calls, so, first of all, implement the marshal and umnarshal
    mehods needed by all serializable classes.
    
    Also a thing to note is that operations requested to Publications must be
    *as fast as posible*. The operations executes in a separated thread,
    and so it cant take a bit more time to execute, but it's recommended that
    the operations executes as fast as posible, and, if it will take a long time,
    split operation so we can keep track of state.
    
    This means that, if we have "slow" operations, we must
    
    We first of all declares an estimation of how long a publication will take.
    This value is instance based, so if we override it in our class, the suggested
    time could change.
    
    The class attribute that indicates this suggested time is "suggestedTime", and
    it's expressed in seconds, (i.e. "suggestedTime = 10") 
    '''
    suggestedTime = 5 #: Suggested recheck time if publication is unfinished in seconds
    
    def initialize(self):
        '''
        This method will be invoked by default __init__ of base class, so it gives
        us the oportunity to initialize whataver we need here. 
        
        In our case, we setup a few attributes..
        '''
        
        # We do not check anything at marshal method, so we ensure that
        # default values are correctly handled by marshal.
        self._name = ''
        self._reason = '' # No error, no reason for it
        self._queue = []
        self._template_id = ''
    
    # Serializable needed methods
    def marshal(self):
        data = {
            'version': 2,
            'name': self._name,
            'reason': self._reason,
            'queue': self._queue,
            'template_id': self._template_id,
        }
        
        return json.dumps(data)
        

    def unmarshal(self, str_):
        if isinstance(str_, bytes):
            str_ = str_.decode('utf-8')

        data = json.loads(str_)
        if data['version']==2:
            self._name = data['name']
            self._reason = data['reason']
            self._queue = data['queue']
            self._template_id = data['template_id']
        else:
            raise Exception("Unsupported serialized data version {}".format(data['version']))
    
    def publish(self):
        return State.FINISHED
    
    def destroy(self):
        '''
        This is called once a publication is no more needed.

        This method do whatever needed to clean up things, such as
        removing created "external" data (environment gets cleaned by core),
        etc..

        The retunred value is the same as when publishing, State.RUNNING,
        State.FINISHED or State.ERROR.
        '''
        
        # Шаблоны в vAir удаляются пока вручную
        
#         try:
#             if not self._template_id:
#                 self._cancelled = True
#                 return State.RUNNING
#             
#             if self._template_id:
#                 self.service().removeTemplate(self._template_id)
#             elif self._clone_id:
#                 self.service().removeTemplate(self._clone_id)
#                 
#         except Exception as e:
#             self._state = 'error'
#             self._reason = str(e)
#             return State.ERROR

        return State.FINISHED

    def cancel(self):
        '''
        Do same thing as destroy
        '''
        return self.destroy()