"""
Created on Feb 20, 2020

@author: dvd

Updated on Jan 10, 2022 by rz@uveon.ru
"""
# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import logging
import threading
import random
import requests

logger = logging.getLogger(__name__)
context = threading.local()
context.token = None

lock = threading.Lock()

REQUEST_TIMEOUT = 30  # Seconds

GET = 'GET'
POST = 'POST'

VM_NOT_FOUND = -1
VM_UP = 1
VM_DOWN = 0
VM_SUSPENDED = 3

SUCCESS = "success"


def connect_vair(method):
    """
    Контекстный декоратор для подключения к vAir API
    Сохраняет токен в локальную контекстную переменную текущего потока для использования в методах клиента
    Поскольку действительный API ключ у пользователя может быть только один, а доступ может осуществляться
    из разных потоков одновременно, то необходимо использовать механизм блокировки (Lock)

    Примечание: Для координации доступа из разных процессов (например, метод test) нужно что-то еще...
                Этот вопрос пока не решен.
    """
    def wrapper(*args, **kwargs):
        instance = args[0]

        try:
            lock.acquire(True)
            context.token = instance.login()
            return method(*args, **kwargs)
        except Exception as exc:
            logger.exception(exc)
            raise
        finally:
            lock.release()
    return wrapper


class ApiAccessError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class OperationError(Exception):
    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class Client:

    def __init__(self, host, port, ssl, username, password):
        self._host = host
        self._port = port
        self._ssl = ssl
        self._username = username
        self._password = password
        self._proto = 'http'

        if self._ssl:
            self._proto = 'https'

        self._base_url = "{proto}://{host}:{port}"\
            .format(proto=self._proto, host=self._host, port=self._port)

        logger.debug("Initialized for %s", self._base_url)

    def test(self):

        logger.debug("Test connection to %s:%s for user %s", self._host, self._port, self._username)
        try:
            # На всякий случай получаем блокировку.
            # Данный метод обычно вызывается в контексте процесса веб-сервера, а не фонового рабочего.
            # В этом случае блокировка потока здесь ничего не даст.
            lock.acquire(True)
            self.login()
            return True
        except Exception as e:
            logger.exception(e)
        finally:
            lock.release()
        return False

    def login(self):
        url = "{base}/api/v2/auth/login/".format(base=self._base_url)
        params = {
            'login': self._username,
            'password': self._password,
        }
        response = self._call_api(url, params, method=POST)

        token = response['token']

        return token

    @connect_vair
    def get_task_info(self, task_id):
        token = context.token
        url = "{base}/api/v2/task/{id}/".format(base=self._base_url, id=task_id)
        response = self._call_api(url, token=token)

        return response

    @connect_vair
    def get_tasks_list(self):
        token = context.token
        url = "{base}/api/v2/tasks/".format(base=self._base_url)
        response = self._call_api(url, token=token)

        return response

    @connect_vair
    def get_vm_list(self):
        token = context.token
        url = "{base}/api/v2/virtual-machines/".format(base=self._base_url)
        params = {
            "token": token,
        }
        response = self._call_api(url, **params)

        result = []
        for item in response['data']:
            vm_info = {
                'id': item['id'],
                'name': item['name'],
                'state': item['state']['state'],
            }
            result.append(vm_info)

        return result

    @connect_vair
    def get_vm_state(self, vm_id):
        token = context.token
        url = "{base}/api/v2/virtual-machines/{id}/".format(base=self._base_url, id=vm_id)
        params = {
            "token": token,
        }
        response = self._call_api(url, **params)

        vm_info = response['data'][0]
        state = {
            'code': vm_info['state']['state'],
            'desc': vm_info['state']['state'],
        }

        return state

    @connect_vair
    def get_vm_info(self, vm_id):
        token = context.token
        url = "{base}/api/v2/virtual-machines/{id}/".format(base=self._base_url, id=vm_id)
        params = {
            "token": token,
        }
        response = self._call_api(url, **params)

        vm_info = response['data'][0]
        disks = [disk["id"] for disk in vm_info["config"]["disks"] if disk["device"] != "cdrom"]
        result = {
            'node': vm_info['state']['node'],
            # FIXME: берется mac-адрес первого интерфейса,
            #  а что если интерфейсов несколько или ни одного?
            'mac_address': vm_info['config']['networks'][0]['mac'],
            'vm_state': vm_info['state']['state'],
            'vnc_port': vm_info['state']['graphics_port_internal'],
            'vm_name': vm_info['name'],
            'vm_disks': disks,
        }

        return result

    @connect_vair
    def get_vm_template_list(self):
        token = context.token
        response = self._get_templates(token)

        result = []
        for item in response['data']:
            vm_info = {
                'id': item['id'],
                'name': item['name'],
                'state': item['state']['state'],
            }
            result.append(vm_info)

        return result

    def _get_templates(self, token):
        url = "{base}/api/v2/virtual-machines-templates/".format(base=self._base_url)
        response = self._call_api(url, token=token)

        return response

    def get_template_by_id(self, template_id, token):
        templates = self._get_templates(token)["data"]
        template_by_id = None
        for template in templates:
            if template["id"] == template_id:
                template_by_id = template

        return template_by_id

    @connect_vair
    def request_deploy_clone(self, template_id, vm_name, suffix_len):
        token = context.token
        url = "{base}/api/v2/virtual-machines-templates/make-vm/".format(base=self._base_url)

        template = self.get_template_by_id(template_id, token)

        params = {
            "cpu": template["config"]["cpu"],
            "disks": [
                {
                    "id": disk["id"],
                    "iotune_rbs": disk["iotune_rbs"],
                    "iotune_riops": disk["iotune_riops"],
                    "iotune_wbs": disk["iotune_wbs"],
                    "iotune_wiops": disk["iotune_wiops"],
                    "name": disk["name"],
                    "pool_name": disk["name"].split('/')[0]
                } for disk in template["config"]["disks"]
            ],
            "id": template["id"],
            "networks": template["config"]["networks"],
            "quantity": "1",
            "ram": template["config"]["memory"],
            "start_value": "0",
            "suffix_len": suffix_len,
            "vm_name": vm_name
        }
        response = self._call_api(url, params, method=POST, token=token)

        task_id = response.get('id')
        return task_id

    @connect_vair
    def request_poweron_vm(self, vm_id):
        token = context.token
        url = "{base}/api/v2/virtual-machines/up/".format(base=self._base_url)

        nodes = self._get_nodes(token)
        node_names = list(nodes.keys())
        selected_node = random.choice(node_names)

        params = {
            'id': vm_id,
            'node': selected_node,
        }
        response = self._call_api(url, params, method=POST, token=token)

        task_id = response.get('id')
        return task_id

    @connect_vair
    def request_suspend_vm(self, vm_id):
        token = context.token
        url = "{base}/api/v2/virtual-machines/suspend/".format(base=self._base_url)

        params = {
            'id': vm_id,
        }
        response = self._call_api(url, params, method=POST, token=token)

        task_id = response.get('id')
        return task_id

    @connect_vair
    def request_resume_vm(self, vm_id):
        token = context.token
        url = "{base}/api/v2/virtual-machines/resume/".format(base=self._base_url)

        params = {
            'id': vm_id,
        }
        response = self._call_api(url, params, method=POST, token=token)

        task_id = response.get('id')
        return task_id

    @connect_vair
    def request_delete_vm(self, vm_id, delete_disk="no"):
        token = context.token
        url = "{base}/api/v2/virtual-machines/delete/".format(base=self._base_url)

        params = {
            'id': vm_id,
            'disk': delete_disk,
        }
        response = self._call_api(url, params, method=POST, token=token)

        try:
            if response['status'] != SUCCESS:
                raise Exception(response.get('error', str(response)))

            task_id = response['id']
        except KeyError as e:
            logger.exception("Error %s while parsing request_delete_vm response: %s", str(e), str(response))
            raise Exception("Error parsing delete VM response: %s", str(e))

        return task_id

    @connect_vair
    def request_delete_disk(self, disk_id):
        token = context.token
        url = "{base}/api/v2/volumes/delete/".format(base=self._base_url)

        params = {
            'id': disk_id,
        }
        response = self._call_api(url, params, method=POST, token=token)

        try:
            if response['status'] != SUCCESS:
                raise Exception(response.get('error', str(response)))

            task_id = response['id']
        except KeyError as e:
            logger.exception("Error %s while parsing request_delete_disk response: %s", str(e), str(response))
            raise Exception("Error parsing delete VM response: %s", str(e))

        return task_id

    @connect_vair
    def get_display_connection(self, vm_id):
        token = context.token
        url = "{base}/api/v2/virtual-machines/{id}/".format(base=self._base_url, id=vm_id)
        response = self._call_api(url, token=token)
        vm_info = response['data'][0]
        if vm_info["state"]["state"] != VM_UP:
            raise OperationError("The virtual machine '%s' is stopped or suspended." % vm_id)

        try:
            node_name = vm_info['state']['node']
            nodes = self._get_nodes(token)
            host_ip = nodes.get(node_name)

            display_type = vm_info['config']['graphics']['type']
            password = vm_info['config']['graphics']['password']

            port = vm_info['state']['graphics_port_internal']
            try:
                port = int(port)
            except TypeError:
                pass

        except Exception as e:
            logger.exception(
                "get_display_connection Failed to process response data for %s. Error: %s Response: %s",
                vm_id, str(e), str(response)
            )
            raise Exception("Failed to process display connection info. Error: %s" % str(e))

        return {
            'type': display_type,
            'host': host_ip,
            'port': port,
            'password': password
        }

    def _get_nodes(self, token):
        url = "{base}/api/v2/cluster-fs/".format(base=self._base_url)
        params = {
            "token": token,
        }
        response = self._call_api(url, **params)

        nodes = {}
        for dc in response['data']:
            nodes[dc["name"]] = dc["mgr_ip"]

        logger.debug("Nodes: %s", nodes)
        return nodes

    def _call_api(self, url, params=None, method=GET, token='', *args, **kwargs):
        headers = {"Content-Type": "application/json", "Accept": "application/json"}
        cookies = {'access_token_cookie': token}

        if params is None:
            params = {}

        if method not in [GET, POST]:
            raise Exception("Unsupported method type: {}".format(method))

        try:
            logger.debug("_call_api Request: %s %s %s", method, url, params)
            request_params = dict(
                headers=headers,
                json=params,
                timeout=REQUEST_TIMEOUT,
                cookies=cookies,
                **kwargs
            )
            if method == GET:
                r = requests.get(url, **request_params)

            else:
                # if method is POST
                r = requests.post(url, **request_params)

            logger.debug("_call_api Response code: %s content: %s",
                         r.status_code, r.content.decode('unicode-escape'))
        except Exception as e:
            logger.exception("Error %s calling vAir api. Method: %s URL: %s Params: %s",
                             str(e), method, url, str(params))
            raise ApiAccessError("vAir API access error: {}".format(str(e)))

        if not r.ok:
            msg = "Error calling vAir api. HTTP code: %s Method: %s URL: %s Params: %s"
            context = [r.status_code, method, url, str(params)]
            logger.error(msg, *context)
            raise ApiAccessError("vAir API access error. HTTP error: {}".format(r.status_code))

        try:
            response = r.json()
        except Exception as e:
            logger.exception("_call_api Failed to parse response JSON: %s", r.text)
            raise ApiAccessError("vAir API access error. JSON parse error: {}".format(e))

        if response['status'] != 'success':
            raise OperationError(response.get('error', r.text))

        return response
