# -*- coding: utf-8 -*-
'''
Created on Jan 20, 2020

@author: dvd
'''
from __future__ import unicode_literals

import codecs
from uds.core.services import UserDeployment
from uds.core.util.State import State
from uds.core.util import log
import logging
import json
import time

from .VMwareJobs import VMwareDeferredRemoval

logger = logging.getLogger(__name__)

opError, opFinish, opRequestDeployVm, opWaitDeployVm, opGetDeployedVm, opRequestStartVm, \
opWaitReadyFromOsManager, opRequestSuspendVm, opWaitSuspendVm, opRequestDestroyVm, opWaitDestroyVm, \
opRequestPowerOnVm, opWaitPowerOnVm, opRequestPowerOffVm, opWaitPowerOffVm, opWaitTask = range(16) 

NO_MORE_NAMES = 'NO-NAME-ERROR'

OP_LABELS = {
    opError: 'error',
    opFinish: 'finish',
    opRequestDeployVm: 'requestDeployVm',
    opWaitDeployVm: 'waitDeployVm', 
    opGetDeployedVm: 'getDeployedVm',
    opRequestStartVm: 'requestStartVm',
    opWaitReadyFromOsManager: 'waitReadyFromOsManager',
    opRequestSuspendVm: 'requestSuspendVm',
    opWaitSuspendVm: 'waitSuspendVm',
    opRequestDestroyVm: 'requestDestroyVm',
    opWaitDestroyVm: 'waitDestroyVm',
    opRequestPowerOnVm: 'requestPowerOnVm',
    opWaitPowerOnVm: 'waitPowerOnVm',
    opRequestPowerOffVm: 'requestPowerOffVm',
    opWaitPowerOffVm: 'waitPowerOffVm',
    opWaitTask: 'waitTask',
}

class VMwareLinkedDeployment(UserDeployment):
    '''
    This class generates the user consumable elements of the service tree.

    After creating at administration interface an Deployed Service, UDS will
    create consumable services for users using UserDeployment class as
    provider of this elements.


    At class instantiation, this will receive an environment with"generator",
    that are classes that provides a way to generate unique items.

    The generators provided right now are 'mac' and 'name'. To get more info
    about this, look at py:class:`uds.core.util.UniqueMacGenerator.UniqueNameGenerator`
    and py:class:`uds.core.util.UniqueNameGenerator.UniqueNameGenerator`

    This first sample do not uses cache. To see one with cache, see
    SampleUserDeploymentTwo. The main difference are the "...Cache".." methods,
    that here are not needed.

    As sample also of environment storage usage, wi will use here the provider
    storage to keep all our needed info, leaving marshal and unmarshal (needed
    by Serializble classes, like this) empty (that is, returns '' first and does
    nothing the second one)

    Also Remember, if you don't include this class as the deployedType of the
    SampleServiceOne, or whenever you trie to access a service of SampleServiceOne,
    you will get an excetion that says that you havent included the deployedType.
    '''
    
    suggestedTime = 5 #: Suggested recheck time if publication is unfinished in seconds
    
    def initialize(self):
        '''
        This method will be invoked by default __init__ of base class, so it gives
        us the oportunity to initialize whataver we need here. 
        
        In our case, we setup a few attributes..
        '''
        
        # We do not check anything at marshal method, so we ensure that
        # default values are correctly handled by marshal.
        self._name = ''
        self._ip = ''
        self._mac = ''
        self._vm_id = ''
        self._reason = ''
        self._queue = []
        self._task_id = ''
        self._cancelled = False
    
    # Serializable needed methods
    def marshal(self):
        data = {
            'version': 1,
            'name': self._name,
            'ip': self._ip,
            'mac': self._mac,
            'vm_id': self._vm_id,
            'reason': self._reason,
            'queue': self._queue,
            'task_id': self._task_id,
            'cancelled': self._cancelled,
        }
        
        return json.dumps(data)
        

    def unmarshal(self, str_):
        if isinstance(str_, bytes):
            str_ = str_.decode('utf-8')
        
        data = json.loads(str_)
        if data['version']==1:
            self._name = data['name']
            self._ip = data['ip']
            self._mac = data['mac']
            self._vm_id = data['vm_id']
            self._reason = data['reason']
            self._queue = data['queue']
            self._task_id = data['task_id']
            self._cancelled = data.get('cancelled', False)
        else:
            raise Exception("Unsupported serialized data version {}".format(data['version']))
    
    def getName(self):
        '''
        We override this to return a name to display. Default inplementation
        (in base class), returns getUniqueIde() value
        This name will help user to identify elements, and is only used
        at administration interface.

        We will use here the environment name provided generator to generate
        a name for this element.

        The namaGenerator need two params, the base name and a length for a
        numeric incremental part for generating unique names. This are unique for
        all UDS names generations, that is, UDS will not generate this name again
        until this name is freed, or object is removed, what makes its environment
        to also get removed, that makes all uniques ids (names and macs right now)
        to also get released.

        Every time get method of a generator gets called, the generator creates
        a new unique name, so we keep the first generated name cached and don't
        generate more names. (Generator are simple utility classes)
        '''
        if self._name == '':
            try:
                self._name = self.nameGenerator().get(self.service().getBaseName(), self.service().getLenName())
            except KeyError:
                return NO_MORE_NAMES
        return self._name

    def setIp(self, ip):
        '''
        In our case, there is no OS manager associated with this, so this method
        will never get called, but we put here as sample.

        Whenever an os manager actor notifies the broker the state of the service
        (mainly machines), the implementation of that os manager can (an probably will)
        need to notify the IP of the deployed service. Remember that UDS treats with
        IP services, so will probable needed in every service that you will create.
        :note: This IP is the IP of the "consumed service", so the transport can
               access it.
        '''
        logger.debug('Setting IP to {}'.format(ip))
        self._ip = ip
    
    def setReady(self):
        '''
        The method is invoked whenever a machine is provided to an user, right
        before presenting it (via transport rendering) to the user.
        '''
        if self.cache.get('ready') == '1':
            return State.FINISHED
        info = {}
        logger.debug('setReady Checking vm {} state'.format(self._vm_id))
        try:
            info = self.service().get_vm_info(self._vm_id)
            logger.debug('setReady Vm {} state {}'.format(self._vm_id, info))
            
            if info.get('power_state','') != 'poweredOn':
                logger.debug('setReady Vm {} needs power on'.format(self._vm_id))
                self._queue = [opRequestPowerOnVm, opWaitPowerOnVm, opFinish]
                # Для ускорения выполняем первый шаг из очереди сейчас
                return self.checkState()
        except Exception as e:
            return self.__error('Machine is not available anymore: {}'.format(e))
        
        self.cache.put('ready', '1')
        self.doLog(log.INFO, "Machine is ready")
        return State.FINISHED
    
    def getUniqueId(self):
        '''
        Return and unique identifier for this service.
        In our case, we will generate a mac name, that can be also as sample
        of 'mac' generator use, and probably will get used something like this
        at some services.

        The get method of a mac generator takes one param, that is the mac range
        to use to get an unused mac.
        '''
        return self._mac.upper()

    def getIp(self):
        '''
        We need to implement this method, so we can return the IP for transports
        use. If no IP is known for this service, this must return None

        If our sample do not returns an IP, IP transport will never work with
        this service. Remember in real cases to return a valid IP address if
        the service is accesible and you alredy know that (for example, because
        the IP has been assigend via setIp by an os manager) or because
        you get it for some other method.

        Storage returns None if key is not stored.

        :note: Keeping the IP address is responsibility of the User Deployment.
               Every time the core needs to provide the service to the user, or
               show the IP to the administrator, this method will get called

        '''
        return self._ip
    
    def deployForCache(self, cacheLevel):
        '''
        Deploys an service instance for cache
        '''
        self.__initQueueForDeploy(cacheLevel == self.L2_CACHE)
        return State.RUNNING
    
    def deployForUser(self, user):
        '''
        Deploys an service instance for an user.
        '''
        logger.debug('Deploying for user')
        self.__initQueueForDeploy(False)
        
        # Execute first step in queue now
        return self.checkState()
    
    def __initQueueForDeploy(self, forLevel2=False):

        if forLevel2 is False:
            self._queue = [opRequestDeployVm, opWaitDeployVm, opGetDeployedVm, opWaitReadyFromOsManager, opFinish]
        else:
            self._queue = [opRequestDeployVm, opWaitDeployVm, opGetDeployedVm, opWaitReadyFromOsManager, opRequestSuspendVm, opWaitSuspendVm, opFinish]
    
    def checkState(self):
        '''
        Check what operation is going on, and acts accordingly to it
        '''
        try:
            self.__debug('checkState')
            op = self.__getCurrentOp()
    
            if op == opError:
                return State.ERROR
    
            if op == opFinish:
                return State.FINISHED
    
            fncs = {
                opRequestDeployVm: self.__request_deploy_vm,
                opWaitDeployVm: self.__wait_deploy_vm,
                opGetDeployedVm: self.__get_deployed_vm,
                opRequestStartVm: self.__request_start_vm,
                opWaitReadyFromOsManager: self.__wait_agent_registration,
                opRequestSuspendVm: self.__request_suspend_vm,
                opWaitSuspendVm: self.__wait_suspend_vm,
                opRequestDestroyVm: self.__request_destroy_vm,
                opWaitDestroyVm: self.__wait_destroy_vm,
                opRequestPowerOnVm: self.__request_poweron_vm,
                opWaitPowerOnVm: self.__wait_poweron_vm,
                opRequestPowerOffVm: self.__request_poweroff_vm,
                opWaitPowerOffVm: self.__wait_poweroff_vm,
                opWaitTask: self.__wait_task,
            }

        
            chkFnc = fncs.get(op, None)

            if chkFnc is None:
                return self.__error('Unknown operation found at check queue ({0})'.format(op))

            state = chkFnc()
            if state == State.FINISHED:
                self.__popCurrentOp()  # Remove runing op
                # Для ускорения выполняем следующий шаг из очереди сейчас
                return self.checkState()

            return state
        except Exception as e:
            logger.error('checkState Operation %s error', VMwareLinkedDeployment.__op2str(op))
            logger.exception(e)
            return self.__error(str(e))
    
    def finish(self):
        '''
        In our case, finish does nothing
        '''
        pass

    def reasonOfError(self):
        '''
        If a publication produces an error, here we must notify the reason why
        it happened. This will be called just after publish or checkState
        if they return State.ERROR

        Returns an string, in our case, set at checkState
        '''
        return self._reason

    def destroy(self):
        '''
        This is called once a publication is no more needed.

        This method do whatever needed to clean up things, such as
        removing created "external" data (environment gets cleaned by core),
        etc..

        The retunred value is the same as when publishing, State.RUNNING,
        State.FINISHED or State.ERROR.
        '''

        try:
            if not self._vm_id:
                self._queue.pop() # Remove opFinish
                self._queue.extend([opRequestPowerOffVm, opWaitTask, opRequestDestroyVm, opWaitDestroyVm, opFinish])
                return State.RUNNING
            
            if self._vm_id:
                self._queue = [opRequestPowerOffVm, opWaitTask, opRequestDestroyVm, opWaitDestroyVm, opFinish]
                return State.RUNNING
                
        except Exception as e:
            self._state = 'error'
            self._reason = str(e)
            return State.ERROR

        return State.FINISHED

    def cancel(self):
        '''
        Do same thing as destroy
        '''
        self._cancelled = True
        self.doLog(log.INFO, 'Cancel machine')
        return self.destroy()
    
    def notifyReadyFromOsManager(self, data):
        '''
          Метод прерывает шаг ожидания подготовки рабочего места и 
          переводит его на следующий шаг.
          Ожидание прерывается по сигналу готовности от агента на ВМ, 
          передаваемого менеджером ОС. 
        '''
        if self.__getCurrentOp() == opWaitReadyFromOsManager:
            logger.debug('Machine %s is ready.', self._name)
            self.doLog(log.INFO, 'OS Manager reports ready')
            if opRequestSuspendVm in self._queue:
                logger.debug('Moving machine {0} to level 2'.format(self._name))
                self.doLog(log.INFO, "Moving machine to level 2")
            
            # Т.к. метод notifyReadyFromOsManager вызывается не из обработчика 
            # очереди, то форсируем переход к следующему шагу очереди, 
            # чтобы не зависнуть в состоянии opWait
            self.__popCurrentOp()
            return self.checkState()

        # Метод вызван повторно, шаг очереди opWait уже пройден.
        # Ничего не делаем.
        return State.FINISHED
    
    def assignToUser(self, user):
        '''
        This method is invoked whenever a cache item gets assigned to an user.
        This gives the User Deployment an oportunity to do whatever actions
        are required so the service puts at a correct state for using by a service.
        '''
        self.doLog(log.INFO, "Machine assigned to user: {}".format(user))
    
    def moveToCache(self, newLevel):
        '''
        Перемещает ВМ в кеш первого уровня
        '''

        if newLevel == self.L1_CACHE:
            logger.debug('Moving VM %s to L1 cache.', self._name)
            self.doLog(log.INFO, "Moving machine from level 2 to level 1")
            self._queue = [opRequestPowerOnVm, opWaitPowerOnVm, opFinish]

        return State.RUNNING
    
    def userLoggedIn(self, user):
        '''
        This method must be available so os managers can invoke it whenever
        an user get logged into a service.

        The user provided is just an string, that is provided by actor.
        '''
        # We store the value at storage, but never get used, just an example
        self.doLog(log.INFO, 'Login user: {}'.format(user))

    def userLoggedOut(self, user):
        '''
        This method must be available so os managers can invoke it whenever
        an user get logged out if a service.

        The user provided is just an string, that is provided by actor.
        '''
        self.doLog(log.INFO, 'Logout user: {}'.format(user))

    def replace(self, data):
        self._name = data.get('name')
        self._ip = ''

        # MAC адрес должен быть в верхнем регистре
        self._mac = data.get('unique_id').upper()

        self._vmid = data.get('vm_id')

    def setupQueueForReplacement(self):
        self._queue = [opRequestPowerOffVm, opRequestPowerOnVm, opWaitPowerOnVm, opFinish]
    
    def __request_deploy_vm(self):
        template_id = self.publication().get_template_id()
        template_name = self.publication().get_name()
        name = self.getName()
        if name == NO_MORE_NAMES:
            raise Exception('No more names available for this service. (Increase digits for this service to fix)')

        name = self.service().sanitizeVmName(name)  # oVirt don't let us to create machines with more than 15 chars!!!

        self._task_id = self.service().request_deploy_linked_clone(self._name, template_id)
        self.doLog(
            log.INFO,
            "Request to deploy a linked clone named {} from template {}".format(self._name, template_name)
        )
        return State.FINISHED
    
    def __request_start_vm(self):
        self._task_id = self.service().request_start_vm(self._vm_id)
        self.doLog(log.INFO, "Request to start machine")
        return State.FINISHED
    
    def __request_suspend_vm(self):
        self._task_id = self.service().request_suspend_vm(self._vm_id)
        self.doLog(log.INFO, 'Request to suspend machine')
        return State.FINISHED
    
    def __request_poweron_vm(self):
        self._task_id = self.service().request_poweron_vm(self._vm_id)
        self.doLog(log.INFO, 'Request to power on machine')
        return State.FINISHED
    
    def __request_poweroff_vm(self):
        self._task_id = self.service().request_poweroff_vm(self._vm_id)
        self.doLog(log.INFO, 'Request to power off machine')
        return State.FINISHED
    
    def __wait_deploy_vm(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось развернуть ВМ".encode('unicode-escape'))
        
        return State.FINISHED  
    
    def __wait_suspend_vm(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось перевести ВМ в режим 'suspend'".encode('unicode-escape'))
        
        self.doLog(log.INFO, 'Machine suspended')
        return State.FINISHED
    
    def __wait_poweron_vm(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось запустить ВМ'".encode('unicode-escape'))
        
        self.doLog(log.INFO, 'Machine powered on')
        return State.FINISHED
    
    def __wait_poweroff_vm(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось выключить ВМ".encode('unicode-escape'))
        
        self.doLog(log.INFO, 'Machine powered off')
        return State.FINISHED
    
    def __wait_destroy_vm(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось удалить ВМ".encode('unicode-escape'))
        
        self.doLog(log.INFO, 'Machine removed')
        return State.FINISHED
    
    def __wait_task(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        return State.FINISHED   
    
    def __wait_agent_registration(self):
        '''
        Executes opWait, it simply waits something "external" to end
        '''
        if self._cancelled:
            # Если ВМ отменена, то уведомление о регистрации не приходит
            return State.FINISHED
        else:
            # Ожидание регистрации агента ВМ на брокере
            return State.RUNNING
    
    def __get_deployed_vm(self):
        task_result = self.service().get_vm_task_result(self._task_id)
        self._vm_id = task_result['vm_id']
        self._mac = task_result['mac_address']
        self.doLog(
            log.INFO,
            "Machine assigned mac {}".format(self._mac)
        )
        return State.FINISHED
    
    def __request_destroy_vm(self):
        self._task_id = self.service().request_destroy_vm(self._vm_id)
        self.doLog(log.INFO, 'Request to remove machine')
        return State.FINISHED
    
    def __getCurrentOp(self):
        if len(self._queue) == 0:
            return opFinish

        return self._queue[0]

    def __popCurrentOp(self):
        if len(self._queue) == 0:
            return opFinish

        res = self._queue.pop(0)
        return res

    def __pushFrontOp(self, op):
        self._queue.insert(0, op)

    def __pushBackOp(self, op):
        self._queue.append(op)

    def __error(self, reason):
        '''
        Internal method to set object as error state

        Returns:
            State.ERROR, so we can do "return self.__error(reason)"
        '''
        try:
            if self._vm_id != '':
                if self._is_dev_mode():
                    self.add_error_log_about_trying_delete_in_debug_mode(logger, reason)
                else:
                    VMwareDeferredRemoval.remove(self.service().parent(), self._vm_id)
            
            self._queue = [opError]
            self._reason = codecs.decode(str(reason),'unicode-escape')
            logger.debug('Setting error state, reason: {0}'.format(self._reason))
        except Exception as e:
            logger.exception(e)
        finally:
            return State.ERROR
    
    @staticmethod
    def __op2str(op):
        return OP_LABELS.get(op, '????')

    def __debug(self, txt):
        logger.debug('_name {0}: {1}'.format(txt, self._name))
        logger.debug('_ip {0}: {1}'.format(txt, self._ip))
        logger.debug('_mac {0}: {1}'.format(txt, self._mac))
        logger.debug('_vm_id {0}: {1}'.format(txt, self._vm_id))
        logger.debug('_task_id {0}: {1}'.format(txt, self._task_id))
        logger.debug('_cancelled {0}: {1}'.format(txt, self._cancelled))
        logger.debug('Queue at {0}: {1}'.format(txt, [VMwareLinkedDeployment.__op2str(op) for op in self._queue]))