# -*- coding: utf-8 -*-
'''
Created on Jan 20, 2020

@author: dvd
'''
from __future__ import unicode_literals

from django.utils.translation import ugettext_noop as _
from uds.core.services import Service, types as serviceTypes
from uds.core.transports import protocols

from uds.core.ui import gui

import logging

logger = logging.getLogger(__name__)

from .VMwarePublication import VMwarePublication
from .VMwareLinkedDeployment import VMwareLinkedDeployment
from .helpers import VMwareHelpers

class VMwareLinkedService(Service):
    '''
    Basic service, the first part (variables) include the description of the service.

    Remember to fill all variables needed, but at least you must define:
        * typeName
        * typeType
        * typeDescription
        * iconFile (defaults to service.png)
        * publicationType, type of publication in case it needs publication.
          If this is not provided, core will assume that the service do not
          needs publishing.
        * deployedType, type of deployed user service. Do not forget this!!!

    The rest of them can be ommited, but its recommended that you fill all
    declarations shown in this sample (that in fact, are all)

    This description informs the core what this service really provides,
    and how this is done. Look at description of class variables for more
    information.

    '''
    # : Name to show the administrator. This string will be translated BEFORE
    # : sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    typeName = _('Связанный клон')
    # : Type used internally to identify this provider
    typeType = 'vmwareLinkedService'
    # : Description shown at administration interface for this provider
    typeDescription = _('Рабочее место на основе связанного клона ВМ')
    # : Icon file used as icon for this provider. This string will be translated
    # : BEFORE sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    iconFile = 'service.png'

    # Functional related data

    # : If the service provides more than 1 "deployed user" (-1 = no limit,
    # : 0 = ???? (do not use it!!!), N = max number to deploy
    maxDeployed = -1
    # : If we need to generate "cache" for this service, so users can access the
    # : provided services faster. Is usesCache is True, you will need also
    # : set publicationType, do take care about that!
    usesCache = True
    # : Tooltip shown to user when this item is pointed at admin interface, none
    # : because we don't use it
    cacheTooltip = _('Желаемое количество активных ВМ в ожидании подключения')
    # : If we need to generate a "Level 2" cache for this service (i.e., L1
    # : could be running machines and L2 suspended machines)
    usesCache_L2 = True
    # : Tooltip shown to user when this item is pointed at admin interface, None
    # : also because we don't use it
    cacheTooltip_L2 = _('Желаемое количество спящих ВМ в ожидании подключения')

    # : If the service needs a s.o. manager (managers are related to agents
    # : provided by services itselfs, i.e. virtual machines with actors)
    needsManager = True
    # : If true, the system can't do an automatic assignation of a deployed user
    # : service from this service
    mustAssignManually = False

    # : Types of publications (preparated data for deploys)
    # : In our case, we do no need a publication, so this is None
    publicationType = VMwarePublication
    # : Types of deploys (services in cache and/or assigned to users)
    deployedType = VMwareLinkedDeployment
    
    allowedProtocols = protocols.GENERIC
    servicesTypeProvided = (serviceTypes.VDI,)
    
    # Параметры формы
    
    datacenter = gui.ChoiceField(
        label=_("Датацентр"),
        order=100,
        fills={
            'callbackName': 'VMwareFillDatastoreClustersFromDC',
            'function': VMwareHelpers.get_datacenter_resources,
            'parameters': ['datacenter', 'ov', 'ev']
        },
        tooltip=_("Датацентр с ВМ"), required=True
    )
    
    datapool = gui.ChoiceField(
        label=_("Кластер данных"),
        rdonly=False,
        order=101,
        tooltip=_('Кластер данных для публикации и размещения ВМ'),
        required=True
    )
    
    computepool = gui.ChoiceField(
        label=_("Кластер ресурсов"),
        rdonly=False,
        order=102,
        fills={
            'callbackName': 'VMwareFillResourcePoolsFromCluster',
            'function': VMwareHelpers.get_resource_pools,
            'parameters': ['computepool', 'ov', 'ev']
        },
        tooltip=_('Кластер вычислительных ресурсов для работы ВМ'),
        required=True
    )
    
    resourcepool = gui.ChoiceField(
        label=_("Пул ресурсов"),
        rdonly=False,
        order=103,
        tooltip=_('Пул вычислительных ресурсов для работы ВМ'),
        required=True
    )
    
    folder = gui.ChoiceField(
        label=_("Размещение ВМ"),
        rdonly=False,
        order=104,
        tooltip=_('Папка размещения создаваемых ВМ'),
        required=True
    )
    
    minSpaceGB = gui.NumericField(
        length=3,
        label=_('Место (Гб)'),
        defvalue='32',
        order=105,
        tooltip=_('Минимальный размер свободного места в Гб'),
        required=True
    )

    machine = gui.ChoiceField(
        label=_("Базовая ВМ"),
        order=110,
        tooltip=_('Базовый шаблон ВМ для рабочего места'),
        tab=_('Machine'),
        required=True
    )
    
    baseName = gui.TextField(
        label=_('Базовое имя'),
        rdonly=False,
        order=115,
        tooltip=('Базовое имя для клонов данной машины'),
        tab=_('Machine'),
        required=True
    )

    lenName = gui.NumericField(
        length=1,
        label=_('Длина суффикса'),
        defvalue=5,
        order=116,
        tooltip=_('Длина суффикса для формирования имени ВМ - от 3 до 6'),
        tab=_('Machine'),
        required=True
    )
    
    ov = gui.HiddenField(value=None)
    ev = gui.HiddenField(value=None)  # We need to keep the env so we can instantiate the Provider
    
    def initGui(self):
        '''
        Loads required values inside
        '''

        # Here we have to use "default values", cause values aren't used at form initialization
        # This is that value is always '', so if we want to change something, we have to do it
        # at defValue
        self.ov.defValue = self.parent().serialize()
        self.ev.defValue = self.parent().env.key
 
        # parent() - отсылка на экземпляр VMwareProvider
        datacenters = self.parent().get_datacenters()
        vals = []
        for dc in datacenters:
            vals.append(gui.choiceItem(dc['id'], dc['name']))

        self.datacenter.setValues(vals)
        
    def check_datapool_has_space(self):
        logger.debug('Checking datapool space for {0}'.format(self.datapool.value))
        info = self.parent().get_datastore_cluster_info(self.datapool.value)
        logger.debug('Datapool info: {0}'.format(info))
        availableGB = info['free_space'] / 1024 / 1024 / 1024 # Gb
        if availableGB < self.minSpaceGB.num():
            raise Exception('Not enough free space available: (Needs at least {0} GB and there is only {1} GB '.format(self.minSpaceGB.num(), availableGB))
    
    def sanitizeVmName(self, name):
        import re
        return re.sub("[^a-zA-Z0-9_-]", "_", name)
    
    def getBaseName(self):
        '''
        Returns the base name
        '''
        return self.baseName.value

    def getLenName(self):
        '''
        Returns the length of numbers part
        '''
        return int(self.lenName.value)
    
    def request_clone_vm(self, name):
        self.check_datapool_has_space()
        return self.parent().request_clone_vm(
            name, 
            self.datacenter.value, 
            self.datapool.value,
            self.computepool.value,
            self.resourcepool.value,
            self.folder.value,
            self.machine.value
        )
    
    def request_deploy_linked_clone(self, name, template_id):
        self.check_datapool_has_space()
        return self.parent().request_deploy_linked_clone(
            name, 
            self.datacenter.value, 
            self.datapool.value,
            self.computepool.value,
            self.resourcepool.value,
            self.folder.value,
            template_id,
            power_on=True,
        )
    
    def get_clone_template_result(self, task_id):
        '''
        Invokes get_clone_template_result from parent provider
        '''
        return self.parent().get_clone_template_result(task_id)
    
    def get_vm_task_result(self, task_id):
        '''
        Invokes get_vm_task_result from parent provider
        '''
        return self.parent().get_vm_task_result(task_id)
    
    def get_vm_info(self, vm_id):
        '''
        Invokes get_vm_info from parent provider
        '''
        return self.parent().get_vm_info(vm_id)
    
    def request_create_snapshot(self, vm_id):
        '''
        Invokes request_create_snapshot from parent provider
        '''
        return self.parent().request_create_snapshot(vm_id)
    
    def request_poweron_vm(self, vm_id):
        '''
        Invokes request_poweron_vm from parent provider
        '''
        return self.parent().request_poweron_vm(vm_id)
    
    def request_poweroff_vm(self, vm_id):
        '''
        Invokes request_poweroff_vm from parent provider
        '''
        return self.parent().request_poweroff_vm(vm_id)
    
    def request_suspend_vm(self, vm_id):
        '''
        Invokes request_suspend_vm from parent provider
        '''
        return self.parent().request_suspend_vm(vm_id)
    
    def request_destroy_vm(self, vm_id):
        '''
        Invokes request_destroy_vm from parent provider
        '''
        return self.parent().request_destroy_vm(vm_id)
    
    def check_task(self, task_id):
        '''
        invokes check_task from parent provider
        '''
        return self.parent().check_task(task_id)
