# -*- coding: utf-8 -*-
'''
Created on Jan 20, 2020

@author: dvd
'''
from __future__ import unicode_literals

from django.utils.translation import ugettext_noop as _
from uds.core.services import ServiceProvider
from uds.core.ui import gui
from uds.core.util import validators

from . import client
from .VMwareLinkedService import VMwareLinkedService
import logging

logger = logging.getLogger(__name__)

class Provider(ServiceProvider):
    
    # : What kind of services we offer, this are classes inherited from Service
    offers = [VMwareLinkedService]
    # : Name to show the administrator. This string will be translated BEFORE
    # : sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    typeName = _('Платформа VMware')
    # : Type used internally to identify this provider
    typeType = 'vmwarePlatform'
    # : Description shown at administration interface for this provider
    typeDescription = _('Поставщик ресурсов на базе VMware')
    # : Icon file used as icon for this provider. This string will be translated
    # : BEFORE sending it to administration interface, so don't forget to
    # : mark it as _ (using ugettext_noop)
    iconFile = 'provider.png'

    # now comes the form fields
    # There is always two fields that are requested to the admin, that are:
    # Service Name, that is a name that the admin uses to name this provider
    # Description, that is a short description that the admin gives to this provider
    # Now we are going to add a few fields that we need to use this provider
    # Remember that these are "dummy" fields, that in fact are not required
    # but used for sample purposes
    # If we don't indicate an order, the output order of fields will be
    # "random"
    platformVersion = gui.ChoiceField(order=1,
        label='Версия платформы',
        tooltip='Версия сервера VMware',
        # In this case, the choice can have none value selected by default
        required=True,
        rdonly=False,
        values=[
            gui.choiceItem('6', '6.0 - 6.7.3'),
        ],
        defvalue='6'  # Default value is the ID of the choicefield
    )

    host = gui.TextField(length=64, label='Адрес сервера', order=2, tooltip=_('IP-адрес или доменное имя сервера VMware'), required=True)
    port = gui.NumericField(length=5, label='Порт', defvalue='443', order=3, tooltip=_('Порт для подключения к api'), required=True)
    ssl = gui.CheckBoxField(label=_('Использовать SSL'), defvalue='true', order=4, tooltip=_('Форсировать использование SSL (не будет работать, если сервер не поддерживает SSL)'))
    
    username = gui.TextField(length=32, label=_('Логин'), order=5, tooltip=_('Пользователь с необходимыми правами в VMware'), required=True, defvalue='')
    password = gui.PasswordField(lenth=32, label=_('Пароль'), order=6, tooltip=_('Пароль пользователя'), required=True)
    
    maxPreparingServices = gui.NumericField(length=7, label=_('Подготавливать ВМ одновременно'), defvalue='10', minValue=1, maxValue=65536, order=50, tooltip=_('Максимальное число одновременно подготавливаемых виртуальных машин'), required=True, tab=gui.ADVANCED_TAB)
    maxRemovingServices = gui.NumericField(length=8, label=_('Удалять ВМ одновременно'), defvalue='5', minValue=1, maxValue=65536, order=51, tooltip=_('Макимальное число одновременно удаляемых виртуальных машин'), required=True, tab=gui.ADVANCED_TAB)

    # Own variables
    _api = None

    def __getApi(self):
        '''
        Returns the connection API object for VMware
        '''
        
        timeout = 10 # unused param for now
        if self._api is None:
            ClientClass = self._api = client.getClient(self.platformVersion.value)
            self._api = ClientClass(self.host.value, self.port.value, self.ssl.isTrue(), self.username.value, self.password.value, timeout, self.cache)
        return self._api

    # There is more fields type, but not here the best place to cover it
    def initialize(self, values=None):
        '''
        We will use the "autosave" feature for form fields
        '''

        # Just reset _api connection variable
        self._api = None

    def testConnection(self):
        '''
        Test that conection to server is fine

        Returns

            True if all went fine, false if id didn't
        '''

        return self.__getApi().test()
    
    @staticmethod
    def test(env, data):
        '''
        Test VMware Connectivity

        Args:
            env: environment passed for testing (temporal environment passed)

            data: data passed for testing (data obtained from the form
            definition)

        Returns:
            Array of two elements, first is True of False, depending on test
            (True is all right, false is error),
            second is an String with error, preferably internacionalizated..

        '''
        prov = Provider(env, data)
        if prov.testConnection() is True:
            return [True, _("Подключение прошло успешно.")]

        return [False, _("Не удалось подключиться. Проверьте параметры подключения.")]
    
    def get_datacenters(self, force=False):
        '''
        Obtains the list of datacenters inside vCenter.

        Args:
            force: If true, force to update the cache, if false, tries to first
            get data from cache and, if valid, return this.

        Returns
            Filters out clusters not attached to any datacenter
            An array of dictionaries, containing:
                'name'
                'id'
        '''

        return self.__getApi().get_datacenters()
    
    def get_datastore_clusters(self, datacenter_id):
        '''
        Obtains the list of datastore clusters inside VMware datacenter.

        Args:
            datacenter_id - ManagedObjectReference id of datacenter entity

        Returns
            An array of dictionaries, containing:
                'name'
                'id'
                'capacity' in bytes
                'free_space' in bytes
        '''

        return self.__getApi().get_datastore_clusters(datacenter_id)
    
    def get_datastore_cluster_info(self, cluster_id):
        '''
        Obtains info about datastore cluster inside VMware datacenter.

        Args:
            cluster_id - ManagedObjectReference id of StoragePod entity

        Returns
            An array of dictionaries, containing:
                'name'
                'id'
                'capacity' in bytes
                'free_space' in bytes
        '''

        return self.__getApi().get_datastore_cluster_info(cluster_id)
    
    def get_compute_clusters(self, datacenter_id):
        '''
        Obtains the list of compute clusters inside VMware datacenter.

        Args:
            datacenter_id - ManagedObjectReference id of datacenter entity
            
        Returns
            An array of dictionaries, containing:
                'name'
                'id'
        '''

        return self.__getApi().get_compute_clusters(datacenter_id)
    
    def get_vms(self, datacenter_id):
        '''
        Obtains the list of virtual machines inside VMware datacenter.

        Args:
            datacenter_id - ManagedObjectReference id of datacenter entity
            
        Returns
            An array of dictionaries, containing:
                'name'
                'id'
                'uuid'
        '''
        return self.__getApi().get_vms(datacenter_id)
    
    def get_templates(self, datacenter_id):
        '''
        Obtains the list of virtual machines inside VMware datacenter.

        Args:
            datacenter_id - ManagedObjectReference id of datacenter entity
            
        Returns
            An array of dictionaries, containing:
                'name'
                'id'
                'uuid'
        '''
        return self.__getApi().get_templates(datacenter_id)
    
    def get_folders(self, datacenter_id):
        '''
        Obtains the list of folders inside VMware datacenter.

        Args:
            datacenter_id - ManagedObjectReference id of datacenter entity
            
        Returns
            An array of dictionaries, containing:
                'name'
                'id'
                'parent'
        '''
        return self.__getApi().get_folders(datacenter_id)
    
    def get_resource_pools(self, cluster_id):
        '''
        Obtains the list of resource pools inside VMware ClusterComputeResource.

        Args:
            cluster_id - ManagedObjectReference id of ClusterComputeResource entity
            
        Returns
            An array of dictionaries, containing:
                'name'
                'id'
                'parent
        '''

        return self.__getApi().get_resource_pools(cluster_id)
    
    def request_clone_vm(self, name, datacenter_id, datapool_id, computepool_id, resourcepool_id, folder_id, vm_id, make_template=False, power_on=False):
        '''
        Requests clone vm in vcenter.

        Args:
            name            - name of new template vm
            datacenter_id   - ManagedObjectReference id of Datacenter entity
            datapool_id     - ManagedObjectReference id of StoragePod entity
            computepool_id  - ManagedObjectReference id of ClusterComputeResource entity
            resourcepool_id - ManagedObjectReference id of ResourcePool entity
            vm_id           - ManagedObjectReference id of VirtualMachine entity
            
        Returns
            task_id - ManagedObjectReference id of Task entity
        '''

        return self.__getApi().request_clone_vm(name, datacenter_id, datapool_id, computepool_id, resourcepool_id, folder_id, vm_id, make_template=False, power_on=False)
    
    def request_deploy_linked_clone(self, name, datacenter_id, datapool_id, computepool_id, resourcepool_id, folder_id, template_id, power_on=False):
        '''
        Requests clone vm in vcenter.

        Args:
            name            - name of new template vm
            datacenter_id   - ManagedObjectReference id of Datacenter entity
            datapool_id     - ManagedObjectReference id of StoragePod entity
            computepool_id  - ManagedObjectReference id of ClusterComputeResource entity
            resourcepool_id - ManagedObjectReference id of ResourcePool entity
            template_id     - ManagedObjectReference id of VirtualMachine entity
            
        Returns
            task_id - ManagedObjectReference id of Task entity
        '''

        return self.__getApi().request_deploy_linked_clone(name, datacenter_id, datapool_id, computepool_id, resourcepool_id, folder_id, template_id, power_on=power_on)
    
    def get_clone_template_result(self, task_id):
        '''
        Requests cloned template info from task results.

        Args:
            task_id   - ManagedObjectReference id of Task entity
            
        Returns
            {
                template_id - ManagedObjectReference id of created template (VirtualMachine entity)
            }
        '''
        
        return self.__getApi().get_clone_template_result(task_id)
    
    def get_vm_task_result(self, task_id):
        '''
        Requests cloned vm info from task results.

        Args:
            task_id   - ManagedObjectReference id of Task entity
            
        Returns
            {
                vm_id - ManagedObjectReference id of created template (VirtualMachine entity)
                mac_address - string
            }
        '''
        
        return self.__getApi().get_vm_task_result(task_id)
    
    def get_vm_info(self, vm_id):
        '''
        Requests vm info from VMware

        Args:
            task_id   - ManagedObjectReference id of Task entity
            
        Returns
            {
                vm_id - ManagedObjectReference id of created template (VirtualMachine entity)
                name - string
                mac_address - string
                power_state - string (poweredOff, poweredOn, suspended)
                paused - boolean
            }
        '''
        
        return self.__getApi().get_vm_info(vm_id)
    
    def request_create_snapshot(self, vm_id):
        '''
        Requests create snapshot for vm in vcenter.

        Args:
            vm_id           - ManagedObjectReference id of VirtualMachine entity
            
        Returns
            task_id - ManagedObjectReference id of Task entity
        '''

        return self.__getApi().request_create_snapshot(vm_id)
    
    def request_destroy_vm(self, vm_id):
        '''
        Requests destroy vm in vcenter.

        Args:
            vm_id           - ManagedObjectReference id of VirtualMachine entity
            
        Returns
            task_id - ManagedObjectReference id of Task entity
        '''

        return self.__getApi().request_destroy_vm(vm_id)
    
    def request_poweron_vm(self, vm_id):
        '''
        Requests poweron vm in vcenter.

        Args:
            vm_id           - ManagedObjectReference id of VirtualMachine entity
            
        Returns
            task_id - ManagedObjectReference id of Task entity
        '''

        return self.__getApi().request_poweron_vm(vm_id)
    
    def request_poweroff_vm(self, vm_id):
        '''
        Requests poweroff vm in vcenter.

        Args:
            vm_id           - ManagedObjectReference id of VirtualMachine entity
            
        Returns
            task_id - ManagedObjectReference id of Task entity
        '''

        return self.__getApi().request_poweroff_vm(vm_id)
    
    def request_suspend_vm(self, vm_id):
        '''
        Requests suspend vm in vcenter.

        Args:
            vm_id           - ManagedObjectReference id of VirtualMachine entity
            
        Returns
            task_id - ManagedObjectReference id of Task entity
        '''

        return self.__getApi().request_suspend_vm(vm_id)
    
    def check_task(self, task_id):
        '''
        Requests task state from vCenter.

        Args:
            task_id   - ManagedObjectReference id of Task entity
            
        Returns
            string - Returns task's state
        '''
        
        return self.__getApi().check_task(task_id)
    