# -*- coding: utf-8 -*-
'''
Created on Jan 20, 2020

@author: dvd
'''
from __future__ import unicode_literals

from django.utils.translation import ugettext as _
from uds.core.services import Publication
from uds.core.util.State import State
from datetime import datetime
import logging
import json

from .VMwareJobs import VMwareDeferredRemoval

logger = logging.getLogger(__name__)

opError, opFinish, opRequestCloneVm, opWaitCloneVm, opGetCloneResult, \
opRequestDestroyTemplate, opWaitDestroyTemplate, opRequestCreateSnapshot, opWaitCreateSnapshot  = range(9) 

OP_LABELS = {
    opError: 'error',
    opFinish: 'finish',
    opRequestCloneVm: 'requestCloneVm',
    opWaitCloneVm: 'waitCloneVm', 
    opGetCloneResult: 'getCloneResult', 
    opRequestCreateSnapshot: 'requestCreateSnapshot',
    opWaitCreateSnapshot: 'waitCreateSnapshot',
    opRequestDestroyTemplate: 'requestDestroyTemplate',
    opWaitDestroyTemplate: 'waitDestroyTemplate',
}

class VMwarePublication(Publication):
    '''
    This class shows how a publication is developed.
    
    In order to a publication to work correctly, we must provide at least the
    following methods:
        * Of course, the __init__
        * :py:meth:`.publish` 
        * :py:meth:`.checkState`
        * :py:meth:`.finish`
        
    Also, of course, methods from :py:class:`uds.core.Serializable.Serializable`
    
    
    Publication do not have an configuration interface, all data contained
    inside an instance of a Publication must be serialized if you want them between
    method calls.
    
    It's not waranteed that the class will not be serialized/deserialized
    between methods calls, so, first of all, implement the marshal and umnarshal
    mehods needed by all serializable classes.
    
    Also a thing to note is that operations requested to Publications must be
    *as fast as posible*. The operations executes in a separated thread,
    and so it cant take a bit more time to execute, but it's recommended that
    the operations executes as fast as posible, and, if it will take a long time,
    split operation so we can keep track of state.
    
    This means that, if we have "slow" operations, we must
    
    We first of all declares an estimation of how long a publication will take.
    This value is instance based, so if we override it in our class, the suggested
    time could change.
    
    The class attribute that indicates this suggested time is "suggestedTime", and
    it's expressed in seconds, (i.e. "suggestedTime = 10") 
    '''
    suggestedTime = 5 #: Suggested recheck time if publication is unfinished in seconds
    
    def initialize(self):
        '''
        This method will be invoked by default __init__ of base class, so it gives
        us the oportunity to initialize whataver we need here. 
        
        In our case, we setup a few attributes..
        '''
        
        # We do not check anything at marshal method, so we ensure that
        # default values are correctly handled by marshal.
        self._name = ''
        self._reason = '' # No error, no reason for it
        self._queue = []
        self._template_id = ''
        self._task_id = ''
    
    # Serializable needed methods
    def marshal(self):
        data = {
            'version': 1,
            'name': self._name,
            'reason': self._reason,
            'queue': self._queue,
            'template_id': self._template_id,
            'task_id': self._task_id,
        }
        
        return json.dumps(data)
        

    def unmarshal(self, str_):
        if isinstance(str_, bytes):
            str_ = str_.decode('utf-8')

        data = json.loads(str_)
        if data['version']==1:
            self._name = data['name']
            self._reason = data['reason']
            self._queue = data['queue']
            self._template_id = data['template_id']
            self._task_id = data['task_id']
        else:
            raise Exception("Unsupported serialized data version {}".format(data['version']))
        
    def publish(self):
        '''
        This method is invoked whenever the administrator requests a new publication.
        
        The method is not invoked directly (i mean, that the administration request
        do no makes a call to this method), but a DelayedTask is saved witch will
        initiate all publication stuff (and, of course, call this method).
        
        You MUST implement it, so the publication do really something.
        All publications can be synchronous or asynchronous.
        
        The main difference between both is that first do whatever needed, (the
        action must be fast enough to do not block core), returning State.FINISHED.
        
        The second (asynchronous) are publications that could block the core, so
        it have to be done in more than one step.
        
        An example publication could be a copy of a virtual machine, where:
            * First we invoke the copy operation to virtualization provider
            * Second, we kept needed values inside instance so we can serialize
              them whenever requested
            * Returns an State.RUNNING, indicating the core that the publication
              has started but has to finish sometime later. (We do no check
              again the state and keep waiting here, because we will block the
              core untill this operation is finished).
        
        In our example wi will simple assign a name, and set number to 5. We 
        will use this number later, to make a "delay" at check if the publication
        has finished. (see method checkState)
        
        We also will make this publication an "stepped one", that is, it will not
        finish at publish call but a later checkState call
        
        Take care with instantiating threads from here. Whenever a publish returns
        "State.RUNNING", the core will recheck it later, but not using this instance
        and maybe that even do not use this server. 
        
        If you want to use threadings or somethin likt it, use DelayedTasks and
        do not block it. You also musht provide the mechanism to allow those
        DelayedTask to communicate with the publication.
        
        One sample could be, for example, to copy a bunch of files, but we know
        that this copy can take a long time and don't want it to take make it
        all here, but in a separate task. Now, do you remember that "environment"
        that is unique for every instance?, well, we can create a delayed task,
        and pass that environment (owned by this intance) as a mechanism for
        informing when the task is finished. (We insert at delayed tasks queue 
        an instance, not a class itself, so we can instantiate a class and
        store it at delayed task queue.
        
        Also note that, in that case, this class can also acomplish that by simply
        using the suggestedTime attribute and the checkState method in most cases.
        '''
        self._name = self.service().sanitizeVmName('TDSK ' + self.dsName() + "-" + str(self.revision()))
        comments = _('Termidesk pub for {0} at {1}').format(self.dsName(), str(datetime.now()).split('.')[0])
        self._queue = [opRequestCloneVm, opWaitCloneVm, opGetCloneResult, opRequestCreateSnapshot, opWaitCreateSnapshot, opFinish]
        
        return State.RUNNING
    
    def checkState(self):
        '''
        Check what operation is going on, and acts accordingly to it
        '''
        try:
            self.__debug('checkState')
            op = self.__getCurrentOp()
    
            if op == opError:
                return State.ERROR
    
            if op == opFinish:
                return State.FINISHED
    
            fncs = {
                opRequestCloneVm: self.__request_clone_template,
                opWaitCloneVm: self.__wait_clone_template,
                opGetCloneResult: self.__get_clone_template_result,
                opRequestCreateSnapshot: self.__request_create_snapshot,
                opWaitCreateSnapshot: self.__wait_create_snapshot,
                opRequestDestroyTemplate: self.__request_destroy_template,
                opWaitDestroyTemplate: self.__wait_destroy_template,
            }

        
            chkFnc = fncs.get(op, None)

            if chkFnc is None:
                return self.__error('Unknown operation found at check queue ({0})'.format(op))

            state = chkFnc()
            if state == State.FINISHED:
                self.__popCurrentOp()  # Remove runing op
                # Для ускорения выполняем следующий шаг из очереди сейчас
                return self.checkState()

            return state
        except Exception as e:
            logger.error('checkState Operation %s error', VMwarePublication.__op2str(op))
            logger.exception(e)
            return self.__error(str(e))
    
    def finish(self):
        '''
        In our case, finish does nothing
        '''
        pass

    def reasonOfError(self):
        '''
        If a publication produces an error, here we must notify the reason why
        it happened. This will be called just after publish or checkState
        if they return State.ERROR

        Returns an string, in our case, set at checkState
        '''
        return self._reason

    def destroy(self):
        '''
        This is called once a publication is no more needed.

        This method do whatever needed to clean up things, such as
        removing created "external" data (environment gets cleaned by core),
        etc..

        The retunred value is the same as when publishing, State.RUNNING,
        State.FINISHED or State.ERROR.
        '''

        try:
            if not self._template_id:
                self._queue.pop() # Remove opFinish
                self._queue.extend([opRequestDestroyTemplate, opWaitDestroyTemplate, opFinish])
                return State.RUNNING
            
            if self._template_id:
                self._queue = [opRequestDestroyTemplate, opWaitDestroyTemplate, opFinish]
                return State.RUNNING
                
        except Exception as e:
            self._state = 'error'
            self._reason = str(e)
            return State.ERROR

        return State.FINISHED

    def cancel(self):
        '''
        Do same thing as destroy
        '''
        return self.destroy()
    
    def __request_clone_template(self):
        self._task_id = self.service().request_clone_vm(self._name)
        return State.FINISHED
    
    def __wait_clone_template(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось клонировать мастер образ".encode('unicode-escape'))
        
        return State.FINISHED    
    
    def __get_clone_template_result(self):
        task_result = self.service().get_clone_template_result(self._task_id)
        self._template_id = task_result['template_id']
        return State.FINISHED
    
    def __request_create_snapshot(self):
        self._task_id = self.service().request_create_snapshot(self._template_id)
        return State.FINISHED
    
    def __wait_create_snapshot(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось создать снимок состояния ВМ публикации".encode('unicode-escape'))
        
        return State.FINISHED
    
    def __request_destroy_template(self):
        self._task_id = self.service().request_destroy_vm(self._template_id)
        return State.FINISHED
    
    def __wait_destroy_template(self):
        end_states = ['success', 'error']
        state = self.service().check_task(self._task_id)
        
        if state not in end_states:
            return State.RUNNING
        
        if state == 'error':
            return self.__error("Не удалось удалить ВМ публикации".encode('unicode-escape'))
        
        return State.FINISHED
    
    def __getCurrentOp(self):
        if len(self._queue) == 0:
            return opFinish

        return self._queue[0]

    def __popCurrentOp(self):
        if len(self._queue) == 0:
            return opFinish

        res = self._queue.pop(0)
        return res

    def __pushFrontOp(self, op):
        self._queue.insert(0, op)

    def __pushBackOp(self, op):
        self._queue.append(op)

    def __error(self, reason):
        '''
        Internal method to set object as error state

        Returns:
            State.ERROR, so we can do "return self.__error(reason)"
        '''
        try:
            logger.debug('Setting error state, reason: {0}'.format(reason.decode('unicode-escape')))
            
            if self._template_id != '':
                VMwareDeferredRemoval.remove(self.service().parent(), self._template_id)
            
            self._queue = [opError]
            self._reason = reason.decode('unicode-escape')
        except Exception as e:
            logger.exception(e)
        finally:
            return State.ERROR
    
    @staticmethod
    def __op2str(op):
        return OP_LABELS.get(op, '????')

    def __debug(self, txt):
        logger.debug('_name {0}: {1}'.format(txt, self._name))
        logger.debug('_template_id {0}: {1}'.format(txt, self._template_id))
        logger.debug('_task_id {0}: {1}'.format(txt, self._task_id))
        logger.debug('Queue at: {0}'.format(str(self._queue)))
        logger.debug('Queue at {0}: {1}'.format(txt, [VMwarePublication.__op2str(op) for op in self._queue]))
    
    def get_template_id(self):
        return self._template_id
    
    def get_name(self):
        return self._name