# -*- coding: utf-8 -*-
'''
Created on Jan 20, 2020

@author: dvd
'''
from __future__ import unicode_literals

import logging
import threading

logger = logging.getLogger(__name__)
context = threading.local()
context.service_instance = None

from pyVim import connect
from pyVmomi import vmodl
from pyVmomi import vim

def connectVMware(method):
    '''
    Контекстный декоратор для подключения к VMware API
    Сохраняет service_instance в локальную контекстную переменную текущего потока для использования в методах клиента
    '''
    def wrapper(*args, **kwargs):
        instance = args[0]
        
        try:
            context.service_instance = instance._connect()
            return method(*args, **kwargs)
        except vmodl.MethodFault as error:
            logger.error("Caught vmodl fault: %s", error.msg)
            raise
        except Exception as e:
            logger.exception(e)
            raise
        finally:
            instance._disconnect(context.service_instance)
            context.service_instance = None
        
    return wrapper

class Client(object):
    
    def __init__(self, host, port, ssl, username, password, timeout, cache):
        self._host = host
        self._port = port
        self._ssl = ssl
        self._username = username
        self._password = password
        self._timeout = int(timeout)
        self._cache = cache
        self._proto = 'http'

        if self._ssl:
            self._proto = 'https'
        
        self._base_url = "{proto}://{host}:{port}".format(proto=self._proto, host=self._host, port=self._port)
            
        logger.debug("Initialized for %s", self._base_url)
    
    def test(self):
        
        logger.debug("Test connection to {}:{} for user {}".format(self._host, self._port, self._username))
        try:
            service_instance = connect.SmartConnect(host=self._host,
                                                user=self._username,
                                                pwd=self._password,
                                                port=int(self._port))
            
            # If we get here then we successfully authenticated into vCenter
            # Only a successfully authenticated session has a
            # session key aka session id.
            session_id = service_instance.content.sessionManager.currentSession.key
            logger.debug("Obtained session id: %s", session_id)
            self._disconnect(service_instance)
            return True
        except vmodl.MethodFault as error:
            logger.error("Caught vmodl fault: %s", error.msg)
        except Exception as e:
            logger.exception(e)
        return False
    
    @connectVMware
    def get_datacenters(self):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        # Search for all Datacenters
        obj_view = content.viewManager.CreateContainerView(content.rootFolder,
                                                           [vim.Datacenter],
                                                           True)
        datacenter_list = obj_view.view
        obj_view.Destroy()
        
        data = []
        for datacenter in datacenter_list:
            data.append({'name': datacenter.name, 'id': datacenter._moId})
        
        return data
    
    @connectVMware
    def get_datastore_clusters(self, datacenter_id):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        dc = vim.Datacenter(datacenter_id)
        dc._stub = service_instance._stub
        
        # Search for all Datastore Clusters aka StoragePod
        obj_view = content.viewManager.CreateContainerView(dc.datastoreFolder,
                                                           [vim.StoragePod],
                                                           True)
        ds_cluster_list = obj_view.view
        obj_view.Destroy()
        
        data = []
        for ds_cluster in ds_cluster_list:
            data.append(
                {
                    'name': ds_cluster.name, 
                    'id': ds_cluster._moId,
                    'capacity': ds_cluster.summary.capacity,
                    'free_space': ds_cluster.summary.freeSpace,
                }
            )
        
        return data
    
    @connectVMware
    def get_datastore_cluster_info(self, cluster_id):
        service_instance = context.service_instance
        
        ds_cluster = vim.StoragePod(cluster_id)
        ds_cluster._stub = service_instance._stub
        
        data = {
            'name': ds_cluster.name, 
            'id': ds_cluster._moId,
            'capacity': ds_cluster.summary.capacity,
            'free_space': ds_cluster.summary.freeSpace,
        }
        
        return data
    
    @connectVMware
    def get_compute_clusters(self, datacenter_id):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        dc = vim.Datacenter(datacenter_id)
        dc._stub = service_instance._stub
        
        obj_view = content.viewManager.CreateContainerView(dc.hostFolder,
                                                           [vim.ClusterComputeResource],
                                                           True)
        resource_pool_list = obj_view.view
        obj_view.Destroy()
        
        data = []
        for item in resource_pool_list:
            data.append({'name': item.name, 'id': item._moId})
        
        return data
    
    @connectVMware
    def get_vms(self, datacenter_id):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        dc = vim.Datacenter(datacenter_id)
        dc._stub = service_instance._stub
        
        obj_view = content.viewManager.CreateContainerView(dc.vmFolder,
                                                           [vim.VirtualMachine],
                                                           True)
        vm_list = obj_view.view
        obj_view.Destroy()
        
        data = []
        for item in vm_list:
            try:
                vm_config = item.config
            
                data.append(
                    {
                        'name': item.name, 
                        'id': item._moId,
                        'uuid': vm_config.uuid,
                    }
                )
            except Exception as e:
                # Likely config property is not available, skipping vm
                logger.exception(e)
        
        return data
    
    @connectVMware
    def get_templates(self, datacenter_id):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        dc = vim.Datacenter(datacenter_id)
        dc._stub = service_instance._stub
        
        obj_view = content.viewManager.CreateContainerView(dc.vmFolder,
                                                           [vim.VirtualMachine],
                                                           True)
        vm_list = obj_view.view
        obj_view.Destroy()
        
        data = []
        for item in vm_list:
            try:
                vm_config = item.config
                if not vm_config.template:
                    continue
            
                data.append(
                    {
                        'name': item.name, 
                        'id': item._moId,
                        'uuid': vm_config.uuid,
                    }
                )
            except Exception as e:
                # Likely config property is not available, skipping vm
                logger.exception(e)
        
        return data
    
    @connectVMware
    def get_folders(self, datacenter_id):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        dc = vim.Datacenter(datacenter_id)
        dc._stub = service_instance._stub
        
        obj_view = content.viewManager.CreateContainerView(dc.vmFolder,
                                                           [vim.Folder],
                                                           True)
        folder_list = obj_view.view
        obj_view.Destroy()
        
        data = []
        for item in folder_list:
            parent_id = None
            if item.parent:
                parent_id = item.parent._moId
            data.append(
                {
                    'name': item.name, 
                    'id': item._moId,
                    'parent': parent_id
                }
            )
        
        return data
    
    @connectVMware
    def get_resource_pools(self, cluster_id):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        compute_cluster = vim.ClusterComputeResource(cluster_id)
        compute_cluster._stub = service_instance._stub
        
        obj_view = content.viewManager.CreateContainerView(compute_cluster,
                                                           [vim.ResourcePool],
                                                           True)
        resource_pool_list = obj_view.view
        obj_view.Destroy()
        
        data = []
        for item in resource_pool_list:
            parent_id = None
            if item.parent:
                parent_id = item.parent._moId
            data.append({'name': item.name, 'id': item._moId, 'parent': parent_id})
        
        return data
    
    @connectVMware
    def request_clone_vm(self, name, datacenter_id, datapool_id, computepool_id, resourcepool_id, folder_id, vm_id, make_template=False, power_on=False):
            service_instance = context.service_instance
            content = service_instance.RetrieveContent()
            
            dc = vim.Datacenter(datacenter_id)
            dc._stub = service_instance._stub
            
            datastore_pod = vim.StoragePod(datapool_id)
            datastore_pod._stub = service_instance._stub
            
            cluster = vim.ClusterComputeResource(computepool_id)
            cluster._stub = service_instance._stub
            
            resource_pool = vim.ResourcePool(resourcepool_id)
            resource_pool._stub = service_instance._stub
            
            folder = vim.Folder(folder_id)
            folder._stub = service_instance._stub
            
            vm = vim.VirtualMachine(vm_id)
            vm._stub = service_instance._stub
            
            vmconf = vim.vm.ConfigSpec()
            
            podsel = vim.storageDrs.PodSelectionSpec()
            podsel.storagePod = datastore_pod

            relospec = vim.vm.RelocateSpec()
        
            clonespec = vim.vm.CloneSpec()
            clonespec.location = relospec
    
            storagespec = vim.storageDrs.StoragePlacementSpec()
            storagespec.podSelectionSpec = podsel
            storagespec.type = 'clone'
            storagespec.folder = folder
            storagespec.configSpec = vmconf
            storagespec.cloneName = name
            storagespec.cloneSpec = clonespec
            storagespec.vm = vm
    
            rec = content.storageResourceManager.RecommendDatastores(
                storageSpec=storagespec)
            rec_action = rec.recommendations[0].action[0]
            datastore = rec_action.destination
            
            # set relospec
            relospec = vim.vm.RelocateSpec()
            relospec.datastore = datastore
            relospec.pool = resource_pool
        
            clonespec = vim.vm.CloneSpec()
            clonespec.location = relospec
            clonespec.template = make_template
            clonespec.powerOn = power_on
            
            task = vm.Clone(folder=folder, name=name, spec=clonespec)
            
            return task._moId
    
    @connectVMware
    def request_deploy_linked_clone(self, name, datacenter_id, datapool_id, computepool_id, resourcepool_id, folder_id, vm_id, power_on=False):
        service_instance = context.service_instance
        content = service_instance.RetrieveContent()
        
        dc = vim.Datacenter(datacenter_id)
        dc._stub = service_instance._stub
        
        datastore_pod = vim.StoragePod(datapool_id)
        datastore_pod._stub = service_instance._stub
        
        cluster = vim.ClusterComputeResource(computepool_id)
        cluster._stub = service_instance._stub
        
        resource_pool = vim.ResourcePool(resourcepool_id)
        resource_pool._stub = service_instance._stub
        
        folder = vim.Folder(folder_id)
        folder._stub = service_instance._stub
        
        vm = vim.VirtualMachine(vm_id)
        vm._stub = service_instance._stub
        
        vmconf = vim.vm.ConfigSpec()
        
        podsel = vim.storageDrs.PodSelectionSpec()
        podsel.storagePod = datastore_pod

        relospec = vim.vm.RelocateSpec()
    
        clonespec = vim.vm.CloneSpec()
        clonespec.location = relospec

        storagespec = vim.storageDrs.StoragePlacementSpec()
        storagespec.podSelectionSpec = podsel
        storagespec.type = 'clone'
        storagespec.folder = folder
        storagespec.configSpec = vmconf
        storagespec.cloneName = name
        storagespec.cloneSpec = clonespec
        storagespec.vm = vm

        rec = content.storageResourceManager.RecommendDatastores(
            storageSpec=storagespec)
        rec_action = rec.recommendations[0].action[0]
        datastore = rec_action.destination
        
        # set relospec
        relospec = vim.vm.RelocateSpec()
        relospec.diskMoveType = 'createNewChildDiskBacking'
        relospec.datastore = datastore
        relospec.pool = resource_pool
    
        clonespec = vim.vm.CloneSpec()
        clonespec.location = relospec
        clonespec.powerOn = power_on
        clonespec.snapshot=vm.snapshot.rootSnapshotList[0].snapshot
        
        task = vm.Clone(folder=folder, name=name, spec=clonespec)
        
        return task._moId
            
    @connectVMware
    def get_clone_template_result(self, task_id):
        service_instance = context.service_instance
        
        task = vim.Task(task_id)
        task._stub = service_instance._stub
        template_vm = task.info.result

        return {
            'template_id': template_vm._moId,
        }
    
    @connectVMware
    def get_vm_task_result(self, task_id):
        service_instance = context.service_instance
        
        task = vim.Task(task_id)
        task._stub = service_instance._stub
        vm = task.info.result
        mac = ''
        for dev in vm.config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualEthernetCard):
                mac = dev.macAddress
        
        return {
            'vm_id': vm._moId,
            'mac_address': mac,
        }
    
    @connectVMware
    def get_vm_info(self, vm_id):
        service_instance = context.service_instance
        
        vm = vim.VirtualMachine(vm_id)
        vm._stub = service_instance._stub

        mac = ''
        for dev in vm.config.hardware.device:
            if isinstance(dev, vim.vm.device.VirtualEthernetCard):
                mac = dev.macAddress
        
        return {
            'vm_id': vm._moId,
            'name': vm.name,
            'mac_address': mac,
            'power_state': vm.runtime.powerState,
            'paused': vm.runtime.paused,
        }
    
    @connectVMware
    def request_poweron_vm(self, vm_id):
        service_instance = context.service_instance
        
        vm = vim.VirtualMachine(vm_id)
        vm._stub = service_instance._stub

        task = vm.PowerOnVM_Task()
        
        return task._moId
    
    @connectVMware
    def request_poweroff_vm(self, vm_id):
        service_instance = context.service_instance
        
        vm = vim.VirtualMachine(vm_id)
        vm._stub = service_instance._stub

        task = vm.PowerOffVM_Task()
        
        return task._moId
    
    @connectVMware
    def request_suspend_vm(self, vm_id):
        service_instance = context.service_instance
        
        vm = vim.VirtualMachine(vm_id)
        vm._stub = service_instance._stub

        task = vm.SuspendVM_Task()
        
        return task._moId
    
    @connectVMware
    def request_destroy_vm(self, vm_id):
        service_instance = context.service_instance
        
        vm = vim.VirtualMachine(vm_id)
        vm._stub = service_instance._stub

        task = vm.Destroy_Task()
        
        return task._moId
    
    @connectVMware
    def request_create_snapshot(self, vm_id):
        service_instance = context.service_instance
        
        vm = vim.VirtualMachine(vm_id)
        vm._stub = service_instance._stub

        task = vm.CreateSnapshot_Task(name='TDSK_SNAPSHOT',
                                  memory=False,
                                  quiesce=False)
        
        return task._moId
    
    @connectVMware
    def check_task(self, task_id):
            service_instance = context.service_instance
            
            task = vim.Task(task_id)
            task._stub = service_instance._stub
            return task.info.state
        
    def _connect(self):
        service_instance = connect.SmartConnect(host=self._host,
                                                user=self._username,
                                                pwd=self._password,
                                                port=int(self._port))
        
        return service_instance
    
    def _disconnect(self, service_instance):
        try:
            connect.Disconnect(service_instance)
        except:
            pass