"""
Модуль содержит теги django-шаблонов для проверки пользовательских разрешений.

Использование тегов:
1. Реализовать в данном модуле тег.
2. Импортировать модуль `user_permissions` в шаблоне внутри переопределяемого блока, например,
   ```
   {% block maintab %}
   {% load user_permissions %}
   ```.
3. После команды импорта определить новую переменную с результатом вычислений, например,
   `{% can_create_provider request.user as user_can_create_provider %}`
4. Использовать введенную переменную в операциях целевого шаблона, например,
   `{% if user_can_create_provider %}`

Итоговый вариант будет выглядеть примерно так
```
{# providers.html #}
{% block maintab %}
{% load user_permissions %}
{% can_create_provider request.user as user_can_create_provider %}
{% if user_can_create_provider %}
<div> some html block </div>
{% endif %}
```.

"""
# -*- encoding: utf-8 -*-
from django import template

from termidesk.core.auths.utils import check_exists_permission_for_user
from termidesk.models.UserPermission import LOG_VIEW, CONFIG_VIEW, SERVICE_POOL_GROUP_VIEW, \
    SERVICE_POOL_GROUP_CREATE, SERVICE_POOL_GROUP_CHANGE, SERVICE_POOL_GROUP_DELETE, \
    IMAGE_CREATE, IMAGE_VIEW, IMAGE_CHANGE, IMAGE_DELETE, LICENSE_VIEW, LICENSE_UPLOAD, USER_ROLES_MANAGE, \
    OSMANAGER_CREATE, OSMANAGER_VIEW, OSMANAGER_CHANGE, OSMANAGER_DELETE, \
    TRANSPORT_VIEW, TRANSPORT_DELETE, TRANSPORT_CREATE, TRANSPORT_CHANGE, NETWORK_CREATE, NETWORK_VIEW, \
    NETWORK_CHANGE, NETWORK_DELETE, AUTHENTICATOR_CREATE, AUTHENTICATOR_VIEW, AUTHENTICATOR_CHANGE, \
    AUTHENTICATOR_DELETE, DEPLOYED_SERVICE_CREATE, DEPLOYED_SERVICE_VIEW, DEPLOYED_SERVICE_CHANGE, \
    DEPLOYED_SERVICE_DELETE, PROVIDER_VIEW, PROVIDER_CREATE, PROVIDER_CHANGE, PROVIDER_DELETE, \
    AUTHENTICATOR_GROUP_MANAGE, DS_USER_SERVICE_MANAGE, DS_USER_MANAGE, DS_GROUP_MANAGE, \
    DS_TRANSPORT_MANAGE, DS_PUBLICATION_MANAGE, PROVIDER_SERVICE_MANAGE, \
    USER_WORKPLACE_SESSION_VIEW, USER_WORKPLACE_SESSION_MANAGE, AUTHENTICATOR_USER_MANAGE
from uds.models.ServicesPool import ServicePool

register = template.Library()


@register.simple_tag
def can_view_logs(user):
    """Проверка наличия у пользователя разрешения на чтение логов"""
    return user.has_perm(LOG_VIEW)


@register.simple_tag
def can_view_configs(user):
    """Проверка наличия у пользователя разрешения на просмотр настроек"""
    return user.has_perm(CONFIG_VIEW)


@register.simple_tag
def can_create_spgroups(user):
    """Проверка наличия у пользователя разрешения на создание групп рабочих мест"""
    return user.has_perm(SERVICE_POOL_GROUP_CREATE)


@register.simple_tag
def can_view_spgroups(user):
    """Проверка наличия у пользователя разрешения на просмотр групп рабочих мест"""
    return user.has_perm(SERVICE_POOL_GROUP_VIEW)


@register.simple_tag
def can_change_spgroups(user):
    """Проверка наличия у пользователя разрешения на изменение групп рабочих мест"""
    return user.has_perm(SERVICE_POOL_GROUP_CHANGE)


@register.simple_tag
def can_delete_spgroups(user):
    """Проверка наличия у пользователя разрешения на удаление групп рабочих мест"""
    return user.has_perm(SERVICE_POOL_GROUP_DELETE)


@register.simple_tag
def can_create_images(user):
    """Проверка наличия у пользователя разрешения на создание изображений из галереи"""
    return user.has_perm(IMAGE_CREATE)


@register.simple_tag
def can_view_images(user):
    """Проверка наличия у пользователя разрешения на чтение изображений из галереи"""
    return user.has_perm(IMAGE_VIEW)


@register.simple_tag
def can_change_images(user):
    """Проверка наличия у пользователя разрешения на изменение изображений из галереи"""
    return user.has_perm(IMAGE_CHANGE)


@register.simple_tag
def can_delete_images(user):
    """Проверка наличия у пользователя разрешения на удаление изображений из галереи"""
    return user.has_perm(IMAGE_DELETE)


@register.simple_tag
def can_view_license(user):
    """Проверка наличия у пользователя разрешения на просмотр лицензии"""
    return user.has_perm(LICENSE_VIEW)


@register.simple_tag
def can_upload_license(user):
    """Проверка наличия у пользователя разрешения на загрузку лицензии"""
    return user.has_perm(LICENSE_UPLOAD)


@register.simple_tag
def can_manage_roles(user):
    """Проверка наличия у пользователя разрешения на управление ролями"""
    return user.has_perm(USER_ROLES_MANAGE)


@register.simple_tag
def can_view_settings_section(user):
    """Проверка наличия у пользователя разрешений на просмотр секции Настройки"""
    _can_view_settings_section = any([
        user.has_perm(IMAGE_VIEW),
        user.has_perm(SERVICE_POOL_GROUP_VIEW),
        user.has_perm(CONFIG_VIEW),
        user.has_perm(LICENSE_VIEW),
        user.has_perm(USER_ROLES_MANAGE),
    ])
    return _can_view_settings_section


@register.simple_tag
def can_create_osmanagers(user):
    """Проверка наличия у пользователя разрешения на создание параметров гостевых ОС"""
    return user.has_perm(OSMANAGER_CREATE)


@register.simple_tag
def can_view_osmanagers(user):
    """Проверка наличия у пользователя разрешения на чтение параметров гостевых ОС"""
    return user.has_perm(OSMANAGER_VIEW)


@register.simple_tag
def can_change_osmanagers(user):
    """Проверка наличия у пользователя разрешения на изменение параметров гостевых ОС"""
    return user.has_perm(OSMANAGER_CHANGE)


@register.simple_tag
def can_delete_osmanagers(user):
    """Проверка наличия у пользователя разрешения на удаление параметров гостевых ОС"""
    return user.has_perm(OSMANAGER_DELETE)


@register.simple_tag
def can_create_transports(user):
    """Проверка наличия у пользователя разрешения на создание протокола доставки"""
    return user.has_perm(TRANSPORT_CREATE)


@register.simple_tag
def can_view_transports(user):
    """Проверка наличия у пользователя разрешения на чтение протокола доставки"""
    return user.has_perm(TRANSPORT_VIEW)


@register.simple_tag
def can_change_transports(user):
    """Проверка наличия у пользователя разрешения на изменение протокола доставки"""
    return user.has_perm(TRANSPORT_CHANGE)


@register.simple_tag
def can_delete_transports(user):
    """Проверка наличия у пользователя разрешения на удаление протокола доставки"""
    return user.has_perm(TRANSPORT_DELETE)


@register.simple_tag
def can_create_networks(user):
    """Проверка наличия у пользователя разрешения на создание сети"""
    return user.has_perm(NETWORK_CREATE)


@register.simple_tag
def can_view_networks(user):
    """Проверка наличия у пользователя разрешения на чтение сети"""
    return user.has_perm(NETWORK_VIEW)


@register.simple_tag
def can_change_networks(user):
    """Проверка наличия у пользователя разрешения на изменение сети"""
    return user.has_perm(NETWORK_CHANGE)


@register.simple_tag
def can_delete_networks(user):
    """Проверка наличия у пользователя разрешения на удаление сети"""
    return user.has_perm(NETWORK_DELETE)


@register.simple_tag
def can_create_authenticators(user):
    """Проверка наличия у пользователя разрешения на создание домена аутентификации"""
    return user.has_perm(AUTHENTICATOR_CREATE)


@register.simple_tag
def can_view_authenticators(user):
    """Проверка наличия у пользователя разрешения на чтение домена аутентификации"""
    return user.has_perm(AUTHENTICATOR_VIEW)


@register.simple_tag
def can_change_authenticators(user):
    """Проверка наличия у пользователя разрешения на изменение домена аутентификации"""
    return user.has_perm(AUTHENTICATOR_CHANGE)


@register.simple_tag
def can_delete_authenticators(user):
    """Проверка наличия у пользователя разрешения на удаление домена аутентификации"""
    return user.has_perm(AUTHENTICATOR_DELETE)


@register.simple_tag
def can_create_deployed_services(user):
    """Проверка наличия у пользователя разрешения на создание фонда рабочих мест"""
    return user.has_perm(DEPLOYED_SERVICE_CREATE)


@register.simple_tag
def can_view_deployed_services(user):
    """Проверка наличия у пользователя разрешения на чтение фонда рабочих мест"""
    return user.has_perm(DEPLOYED_SERVICE_VIEW)


@register.simple_tag
def can_change_deployed_services(user):
    """Проверка наличия у пользователя разрешения на изменение фонда рабочих мест"""
    return user.has_perm(DEPLOYED_SERVICE_CHANGE)


@register.simple_tag
def can_delete_deployed_services(user):
    """Проверка наличия у пользователя разрешения на удаление фонда рабочих мест"""
    return user.has_perm(DEPLOYED_SERVICE_DELETE)


@register.simple_tag
def can_view_provider(user):
    """Проверка наличия у пользователя разрешения на просмотр поставщиков ресурсов"""
    return user.has_perm(PROVIDER_VIEW)


@register.simple_tag
def can_create_provider(user):
    """Проверка наличия у пользователя разрешения на создание поставщикa ресурсов"""
    return user.has_perm(PROVIDER_CREATE)


@register.simple_tag
def can_change_provider(user):
    """Проверка наличия у пользователя разрешения на редактирование поставщикa ресурсов"""
    return user.has_perm(PROVIDER_CHANGE)


@register.simple_tag
def can_delete_provider(user):
    """Проверка наличия у пользователя разрешения на удаление поставщикa ресурсов"""
    return user.has_perm(PROVIDER_DELETE)


@register.simple_tag
def can_manage_group(user):
    """Проверка наличия у пользователя разрешения на управление группами домена аутентификации."""
    return user.has_perm(AUTHENTICATOR_GROUP_MANAGE)


@register.simple_tag
def can_manage_user_service_in_ds(user, uuid):
    """Проверка наличия у пользователя разрешения на управление кешем фонда рабочих мест. """
    return check_exists_permission_for_user(user, ServicePool, uuid, DS_USER_SERVICE_MANAGE)


@register.simple_tag
def can_manage_user_in_ds(user, uuid):
    """Проверка наличия у пользователя разрешения на управление пользователями фонда рабочих мест. """
    return check_exists_permission_for_user(user, ServicePool, uuid, DS_USER_MANAGE)


@register.simple_tag
def can_manage_group_in_ds(user, uuid):
    """Проверка наличия у пользователя разрешения на управление пользовательскими группами фонда рабочих мест. """
    return check_exists_permission_for_user(user, ServicePool, uuid, DS_GROUP_MANAGE)


@register.simple_tag
def can_manage_transport_in_ds(user, uuid):
    """Проверка наличия у пользователя разрешения на управление протоколами доставки фонда рабочих мест. """
    return check_exists_permission_for_user(user, ServicePool, uuid, DS_TRANSPORT_MANAGE)


@register.simple_tag
def can_manage_publication_in_ds(user, uuid):
    """Проверка наличия у пользователя разрешения на управление публикациями фонда рабочих мест. """
    return check_exists_permission_for_user(user, ServicePool, uuid, DS_PUBLICATION_MANAGE)


@register.simple_tag
def can_manage_services(user):
    """Проверка наличия у пользователя разрешения на управление Шаблонами ВМ"""
    return user.has_perm(PROVIDER_SERVICE_MANAGE)


@register.simple_tag
def can_view_workplace_session(user):
    """Проверка наличия у пользователя разрешения на просмотр пользовательских сессий."""
    return user.has_perm(USER_WORKPLACE_SESSION_VIEW)


@register.simple_tag
def can_manage_workplace_session(user):
    """Проверка наличия у пользователя разрешения на управление пользовательскими сессиями."""
    return user.has_perm(USER_WORKPLACE_SESSION_MANAGE)


@register.simple_tag
def can_manage_authenticator_user(user):
    """Проверка наличия у пользователя разрешения на управление пользователями домена аутентификации"""
    return user.has_perm(AUTHENTICATOR_USER_MANAGE)


@register.simple_tag
def is_admin(user):
    """Проверка является ли пользователь администратором"""
    return user.is_superuser


@register.simple_tag
def can_view_components_section(user):
    """Проверка наличия у пользователя разрешений на просмотр секции Компоненты"""
    _can_view_components_section = any([
        user.has_perm(PROVIDER_VIEW),
        user.has_perm(AUTHENTICATOR_VIEW),
        user.has_perm(OSMANAGER_VIEW),
        user.has_perm(TRANSPORT_VIEW),
        user.has_perm(NETWORK_VIEW),
    ])
    return _can_view_components_section


@register.simple_tag
def can_view_workplaces_section(user):
    """Проверка наличия у пользователя разрешений на просмотр секции Рабочие места"""
    _can_view_components_section = any([
        user.has_perm(DEPLOYED_SERVICE_VIEW),
        user.has_perm(USER_WORKPLACE_SESSION_VIEW),
    ])
    return _can_view_components_section
