# -*- encoding: utf-8 -*-
from __future__ import unicode_literals
import logging
from termidesk.core.util import wsptoken
import config

from django.utils.translation import ugettext_noop as _

logger = logging.getLogger(__name__)


errors_map = {
    '[Errno 113] No route to host': 'Проверьте адрес в URL шлюза.',
    'Bad Gateway': 'Проверьте /path в URL шлюза.',
    '[Errno 111] Connection refused': 'Проверьте протокол (ws/wss) в URL шлюза.',
}


def check_connection_to_web_socket_server(ws_uri):
    """Проверка соединения с веб сокет прокси сервером, по адресу, указанному в ws_uri"""

    # Для соединения необходимо подготовить токен. В токене должен присутствовать словарь
    # test_connection, в котором будет записан HTTP-код ответа 100 и ответное сообщение
    # host, port, ssl_target игнорируются, если присутствует test_connection
    # FIXME: код ответа и сообщение перенести в глобальные настройки
    test_http_code = 100
    test_http_message = "Testing connection to wsproxy"

    wstoken_data = {
        'host': '127.0.0.1',
        'port': 80,
        'ssl_target': False,
        'test_connection': {
            'test_code': test_http_code,
            'test_message': test_http_message,
        },
    }
    try:
        # Для подписи токена необходимо прочитать закрытый ключ из файла
        with open(config.WSPROXY_PRIV_KEY_FILE) as f:
            sign_key = f.read()
    except FileNotFoundError:
        return [False, 'Не найден файл ключа {key_file}'.format(key_file=config.WSPROXY_PRIV_KEY_FILE)]

    # Время действия токена, по умолчанию 20 секунд. Если токен будет проверен позже этого времени,
    # то wsproxy будет считать его устаревшим и потому не действительным.
    token = wsptoken.TokenEncoder(sign_key, lifetime=20).encode(wstoken_data)

    import websockify
    ws = websockify.WebSocket()
    response_code, response_message = "", ""

    # Адрес подключения состоит из двух частей: адреса и токена.
    # Например, ws://wsproxy.domain.net/token_in_base64
    # Адрес указывается через форму, токен формируется динамически

    # Если ws_uri не заканчивается слэшем, то при добавлении токена необходимо добавить и слэш
    if ws_uri[-1] != '/':
        delimiter = '/'  # Слэш
    else:
        delimiter = ''  # Пустая строка

    # Готовый адрес для соединения (адрес+разделитель+токен)
    ws_uri = delimiter.join((ws_uri, token,))
    error = ''

    try:
        ws.connect(uri=ws_uri)
        logger.debug("Exception not raises. Check the wsproxy log file.")

    except OSError as e:
        # wrong path
        error = str(e)

    except ConnectionRefusedError as e:
        # wrong scheme
        error = str(e)

    except Exception as e:
        # Ответ от web socket сервера обрабатывается библиотекой websockify и если код ответа не равен 101,
        # возникает исключение, с примерно следующим содержанием 'WebSocket request denied: 100 Test web socket'.
        # Это стандартное поведение websockify. (Это поведение можно изменить, переопределив WebSocket.connect()).
        #
        # Для определения успешности соединения из данной строки необходимо выделить код ответа и сообщение:
        # '100' и 'Test web socket'. Если код ответа равен '100' и сообщение соответствует отправленному, тогда
        # тест пройден успешно.

        # Из строки с оветом оставляется часть после ': ', то есть '100 Test web socket'
        response_status_string = str(e).split(': ')[1]
        response_split = response_status_string.split(" ")
        response_code = response_split[0]
        error = ' '.join(response_split[1:])

    finally:
        response_message = errors_map.get(error, error)
        try:
            ws.close()
        except:
            pass

    # При успешном подключении первый элемент списка должен быть True
    return [str(test_http_code) == response_code, _(response_message)]
