# -*- coding: utf-8 -*-

#
# Copyright (c) 2012 Virtual Cable S.L.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
@author: Adolfo Gómez, dkmaster at dkmon dot com
'''
from __future__ import unicode_literals

import sys

from django.conf import settings
from Crypto.PublicKey import RSA
from OpenSSL import crypto
from Crypto.Random import atfork
import hashlib
import array
import uuid
import datetime
import codecs
import random
import string

import logging
import six

logger = logging.getLogger(__name__)

# To generate an rsa key, first we need the crypt module
# next, we do:
# from Crypto.PublicKey import RSA
# import os
# RSA.generate(1024, os.urandom).exportKey()


class CryptoManager(object):
    CODEC = 'base64'

    instance = None

    def __init__(self):
        self._rsa = RSA.importKey(settings.RSA_KEY)
        self._namespace = uuid.UUID('627a37a5-e8db-431a-b783-73f7d20b4934')
        self._counter = 0

    @staticmethod
    def manager():
        if CryptoManager.instance is None:
            CryptoManager.instance = CryptoManager()
        return CryptoManager.instance

    def encrypt(self, value):
        if isinstance(value, six.text_type):
            value = value.encode('utf-8')

        atfork()

        # For python 2.7 CryptoManager.CODEC must be a str type, not unicode,
        # but for python 3.5 CryptoManager.CODEC can be a str type
        if not isinstance(CryptoManager.CODEC, str) and isinstance(CryptoManager.CODEC, six.text_type):
            codec = six.binary_type(CryptoManager.CODEC)
        else:
            codec = CryptoManager.CODEC

        rsa_encrypted_string = self._rsa.encrypt(value, six.b(''))[0]

        # may be it is needed to pass 'codec' instead  'CryptoManager.CODEC'
        encoded_string = codecs.encode(rsa_encrypted_string, CryptoManager.CODEC)
        return six.text_type(encoded_string, 'utf-8')

    def decrypt(self, value):
        # type: (six.text_type) -> six.text_type
        if isinstance(value, six.text_type):
            value = value.encode('utf-8')
        # import inspect
        try:
            atfork()

            # For python 2.7 CryptoManager.CODEC must be a str type, not unicode,
            # but for python 3.5 CryptoManager.CODEC can be a str type
            if not isinstance(CryptoManager.CODEC, str) and isinstance(CryptoManager.CODEC, six.text_type):
                codec = six.binary_type(CryptoManager.CODEC)
            else:
                codec = CryptoManager.CODEC

            decoded_value = codecs.decode(value, codec)
            rsa_decrypted_value = self._rsa.decrypt(decoded_value).decode('utf-8')
            return six.text_type(rsa_decrypted_value)

        except Exception:
            logger.exception('Decripting: {0}'.format(value))
            # logger.error(inspect.stack())
            return six.text_type('decript error')

    @staticmethod
    def xor(s1: bytes, s2: bytes) -> bytes:
        """
        Вычисляет Исключающее ИЛИ (xor) между двумя bytes-строками через массивы беззнаковых целых чисел
        @param s1: строка с данными
        @param s2: строка, содержащая специальные salt-данные
        @return: bytes-строка - результат вычисления xor
        """
        if isinstance(s1, six.text_type):
            s1 = s1.encode('utf-8')
        if isinstance(s2, six.text_type):
            s2 = s2.encode('utf-8')

        # Множитель для увеличения длины salt-строки, если длина s2 меньше длины s1
        # В случае если s1 длиньше s2, то s2 повторяется еще раз (s2 * mult) столько раз
        # чтобы в итоге длина s2 стала больше или равной длине s1
        mult = int(len(s1) / len(s2)) + 1

        # Строки s1 и s2 преобразовываются в массивы,
        # где каждый элемент имеет тип unsigned char
        unsigned_char = 'B'
        s1 = array.array(unsigned_char, s1)
        s2 = array.array(unsigned_char, s2 * mult)

        # Вычисление xor между смежными элементами массивов
        xor_result = (s1[i] ^ s2[i] for i in range(len(s1)))
        # Преобразование результата в массив беззнаковых целых
        _array = array.array(unsigned_char, xor_result)
        # Преобразование массива в bytes-строку, чтобы вернуть результат с правильным типом
        bytes_result = _array.tobytes()
        return bytes_result

    def loadPrivateKey(self, rsaKey):
        try:
            pk = RSA.importKey(rsaKey)
        except Exception as e:
            raise e
        return pk

    def loadCertificate(self, certificate):
        try:
            cert = crypto.load_certificate(crypto.FILETYPE_PEM, certificate)
        except crypto.Error as e:
            raise Exception(e.message[0][2])
        return cert

    def certificateString(self, certificate):
        return certificate.replace('-----BEGIN CERTIFICATE-----', '').replace('-----END CERTIFICATE-----', '').replace('\n', '')

    def hash(self, value):
        if isinstance(value, six.text_type):
            value = value.encode('utf-8')

        if value is '' or value is None:
            return ''

        return six.text_type(hashlib.sha1(value).hexdigest())

    def uuid(self, obj=None):
        '''
        Generates an uuid from obj. (lower case)
        If obj is None, returns an uuid based on current datetime + counter
        '''
        if obj is None:
            obj = six.text_type(datetime.datetime.now()) + six.text_type(self._counter)
            self._counter += 1

        if six.PY3:
            # For python 3 obj must be a 'str'
            if not isinstance(obj, str):
                if isinstance(obj, datetime.datetime):
                    obj = obj.isoformat()
                else:
                    obj = six.text_type(obj, 'utf-8')

        else:
            # For python 2
            if not isinstance(obj, str):
                obj = six.binary_type(obj)

        # I believe uuid returns a lowercase uuid always, but in case... :)
        return six.text_type(uuid.uuid5(self._namespace, obj)).lower()

    def randomString(self, length=40):
        return ''.join(random.choice(string.ascii_lowercase + string.digits) for _ in range(length))
