# -*- coding: utf-8 -*-

#
# Copyright (c) 2012 Virtual Cable S.L.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
@author: Adolfo Gómez, dkmaster at dkmon dot com
'''
import codecs

import six

from uds.core.Serializable import Serializable
import pickle
import timeit


class Attribute(object):

    def __init__(self, theType, value=None):
        self._type = theType
        self.setValue(value)

    def getType(self):
        return self._type

    def getValue(self):
        return self._value

    def getStrValue(self):
        return str(self._value)

    def setValue(self, value):
        if value is None:
            self._value = self._type()
        else:
            self._value = self._type(value)


class AutoAttributes(Serializable):
    '''
    Easy creation of attributes to marshal & unmarshal at modules
    usage as base class (First class so yours inherits this "marshal" and "unmarshal"
    initialize at init with super(myclass,self).__init__(attr1=type, attr2=type, ...)
    or with declare(attr1=type,attr2=type,..)
    Access attrs as "self._attr1, self._attr2"
    '''

    # : This codec is not intended to override Serializable codec
    # : Serializable codec is for encoding marshaled data,
    # : while this codec is for encoding pickled data from autoattributes
    ACODEC = 'zip'

    def __init__(self, **kwargs):
        self.declare(**kwargs)

    def __getattribute__(self, name):
        if name.startswith('_') and name[1:] in self.dict:
            return self.dict[name[1:]].getValue()
        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        if name.startswith('_') and name[1:] in self.dict:
            self.dict[name[1:]].setValue(value)
        else:
            object.__setattr__(self, name, value)

    def declare(self, **kwargs):
        d = {}
        # 2to3: iteritems() changed to items()
        for key, typ in kwargs.items():
            d[key] = Attribute(typ)
        self.dict = d

    def marshal(self):
        # type: () -> six.binary_type
        if six.PY3:
            group_delimiter = b'\2'
            key_value_delimiter = b'\1'
        else:
            group_delimiter = '\2'
            key_value_delimiter = '\1'

        items_list = []
        for key, value in self.dict.items():

            if six.PY3 and isinstance(key, str):
                key = key.encode('utf-8')

            p_value = pickle.dumps(value, protocol=0)

            item = key_value_delimiter.join((key, p_value))
            items_list.append(item)

        marshaled_result = group_delimiter.join(items_list)

        marshaled_result = codecs.encode(marshaled_result, AutoAttributes.ACODEC)

        return marshaled_result

    def unmarshal(self, data):
        # type: (six.binary_type) -> ()

        if six.PY3:
            group_delimiter = b'\2'
            key_value_delimiter = b'\1'
        else:
            group_delimiter = '\2'
            key_value_delimiter = '\1'

        if data == '':  # Can be empty
            return
        # We keep original data (maybe incomplete)
        _data = codecs.decode(data, AutoAttributes.ACODEC)

        for pair in _data.split(group_delimiter):

            k, v = pair.split(key_value_delimiter)
            if six.PY3 and isinstance(k, six.binary_type):
                k = k.decode('utf-8')

            self.dict[k] = pickle.loads(v)

    def __str__(self):
        str_ = '<AutoAttribute '
        # 2to3: iteritems() changed to items()
        for k, v in self.dict.items():
            str_ += "%s (%s) = %s" % (k, v.getType(), v.getStrValue())
        return str_ + '>'
