# -*- coding: utf-8 -*-

#
# Copyright (c) 2012 Virtual Cable S.L.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
@author: Adolfo Gómez, dkmaster at dkmon dot com
'''
from __future__ import unicode_literals
from server.celery import app as celery
from django.db import transaction
from uds.core.util.Config import GlobalConfig
from uds.models import DeployedService, getSqlDatetime
from uds.core.util.State import State
from uds.core.jobs.Job import Job
from datetime import timedelta
import logging

logger = logging.getLogger(__name__)


# Класс должен быть наследником классов celery.Task и Job,
# для того, чтобы он имел возможность запускаться с помощью celery и TaskManager
class DeployedServiceInfoItemsCleaner(celery.Task, Job):
    # Частота запуска задачи в секундах
    frecuency = 3607
    frecuency_cfg = GlobalConfig.CLEANUP_CHECK  # Request run cache "info" cleaner every configured seconds. If config value is changed, it will be used at next reload
    # Имя задачи для отображения в интерфейсе
    friendly_name = 'Deployed Service Info Cleaner'
    # Имя задачи, которое будет зарегистрировано в реестре задач celery
    # Данное имя должно быть уникально для каждой задачи
    celery_name = 'J.uds.core.workers.DeployedServiceCleaner.DeployedServiceInfoItemsCleaner'
    # Переопределение атрибута celery.Task.name
    name = celery_name
    # Возможность выполнения в celery
    celery_compatible = True

    def __init__(self, environment=None):
        # Для TaskManager, environment инициализируется из модели.
        # В celery модель не привязана к задаче, поэтому для обеспечения
        # уникальности инициализируется environment c celery_name в качестве unique_key
        if environment is None:
            from uds.core.Environment import Environment
            environment = Environment(self.celery_name)
        super(DeployedServiceInfoItemsCleaner, self).__init__(environment)

    def __call__(self, *args, **kwargs):
        # Необходимо переопределить __call__(), если метод self.run() не принимает аргументы.
        # В противном случае, выполнение задачи вызывает исключение.
        return self.run()

    def run(self):
        # Метод run является телом задачи для celery.
        # При необходимости он может вызывать другие методы класса
        removeFrom = getSqlDatetime() - timedelta(seconds=GlobalConfig.KEEP_INFO_TIME.getInt())
        DeployedService.objects.filter(state__in=State.INFO_STATES, state_date__lt=removeFrom).delete()


# Регистрация класса в качестве задачи в реестре celery.
celery.tasks.register(DeployedServiceInfoItemsCleaner)


# Класс должен быть наследником классов celery.Task и Job,
# для того, чтобы он имел возможность запускаться с помощью celery и TaskManager
class DeployedServiceRemover(celery.Task, Job):
    # Частота запуска задачи в секундах
    frecuency = 31
    frecuency_cfg = GlobalConfig.REMOVAL_CHECK  # Request run publication "removal" every configued seconds. If config value is changed, it will be used at next reload
    # Имя задачи для отображения в интерфейсе
    friendly_name = 'Deployed Service Cleaner'
    # Имя задачи, которое будет зарегистрировано в реестре задач celery
    # Данное имя должно быть уникально для каждой задачи
    celery_name = 'J.uds.core.workers.DeployedServiceCleaner.DeployedServiceRemover'
    # Переопределение атрибута celery.Task.name
    name = celery_name
    # Возможность выполнения в celery
    celery_compatible = True

    def __init__(self, environment=None):
        # Для TaskManager, environment инициализируется из модели.
        # В celery модель не привязана к задаче, поэтому для обеспечения
        # уникальности инициализируется environment c celery_name в качестве unique_key
        if environment is None:
            from uds.core.Environment import Environment
            environment = Environment(self.celery_name)
        super(DeployedServiceRemover, self).__init__(environment)

    def __call__(self, *args, **kwargs):
        # Необходимо переопределить __call__(), если метод self.run() не принимает аргументы.
        # В противном случае, выполнение задачи вызывает исключение.
        return self.run()

    def startRemovalOf(self, ds):
        if ds.service is None:  # Maybe an inconsistent value?
            logger.error('Found service pool {} without service')
            ds.delete()
            return

        # Get publications in course...., that only can be one :)
        logger.debug('Removal process of {0}'.format(ds))

        publishing = ds.publications.filter(state=State.PREPARING)
        for p in publishing:
            p.cancel()
        # Now all publishments are canceling, let's try to cancel cache and assigned
        uServices = ds.userServices.filter(state=State.PREPARING)
        for u in uServices:
            logger.debug('Canceling {0}'.format(u))
            u.cancel()
        # Nice start of removal, maybe we need to do some limitation later, but there should not be too much services nor publications cancelable at once
        ds.state = State.REMOVING
        #ds.name = ds.name + ' (removed)'
        ds.save()

    def continueRemovalOf(self, ds):
        # Recheck that there is no publication created in "bad moment"
        try:
            for p in ds.publications.filter(state=State.PREPARING):
                p.cancel()
        except Exception:
            pass

        try:
            # Now all publishments are canceling, let's try to cancel cache and assigned
            uServices = ds.userServices.filter(state=State.PREPARING)
            for u in uServices:
                logger.debug('Canceling {0}'.format(u))
                u.cancel()
        except Exception:
            pass

        # First, we remove all publications and user services in "info_state"
        with transaction.atomic():
            ds.userServices.select_for_update().filter(state__in=State.INFO_STATES).delete()

        # Mark usable user services as removable
        now = getSqlDatetime()

        with transaction.atomic():
            ds.userServices.select_for_update().filter(state=State.USABLE).update(state=State.REMOVABLE, state_date=now)

        # When no service is at database, we start with publications
        if ds.userServices.all().count() == 0:
            try:
                logger.debug('All services removed, checking active publication')
                if ds.activePublication() is not None:
                    logger.debug('Active publication found, unpublishing it')
                    ds.unpublish()
                else:
                    logger.debug('No active publication found, removing info states and checking if removal is done')
                    ds.publications.filter(state__in=State.INFO_STATES).delete()
                    if ds.publications.count() is 0:
                        ds.removed()  # Mark it as removed, clean later from database
            except Exception:
                logger.exception('Cought unexpected exception at continueRemovalOf: ')

    def run(self):
        # Метод run является телом задачи для celery.
        # При необходимости он может вызывать другие методы класса

        # First check if there is someone in "removable" estate
        rems = DeployedService.objects.filter(state=State.REMOVABLE)[:10]
        if len(rems) > 0:
            logger.debug('Found a deployed service marked for removal. Starting removal of {0}'.format(rems))
            for ds in rems:
                try:
                    # Skips checking deployed services in maintenance mode
                    if ds.isInMaintenance() is False:
                        self.startRemovalOf(ds)
                except Exception as e1:
                    logger.error('Error removing {}: {}'.format(ds, e1))
                    try:
                        ds.delete()
                    except Exception as e2:
                        logger.error('Could not delete {}'.format(e2))


        rems = DeployedService.objects.filter(state=State.REMOVING)[:10]
        if len(rems) > 0:
            logger.debug('Found a deployed service in removing state, continuing removal of {0}'.format(rems))
            for ds in rems:
                try:
                    # Skips checking deployed services in maintenance mode
                    if ds.isInMaintenance() is False:
                        self.continueRemovalOf(ds)
                except Exception:
                    logger.exception('Removing deployed service')


# Регистрация класса в качестве задачи в реестре celery.
celery.tasks.register(DeployedServiceRemover)
