# -*- coding: utf-8 -*-

#
# Copyright (c) 2012 Virtual Cable S.L.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright notice,
#      this list of conditions and the following disclaimer in the documentation
#      and/or other materials provided with the distribution.
#    * Neither the name of Virtual Cable S.L. nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''
@author: Adolfo Gómez, dkmaster at dkmon dot com

modified by Dmitry Dvoeglazov 2019-09-12
'''
from __future__ import unicode_literals
from itertools import filterfalse as exclude
from server.celery import app as celery
from django.db import transaction
from django.db.models import Q
from uds.core.util.Config import GlobalConfig
from uds.core.util.State import State
from uds.core.managers.UserServiceManager import UserServiceManager
from uds.core.services.Exceptions import MaxServicesReachedError
from uds.models import DeployedService
from uds.core import services
from uds.core.util import log
from uds.core.jobs.Job import Job
import logging

logger = logging.getLogger(__name__)


def _no_provider_implementation(sp):
    if sp.service.provider.getType() is None:
        # Пропускаем фонды(сервис пулы) для провайдеров с отсутсвующей реализацией в системе
        # Например, провайдер был реализован в плагине, а плагин удалили
        # Deprecated 'warn' replaced to 'warning'
        logger.warning("Missing provider %s implementation for %s.", sp.service.provider, sp)
        return True
    return False


def _no_active_publication(sp):
    # If this deployedService don't have a publication active and needs it, ignore it
    spServiceInstance = sp.service.getInstance()

    if sp.activePublication() is None and spServiceInstance.publicationType is not None:
        logger.debug('{} Needs publication but do not have one, cache test ignored'.format(sp))
        return True

    return False


def _publication_in_progress(sp):
    # If it has any running publication, do not generate cache anymore
    if sp.publications.filter(state=State.PREPARING).count() > 0:
        logger.debug('Stopped cache generation for deployed service with publication running: {0}'.format(sp))
        return True

    return False


def _is_restrained(sp):
    if sp.isRestrained():
        return True

    return False


# Класс должен быть наследником классов celery.Task и Job,
# для того, чтобы он имел возможность запускаться с помощью celery и TaskManager
class ServiceCacheUpdater(celery.Task, Job):
    '''
    Cache updater is responsible of keeping up to date the cache for different deployed services configurations requested
    We only process items that are "cacheables", to speed up process we will use the fact that initialServices = preparedServices = maxServices = 0
    if cache is not needed.
    This is included as a scheduled task that will run every X seconds, and scheduler will keep it so it will be only executed by one backend at a time
    '''
    # Частота запуска задачи в секундах
    frecuency = 19
    frecuency_cfg = GlobalConfig.CACHE_CHECK_DELAY  # Request run cache manager every configured seconds (defaults to 20 seconds).
    # Имя задачи для отображения в интерфейсе
    friendly_name = 'Service Cache Updater'
    # Имя задачи, которое будет зарегистрировано в реестре задач celery
    # Данное имя должно быть уникально для каждой задачи
    celery_name = 'J.uds.core.workers.ServiceCacheUpdater.ServiceCacheUpdater'
    # Переопределение атрибута celery.Task.name
    name = celery_name
    # Возможность выполнения в celery
    celery_compatible = True

    def __init__(self, environment=None):
        # Для TaskManager, environment инициализируется из модели.
        # В celery модель не привязана к задаче, поэтому для обеспечения
        # уникальности инициализируется environment c celery_name в качестве unique_key
        if environment is None:
            from uds.core.Environment import Environment
            environment = Environment(self.celery_name)
        super(ServiceCacheUpdater, self).__init__(environment)

    def __call__(self, *args, **kwargs):
        # Необходимо переопределить __call__(), если метод self.run() не принимает аргументы.
        # В противном случае, выполнение задачи вызывает исключение.
        return self.run()

    @staticmethod
    def calcProportion(max_, actual):
        return actual * 10000 / (max_ or 1)

    def _notifyRestrain(self, deployedService):
        log.doLog(deployedService, log.WARN, 'Service Pool is restrained due to errors', log.INTERNAL)
        logger.info(' {0} is restrained, will check this later'.format(deployedService.name))

    def _get_processable_pools(self):
        '''
        Получение списка фондов (:class:`uds.models.ServicesPool.DeployedService`)
        для которых требуется проверить изменения в кэшах.

        :return: Список объектов класса :class:`uds.models.ServicesPool.DeployedService`
        :rtype: list
        '''

        filter_query = DeployedService.objects.all()

        # В фондах должен быть задан L1 кэш
        filter_query = filter_query.filter(cache_l1_srvs__gte=0)

        # Исключаем фонды с 0-м максимальным кол-вом машин
        # Странное правило, похоже приводит к тому, что при изменении максимума
        # до 0 на существующем фонде он не очищается, а перестает обновляться.
        filter_query = filter_query.filter(max_srvs__gt=0)

        # Исключаем неактивные фонды или фонды в других состояниях
        filter_query = filter_query.filter(state=State.ACTIVE)

        # Исключаем фонды, чьи поставщики ресурсов находятся в режиме техобслуживания
        filter_query = filter_query.filter(service__provider__maintenance_mode=False)

        # Возвращаем итоговый список фондов
        return filter_query[:]

    def servicesPoolsNeedingCacheUpdate(self):
        # State filter for cached and inAssigned objects
        # First we get all deployed services that could need cache generation
        DeployedService.objects.update()

        # We start filtering out the deployed services that do not need caching at all.
        whichNeedsCaching = self._get_processable_pools()

        # Исключаем из списка фонды с провайдерами без реализации
        whichNeedsCaching = exclude(_no_provider_implementation, whichNeedsCaching)

        # Исключаем из списка фонды без публикаций.
        # В таких фондах рабочие места создавать нельзя.
        whichNeedsCaching = exclude(_no_active_publication, whichNeedsCaching)

        # Исключаем из списка фонды с готовящимися публикациями.
        # Без действительной публикации места создавать нельзя.
        # Также, нет смысла создавать места для старых публикаций.
        whichNeedsCaching = list(exclude(_publication_in_progress, whichNeedsCaching))

        # Логируем информацию об ограниченных фондах
        restrained_pools = list(filter(_is_restrained, whichNeedsCaching))
        for sp in restrained_pools:
            self._notifyRestrain(sp)

        # Исключаем из списка ограниченные фонды.
        # В таких фондах изменения в кеше производятся реже.
        whichNeedsCaching = exclude(_is_restrained, whichNeedsCaching)

        # We will get the one that proportionally needs more cache
        servicesPools = []
        for sp in whichNeedsCaching:
            sp.userServices.update()  # Cleans cached queries

            # Get data related to actual state of cache
            inCacheL1 = sp.cachedUserServices().filter(UserServiceManager.getCacheStateFilter(services.UserDeployment.L1_CACHE)).count()
            inCacheL2 = sp.cachedUserServices().filter(UserServiceManager.getCacheStateFilter(services.UserDeployment.L2_CACHE)).count()
            inAssigned = sp.assignedUserServices().filter(UserServiceManager.getStateFilter()).count()
            # if we bypasses max cache, we will reduce it in first place. This is so because this will free resources on service provider
            logger.debug("Examining {0} with {1} in cache L1 and {2} in cache L2, {3} inAssigned".format(
                         sp, inCacheL1, inCacheL2, inAssigned))
            totalL1Assigned = inCacheL1 + inAssigned
            # We have more than we want
            if totalL1Assigned > sp.max_srvs:
                logger.debug('We have more services than max configured')
                servicesPools.append((sp, inCacheL1, inCacheL2, inAssigned))
                continue
            # We have more in L1 cache than needed
            if inCacheL1 > sp.cache_l1_srvs:
                logger.debug('We have more services in cache L1 than configured')
                servicesPools.append((sp, inCacheL1, inCacheL2, inAssigned))
                continue

            # If we have more in L2 cache than needed, decrease L2 cache, but int this case, we continue checking cause L2 cache removal
            # has less priority than l1 creations or removals, but higher. In this case, we will simply take last l2 oversized found and reduce it
            if inCacheL2 > sp.cache_l2_srvs:
                logger.debug('We have more services in L2 cache than configured, decreasing it')
                servicesPools.append((sp, inCacheL1, inCacheL2, inAssigned))
                continue

            # If this service don't allows more starting user services, continue
            if UserServiceManager.manager().canInitiateServiceFromDeployedService(sp) is False:
                logger.debug('This provider has the max allowed starting services running: {0}'.format(sp))
                provider_limit = sp.service.getInstance().parent().getMaxPreparingServices()
                logger.debug('Provider for {spname} has the max allowed starting services limit [{plimit}] reached'.format(spname=sp.name, plimit=provider_limit))
                continue

            # If wee need to grow l2 cache, annotate it
            # Whe check this before checking the total, because the l2 cache is independent of max services or l1 cache.
            # It reflects a value that must be keeped in cache for futre fast use.
            if inCacheL2 < sp.cache_l2_srvs:
                logger.debug('Needs to grow L2 cache for {}'.format(sp))
                servicesPools.append((sp, inCacheL1, inCacheL2, inAssigned))
                continue

            # We skip it if already at max
            if totalL1Assigned == sp.max_srvs:
                continue

            if inCacheL1 < sp.cache_l1_srvs:
                logger.debug('Needs to grow L1 cache for {}'.format(sp))
                servicesPools.append((sp, inCacheL1, inCacheL2, inAssigned))

        # We also return calculated values so we can reuse then
        return servicesPools

    def growL1Cache(self, sp, cacheL1, cacheL2, assigned):
        '''
        This method tries to enlarge L1 cache.

        If for some reason the number of deployed services (Counting all, ACTIVE
        and PREPARING, assigned, L1 and L2) is over max allowed service deployments,
        this method will not grow the L1 cache
        '''
        logger.debug("Growing L1 cache creating a new service for {0}".format(sp))
        # First, we try to assign from L2 cache
        if cacheL2 > 0:
            valid = None
            with transaction.atomic():
                for n in sp.cachedUserServices().select_for_update().filter(UserServiceManager.getCacheStateFilter(services.UserDeployment.L2_CACHE)).order_by('creation_date'):
                    if n.needsOsManager():
                        if State.isUsable(n.state) is False or State.isUsable(n.os_state):
                            valid = n
                            break
                    else:
                        valid = n
                        break

            if valid is not None:
                valid.moveToLevel(services.UserDeployment.L1_CACHE)
                return
        try:
            UserServiceManager.manager().createCacheFor(sp.activePublication(), services.UserDeployment.L1_CACHE)
        except MaxServicesReachedError as e:
            log.doLog(sp, log.ERROR, 'Max number of services reached for this service', log.INTERNAL)
            logger.error(str(e))
        except:
            logger.exception('Exception')

    def growL2Cache(self, sp, cacheL1, cacheL2, assigned):
        '''
        Tries to grow L2 cache of service.

        If for some reason the number of deployed services (Counting all, ACTIVE
        and PREPARING, assigned, L1 and L2) is over max allowed service deployments,
        this method will not grow the L1 cache
        '''
        logger.debug("Growing L2 cache creating a new service for {0}".format(sp))
        try:
            UserServiceManager.manager().createCacheFor(sp.activePublication(), services.UserDeployment.L2_CACHE)
        except MaxServicesReachedError as e:
            logger.error(str(e))
            # TODO: When alerts are ready, notify this

    def reduceL1Cache(self, sp, cacheL1, cacheL2, assigned):
        logger.debug("Reducing L1 cache erasing a service in cache for {0}".format(sp))
        # We will try to destroy the newest cacheL1 element that is USABLE if the deployer can't cancel a new service creation
        cacheItems = sp.cachedUserServices().filter(UserServiceManager.getCacheStateFilter(services.UserDeployment.L1_CACHE)).order_by('-creation_date')
        if len(cacheItems) == 0:
            logger.debug('There is more services than configured, but could not reduce cache cause its already empty')
            return

        if cacheL2 < sp.cache_l2_srvs:
            valid = None
            for n in cacheItems:
                if n.needsOsManager():
                    if State.isUsable(n.state) is False or State.isUsable(n.os_state):
                        valid = n
                        break
                else:
                    valid = n
                    break

            if valid is not None:
                valid.moveToLevel(services.UserDeployment.L2_CACHE)
                return

        cache = cacheItems[0]
        cache.removeOrCancel()

    def reduceL2Cache(self, sp, cacheL1, cacheL2, assigned):
        logger.debug("Reducing L2 cache erasing a service in cache for {0}".format(sp))
        if cacheL2 > 0:
            cacheItems = sp.cachedUserServices().filter(UserServiceManager.getCacheStateFilter(services.UserDeployment.L2_CACHE)).order_by('creation_date')
            # TODO: Look first for non finished cache items and cancel them
            cache = cacheItems[0]
            cache.removeOrCancel()

    def run(self):
        # Метод run является телом задачи для celery.
        # При необходимости он может вызывать другие методы класса

        logger.debug('Starting cache checking')
        logger.debug("Services cache check STARTED")
        # We need to get
        servicesThatNeedsUpdate = self.servicesPoolsNeedingCacheUpdate()
        logger.debug("Services that needs update {}: {}".format(len(servicesThatNeedsUpdate), [s[0].name for s in servicesThatNeedsUpdate]))
        for sp, cacheL1, cacheL2, assigned in servicesThatNeedsUpdate:
            # We have cache to update??
            logger.debug("Updating cache for {0}".format(sp))
            logger.debug("Checking {}".format(sp.name))
            totalL1Assigned = cacheL1 + assigned
            logger.debug("[{}] Total L1 + Assigned: {}".format(sp.name, totalL1Assigned))
            # We try first to reduce cache before tring to increase it.
            # This means that if there is excesive number of user deployments
            # for L1 or L2 cache, this will be reduced untill they have good numbers.
            # This is so because service can have limited the number of services and,
            # if we try to increase cache before having reduced whatever needed
            # first, the service will get lock until someone removes something.
            if totalL1Assigned > sp.max_srvs:
                logger.debug("[{}] Total L1 + Assigned [{}] > Max services [{}]. Reduce L1 cache".format(sp.name, totalL1Assigned, sp.max_srvs))
                self.free_cache_l1(sp, cacheL1, cacheL2, assigned)
            elif cacheL1 > sp.cache_l1_srvs:
                logger.debug("[{}] Current L1 cache [{}] > Max L1 cache [{}]. Reduce L1 cache".format(
                    sp.name, totalL1Assigned, cacheL1, sp.cache_l1_srvs))
                self.free_cache_l1(sp, cacheL1, cacheL2, assigned)
            elif cacheL2 > sp.cache_l2_srvs:  # We have excesives L2 items
                logger.debug("[{}] Current L2 cache [{}] > Max L2 cache [{}]. Reduce L2 cache".format(sp.name, cacheL2, sp.cache_l2_srvs))
                self.free_cache_l2(sp, cacheL1, cacheL2, assigned)
            elif totalL1Assigned < sp.max_srvs and cacheL1 < sp.cache_l1_srvs:  # We need more services
                logger.debug("[{}] Total L1 + Assigned [{}] < Max services [{}] AND Current L1 cache [{}] < Max L1 cache [{}]. Grow L1 cache".format(
                    sp.name, totalL1Assigned, sp.max_srvs, cacheL1, sp.cache_l1_srvs))
                self.fill_cache_l1(sp, cacheL1, cacheL2, assigned)
            elif cacheL2 < sp.cache_l2_srvs:  # We need more L2 items
                logger.debug("[{}] Current L2 cache [{}] < Max L2 cache [{}]. Grow L2 cache".format(sp.name, cacheL2, sp.cache_l2_srvs))
                self.fill_cache_l2(sp, cacheL1, cacheL2, assigned)
            else:
                logger.info("We have more services than max requested for {0}, but can't erase any of then cause all of them are already assigned".format(sp))

        logger.debug("Services cache check ENDED")

    def fill_cache_l1(self, sp, cacheL1, cacheL2, assigned):
        logger.debug('[{spn}] [L1] Filling L1 cache (L1={l1}, L2={l2}, assigned={a})'
                     .format(spn=sp.name, l1=cacheL1, l2=cacheL2, a=assigned))
        services_count = cacheL1
        max_cache_capacity = sp.cache_l1_srvs

        vms_needed = max_cache_capacity-services_count
        vms_to_create = min(self.count_vms_to_create(sp), vms_needed)

        logger.debug('[{sp}] [L1] There are {d} workplace(s) in L1 cache'.format(sp=sp.name,d=services_count))
        logger.debug('[{sp}] [L1] The maximum number of workplaces in a pool is {n}'.format(sp=sp.name,n=max_cache_capacity))
        logger.debug('[{sp}] [L1] Need to add {f} workplace(s).'.format(sp=sp.name,f=vms_needed))
        logger.debug('[{sp}] [L1] {f} workplace(s) will be added'.format(sp=sp.name,f=vms_to_create))

        for n in range(0, vms_to_create):
            self.growL1Cache(sp, cacheL1, cacheL2, assigned)

        logger.info('[{sp}] [L1] Added {n} workplace(s) to L1 cache'.format(sp=sp.name,n=vms_to_create))

    def fill_cache_l2(self, sp, cacheL1, cacheL2, assigned):
        logger.debug('[{spn}] [L2] Filling L2 cache for {sp}(L1={l1}, L2={l2}, assigned={a})'
                     .format(spn=sp.name, sp=sp, l1=cacheL1, l2=cacheL2, a=assigned))
        services_count = cacheL2
        max_cache_capacity = sp.cache_l2_srvs

        vms_needed = max_cache_capacity-services_count
        vms_to_create = min(self.count_vms_to_create(sp), vms_needed)
        logger.debug('[{sp}] [L2] There are {d} workplace(s) in L2 cache'.format(sp=sp.name,d=services_count))
        logger.debug('[{sp}] [L2] The maximum number of workplaces in L2 cache is {n}'.format(sp=sp.name,n=max_cache_capacity))
        logger.debug('[{sp}] [L2] Need to add {f} workplace(s).'.format(sp=sp.name,f=vms_needed))
        logger.debug('[{sp}] [L2] {f} workplace(s) will be added'.format(sp=sp.name,f=vms_to_create))

        for n in range(0, vms_to_create):
            self.growL2Cache(sp, cacheL1, cacheL2, assigned)

        logger.info('[{sp}] [L2] {n} - Added {n} workplace(s) to L2 cache'.format(sp=sp.name,n=vms_to_create))

    def free_cache_l1(self, sp, cacheL1, cacheL2, assigned):
        logger.debug('[{spn}] [L1] Freeing L1 cache for {sp}(L1={l1}, L2={l2}, assigned={a})'
                     .format(spn=sp.name, sp=sp, l1=cacheL1, l2=cacheL2, a=assigned))
        services_count = cacheL1
        max_cache_capacity = min(sp.cache_l1_srvs, sp.max_srvs)
        
        vms_needed = services_count-max_cache_capacity
        vms_to_remove = min(self.count_vms_to_remove(sp), vms_needed)

        logger.debug('[{sp}] [L1] There are {d} workplace(s) in L1 cache'.format(sp=sp.name,d=services_count))
        logger.debug('[{sp}] [L1] The maximum number of workplaces in L1 cache is {n}'.format(sp=sp.name,n=max_cache_capacity))
        logger.debug('[{sp}] [L1] Need to remove {f} workplace(s).'.format(sp=sp.name,f=vms_needed))
        logger.debug('[{sp}] [L1] {f} workplace(s) will be removed'.format(sp=sp.name,f=vms_to_remove))

        for n in range(0, vms_to_remove):
            self.reduceL1Cache(sp, cacheL1, cacheL2, assigned)

        logger.info('[{sp}] [L1] Removed {n} workplaces from L1 cache'.format(sp=sp.name,n=vms_to_remove))

    def free_cache_l2(self, sp, cacheL1, cacheL2, assigned):
        logger.debug('[{spn}] [L1] Freeing L2 cache for {sp}(L1={l1}, L2={l2}, assigned={a})'
                     .format(spn=sp.name, sp=sp, l1=cacheL1, l2=cacheL2, a=assigned))
        services_count = cacheL2
        max_cache_capacity = min(sp.cache_l2_srvs, sp.max_srvs)

        vms_needed = services_count-max_cache_capacity
        vms_to_remove = min(self.count_vms_to_remove(sp), vms_needed)

        logger.debug('[{sp}] [L2] There are {d} workplace(s) in L2 cache'.format(sp=sp.name,d=services_count))
        logger.debug('[{sp}] [L2] The maximum number of workplaces in L2 cache is {n}'.format(sp=sp.name,n=max_cache_capacity))
        logger.debug('[{sp}] [L2] Need to remove {f} workplace(s).'.format(sp=sp.name,f=vms_needed))
        logger.debug('[{sp}] [L2] {f} workplace(s) will be removed'.format(sp=sp.name,f=vms_to_remove))

        for n in range(0, vms_to_remove):
            self.reduceL2Cache(sp, cacheL1, cacheL2, assigned)

        logger.info('[{sp}] [L2] Removed {n} workplaces from L2 cache'.format(sp=sp.name,n=vms_to_remove))

    @staticmethod
    def count_vms_to_create(service_pool):
        """Расчет допустимого количества создаваемых ВМ"""
        # free_slots = Создавать ВМ одновременно - Кол-во машин в состоянии подготовки для провайдера
        # K = min(Создавать ВМ одновременно / кол-во фондов этого провайдера, free_slots)

        provider_id = service_pool.service.provider_id
        # Количество ВМ создаваемых одновременно
        max_preparing_vms = service_pool.service.getInstance().parent().getMaxPreparingServices()
        # Кол-во машин в состоянии подготовки для провайдера
        preparing_vms_for_provider = UserServiceManager.manager().\
            getServicesInStateForProvider(provider_id, State.PREPARING)
        # Количество "свободных мест" в фондах
        free_slots = max_preparing_vms - preparing_vms_for_provider

        # Количество фондов для поставщика ресурсов
        number_of_pools_for_provider = DeployedService.objects.filter(service_id__provider_id=provider_id)\
            .count()
        slots_per_pool = max(1, max_preparing_vms//number_of_pools_for_provider)
        # Количество машин, которое может быт создано на данной итерации
        vms_available_to_create = min(slots_per_pool, free_slots)

        # Ограничение максимального количества одновременно создаваемых ВМ за проход
        max_vms_to_create = 1000

        logger.debug('[{sp}] Provider allows {d} workplace(s) in preparation state'
                     .format(sp=service_pool.name, d=max_preparing_vms))
        logger.debug('[{sp}] Provider has {d} workplace(s) in "PREPARING" state'
                     .format(sp=service_pool.name, d=preparing_vms_for_provider))
        logger.debug('[{sp}] Provider has {d} free slots for adding VM(s)'
                     .format(sp=service_pool.name, d=free_slots))
        logger.debug('[{sp}] Provider has {d} service pools, {n} slots per pool'
                     .format(sp=service_pool.name, d=number_of_pools_for_provider,
                             n=slots_per_pool))
        logger.debug('[{sp}] Can add {d} workplace(s) to service pool on this step'
                     .format(sp=service_pool.name, d=vms_available_to_create))

        vms_to_create = min(vms_available_to_create, max_vms_to_create)

        if vms_available_to_create > max_vms_to_create:
            logger.debug('[{sp}] Caution: The maximum number of VMs available to create '
                         'is hard limited up to {d}'
                         .format(sp=service_pool.name, d=max_vms_to_create))

        return vms_to_create


    @staticmethod
    def count_vms_to_remove(service_pool):
        """Расчет допустимого количества удаляемых ВМ"""
        # free_slots = Удалять ВМ одновременно - Кол-во машин в состоянии удаления для провайдера
        # K = min(Удалять ВМ одновременно / кол-во фондов этого провайдера, free_slots)

        provider_id = service_pool.service.provider_id
        # Количество ВМ удаляемых одновременно
        max_removing_vms = service_pool.service.getInstance().parent().getMaxRemovingServices()
        # Кол-во машин в состоянии удаления для провайдера
        removing_vms_for_provider = UserServiceManager.manager().\
            getServicesInStateForProvider(provider_id, State.REMOVING)
        # Количество "свободных мест" в фондах
        free_slots = max_removing_vms - removing_vms_for_provider

        # Количество фондов для поставщика ресурсов
        number_of_pools_for_provider = DeployedService.objects.filter(service_id__provider_id=provider_id)\
            .count()
        slots_per_pool = max(1, max_removing_vms//number_of_pools_for_provider)
        # Количество машин, которое может быт удалено на данной итерации
        vms_available_to_remove = min(slots_per_pool, free_slots)

        # Ограничение максимального количества одновременно создаваемых ВМ за проход
        max_vms_to_remove = 1000

        logger.debug('[{sp}] Provider allows {d} workplace(s) in removal state'
                     .format(sp=service_pool.name, d=max_removing_vms))
        logger.debug('[{sp}] Provider has {d} workplace(s) in "REMOVING" state'
                     .format(sp=service_pool.name, d=removing_vms_for_provider))
        logger.debug('[{sp}] Provider has {d} free slots for removing VM(s)'
                     .format(sp=service_pool.name, d=free_slots))
        logger.debug('[{sp}] Provider has {d} service pools, {n} slots per pool'
                     .format(sp=service_pool.name, d=number_of_pools_for_provider,
                             n=slots_per_pool))
        logger.debug('[{sp}] Can remove {d} workplace(s) from service pool on this step'
                     .format(sp=service_pool.name, d=vms_available_to_remove))

        vms_to_remove = min(vms_available_to_remove, max_vms_to_remove)

        if vms_available_to_remove > max_vms_to_remove:
            logger.debug('[{sp}] Caution: The maximum number of VMs available to remove '
                         'is hard limited up to {d}'
                         .format(sp=service_pool.name, d=max_removing_vms))

        return vms_to_remove


# Регистрация класса в качестве задачи в реестре celery.
celery.tasks.register(ServiceCacheUpdater)
