# -*- coding: utf-8 -*-

#
# Copyright (c) 2012 Virtual Cable S.L.
# All rights reserved.
#

'''
@author: Adolfo Gómez, dkmaster at dkmon dot com
'''
from __future__ import unicode_literals

from django.utils.translation import ugettext_noop as _
from django.conf import settings

from uds.core.osmanagers.BaseOsManager import idle_normalize
from uds.core.services import types as serviceTypes
from uds.core.ui.UserInterface import gui
from uds.core import osmanagers
from uds.core.util.State import State
from uds.core.util import log
from termidesk.license.base import OS_WINDOWS
import six

import logging

logger = logging.getLogger(__name__)


def scrambleMsg(data):
    '''
    Simple scrambler so password are not seen at source page
    '''
    if isinstance(data, six.text_type):
        data = data.encode('utf8')
    res = []
    n = 0x32
    for c in data[::-1]:
        res.append(chr(ord(c) ^ n))
        n = (n + ord(c)) & 0xFF
    return six.text_type(b''.join(res).encode('hex'))


class WindowsOsManager(osmanagers.OSManager):
    #typeName = _('Windows Basic OS Manager')
    typeName = _('ОС Windows')
    typeType = 'WindowsManager'
    #typeDescription = _('Os Manager to control windows machines without domain.')
    typeDescription = _('Параметры управления ВМ на ОС Windows без домена')
    iconFile = 'wosmanager.png'
    servicesType = (serviceTypes.VDI,)

    onLogout = gui.ChoiceField(
        label=_('При выходе из ОС'),
        order=10,
        rdonly=True,
        tooltip=_('Что делать при выходе пользователя из ОС'),
        values=[
            {'id': 'keep', 'text': _('Сохранять рабочее место')},
            {'id': 'remove', 'text': _('Удалять рабочее место')}
        ],
        defvalue='keep'
    )

    idle = gui.NumericField(
        label=_("Время простоя"),
        length=4,
        defvalue=-1,
        rdonly=False,
        order=11,
        tooltip=_('Время простоя (в секундах) до закрытия сессии пользователя '
                  '(-1 - не ограничивать по времени или от 60 до 86400 секунд).'),
        required=True
    )

    # TSRV-447: Для корректного построения описания интерфейса,
    # необходимо явно наследовать `maxSessionTime`
    maxSessionTime = osmanagers.OSManager.maxSessionTime
    
    os_type = OS_WINDOWS

    @staticmethod
    def validateLen(length):
        try:
            length = int(length)
        except Exception:
            raise osmanagers.OSManager.ValidationException(_('Length must be numeric!!'))
        if length > 6 or length < 1:
            raise osmanagers.OSManager.ValidationException(_('Length must be betwen 1 and 6'))
        return length

    def __setProcessUnusedMachines(self):
        self.processUnusedMachines = self._onLogout == 'remove'

    def __init__(self, environment, values):
        super(WindowsOsManager, self).__init__(environment, values)
        if values is not None:
            self._onLogout = values['onLogout']
            self._idle = int(values['idle'])
        else:
            self._onLogout = ''
            self._idle = -1

        self.__setProcessUnusedMachines()

    def release(self, service):
        pass

    def getName(self, service):
        '''
        gets name from deployed
        '''
        return service.getName()

    def infoVal(self, service):
        return 'rename:' + self.getName(service)

    def infoValue(self, service):
        return 'rename\r' + self.getName(service)

    def notifyIp(self, uid, service, data):
        si = service.getInstance()

        ip = ''
        # Notifies IP to deployed
        pairs = data.split(',')
        for p in pairs:
            key, val = p.split('=')
            if key.lower() == uid.lower():
                si.setIp(val)
                ip = val
                break

        self.logKnownIp(service, ip)
        service.updateData(si)

    def doLog(self, service, data, origin=log.OSMANAGER):
        # Stores a log associated with this service
        try:
            msg, level = data.split('\t')
            try:
                level = int(level)
            except Exception:
                logger.debug('Do not understand level {}'.format(level))
                level = log.INFO

            log.doLog(service, level, msg, origin)
        except Exception:
            logger.exception('WindowsOs Manager message log: ')
            log.doLog(service, log.ERROR, "do not understand {0}".format(data), origin)

    def process_login(self, userService, msg, data, options, ret):
        if '\\' not in data:
            self.loggedIn(userService, data, False)
        userService.setInUse(True)
        # We get the userService logged hostname & ip and returns this
        ip, hostname = userService.getConnectionSource()
        deadLine = self.get_dead_line()
        if userService.getProperty('actor_version', None) >= '2.0.0':
            ret = "{0}\t{1}\t{2}".format(ip, hostname, deadLine)
        else:
            ret = "{0}\t{1}".format(ip, hostname)
        return ret

    def process_prefinish(self, userService, msg, data, options, ret):
        ret = super().process_prefinish(userService, msg, data, options, ret)
        if options is not None and options.get('scramble', True) is False:
            return ret
        return scrambleMsg(ret)

    def processUnused(self, userService):
        '''
        This will be invoked for every assigned and unused user service that has been in this state at least 1/2 of Globalconfig.CHECK_UNUSED_TIME
        This function can update userService values. Normal operation will be remove machines if this state is not valid
        '''
        if self._onLogout == 'remove':
            userService.remove()

    def checkState(self, service):
        logger.debug('Checking state for service {0}'.format(service))
        return State.RUNNING

    def maxIdle(self):
        '''
        On production environments, will return no idle for non removable machines
        '''
        if self._idle <= 0:  # or (settings.DEBUG is False and self._onLogout != 'remove'):
            return None

        return self._idle

    def marshal(self):
        # type: () -> str
        '''
        Serializes the os manager data so we can store it in database
        '''
        return '\t'.join([
            'v3',
            self._onLogout,
            str(idle_normalize(int(self._idle))),
            str(self._max_session_time),
        ])

    def unmarshal(self, s):
        # type: (str) -> ()
        if not isinstance(s, str):
            s = s.decode('utf-8')
        data = s.split('\t')
        try:
            if data[0] == 'v1':
                self._onLogout = data[1]
                self._idle = -1
            elif data[0] == 'v2':
                self._onLogout, self._idle = data[1], idle_normalize(int(data[2]))
            elif data[0] == 'v3':
                self._onLogout, self._idle = data[1], idle_normalize(int(data[2]))
                # Длительность сессии пользователя
                self._max_session_time = data[3]
        except Exception:
            logger.exception('Exception unmarshalling. Some values left as default ones')

        self.__setProcessUnusedMachines()

    def valuesDict(self):
        return {'onLogout': self._onLogout, 'idle': self._idle, 'maxSessionTime': self._max_session_time, }
