# -*- coding: utf-8 -*-
"""
  Модуль содержит текущую версию сервера и вспомогательные функции
  для ее строкового представления.
"""

from __future__ import unicode_literals
import datetime
import os
import subprocess


def get_version(version=None):
    """Формирует строку с полной версией сервера"""
    version = get_complete_version(version)

    # Формируем части номера версии ПО:
    # main = X.Y[.Z][.N]
    # sub = -dev-<git branch>-<utc timestamp>-<id> - для версий разработчиков
    #     | -{a|b|rc}N - для альфа, бета и других релизов

    main = get_main_version(version)

    sub = ''
    if version[4] == 'devel':
        git_changeset = get_git_changeset()
        if git_changeset:
            sub = '-dev-%s' % git_changeset

    elif version[4] != 'final':
        mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'rc'}
        sub = '-{}{}'.format(mapping[version[4]], str(version[5]))

    return "{}{}".format(main, sub)


def get_main_version(version=None):
    """Return main version (X.Y[.Z]) from VERSION."""
    version = get_complete_version(version)
    parts = 4
    if version[3] == 0:
        parts = 3
        if version[2] == 0:
            parts = 2
    return '.'.join(str(x) for x in version[:parts])


def get_complete_version(version=None):
    """
    Return a tuple of the app version. If version argument is non-empty,
    check for correctness of the tuple provided.
    """
    if version is None:
        version = VERSION
    else:
        assert len(version) == 6
        assert version[4] in ('devel', 'alpha', 'beta', 'rc', 'final')

    return version


# @functools.lru_cache()
def get_git_changeset():
    """ Возвращает строку <git branch>.<utc timestamp>.<id> для текущего коммита в git.
        Не предполагается вызов данной функции в рабочей среде, т.е. только для среды разработки
    """
    repo_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
    git_branch_proc = subprocess.Popen(
        'git rev-parse --abbrev-ref HEAD',
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True, cwd=repo_dir, universal_newlines=True,
    )
    git_branch = git_branch_proc.communicate()[0].strip()

    git_short_sha_proc = subprocess.Popen(
        'git rev-parse --short HEAD',
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True, cwd=repo_dir, universal_newlines=True,
    )
    git_short_sha = git_short_sha_proc.communicate()[0].strip()

    git_log = subprocess.Popen(
        'git log --pretty=format:%ct --quiet -1 HEAD',
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        shell=True, cwd=repo_dir, universal_newlines=True,
    )
    timestamp = git_log.communicate()[0]
    try:
        timestamp = datetime.datetime.utcfromtimestamp(int(timestamp))
    except ValueError:
        return None
    normalized_git_branch = git_branch.replace("/", "-")
    normalized_git_branch = normalized_git_branch.replace("@", "-at-")
    sub = "{}-{}-{}".format(normalized_git_branch, timestamp.strftime('%Y%m%d%H%M%S'), git_short_sha)
    return sub


VERSION = (3, 1, 0, 0, 'final', 0)
__version__ = '3.1'

if __name__ == "__main__":
    print(get_version(VERSION))
