# -*- coding: utf-8 -*-
'''
Created on 16 янв. 2019 г.

@author: dvd
'''
from __future__ import unicode_literals

# 2to3: Optimized imports: deleted json, websocket, socket, urllib

import websockify
import logging
import requests
from json import loads as json_loads
from termidesk.core.util import wsptoken
from rest_framework import status
import argparse

WORKMODE_LOCAL = 'local'
WORKMODE_REMOTE = 'remote'


class CommonProxyRequestHandler(websockify.ProxyRequestHandler):
    """ Обработчик запросов с общими методами. """
    def _get_host_and_port(self, connection_data):
        self.logger.debug("Decoded token data: %s", connection_data)
        if 'test_connection' in connection_data:
            self.logger.info("Got request for check connection to wsproxy.")
            # Для отправки ответа через websockify необходимо явно задать протокол 'HTTP/1.1' и код ответа 100
            # Сообщение HTTP-ответа заменяется на 'test_connection', чтобы принимающая сторона была уверена, что
            # ответ пришел от Termidesk
            self.protocol_version = 'HTTP/1.1'
            self.code = connection_data['test_connection']['test_code']
            self.message = connection_data['test_connection']['test_message']
            # Важно! Необходимо вызвать send_error с измененными кодом и сообщением в качестве ответа на запрос
            # соединения, для того, чтобы тестирующий метод понял, что тестирование прошло успешно.
            # В штатном режиме, ответ содержит код 101 - это значит соединение веб-сокетом успешно установлено.
            # Любой другой код, отличный от 101 считается ошибкой, поэтому для тестирования отправляется произвольный
            # код ответа, а тестирующий метод ожидает возникновение исключения и анализирует код ответа.
            self.send_error(int(self.code), self.message)
            self.logger.info(
                "Send response: '{proto} {code} {msg}'".format(
                    proto=self.protocol_version, code=self.code, msg=self.message))
            # Чтобы не оставлять после себя открытый сокет, указываются пустой host и нулевой порт
            # При этом в логе wsproxy будет присутствовать запись
            # 'Failed to connect to :0' - это нормально
            (target_host, target_port) = ("", 0)
        else:
            target_host = connection_data['host']
            target_port = connection_data['port']
            self.server.ssl_target = connection_data['ssl_target']

        return (target_host, target_port)


class TermideskProxyRequestHandler(CommonProxyRequestHandler):
    def __init__(self, retsock, address, proxy, *args, **kwargs):
        self._proxy = proxy
        websockify.ProxyRequestHandler.__init__(self, retsock, address, proxy, *args, **kwargs)

    def get_target(self, target_plugin):
        """
        Parses the path, extracts a token, and looks for a valid
        target for that token in the configuration file(s). Returns
        target_host and target_port if successful and sets an ssl_target
        flag.
        """
        # URL для соединения выглядит следующим образом
        # ws(s)://ip-address-or-domain-name/websocket/<token>
        # <token> - json-строка в кодировке base64
        # Если json-строка содержит структуру 'test_connection', вида
        #     test_connection: {"test_code": 100, "test_message": "test_connection_message"}
        # тогда отправляется HTTP-ответ 'HTTP/1.1 100 test_connection_message'
        # ВАЖНО! При проверке соединения, host и port из токена игнорируются
        # и заменяются на host = "", port = 0
        #
        # Если <token> не содержит структуру 'test_connection', то работаем с токеном как обычно -
        # получаем из него адрес хоста, порт и т.д.
        connection_data = self._proxy._tokenDecoder.decode(self.path[2:])

        return self._get_host_and_port(connection_data)


class APIProxyRequestHandler(CommonProxyRequestHandler):
    def __init__(self, retsock, address, proxy, *args, **kwargs):
        self._proxy = proxy
        websockify.ProxyRequestHandler.__init__(self, retsock, address, proxy,
                                                *args, **kwargs)

    def get_target(self, target_plugin):
        full_path = self._proxy._api_url + self.path[2:]
        self.logger.debug("Send POST request by path: '{path}'".format(path=full_path))

        response = requests.post(full_path)
        if response.status_code != status.HTTP_200_OK:
            message = "Token failed validation by path {}".format(full_path)
            self.logger.error(message)
            raise ValueError(message)

        connection_data = json_loads(response.content)

        self.logger.debug(
            "Accept response with {status} status and data: {data} ".format(
                status=response.status_code, data=response.content.decode("utf-8")
            )
        )

        return self._get_host_and_port(connection_data)


class TermideskWebSocketProxy(websockify.WebSocketProxy):
    """"
    Обертка вокруг websockify прокси для работы с Termidesk
    """

    def __init__(self, tokenDecoder, logger = None, *args, **kwargs):
        self._tokenDecoder = tokenDecoder
        if logger:
            self._logger = logger
        else:
            self._logger = logging.getLogger('TermideskWebSocketProxy')

        super(TermideskWebSocketProxy, self).__init__(*args, **kwargs)

    def get_logger(self):
        return self._logger


class APITermideskWebSocketProxy(websockify.WebSocketProxy):
    """ Обертка вокруг websockify прокси для работы с API Termidesk. """

    def __init__(self, api_url: str, logger=None, *args, **kwargs):
        self._api_url = api_url
        if logger:
            self._logger = logger
        else:
            self._logger = logging.getLogger('APITermideskWebSocketProxy')

        super().__init__(*args, **kwargs)

    def get_logger(self):
        return self._logger


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Termidesk websocket proxy')
    parser.add_argument('--log', default='', help='Log file')
    parser.add_argument('--verbose', dest='verbose', action='store_true', help='Verbose output')
    parser.add_argument('--host', default='127.0.0.1', help='Listen ip address')
    parser.add_argument('--port', default=5099, type=int, help='Listen port')
    parser.add_argument('--verification-key', help='Path to token data verification key')
    parser.add_argument(
        '--workmode',
        choices=[WORKMODE_REMOTE, WORKMODE_LOCAL],
        default=WORKMODE_LOCAL,
        help=('Token verification and decoding mode. '
              '"local" - verify and decode locally using verification RSA key. '
              '"remote" - verify and decode using server REST api endpoint. '
              'Default mode: local')
    )
    parser.add_argument('--url', 
                        required=False, 
                        help=('Base url for token verification and decoding. '
                              'URL syntax: http[s]://{host}[:port]/api/wsproxy/v1/verify/'))

    args = parser.parse_args()

    log_level = logging.INFO
    log_format = '[%(asctime)s] [%(process)s] [%(threadName)s] [%(levelname)s] [%(name)s] %(message)s'
    log_datefmt = '%Y-%m-%d %H:%M:%S %z'
    if args.verbose:
        log_level = logging.DEBUG
    if args.log:
        logging.basicConfig(
            format = log_format,
            datefmt = log_datefmt,
            filename=args.log,
            level=log_level)
    else:
        logging.basicConfig(
            format = log_format,
            datefmt = log_datefmt,
            level=log_level)

    kwargs = {'token_plugin': 'TokenFile'}

    if args.workmode == WORKMODE_REMOTE:
        if not args.url:
            raise ValueError('You need to pass url param for remote workmode.')

        APITermideskWebSocketProxy(
            api_url=args.url,
            listen_host=args.host,
            listen_port=args.port,
            source_is_ipv6=False,
            verbose=args.verbose,
            logger=None,
            daemon=False,
            web=None,
            target_host=None,
            target_port=None,
            wrap_mode='exit',
            wrap_cmd=None,
            RequestHandlerClass=APIProxyRequestHandler,
            **kwargs
        ).start_server()
    else:
        if not args.verification_key:
            raise ValueError('You need to pass verification_key param for local workmode.')

        with open(args.verification_key) as f:
            public_key = f.read()

        TermideskWebSocketProxy(
            listen_host=args.host,
            listen_port=args.port,
            source_is_ipv6=False,
            verbose=args.verbose,
            tokenDecoder=wsptoken.TokenDecoder(public_key),
            logger=None,
            daemon=False,
            web=None,
            target_host=None,
            target_port=None,
            wrap_mode='exit',
            wrap_cmd=None,
            RequestHandlerClass=TermideskProxyRequestHandler,
            **kwargs
        ).start_server()
